#!/usr/bin/env python
"""
generate Prophecy4f input files for all decay channels for given
Higgs mass.

usage: gen-input.py <mh>
"""

import sys

# these parameters are from the LHC Higgs Cross Section Working Group
# recommended Standard Model input parameters (as of 23-feb-2010)
# see https://twiki.cern.ch/twiki/bin/view/LHCPhysics/SMInputParameter
inp = """
outputfile=out.%s.%i             ! name of outputfile
nevents=100000000                ! nevents: number of weighted events

contrib=%i                       ! contrib: 1=best 2=IBA 3=Born

fname(1)=%s                      ! fermion 1
fname(2)=%s                      ! (anti-)fermion 2
fname(3)=%s                      ! fermion 3
fname(4)=%s                      ! (anti-)fermion 4

mh       = %id0                 ! Higgs boson mass
alphas   = 0.119d0              ! strong coupling constant
gf       = 1.16637d-5           ! Fermi constant
mz       = 91.1876d0            ! Z boson mass
mw       = 80.398d0             ! W boson mass
gammaz   = 2.4952               ! on-shell Z-boson width (only to calculate pole mass)
gammaw   = 2.141                ! on-shell W-boson width (only to calculate pole mass)
me       = 0.51099891d-3        ! electron mass
mmu      = 105.658367d-3        ! muon mass
mtau     = 1.77684d0            ! tau mass
md       = 0.190d0              ! d-quark mass
mu       = 0.190d0              ! u-quark mass
ms       = 0.190d0              ! s-quark mass
mc       = 1.4d0                ! c-quark mass
mb       = 4.75d0               ! b-quark mass (pole mass from MSTW2008)
mt       = 172.5d0              ! t-quark mass
"""

# full particle names in prophecy4f convention
procs = {0 : ('nue', 'anti-nue', 'num', 'anti-num'),
         1 : ('e',   'anti-e',   'mu',  'anti-mu'),
         2 : ('nue', 'anti-nue', 'mu',  'anti-mu'),
         3 : ('nue', 'anti-e',   'mu',  'anti-num'),
         4 : ('nue', 'anti-nue', 'nue', 'anti-nue'),
         5 : ('e',   'anti-e',   'e',   'anti-e'),
         6 : ('nue', 'anti-e',   'e',   'anti-nue'),

         10 : ('uq', 'anti-uq', 'cq', 'anti-cq'),
         11 : ('dq', 'anti-dq', 'sq', 'anti-sq'),
         12 : ('uq', 'anti-uq', 'sq', 'anti-sq'),
         13 : ('uq', 'anti-dq', 'sq', 'anti-cq'),
         14 : ('uq', 'anti-uq', 'uq', 'anti-uq'),
         15 : ('dq', 'anti-dq', 'dq', 'anti-dq'),
         16 : ('uq', 'anti-dq', 'dq', 'anti-uq'),

         25 : ('nue', 'anti-nue', 'uq', 'anti-uq'),
         26 : ('nue', 'anti-nue', 'dq', 'anti-dq'),
         27 : ('uq', 'anti-uq', 'e', 'anti-e'),
         28 : ('dq', 'anti-dq', 'e', 'anti-e'),
         29 : ('nue', 'anti-e', 'dq', 'anti-uq')}

procfnames = {0: 'nue.nue.num.num',
              1: 'e.e.mu.mu',
              2: 'nue.nue.mu.mu',
              3: 'nue.e.mu.num',
              4: 'nue.nue.nue.nue',
              5: 'e.e.e.e',
              6: 'nue.e.e.nue',

              10:'uq.uq.cq.cq',
              11:'dq.dq.sq.sq',
              12:'uq.uq.sq.sq',
              13:'uq.dq.sq.cq',
              14:'uq.uq.uq.uq',
              15:'dq.dq.dq.dq',
              16:'uq.dq.dq.uq',

              25:'nue.nue.uq.uq',
              26:'nue.nue.dq.dq',
              27:'uq.uq.e.e',
              28:'dq.dq.e.e',
              29:'nue.e.dq.uq'}


def geninput(mh=130, contrib=1):
    """generates input files for all 19 processes for given Higgs mass
    and contribution.

    contrib: 1=best 2=IBA 3=Born
    """
    for proc in [0,1,2,3,4,5,6, 10,11,12,13,14,15,16, 25,26,27,28,29]:
        part = procs[proc]
        fname = 'input.%s.%i' % (procfnames[proc], mh)

        f = open(fname, 'w')
        f.write(inp % (procfnames[proc], mh, contrib,
                       part[0], part[1], part[2], part[3], mh))
        f.close()

if __name__ == "__main__":
    if len(sys.argv) == 2:
        geninput(mh=int(sys.argv[1]))
    else:
        print __doc__
