************************************************************************
*
*     phase-space integration
*       weight(1):Born
*       weight(2):EW virtual correction, endpoint of subtraction + part of
*                   non-coll-safe case /soft part of slicing, part of LL FSR
*       weight(3):QCD virtual correction, endpoint of subtraction
*                   /soft part of slicing
*       weight(4-7):non-collinear-safe part of 4-particle phase space
*       weight(8-22):LL FSR
*       weight(23):Photon Bremsstrahlung
*       weight(24-35):corresponding subtraction contributions
*       weight(36):Gluon Bremsstrahlung
*       weight(37..48): corresponding subtraction contributions
*       outmom: outgoing momenta generated by subroutine phasespace
*       distmom: momenta after possible recombination and FSR
*                (numbering according to weight)
************************************************************************
      subroutine integration
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/output.inc'
      include 'COMMON/options.inc'                 !11 numbers for phase space
      integer i,nw                                 !1 for channel, 1 for z,
      real*8 weight(maxw),ran1,ran_rcarry,random(18)!1 for unweighted events
      real*8 average,sigma,bornaverage,weightsum,prelim,prelim_si
      real*8 distmom(0:3,5,maxw),dummy,dummy2
      real*8 unwran,weightforun(maxw),maxweight(maxw),glomaxweight
      real*8 bornweight,safetyfactor
      real*8 weightborn,maxbornweight
      integer bornflag,virtflag,realflag,evprop,genunwbornev
      integer genwev,genunwev,novershoot,negev,unweight,bornunweight
      integer nweight(maxw)
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej
c random generator
      integer rangen,ranshift
      common/random/rangen,ranshift
c unweighting
      integer unweighting,wflag(maxw)
      common/unweighting/unweighting,wflag

c initialize
      unweighting = 0  !generate weighted events in first cycle
      average = 0d0
      sigma   = 0d0
      prelim  = 0d0
      prelim_si = 0d0
      do i = 1,maxsubc
        subav(i) = 0d0
        subsi(i) = 0d0
        submax(i) = 0d0
      enddo
      do nw = 1,maxw
         maxweight(nw) = 0d0
         subrej(nw) = 0
      enddo
      nancount = 0
      ewcount  = 0
      qcdcount = 0
      if (nunwevents.gt.0) !check if directories exist
     &     call create_histo(0d0,distmom(0,1,1),1,1)
      call create_histo(0d0,distmom(0,1,1),1,0)

c integrate
      do nev = 1,nevents
        do nw = 1,maxw
          weight(nw) = 0d0
        enddo
        do i = 1,18
          if (rangen.eq.1) then
            random(i) = ran1()
          elseif (rangen.eq.2) then
            random(i) = ran_rcarry()
          endif
        enddo

        if (contrib.ne.4)
     &       call weight_bornvirt(random,weight(1),distmom(0,1,1)) !weight:1-22
        if (contrib.ne.2.and.contrib.ne.3.and.contrib.ne.7
     &       .and.qqcd.le.4)
     &       call weight_real(random,weight(23),distmom(0,1,23))!weight:23-48

        weightsum = 0d0
        do nw = 1,maxw
          weightsum = weightsum + weight(nw)
          call create_histo(weight(nw),distmom(0,1,nw),2,0) !input weight of
                                                            !sub-event nw
        enddo
        call create_histo(weight(1),distmom(0,1,1),3,0)  !input Born
        average = average + weightsum
        sigma  = sigma  + weightsum**2

c maximal weights for unweighting, combine weights 1+2+3
        do nw = 1,maxw
           weightforun(nw) = weight(nw)
        enddo
        bornweight = weight(1)
        weightforun(1) = weightforun(1)+weightforun(2)+weightforun(3)
        weightforun(2) = 0d0
        weightforun(3) = 0d0
        if (dabs(bornweight).gt.maxbornweight)
     &       maxbornweight = dabs(bornweight)
        do nw = 1,maxw
           if (dabs(weightforun(nw)).gt.maxweight(nw)) !max. weight for each
     &          maxweight(nw) = dabs(weightforun(nw)) !subevent for unweighting
        enddo

      enddo
      call create_histo(weight(1),distmom(0,1,1),4,0) !finalize histo

c final result
      average = average /dble(nevents)
      sigma   = sigma  /dble(nevents)**2
      sigma   = dsqrt(abs(sigma-average**2/dble(nevents)))

c sub-contributions and maximal weight for each sub-contribution
c  (only for output)
      do i = 1,maxsubc
        subav(i) = subav(i) /dble(nevents)
        subsi(i) = subsi(i) /dble(nevents)**2
        subsi(i) = dsqrt(abs(subsi(i)-subav(i)**2/dble(nevents)))
        submax (i) = submax(i) /dble(nevents)
      enddo
      bornaverage = subav(1)

      call write_output(average,sigma)

c----------------------------------------------------------------------
c generate unweighted events
      if (nunwevents.gt.0) then
         safetyfactor = 1.5d0
         unweighting = 1
         call create_histo(0d0,distmom(0,1,1),1,1) !hist.with unweighted events
         glomaxweight = 0d0
         do nw = 1,maxw
            if (dabs(maxweight(nw)).gt.glomaxweight) !determine overall
     &           glomaxweight = maxweight(nw)        !maximum weight
         enddo
         do nw = 1,maxw
            if (maxweight(nw).ne.0) then       !calculate weight nw only every
               nweight(nw) = int(glomaxweight/maxweight(nw)) !nweight's time
            else
               nweight(nw) = 0
            endif
         enddo

         genunwev = 0         !number of generated unweighted events
         genunwbornev = 0     !number of generated unweighted tree-level events
         negev = 0            !number of negative events
         novershoot = 0       !number of weights larger than largest weight
                              !in 1st cycle
         do nev = 1,10000*nunwevents
           do nw = 1,maxw      !calculate only if wflag.ne.0
              weight(nw) = 0d0
              if (nweight(nw).ne.0) then
                 if (mod(nev,nweight(nw)).eq.0) then
                    wflag(nw) = 1
                 else
                    wflag(nw) = 0
                 endif
              endif
           enddo
c combine weights 1+2+3 (Born,EW+QCD(endpoint))
           wflag(2) = wflag(1)
           wflag(3) = wflag(1)
c
           bornflag = wflag(1)
           if (contrib.eq.2) then !IBA
              do i = 2,16
                 bornflag = bornflag + wflag(i)
              enddo
              if (bornflag.ne.0) bornflag = 1
           endif
           do i = 2,22
              virtflag = virtflag + wflag(i)
           enddo
           if (virtflag.ne.0) virtflag = 1
           do i = 23,48
              realflag = realflag + wflag(i)
           enddo
           if (realflag.ne.0) realflag = 1

           do i = 1,18
              if (rangen.eq.1) then
                 random(i) = ran1()
              elseif (rangen.eq.2) then
                 random(i) = ran_rcarry()
              endif
           enddo

           if (contrib.ne.4.and.bornflag.eq.1.or.virtflag.eq.1)
     &       call weight_bornvirt(random,weight(1),distmom(0,1,1)) !weight:1-22
           if (contrib.ne.2.and.contrib.ne.3.and.contrib.ne.7
     &          .and.qqcd.le.4.and.realflag.eq.1)
     &       call weight_real(random,weight(23),distmom(0,1,23))  !weight:23-48

           unwran = ran_rcarry()
c combine weights 1+2+3 (Born,EW+QCD(endpoint))
           bornweight = weight(1) *nweight(1)*wflag(1)
           weight(1) = weight(1) + weight(2) + weight(3)
           weight(2) = 0d0
           weight(3) = 0d0
           do nw = 1,maxw
             weight(nw) = weight(nw) *nweight(nw)*wflag(nw)
c generate unweighted events
             if (dabs(weight(nw)).gt.safetyfactor*glomaxweight)
     &            novershoot = novershoot+1
             if (unwran*safetyfactor*glomaxweight.lt.dabs(weight(nw)))
     &            then
                 genunwev = genunwev + 1
                 unweight = +1
                 if (weight(nw).lt.0d0) then
                    negev = negev+1
                    unweight = -1
                 endif
                 if (nw.eq.1.or.nw.eq.2) evprop = 0       !tree
                 if (nw.gt.3) evprop = 1                  !photon
                 if (nw.gt.35) evprop = 2                 !gluon
                 if (contrib.eq.2.and.nw.gt.1) evprop = 1 !IBA
                 if (distmom(0,5,nw).eq.0) evprop = 0     !no photon/gluon
                 call unweighted_output(distmom(0,1,nw),evprop,
     &                                  average*unweight)
                 call create_histo(average*unweight,distmom(0,1,nw),2,1)
              endif
           enddo

c bornweight for testing purpose
           if (unwran*safetyfactor*maxbornweight.lt.dabs(bornweight))
     &          then
              genunwbornev = genunwbornev + 1
              bornunweight = sign(1d0,bornweight)
           else
              bornunweight = 0
           endif
           if (genunwbornev.gt.nunwevents) bornunweight = 0d0
           call create_histo(bornaverage*bornunweight,distmom(0,1,1),
     &                       3,1)   !input Born and final calc. of whole event

           genwev = nev         !number of calls of weighted events needed
           if ((genunwev-2*negev).ge.nunwevents) goto 10 !finished
        enddo

 10     continue

        call create_histo(1d0,distmom(0,1,1),4,1) !finalize histo
        call write_unweighting(genunwev,genwev,novershoot,negev,
     &       maxweight,safetyfactor*glomaxweight)

      endif ! end unweighting

      If(nout.ne.6) close(nout)

      end

************************************************************************
*
*     generate event for 4-particle phase space and calculate weight
*           for Born cross section and virtual corrections
*
*       weight(1):Born
*       weight(2):EW virtual correction, endpoint of subtraction + part of
*                   non-coll-safe case /soft part of slicing, part of LL FSR
*       weight(3):QCD virtual correction, endpoint of subtr.
*                                /soft part of slicing
*       weight(4-7):non-collinear-safe part of 4-particle phase space
*       weight(8-22):LL FSR
*
************************************************************************
      subroutine weight_bornvirt(random,weight,distmom)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      real*8 random(18),weight(22),w,zllfsr(4),wend,wendsli,wendqcd
      integer ngen,nparticles,nocut,disc,i,j,nweights,subc
      integer increqcd,incqcd,incew,incvirt,incsub,incvirtend
      integer incfsr,nwborn,nwew,nwqcd,nwncs,nwfsr
      real*8 inmom(0:3),outmom(0:3,5),distmom(0:3,5,22),sij(7,7)
      real*8 dens(maxchan),denstotal,m2_hvv4f,M2Born,M2IBA
      real*8 born(0:4),z,dM2_virt_EW,dM2_virt_QCD
      real*8 wendsliqcd,w_born,w_ew,w_qcd
      logical misnan
c unit (e.g. MeV)
      real*8 unit
      common/unit/unit
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c m2flip
      real*8 m2flip(4)
      common/m2flip/m2flip
c unweighting
      integer unweighting,wflag(maxw)
      common/unweighting/unweighting,wflag

      ngen = 1              ! number of generator
      nparticles = 4        ! number of outgoing particles
      nweights = 22          ! number of weights in this subroutine
      nwborn = 1
      nwew   = 2
      nwqcd  = 3
      nwncs  = 4
      nwfsr  = 8
c include virtual, EW and (real!) QCD contribution?
      if (contrib.ne.2.and.contrib.ne.3.and.contrib.ne.4) then
         incvirt = 1
      else
         incvirt = 0
      endif
      incew = 0
      incqcd = 0
      increqcd = 0
      if (qqcd.eq.0.or.qqcd.eq.1) incew = 1
      if (qqcd.ge.0) incqcd = 1
      if ((qqcd.eq.1.or.qqcd.eq.2.or.qqcd.eq.3.or.qqcd.eq.4)
     &     .and.(quark(1).eq.1.or.quark(3).eq.1)) increqcd = 1
      incvirtend = 0
      incfsr = 0
      incsub = 0
      if (contrib.eq.1.or.contrib.eq.5) incsub = 1
      if (contrib.eq.1.or.contrib.eq.6) incvirtend = 1
      if (qllfsr.ne.0.and.incew.eq.1.and.(contrib.eq.1.or.
     &     contrib.eq.2.or.contrib.eq.5.or.contrib.eq.7)) incfsr = 1
c define incoming momenta (has to be in 3-direction!)
      inmom(0) = dsqrt(CMenergy2) ! Higgs mass
      do i = 1,3
        inmom(i) = 0d0
      enddo

c-----------------------------------------------------------------------
c choose channel(chan), generate outgoing momenta, apply cuts and calculate
c densities dens(chan) and total density
      call phasespace(random,ngen,inmom,outmom,distmom(0,1,1),sij,dens
     &     ,denstotal,nparticles,nocut,disc)
      do i=0,3
      do j=1,5      !2:EW virtual corr., 3:QCD virtual corr.
        distmom(i,j,2) = distmom(i,j,1)
        distmom(i,j,3) = distmom(i,j,1)
      enddo
      enddo

      z = random(13)  ! fraction of photon energy in new subtr. terms
      do i = 1,4
        zllfsr(i) = random(14+i) !random numbers 15-18 for LL FSR
      enddo

      if (nocut.ne.1.or.disc.eq.1) then
        subrej(nwborn) = subrej(nwborn) +1
        subrej(nwew) = subrej(nwew) +1
        subrej(nwqcd) = subrej(nwqcd) +1
      else
c-----------------------------------------------------------------------

c Born contribution
        call setprods(outmom,nparticles)
c        weight(1)=1d0 /denstotal /(pi/2d0)**3 *12d0 /dsqrt(CMenergy2)**4 !PS-V
        if (contrib.eq.2) then
          call calcsme()
          M2Born = M2IBA()
        else
          M2Born = m2_hvv4f()
        endif
        weight(nwborn) = weight(nwborn) + M2Born
c-----------------------------------------------------------------------

c virtual corrections with Born kinematics
        if (incvirt.eq.1) then
           born(0) = M2Born
           do i = 1,4
              born(i) = m2flip(i)
           enddo
           call calcsme()
           if (qfast.eq.1) call fastinit(0)
c calculate QCD virtual corrections before EW for fast evaluation to work
c if QCD is calculated more often (nvirt(2).lt.nvirt(1))
           if (incqcd.eq.1.and.incvirtend.eq.1) then
              if(mod(nev,nvirt(2)).eq.0.and.wflag(nwqcd).eq.1) then
                                ! calculate only every nvirt'th time
                 weight(nwqcd) = weight(nwqcd) + dM2_virt_QCD()*nvirt(2)
                 if (increqcd.eq.1) then
                    wendqcd = 0d0
                    call sub4_qcd(sij,wendqcd)
                    weight(nwqcd) = weight(nwqcd) + wendqcd *nvirt(2)
                 endif
              endif
           endif
           if (incew.eq.1.and.incvirtend.eq.1) then
              if(mod(nev,nvirt(1)).eq.0.and.wflag(nwew).eq.1) then
                                ! calculate only every nvirt'th time
                 weight(nwew) = weight(nwew) + dM2_virt_EW() *nvirt(1)
                 wend = 0d0
                 call sub4(born,sij,wend)
                 weight(nwew) = weight(nwew) + wend *nvirt(1)
              endif
           endif
        endif

        if (qsoftcoll.eq.2.and.incsub.eq.1) then
c subtract sub-contrib and add slicing-contrib if slicing method is used
           wend = 0d0
           wendsli = 0d0
           wendqcd = 0d0
           wendsliqcd = 0d0
           if (incew.eq.1) then
              call sub4(born,sij,wend)
              call sli_soft(outmom(0,1),born,sij,wendsli)
           endif
           if (increqcd.eq.1) then
              call sub4_qcd(sij,wendqcd)
              call sli_softcoll_qcd(outmom(0,1),sij,wendsliqcd)
           endif
           weight(nwew) = weight(nwew) - wend + wendsli
           weight(nwqcd) = weight(nwqcd) - wendqcd + wendsliqcd
        endif
      endif

c-----------------------------------------------------------------------

c virtual corrections and LL FSR with different kinematics
      if (incew.eq.1.and.incsub.eq.1) then
c non-collinear-safe terms for subtraction, collinear terms for slicing
         if (qsoftcoll.eq.1.and.disc.ne.1) then
            call sub4_ncs(outmom(0,1),distmom(0,1,nwncs),z,born,sij
     &           ,weight(nwew),weight(nwncs),nwncs,nocut)
         elseif (qsoftcoll.eq.2.and.disc.ne.1.and.qrecomb.eq.0) then
            call sli_ncs(outmom(0,1),distmom(0,1,nwncs),z,born,sij
     &           ,weight(nwncs),nwncs)
         endif
      endif

c LL final-state radiation
      if (incfsr.eq.1.and.disc.ne.1) then
         call llfsr(outmom(0,1),born(0),weight(nwew),weight(nwfsr)
     &        ,distmom(0,1,nwfsr),nwfsr,nocut,zllfsr)
      endif

c-----------------------------------------------------------------------
c summing up sub-contributions and adaptive optimization

      do i = 1,nweights
        weight(i) = weight(i) /denstotal /(2d0*dsqrt(CMenergy2))
     &        /(2d0*pi)**8 *unit
      enddo

c discard unreasonable events
      if (disc.eq.0 .and. nocut.eq.1) then
      if (misnan(weight(nwew)) .or. misnan(weight(nwqcd))
     &   .or. (dabs(weight(nwew)/weight(nwborn)) .gt. weightcut)
     &   .or. (dabs(weight(nwqcd)/weight(nwborn)) .gt. weightcut)) then
         if (misnan(weight(nwew)) .or. misnan(weight(nwqcd))) then
            nancount = nancount + 1
         endif
         if (dabs(weight(nwew)/weight(nwborn)) .gt. weightcut) then
            ewcount = ewcount + 1
         endif
         if (dabs(weight(nwqcd)/weight(nwborn)) .gt. weightcut) then
            qcdcount = qcdcount + 1
         endif
         weight(nwew) = 0d0
         weight(nwqcd) = 0d0
      endif
      endif

c Born sub-contribution
      w_born = weight(nwborn)
      subc = 1
      subav(subc) = subav(subc) + w_born
      subsi(subc) = subsi(subc) + w_born**2
c maximal weight
      if (dabs(w_born).gt.dabs(submax(subc))) then
        submax(subc) = w_born
        subch(subc) = chan
        subn(subc) = nev
      endif

c EW sub-contribution
      subc = 2
      w_ew = weight(nwew)
      do i = nwncs,nweights
        w_ew = w_ew + weight(i)
      enddo
      subav(subc) = subav(subc) + w_ew
      subsi(subc) = subsi(subc) + w_ew**2
c maximal weight
      if (dabs(w_ew).gt.dabs(submax(subc))) then
        submax(subc) = w_ew
        subch(subc) = chan
        subn(subc) = nev
      endif

c QCD sub-contribution
      subc = 3
      w_qcd = weight(nwqcd)
      subav(subc) = subav(subc) + w_qcd
      subsi(subc) = subsi(subc) + w_qcd**2
c maximal weight
      if (dabs(w_qcd).gt.dabs(submax(subc))) then
        submax(subc) = w_qcd
        subch(subc) = chan
        subn(subc) = nev
      endif

c adaptive optimization according to Born matrix elements
      if (unweighting.eq.0)
     &     call optimization(w_born,dens,denstotal,ngen,nchannels(ngen))

      end

*******************************************************************
*
*     generate event and calculate weight for photon bremsstrahlung
*        weight(1)/outmom/distmom(1): Bremsstrahlung,
*        weight(2..13)/distmom(2..13): subtraction functions
*******************************************************************
      subroutine weight_real(random,weight,distmom)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      real*8 random(18),weight(26),outmom(0:3,5),distmom(0:3,5,26)
      integer ngen,nparticles,nocut,nocutgluon,disc,i,j,nwabs,nweights
      real*8 inmom(0:3),sij(7,7),dens(maxchan),denstotal,m2_hvv4fga,w
      integer singular,ferm,flag,slicut,check_slicut
      integer incew,increqcd,subc
      real*8 minen,mintheta,out,m2_hvv4fgluon,w_ew,w_qcd
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej
c unit (e.g. MeV)
      real*8 unit
      common/unit/unit
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c unweighting
      integer unweighting,wflag(maxw)
      common/unweighting/unweighting,wflag

      ngen = 2              ! number of generator
      if (nchannels(ngen).eq.0) return
      nparticles = 5        ! number of outgoing particles
      nwabs = 23            ! number of weight 23..48 (global counting)
      nweights = 26         ! number of weights in this subroutine
c include EW and QCD contribution?
      incew = 0
      increqcd = 0
      if (qqcd.eq.0.or.qqcd.eq.1) incew = 1
      if ((qqcd.eq.1.or.qqcd.eq.2.or.qqcd.eq.3.or.qqcd.eq.4)
     &     .and.(quark(1).eq.1.or.quark(3).eq.1)) increqcd = 1
c define incoming momenta (has to be in 3-direction!)
      inmom(0) = dsqrt(CMenergy2) ! Higgs mass
      do i = 1,3
        inmom(i) = 0d0
      enddo
c choose channel(chan), generate outgoing momenta, apply cuts and calculate
c total density
      call phasespace(random,ngen,inmom,outmom,distmom(0,1,1),sij
     &     ,dens,denstotal,nparticles,nocut,disc)
      slicut = 0
      if (qsoftcoll.eq.2) slicut = check_slicut(outmom(0,1))
      call setprods(outmom(0,1),nparticles)
      if (incew.eq.1) then
        if (nocut.eq.1.and.disc.ne.1.and.slicut.eq.0) then
          weight(1) = weight(1) + m2_hvv4fga()
        else
          subrej(nwabs) = subrej(nwabs) +1
        endif
      endif
      if (increqcd.eq.1) then
        if (disc.ne.1) call cut(outmom,distmom(0,1,14),nparticles
     &                          ,nocutgluon,1)
        if (nocutgluon.eq.1.and.disc.ne.1.and.slicut.eq.0) then
          weight(14) = weight(14) + m2_hvv4fgluon()
        else
          subrej(nwabs+13) = subrej(nwabs+13) +1
        endif
      endif

      if (disc.ne.1.and.contrib.ne.4.and.qsoftcoll.eq.1) then  !subtr.functions
        if (incew.eq.1) !weight:2..13
     &        call sub5(outmom,distmom(0,1,2),sij,weight(2),nwabs+1)
        if (increqcd.eq.1) !weight:15..26
     &        call sub5_qcd(outmom,distmom(0,1,15),sij,weight(15)
     &                      ,nwabs+14)
      endif

      do i = 1,nweights
        weight(i) = weight(i) /denstotal /(2d0*dsqrt(CMenergy2))
     &        /(2d0*pi)**11 *unit
      enddo

c check phase space volume
c$$$      weight(1) = weight(1) + 1d0 /denstotal /(pi/2d0)**4 *144d0
c$$$     &      /CMenergy2**3         !PS-volume

c EW subcontribution
      subc = 4
      w_ew = 0d0
      do i = 1,13
        w_ew = w_ew+weight(i)
      enddo
      subav(subc) = subav(subc) + w_ew
      subsi(subc) = subsi(subc) + w_ew**2
c maximal weight
      if (dabs(w_ew).gt.dabs(submax(subc))) then
        submax(subc) = w_ew
        subch(subc) = chan
        subn(subc) = nev
      endif
c QCD subcontribution
      subc = 5
      w_qcd = 0d0
      do i = 14,26
        w_qcd = w_qcd+weight(i)
      enddo
      subav(subc) = subav(subc) + w_qcd
      subsi(subc) = subsi(subc) + w_qcd**2
c maximal weight
      if (dabs(w_qcd).gt.dabs(submax(subc))) then
        submax(subc) = w_qcd
        subch(subc) = chan
        subn(subc) = nev
      endif
c adaptive optimization
      if (unweighting.eq.0) then
         w = w_ew + w_qcd
         call optimization(w,dens,denstotal,ngen,nchannels(ngen))
      endif

c check events in singular regions (turn off mininvmass-condition)
c      ferm=1
c      minen=1d0
c      mintheta=2d0
c      flag=1
c      if (singular(outmom,ferm,minen,mintheta,out,flag).eq.1) then
c        write(13,*),out,dabs(weight(1)/(w-weight(1))),weight(1)
c     &              ,w-weight(1)
c        write(13,*),out,dabs(weight(14)/(w-weight(14))),weight(14)
c     &             ,w-weight(14)
c      endif

      end

************************************************************************
*
*
************************************************************************
      subroutine event_born(random,weight,distmom)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/output.inc'
      real*8 random(18),weight
      integer ngen,nparticles,nocut,disc,i,j,k,nmap,em,sp,nweights,subc
      real*8 inmom(0:3),outmom(0:3,5),distmom(0:3,5),sij(7,7)
      real*8 dens(maxchan),denstotal,m2_hvv4f,jac,w
c unit (e.g. MeV)
      real*8 unit
      common/unit/unit
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels

      ngen = 3              ! number of generator
      nparticles = 4        ! number of outgoing particles
      nweights = 1          ! number of weights in this subroutine
c define incoming momenta (has to be in 3-direction!)
      inmom(0) = dsqrt(CMenergy2) ! Higgs mass
      do i = 1,3
        inmom(i) = 0d0
      enddo
c choose mapping from 5p-phsp to 4p-phsp according to random(14)
      nmap = 0
      do i=1,4
      do j=1,4
        if (i.ne.j.and.q(i).ne.0d0.and.q(j).ne.0d0) nmap = nmap+1
      enddo
      enddo
      k = 0
      do i=1,4
      do j=1,4
        if (i.ne.j.and.q(i).ne.0d0.and.q(j).ne.0d0) then
          if (random(14)*nmap.gt.k.and.random(14)*nmap.lt.k+1) then
            em = i+2
            sp = j+2
          endif
          k = k+1
        endif
      enddo
      enddo
c choose channel(chan), generate outgoing momenta, apply cuts and calculate
c densities dens(chan) and total density
      call phsp_4from5(random,em,sp,ngen,inmom,outmom,distmom,sij,dens
     &     ,denstotal,jac,nocut,disc)

      if (nocut.eq.1.and.disc.ne.1) then
        call setprods(outmom,nparticles)
c        weight = 1d0 /denstotal /(pi/2d0)**3 *12d0 /dsqrt(CMenergy2)**4 !PS-V
        weight = weight + m2_hvv4f() /denstotal *jac
     &       /(2d0*dsqrt(CMenergy2)) /(2d0*pi)**8 *unit
      else
        subrej(1) = subrej(1) +1
      endif

c subcontribution
      w = weight
      subc = 1
      subav(subc) = subav(subc) + w
      subsi(subc) = subsi(subc) + w**2
c maximal weight
      if (dabs(w).gt.dabs(submax(subc))) then
        submax(subc) = w
        subch(subc) = chan
        subn(subc) = nev
      endif

c adaptive optimization
      call optimization(weight,dens,denstotal,ngen,nchannels(ngen))

      end

*******************************************************************
*
*     subtraction functions for the four-particle phase space
*
*******************************************************************
      subroutine sub4(born,sij,weight)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      integer em,sp,i1,i2
      real*8 sij(7,7),weight
      real*8 born(0:4),pf,logs
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2

      do em=1,4
      do sp=1,4
        if(em.ne.sp.and.q(em).ne.0d0.and.q(sp).ne.0d0)then
          pf=(-1d0)**(em+sp)*alpha*q(em)*q(sp)/2d0/pi
          logs=(log(mf2(em)/sij(em+2,sp+2))+1d0)
     &         *log(lambda2/sij(em+2,sp+2))
     &	    -0.5d0*log(mf2(em)/sij(em+2,sp+2))**2
     &      +0.5d0*log(mf2(em)/sij(em+2,sp+2))
          weight = weight - pf*(logs-pi**2/3d0+1d0)*born(0)
     &                          - pf/2d0*born(em)
        endif
      enddo
      enddo

      end

*******************************************************************
*
*     QCD subtraction functions for the four-particle phase space
*       call m2_hvv4f() before
*
*******************************************************************
      subroutine sub4_qcd(sij,weight)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      integer em,sp,i,j
      real*8 sij(7,7),weight,CF,T_dg(4,4),T_dg42(4,4),T_ndg(4,4)
      real*8 pf,logs
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
c m2born (diagonal and non-diagonal contribution)
      real*8 M2B_dg,M2B_dg42,M2B_ndg
      common/m2born/M2B_dg,M2B_dg42,M2B_ndg

      CF = 4d0/3d0
      T_dg(1,2) = CF
      T_dg(3,4) = T_dg(1,2)
      T_dg(1,3) = 0d0
      T_dg(1,4) = T_dg(1,3)
      T_dg(2,3) = T_dg(1,3)
      T_dg(2,4) = T_dg(1,3)

      T_dg42(1,4) = CF
      T_dg42(2,3) = T_dg42(1,4)
      T_dg42(1,2) = 0d0
      T_dg42(1,3) = T_dg42(1,2)
      T_dg42(2,4) = T_dg42(1,2)
      T_dg42(3,4) = T_dg42(1,2)

      T_ndg(1,2) = CF
      T_ndg(1,3) = T_ndg(1,2)
      T_ndg(1,4) = T_ndg(1,2)
      T_ndg(2,3) = T_ndg(1,2)
      T_ndg(2,4) = T_ndg(1,2)
      T_ndg(3,4) = T_ndg(1,2)

      do i=2,4
      do j=1,i-1
        T_dg(i,j) = T_dg(j,i)
        T_dg42(i,j) = T_dg42(j,i)
        T_ndg(i,j) = T_ndg(j,i)
      enddo
      enddo

      do em=1,4
      do sp=1,4
        if(em.ne.sp.and.quark(em).ne.0.and.quark(sp).ne.0)then
          pf=(-1d0)**(em+sp)*alphas/2d0/pi
          logs=(log(mf2(em)/sij(em+2,sp+2))+1d0)
     &         *log(lambda2/sij(em+2,sp+2))
     &	    -0.5d0*log(mf2(em)/sij(em+2,sp+2))**2
     &      +0.5d0*log(mf2(em)/sij(em+2,sp+2))
          weight = weight - pf*(logs-pi**2/3d0+3d0/2d0)
     &             *( T_dg(em,sp)*M2B_dg + T_dg42(em,sp)*M2B_dg42
     &               +T_ndg(em,sp)*M2B_ndg)
        endif
      enddo
      enddo

      end

*******************************************************************
*
*     subtraction functions for the four-particle phase space
*       additional numerical integration over z for non-collinear
*       save observables
*
*******************************************************************
      subroutine sub4_ncs(virtmom,distmom,z,born,sij,weight4f,weight
     &     ,nwabs,nocut4f)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      integer em,sp,nparticles,nw,nwabs,ga,i1,i2,nocut,nocut4f
      real*8 virtmom(0:3,5),distmom(0:3,5,4),sij(7,7),weight4f,weight(4)
      real*8 z,born(0:4),pf,preco(0:3,5)
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej

      if ((1d0-z).lt.sub4cut) then
        do em = 1,4
        do i1=1,4
        do i2=0,3
          distmom(i2,i1,em) = virtmom(i2,i1) !set to any value because weight=0
        enddo
        enddo
        enddo
        nsub4 = nsub4+1
        return
      endif

      if (nocut4f.ne.0) then
        do em=1,4
        do sp=1,4
          if(em.ne.sp.and.q(em).ne.0d0.and.q(sp).ne.0d0)then
            pf=(-1d0)**(em+sp)*alpha*q(em)*q(sp)/2d0/pi
            weight4f = weight4f    !opposite sign in delta+ prescription
     &          + pf*((1d0+z**2)/(1d0-z)*
     &             (log(sij(em+2,sp+2)*z/mf2(em))-1d0)
     &             +(1d0+z)*log(1d0-z)) *born(0)
     &          + pf*(1d0-z) *born(em)
          endif
        enddo
        enddo
      endif

c ncs terms for z.ne.1
      ga=5
      nparticles = 5
      do em = 1,4
        if(q(em).ne.0d0)then
          do i1=1,4
          do i2=0,3
            preco(i2,i1)=virtmom(i2,i1)
          enddo
          enddo
          do i2=0,3
            preco(i2,em)=virtmom(i2,em) *z
            preco(i2,ga)=virtmom(i2,em) *(1d0-z)
          enddo
          call cut(preco,distmom(0,1,em),nparticles,nocut,0)!nocut=0: apply cut
          if(nocut.eq.0)then
            subrej(nwabs-1+em) = subrej(nwabs-1+em) + 1
          else
            do sp = 1,4
              if (sp.ne.em.and.q(sp).ne.0d0) then
                pf=(-1d0)**(em+sp)*alpha*q(em)*q(sp)/2d0/pi
                weight(em) = weight(em)
     &              -pf*((1d0+z**2)/(1d0-z)*
     &                  (log(sij(em+2,sp+2)*z/mf2(em))-1d0)
     &                  +(1d0+z)*log(1d0-z)) *born(0)
     &              -pf*(1d0-z) *born(em)
              endif
            enddo
          endif
        endif
      enddo

      end

*******************************************************************
*
*     subtraction funtions for 5-particle phase space
*         weight(1..12)/outmom(1..12)
*
*******************************************************************
      subroutine sub5(bremsmom,distmom,sij,weight,nwabs)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      real*8 bremsmom(0:3,5),distmom(0:3,5,12),sij(7,7),weight(12)
      integer nw,nwabs
      integer em,sp,ga,nocut,i1,i2,nparticles
      real*8 mom4(0:3,5),mom5(0:3,5),z,y,born,m2_hvv4f
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej

      nw = 1       !number of weight (1..12)
      ga = 5
      do em = 1,4
      do sp = 1,4
      if (em.ne.sp) then
        if (q(em).ne.0d0.and.q(sp).ne.0d0) then
          z = sij(em+2,sp+2)/(sij(em+2,sp+2)+sij(sp+2,ga+2))
          y = sij(em+2,ga+2)/(sij(em+2,sp+2)+sij(em+2,ga+2)
     &                        +sij(sp+2,ga+2))
          do i1=1,5
          do i2=0,3
            mom4(i2,i1) = bremsmom(i2,i1)
            mom5(i2,i1) = bremsmom(i2,i1)
          enddo
          enddo
          do i2=0,3
            mom4(i2,em)  =bremsmom(i2,em)+bremsmom(i2,ga)
     &                           -y/(1d0-y)*bremsmom(i2,sp)     !p^~_i
            mom4(i2,sp)  =1d0/(1d0-y)*bremsmom(i2,sp)           !p^~_j
            mom4(i2,ga)  =0d0
            mom5(i2,em)=mom4(i2,em) *z
            mom5(i2,sp)=mom4(i2,sp)
            mom5(i2,ga)=mom4(i2,em) *(1d0-z)
          enddo
          nparticles = 5
          if (bremsmom(0,5).lt.sub5cut*dsqrt(CMenergy2)
     &         .and.y.lt.sub5cut) then
            do i1 = 1,5
            do i2 = 0,3
              mom5(i2,i1) = bremsmom(i2,i1)
            enddo
            enddo
            nsub5 = nsub5+1
          endif
          call cut(mom5,distmom(0,1,nw),nparticles,nocut,0)!nocut=0:apply cut
          if (nocut.eq.0) then
            subrej(nwabs-1+nw) = subrej(nwabs-1+nw) + 1
          else
            nparticles = 4
            call setprods(mom4,nparticles)
            born = m2_hvv4f()
            weight(nw) = weight(nw) + 8d0*pi*(-1d0)**(em+sp)
     &            *(alpha*q(em)*q(sp)*born)
     &            *(2d0/(1d0-z*(1d0-y))-(1+z))/(1d0-y)/sij(em+2,ga+2)
          endif
        endif
        nw = nw+1
      endif
      enddo
      enddo

      end

*******************************************************************
*
*     QCD subtraction funtions for 5-particle phase space
*         weight(1..12)/outmom(1..12)
*
*******************************************************************
      subroutine sub5_qcd(bremsmom,distmom,sij,weight,nwabs)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      real*8 bremsmom(0:3,5),distmom(0:3,5,12),sij(7,7),weight(12)
      integer nw,nwabs
      integer em,sp,ga,nocut,i1,i2,i,j,nparticles
      real*8 mom4(0:3,5),mom4mod(0:3,5),z,y,born,m2_hvv4f,CF
      real*8 T_dg(4,4),T_dg42(4,4),T_ndg(4,4)
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej
c m2born (diagonal and non-diagonal contribution)
      real*8 M2B_dg,M2B_dg42,M2B_ndg
      common/m2born/M2B_dg,M2B_dg42,M2B_ndg

      CF = 4d0/3d0
      T_dg(1,2) = CF
      T_dg(3,4) = T_dg(1,2)
      T_dg(1,3) = 0d0
      T_dg(1,4) = T_dg(1,3)
      T_dg(2,3) = T_dg(1,3)
      T_dg(2,4) = T_dg(1,3)

      T_dg42(1,4) = CF
      T_dg42(2,3) = T_dg42(1,4)
      T_dg42(1,2) = 0d0
      T_dg42(1,3) = T_dg42(1,2)
      T_dg42(2,4) = T_dg42(1,2)
      T_dg42(3,4) = T_dg42(1,2)

      T_ndg(1,2) = CF
      T_ndg(1,3) = T_ndg(1,2)
      T_ndg(1,4) = T_ndg(1,2)
      T_ndg(2,3) = T_ndg(1,2)
      T_ndg(2,4) = T_ndg(1,2)
      T_ndg(3,4) = T_ndg(1,2)

      do i=2,4
      do j=1,i-1
        T_dg(i,j) = T_dg(j,i)
        T_dg42(i,j) = T_dg42(j,i)
        T_ndg(i,j) = T_ndg(j,i)
      enddo
      enddo

      nw = 1       !number of weight (1..12)
      ga = 5
      nparticles = 4 !collinear safety assumed
      do em = 1,4
      do sp = 1,4
      if (em.ne.sp) then
        if (quark(em).ne.0.and.quark(sp).ne.0) then
          z = sij(em+2,sp+2)/(sij(em+2,sp+2)+sij(sp+2,ga+2))
          y = sij(em+2,ga+2)/(sij(em+2,sp+2)+sij(em+2,ga+2)
     &                        +sij(sp+2,ga+2))
          do i1=1,5
          do i2=0,3
            mom4(i2,i1) = bremsmom(i2,i1)
          enddo
          enddo
          do i2=0,3
            mom4(i2,em)  =bremsmom(i2,em)+bremsmom(i2,ga)
     &                           -y/(1d0-y)*bremsmom(i2,sp)     !p^~_i
            mom4(i2,sp)  =1d0/(1d0-y)*bremsmom(i2,sp)           !p^~_j
            mom4(i2,ga)  =0d0
          enddo
          if (y.lt.sub5cutgl) then !treat weight as Bremsstrahlung
c          if (bremsmom(0,5).lt.sub5cutgl*dsqrt(CMenergy2)
c     &         .and.y.lt.sub5cutgl) then !treat weight as Bremsstrahlung
            nsub5gl = nsub5gl+1          !(only in histograms)
            nparticles = 5
            call cut(bremsmom,distmom(0,1,nw),nparticles,nocut,1)
            nparticles = 4
          else       !"correct" treatment
            call cut(mom4,distmom(0,1,nw),nparticles,nocut,1)
          endif
          if (nocut.eq.0) then
            subrej(nwabs-1+nw) = subrej(nwabs-1+nw) + 1
          else
            call setprods(mom4,nparticles)
            born = m2_hvv4f()
            weight(nw) = weight(nw) + 8d0*(-1d0)**(em+sp)*pi
     &            *alphas*( T_dg(em,sp)*M2B_dg + T_dg42(em,sp)*M2B_dg42
     &                     +T_ndg(em,sp)*M2B_ndg)
     &            *(2d0/(1d0-z*(1d0-y))-(1+z))/(1d0-y)/sij(em+2,ga+2)
          endif
        endif
        nw = nw+1
      endif
      enddo
      enddo

      end

************************************************************************
*
*     integrated soft (and collinear for qrecomb.ne.0)
*     part in the slicing approach
*
************************************************************************
      subroutine sli_soft(virtmom,born,sij,weight)
      implicit none
      real*8 virtmom(0:3,5),born(0:4),sij(7,7)
      real*8 weight,preco(0:3,5)
      real*8 dE,func,coll
      complex*16 cspen
      integer nw,i,j,i1,i2,ga,em,nparticles
      include 'COMMON/options.inc'
      include 'COMMON/max.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/param.inc'
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2

      func(i,j) = 2d0*log(2d0*dE/lambda)*(1d0-log(sij(i+2,j+2)/mf(i)
     &     /mf(j))) -log(4d0*virtmom(0,i)*virtmom(0,j)/mf(i)/mf(j))
     &     +log(2d0*virtmom(0,i)/mf(i))**2+log(2d0*virtmom(0,j)
     &     /mf(j))**2 +pi**2/3d0 +cspen(dcmplx(1d0-4d0*virtmom(0,i)
     &     *virtmom(0,j)/sij(i+2,j+2)))

      dE=deltaE*dsqrt(CMenergy2)

c endpoint contribution (soft)
      do i=1,4
      do j=i+1,4
         if (q(i).ne.0d0.and.q(j).ne.0d0)
     &      weight = weight + alpha/pi*(-1d0)**(i+j)*q(i)*q(j)
     &                             *func(i,j)*born(0)
      enddo
      enddo

c collinear part, analytical integration over z (coll-safe observables)
      if (qrecomb.ne.0) then
      do i=1,4
        if (q(i).ne.0d0) then
          coll = (3d0/2d0+2d0*log(dE/virtmom(0,i)))*(1d0-2d0
     &          *log(deltaTh*virtmom(0,i)/mf(i)))+5d0/2d0-2*pi**2/3d0
          weight = weight + alpha/2d0/pi*q(i)**2
     &                                   *(coll*born(0)+born(i)/2d0)
        endif
      enddo
      endif

      end

************************************************************************
*
*     integrated soft and coll. part (slicing) for real QCD corrections
*       call m2_hvv4f() before
*
************************************************************************
      subroutine sli_softcoll_qcd(virtmom,sij,weight)
      implicit none
      real*8 virtmom(0:3,5),sij(7,7)
      real*8 weight,preco(0:3,5)
      real*8 dE,func,coll,CF,T_dg(4,4),T_dg42(4,4),T_ndg(4,4)
      complex*16 cspen
      integer nw,i,j,i1,i2,ga,em,nparticles
      include 'COMMON/max.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/param.inc'
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
c m2born (diagonal and non-diagonal contribution)
      real*8 M2B_dg,M2B_dg42,M2B_ndg
      common/m2born/M2B_dg,M2B_dg42,M2B_ndg

      func(i,j) = 2d0*log(2d0*dE/lambda)*(1d0-log(sij(i+2,j+2)/mf(i)
     &     /mf(j))) -log(4d0*virtmom(0,i)*virtmom(0,j)/mf(i)/mf(j))
     &     +log(2d0*virtmom(0,i)/mf(i))**2+log(2d0*virtmom(0,j)
     &     /mf(j))**2 +pi**2/3d0 +cspen(dcmplx(1d0-4d0*virtmom(0,i)
     &     *virtmom(0,j)/sij(i+2,j+2)))

      CF = 4d0/3d0
      T_dg(1,2) = CF
      T_dg(3,4) = T_dg(1,2)
      T_dg(1,3) = 0d0
      T_dg(1,4) = T_dg(1,3)
      T_dg(2,3) = T_dg(1,3)
      T_dg(2,4) = T_dg(1,3)

      T_dg42(1,4) = CF
      T_dg42(2,3) = T_dg42(1,4)
      T_dg42(1,2) = 0d0
      T_dg42(1,3) = T_dg42(1,2)
      T_dg42(2,4) = T_dg42(1,2)
      T_dg42(3,4) = T_dg42(1,2)

      T_ndg(1,2) = CF
      T_ndg(1,3) = T_ndg(1,2)
      T_ndg(1,4) = T_ndg(1,2)
      T_ndg(2,3) = T_ndg(1,2)
      T_ndg(2,4) = T_ndg(1,2)
      T_ndg(3,4) = T_ndg(1,2)

      do i=2,4
      do j=1,i-1
        T_dg(i,j) = T_dg(j,i)
        T_dg42(i,j) = T_dg42(j,i)
        T_ndg(i,j) = T_ndg(j,i)
      enddo
      enddo

      dE=deltaE*dsqrt(CMenergy2)

c endpoint contribution (soft)
      do i=1,4
      do j=i+1,4
         if (quark(i).ne.0.and.quark(j).ne.0)
     &      weight = weight + alphas/pi*(-1d0)**(i+j)*func(i,j)
     &        *(  T_dg(i,j)*M2B_dg + T_dg42(i,j)*M2B_dg42
     &          + T_ndg(i,j)*M2B_ndg)
      enddo
      enddo

c collinear part, analytical integration over z (coll-safe observables)
      do i=1,4
        if (quark(i).ne.0) then
          coll = (3d0/2d0+2d0*log(dE/virtmom(0,i)))*(1d0-2d0
     &          *log(deltaTh*virtmom(0,i)/mf(i)))+3d0-2*pi**2/3d0
          weight = weight + CF*alphas/2d0/pi*coll
     &                      *(M2B_dg+M2B_dg42+M2B_ndg)
        endif
      enddo

      end

************************************************************************
*
*     collinear part in the slicing approach
*     (for non-collinear-safe observables, numerical integration over z)
*
************************************************************************
      subroutine sli_ncs(virtmom,distmom,z,born,sij,weight,nwabs)
      implicit none
      integer nwabs
      real*8 virtmom(0:3,5),distmom(0:3,5,4),z,born(0:4),sij(7,7)
      real*8 weight(4),preco(0:3,5)
      real*8 dE,zpr,zmax
      integer nw,i,j,i1,i2,ga,em,nparticles,nocut
      include 'COMMON/max.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/param.inc'
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej

      dE=deltaE*dsqrt(CMenergy2)

      nparticles = 5
      ga=5
      do em=1,4
        if(q(em).ne.0d0)then
          zmax=1d0-dE/virtmom(0,em)
          if(zmax.lt.0d0)zmax=0d0
          zpr=zmax*z           ! integration from 0 to zmax
          do i1=0,3
          do i2=1,4
            preco(i1,i2)=virtmom(i1,i2)
          enddo
          enddo
          do i1=0,3
            preco(i1,em)=virtmom(i1,em) *zpr
            preco(i1,ga)=virtmom(i1,em) *(1d0-zpr)
          enddo
          call cut(preco,distmom(0,1,em),nparticles,nocut,0)
          if(nocut.eq.0)then
            subrej(nwabs-1+em) = subrej(nwabs-1+em) + 1
          elseif(zpr.ne.0d0)then
            weight(em) = weight(em) + alpha/2d0/pi * q(em)**2
     &           *((2d0*log(deltaTh*virtmom(0,em)/mf(em)*zpr)-1d0)
     &               *(1d0+zpr**2)/(1d0-zpr)*born(0)
     &              +(1d0-zpr)*born(em) ) *zmax
          endif
        endif
      enddo


      end

************************************************************************
*
*     generate phase-space point or calculate weight
*
************************************************************************
      subroutine phasespace(random,ngen,inmom,outmom,distmom,sij,dens
     &     ,denstotal,nparticles,nocut,disc)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      integer ngen,nparticles,nocut
      real*8 random(18),inmom(0:3),outmom(0:3,5),distmom(0:3,5)
      real*8 dens(maxchan),denstotal
      integer step,disc,nchan,i,i1,i2
      real*8 sij(7,7),p(0:3,7)
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c adaptopt
      real*8 al(maxchan,maxopt,maxgen),be(0:maxchan,maxgen)
      real*8 wi(maxchan,maxgen),almin(maxgen)
      integer nopt(0:maxopt,maxgen),opt(maxgen)
      common/adaptopt/al,be,wi,almin,nopt,opt
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2

c define ficticious ingoing momenta p1 and p2 (p1+p2=inmom(Higgs))
      p(0,1) = (inmom(0)+inmom(3))/2d0
      p(0,2) = (inmom(0)-inmom(3))/2d0
      p(3,1) = p(0,1)            ! p1 and p2 are lightlike
      p(3,2) = -p(0,2)
      do i=1,2
        p(i,1) = 0d0
        p(i,2) = 0d0
      enddo

c choose channel (depending on values of adaptive optimization)
      do i = 1,nchannels(ngen)
        if(random(1).ge.be(i-1,ngen).and.random(1).le.be(i,ngen))then
          chan = i
        endif
      enddo

c generate momenta, input:random(2..9),ep,ip,nu,ngen; output: p
c   'dens(chan)': dummy variable
      step = 1
      disc = 0  ! do not discard event (if output of topology: disc=1: discard)

      if (topology(chan,ngen).eq.1) then
        call topology_1(random(2),ep(1,chan,ngen),ip(1,chan,ngen)
     &     ,emps(chan,ngen),spps(chan,ngen),nu(1,chan,ngen),p,sij
     &     ,dens(chan),ngen,step,disc)
      elseif (topology(chan,ngen).eq.2) then
        call topology_2(random(2),ep(1,chan,ngen),ip(1,chan,ngen)
     &       ,nu(1,chan,ngen),p,sij,dens(chan),ngen,step,disc)
      endif

c phase-space point for testing purpose
c      call test_phsp(p)
c      call dets(p,sij,4,disc)

c initialize photon momentum
      do i1 = 0,3
        outmom(i1,5) = 0d0
      enddo
c outgoing momenta, p defined for all momenta (incl. 2(!) incoming momenta)
      do i1 = 0,3
      do i2 = 1,nparticles
        outmom(i1,i2) = p(i1,i2+2)
      enddo
      enddo

c call cut/recombination routine
      call cut(outmom,distmom,nparticles,nocut,0) !nocut=0: apply cut, no gluon

c calculate densities (g_k) of all channels (g_tot)
      step = 2
      denstotal = 0d0
      do nchan = 1,nchannels(ngen) !'nchan' runs, 'chan' fixed
        dens(nchan) = 1d0
        if (topology(nchan,ngen).eq.1) then
          call topology_1(random(2),ep(1,nchan,ngen),ip(1,nchan,ngen)
     &     ,emps(nchan,ngen),spps(nchan,ngen),nu(1,nchan,ngen),p,sij
     &     ,dens(nchan),ngen,step,disc)
        elseif (topology(nchan,ngen).eq.2) then
          call topology_2(random(2),ep(1,nchan,ngen),ip(1,nchan,ngen)
     &     ,nu(1,nchan,ngen),p,sij,dens(nchan),ngen,step,disc)
        endif
        denstotal = denstotal + dens(nchan)*al(nchan,opt(ngen),ngen)
      enddo

c technical cut: mininmal invariant mass of fermion+photon
      if (nparticles.eq.5.and.qsoftcoll.eq.1) then
        do i = 1,4
          if (sij(i+2,7).lt.(mininvmass)**2*CMenergy2
     &          .or.outmom(0,5).lt.1d-6*dsqrt(CMenergy2) ) then
            nmininv=nmininv+1
            disc=1
            goto 100
          endif
        enddo
      endif

 100  continue

c technical cut for virtual corrections: energy of one particle
c smaller than 10^-6*CME
      do i = 1,4
        if (outmom(0,i).lt.1d-6*dsqrt(CMenergy2)) then
          nminen=nminen+1
          disc=1
          goto 200
        endif
      enddo

 200  continue

      end

************************************************************************
*
*
************************************************************************
      subroutine phsp_4from5(random,em,sp,ngen,inmom,outmom,distmom,sij4
     &     ,dens,denstotal,jac,nocut,disc)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      integer ngen,nparticles,nocut
      real*8 random(18),inmom(0:3),outmom(0:3,5),distmom(0:3,5)
      real*8 p5part(0:3,7),dens(maxchan),denstotal,jac
      integer step,disc,nchan,i,i1,i2,em,sp
      real*8 sij(7,7),sij4(7,7),p(0:3,7)
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c adaptopt
      real*8 al(maxchan,maxopt,maxgen),be(0:maxchan,maxgen)
      real*8 wi(maxchan,maxgen),almin(maxgen)
      integer nopt(0:maxopt,maxgen),opt(maxgen)
      common/adaptopt/al,be,wi,almin,nopt,opt

      nparticles = 5
c define ficticious ingoing momenta p1 and p2 (p1+p2=inmom(Higgs))
      p(0,1) = (inmom(0)+inmom(3))/2d0
      p(0,2) = (inmom(0)-inmom(3))/2d0
      p(3,1) = p(0,1)            ! p1 and p2 are lightlike
      p(3,2) = -p(0,2)
      do i=1,2
        p(i,1) = 0d0
        p(i,2) = 0d0
      enddo

c choose channel (depending on values of adaptive optimization)
      do i = 1,nchannels(ngen)
        if(random(1).ge.be(i-1,ngen).and.random(1).le.be(i,ngen))then
          chan = i
        endif
      enddo

c generate momenta, input:random(2..9),ep,ip,nu,ngen; output: p
c   'dens(chan)': dummy variable
      step = 1
      disc = 0  ! do not discard event (if output of topology: disc=1: discard)

      if (topology(chan,ngen).eq.1) then
        call topology_1(random(2),ep(1,chan,ngen),ip(1,chan,ngen)
     &     ,emps(chan,ngen),spps(chan,ngen),nu(1,chan,ngen),p,sij
     &     ,dens(chan),ngen,step,disc)
      elseif (topology(chan,ngen).eq.2) then
        call topology_2(random(2),ep(1,chan,ngen),ip(1,chan,ngen)
     &       ,nu(1,chan,ngen),p,sij,dens(chan),ngen,step,disc)
      endif

c calculate densities (g_k) of all channels (g_tot)
      step = 2
      denstotal = 0d0
      do nchan = 1,nchannels(ngen) !'nchan' runs, 'chan' fixed
        dens(nchan) = 1d0
        if (topology(nchan,ngen).eq.1) then
          call topology_1(random(2),ep(1,nchan,ngen),ip(1,nchan,ngen)
     &     ,emps(nchan,ngen),spps(nchan,ngen),nu(1,nchan,ngen),p,sij
     &     ,dens(nchan),ngen,step,disc)
        elseif (topology(nchan,ngen).eq.2) then
          call topology_2(random(2),ep(1,nchan,ngen),ip(1,nchan,ngen)
     &     ,nu(1,nchan,ngen),p,sij,dens(nchan),ngen,step,disc)
        endif
        denstotal = denstotal + dens(nchan)*al(nchan,opt(ngen),ngen)
      enddo

c map 5p-phsp into 4p-phsp
      jac = 1d0
      call map54(sij,sij4,p,p5part,jac,em,sp,2,disc)

c initialize photon momentum
      do i1 = 0,3
        outmom(i1,5) = 0d0
      enddo
c outgoing momenta, p defined for all momenta (incl. 2(!) incoming momenta)
      nparticles = 4
      do i1 = 0,3
      do i2 = 1,nparticles
        outmom(i1,i2) = p(i1,i2+2)
      enddo
      enddo

c call cut or recombination routine
      nparticles = 4
      call cut(outmom,distmom,nparticles,nocut,0) !nocut=0: apply cut, no gluon

      end

*******************************************************************
*
*     initialize phase-space integration
*
*******************************************************************
      subroutine init_integration
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      integer ngen,i1,i2,i,nocuts,generators
      real*8 nups
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c adaptopt
      real*8 al(maxchan,maxopt,maxgen),be(0:maxchan,maxgen)
      real*8 wi(maxchan,maxgen),almin(maxgen)
      integer nopt(0:maxopt,maxgen),opt(maxgen)
      common/adaptopt/al,be,wi,almin,nopt,opt
c numap
      real*8 numap
      common/numap/numap
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c unweighting
      integer unweighting,wflag(maxw)
      common/unweighting/unweighting,wflag

c output of discarded points
      numout = 0
      nmininv = 0
      nminen = 0
      nsub4 = 0
      nsub5 = 0
c technical cuts and parameters
      tequalzero=0d0          ! technical cut in subroutine process
c already defined in input file
c      massmapping=1d-4        ! parameter in h-function (mapping)
c      mininvmass=1d-4         ! cut for real corr.(min.*CMenergy)^2
      sub4cut = 1d-3          ! put events in the same bin for (1-z).lt.cut
      sub5cut = 1d-3          ! put events in the same bin for y and E_ga small
      sub5cutgl = 1d-3        ! put ev. in the same bin for y and E_gluon small
c value of nu in map54/map45
      numap=0.8d0             ! exponant in h(...) (mapping for massless prop.)
c initialize four-particle phase space generator
      ngen=1                  ! number of generator
      generators=1            ! generators=1: only 4-part. mappings
      if (contrib.eq.3) then
         nocuts=0             ! applying cuts for the generation of events
      else
         nocuts=1               ! not cuts for the generation of events
      endif
      nups=0.8d0              ! exponant in h(...) (mapping for massless prop.)
      call init_phasespace(ngen,generators,nocuts,nups)
c initialize five-particle phase space generator
      if (contrib.ne.2.and.contrib.ne.3) then
        ngen=2                ! number of generator
        if (qsoftcoll.eq.1.and.contrib.ne.3) then
          generators=4        ! generators=4: 5-part. and 4-part. mappings
        else
          generators=2        ! generators=2: 5-part. mappings only
        endif
        nups=0.8d0            ! exponant in h(...) (mapping for massless prop.)
        if(contrib.eq.4)then  ! contrib=4: Bremsstrahlung only
          nocuts=0            ! applying cuts for tree-level process H>4f+ga
        else
          nocuts=1            ! no cuts for subtraction functions possible
        endif
        call init_phasespace(ngen,generators,nocuts,nups)
      endif
c initialize generator for unweighted events, 4p-phsp from 5p-phsp
      ngen=4                  ! number of generator
      generators=3            ! generators=3: only 5->4 mappings
      nocuts=0                ! applying cuts for the generation of events
      nups=0.8d0              ! exponant in h(...) (mapping for massless prop.)
      call init_phasespace(ngen,generators,nocuts,nups)

c initialize adaptive optimization
      do i1=1,maxgen                ! i1: loop through generators
        do i2=0,maxopt         ! i2: loop through opt. steps
          nopt(i2,i1)=0
        enddo
        nopt(1,i1)=1000
        nopt(2,i1)=2000
        nopt(3,i1)=3000
        do i2=0,maxopt
          nopt(i2,i1)=nopt(i2,i1)*nchannels(i1)
        enddo
        opt(i1)=1
        be(0,i1)=0d0
        if(nchannels(i1).ne.0)then
          almin(i1)=0.2d0/nchannels(i1)
          do i2=1,nchannels(i1)    ! i2: loop through channels
            wi(i2,i1)=0d0
            al(i2,1,i1)=1d0/nchannels(i1)    ! start value for 1st opt. step
            be(i2,i1)=dble(i2)/nchannels(i1)
          enddo
        endif
      enddo
c initialize unweighting procedure
      do i = 1,maxw
         wflag(i) = 1
      enddo

      end

*******************************************************************
*
*     initialize phase-space generation
*
*******************************************************************
      subroutine init_phasespace(ngen,generators,nocuts,nups)
      implicit none
      integer ngen,generators,nocuts
      integer ep1,ep2,ep3,ep4,ep5,ep6,ep7,ip1,ip2,ip3,ep7min,ip3min
      integer map1,map2,map3,minmap1,minmap2,minmap3
      integer em,sp,nchan,i1,i2,topol,nparticles,exist
      integer topmin,topmax,emmax,spmax
      real*8 nups
      include 'COMMON/max.inc'
      include 'COMMON/processpar.inc'
c cuts
      real*8 ecut(5),scut(5,5),ccut(5,5),precomb(3)
      integer qcuts
      common/cuts/ecut,scut,ccut,precomb,qcuts
c cutsps
      real*8 ccutps(7,7,maxgen),scutps(7,7,maxgen)
      common/cutsps/ccutps,scutps
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width

c initializing cuts in generation of the event (cutsps:1..7)
c 'official' cuts (cuts) only defined for outgoing particles (1..5)
      do i1=1,7
      do i2=1,7
        ccutps(i1,i2,ngen)=1d0
        scutps(i1,i2,ngen)=0d0
      enddo
      enddo
      if(nocuts.eq.0)then
        do i1=1,5
        do i2=1,5
          ccutps(i1+2,i2+2,ngen)=min(ccut(i1,i2),ccut(i2,i1))
          scutps(i1+2,i2+2,ngen)=max(scut(i1,i2),scut(i2,i1),
     &      2d0*ecut(i1)*ecut(i2)*(1d0-ccutps(i1,i2,ngen)))
        enddo
        enddo
      endif

c generators=1:4part.phase space mappings, generators=2:5part.ph.sp. mappings
c generators=3:special mappings for 4pphsp embedded in 5pphsp,
c generators=4:combination of 2+3
      if (generators.eq.1) then
        topmin = 1
        topmax = 1
      elseif (generators.eq.2) then
        topmin = 2
        topmax = 2
      elseif (generators.eq.3) then
        topmin = 1
        topmax = 1
      elseif (generators.eq.4) then
        topmin = 1
        topmax = 2
      endif

c start finding channel number 1
      nchan = 1
c ep: external particles, ip: internal particles
      ep1 = 1
      ep2 = 2
c topology 1 for 4 part.phase space, top.2 for 5 part.ph.sp.
      do topol = topmin,topmax
      if (topol.eq.1) then
        nparticles = 4
        ep7min = 7    ! no photon in final state
        ip3min = 16   ! no third internal particle for top. 1
        if (generators.eq.3.or.generators.eq.4) then
          emmax = 6   !special 5->4 mappings included
          spmax = 6
        else
          emmax = 3   ! NO special 5->4 mappings
          spmax = 3
        endif
        minmap1 = 0   ! 0: both with and without mapping
        minmap2 = 0
        minmap3 = 1   ! only two propagators for top. 1
      elseif (topol.eq.2) then
        nparticles = 5
        ep7min = 3    ! ep7 (photon) varied through all final state particles
        ip3min = -13  ! ip3 varied through all possible particles
        emmax = 3     ! NO special 5->4 mappings for topol.eq.2 (of course not)
        spmax = 3
        minmap1 = 0   ! both with and without mapping
        minmap2 = 0
        minmap3 = 0   ! 3 propagators for top. 2
      endif

      do ep3 = 3,nparticles+2  ! vary ep3,... through all final state particles
      do ep4 = 3,nparticles+2
      if (ep3.ne.ep4) then
      do ep5 = 3,nparticles+2
      if (ep5.ne.ep3.and.ep5.ne.ep4) then
      do ep6 = 3,nparticles+2
      if (ep6.ne.ep3.and.ep6.ne.ep4.and.ep6.ne.ep5) then
      do ep7 = ep7min,7       ! vary ep7 (photon) through 3..7 or set ep=7
      if (ep7.ne.ep3.and.ep7.ne.ep4.and.ep7.ne.ep5.and.ep7.ne.ep6) then
      do ip1 = -13,16         ! vary internal particles ip1 etc.
      do ip2 = -13,16
      do ip3 = ip3min,16

        if (topol.eq.1) then
          call check_topology_1(ep1,ep2,ep3,ep4,ep5,ep6,ip1,ip2,exist)
        elseif (topol.eq.2) then
          call check_topology_2(ep1,ep2,ep3,ep4,ep5,ep6,ep7,ip1,ip2,ip3
     &         ,exist)
        endif
        if (exist.eq.1)then
          do em = 3,emmax  !special 5->4 mappings for em.ne.3 and sp.ne.3
          do sp = 3,spmax
          if ((em.eq.3.and.sp.eq.3.and.emmax.eq.3.and.spmax.eq.3).or.
     &        (em.ne.sp.and.q(em-2).ne.0d0.and.q(sp-2).ne.0d0)) then
          do map1 = minmap1,1
          if (map1.eq.1.or.mass(ip1).ne.0d0) then !massive prop:both map&nomap
          do map2 = minmap2,1
          if (map2.eq.1.or.mass(ip2).ne.0d0) then
          do map3 = minmap3,1
          if (map3.eq.1.or.mass(ip3).ne.0d0) then
            ep(1,nchan,ngen) = ep1
            ep(2,nchan,ngen) = ep2
            ep(3,nchan,ngen) = ep3
            ep(4,nchan,ngen) = ep4
            ep(5,nchan,ngen) = ep5
            ep(6,nchan,ngen) = ep6
            ep(7,nchan,ngen) = ep7
            ip(1,nchan,ngen) = ip1
            ip(2,nchan,ngen) = ip2
            ip(3,nchan,ngen) = ip3
            nu(1,nchan,ngen) = map1 * nups  !mapping for propagator
            nu(2,nchan,ngen) = map2 * nups  !eq.0 or ne.0 for massive prop.
            nu(3,nchan,ngen) = map3 * nups  !=nups for massless prop.
            if (em.eq.3.and.sp.eq.3) then
              emps(nchan,ngen) = 0
              spps(nchan,ngen) = 0
            else
              emps(nchan,ngen) = em
              spps(nchan,ngen) = sp
            endif
            topology(nchan,ngen)  = topol
            nchan = nchan + 1
          endif
          enddo
          endif
          enddo
          endif
          enddo
          endif
          enddo
          enddo
        endif
      enddo
      enddo
      enddo
      endif
      enddo
      endif
      enddo
      endif
      enddo
      endif
      enddo
      enddo
      enddo
      nchannels(ngen) = nchan - 1   ! total number of channels

      end

************************************************************************
*
*     output of multichannel integration
*
************************************************************************
      subroutine write_multichannel
      implicit none
      integer out,ngen,nchan,length
      include 'COMMON/max.inc'
      include 'COMMON/processpar.inc'
c channels
      real*8 nu(3,maxchan,maxgen)
      integer ep(7,maxchan,maxgen),ip(3,maxchan,maxgen)
      integer topology(maxchan,maxgen),emps(maxchan,maxgen)
      integer spps(maxchan,maxgen),nchannels(maxgen)
      common/channels/nu,ep,ip,topology,emps,spps,nchannels
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width
c adaptopt
      real*8 al(maxchan,maxopt,maxgen),be(0:maxchan,maxgen)
      real*8 wi(maxchan,maxgen),almin(maxgen)
      integer nopt(0:maxopt,maxgen),opt(maxgen)
      common/adaptopt/al,be,wi,almin,nopt,opt
c outputfile
      character*90 outputfile
      common/outputfile/outputfile

      out = 7
      open(unit=out,file='multich.dat',status='unknown')
      write(out,*)'  Process:  H ->   '
     &     ,fname(1)(1:length(fname(1))),'(3) '
     &     ,fname(2)(1:length(fname(2))),'(4) '
     &     ,fname(3)(1:length(fname(3))),'(5) '
     &     ,fname(4)(1:length(fname(4))),'(6) '
     &     ,'(+Photon)','(7)'
      write(out,*)
      do ngen = 1,maxgen
        write(out,*)'Phase-space Generator',ngen
        write(out,*)'-----------------------'
        write(out,*)'channel, topol, ext.part(ep), '
     &       ,'virt.part(ip), map1, map2, map3, emit, spect'
        do nchan = 1,nchannels(ngen)
          if (topology(nchan,ngen).eq.1.and.emps(nchan,ngen).eq.0) then
            write(out,'(2i4,3x,4i4,8x,2a8,2f6.2)')
     &         nchan,topology(nchan,ngen)
     &         ,ep(3,nchan,ngen),ep(4,nchan,ngen),ep(5,nchan,ngen)
     &         ,ep(6,nchan,ngen)
     &         ,gname(ip(1,nchan,ngen)),gname(ip(2,nchan,ngen))
     &         ,nu(1,nchan,ngen),nu(2,nchan,ngen)
          elseif (topology(nchan,ngen).eq.2) then
            write(out,'(2i4,3x,5i4,4x,3a8,3f6.2)')
     &         nchan,topology(nchan,ngen)
     &         ,ep(3,nchan,ngen),ep(4,nchan,ngen),ep(5,nchan,ngen)
     &         ,ep(6,nchan,ngen),ep(7,nchan,ngen)
     &         ,gname(ip(1,nchan,ngen)),gname(ip(2,nchan,ngen))
     &         ,gname(ip(3,nchan,ngen))
     &         ,nu(1,nchan,ngen),nu(2,nchan,ngen),nu(3,nchan,ngen)
          else
            write(out,'(2i4,3x,4i4,8x,2a8,2f6.2,2i4)')
     &         nchan,topology(nchan,ngen)
     &         ,ep(3,nchan,ngen),ep(4,nchan,ngen),ep(5,nchan,ngen)
     &         ,ep(6,nchan,ngen)
     &         ,gname(ip(1,nchan,ngen)),gname(ip(2,nchan,ngen))
     &         ,nu(1,nchan,ngen),nu(2,nchan,ngen)
     &         ,emps(nchan,ngen),spps(nchan,ngen)
          endif
        enddo
        write(out,*)
      enddo
c adaptive optimization
      write(out,*)
      write(out,'(a)')' Adaptive optimization:'
      write(out,'(a)')' ----------------------'
      do ngen=1,maxgen
        write(out,'(a)')' '
        write(out,*)' Phase-space generator',ngen
        write(out,*)' ----------------------'
        write(out,'(a)')'  Optimization steps:'
        write(out,'(8i8)')nopt(1,ngen),nopt(2,ngen),nopt(3,ngen),
     &    nopt(4,ngen),nopt(5,ngen),nopt(6,ngen),nopt(7,ngen)
     &       ,nopt(8,ngen)
        write(out,'(a)') '  Channels, a priori weights in % '
        do nchan=1,nchannels(ngen)
          write(out,'(" ",i4,",",8f8.3)')nchan
     &      ,1d-3*nint(1d5*al(nchan,1,ngen))
     &      ,1d-3*nint(1d5*al(nchan,2,ngen))
     &      ,1d-3*nint(1d5*al(nchan,3,ngen))
     &      ,1d-3*nint(1d5*al(nchan,4,ngen))
     &      ,1d-3*nint(1d5*al(nchan,5,ngen))
     &      ,1d-3*nint(1d5*al(nchan,6,ngen))
     &      ,1d-3*nint(1d5*al(nchan,7,ngen))
     &      ,1d-3*nint(1d5*al(nchan,8,ngen))
        enddo
      enddo
      close(out)

      end

************************************************************************
*
*                                       derived from:
*                     xx ep3                                xx ep3
*                   xx                              ip1   xx
*                 xx                 ep1 xxxxxxxxxxxxxxxxx
*          ip1  xx  xx                            x       xx
*             xx      xx ep4                      x         xx ep4
*           xx                                    x
*   -------x                                      x ip3
*           xx                                    x
*             xx      xx ep5                      x         xx ep5
*          ip2  xx  xx                            x       xx
*                 xx                 ep2 xxxxxxxxxxxxxxxxx
*                   xx                              ip2   xx
*                     xx ep6                                xx ep6
*
*     step=1: generate momenta and invariants
*     step=2: calculate local density (=g) of given channel
*
*     no different values of 'nu' necessary since always nu=0
*     in subroutine 'process'
************************************************************************
      subroutine topology_1(random,ep,ip,em,sp,nu,p,sij,dens,ngen
     &     ,step,disc)
      implicit none
      include 'COMMON/max.inc'
      integer ep(7),ip(3),em,sp,ngen,step,disc,nparticles
      real*8 random(11),nu(3),p(0:3,7),p5part(0:3,7),sij(7,7),dens
      real*8 s1,s2,t3,q1(0:3),q2(0:3),qt3(0:3),roots,sij4(7,7)
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width
c cutsps
      real*8 ccutps(7,7,maxgen),scutps(7,7,maxgen)
      common/cutsps/ccutps,scutps

      roots = dsqrt(2d0*(p(0,1)*p(0,2)-p(3,1)*p(3,2)))

      if (step.eq.2) then
        if (em.eq.0) then
          s1=sij(ep(3),ep(4))
          s2=sij(ep(5),ep(6))
          t3=sij(ep(3),ep(4))-sij(ep(1),ep(3))-sij(ep(1),ep(4))
        else                  !special mappings for 5-particle phase space
          call map54(sij,sij4,p,p5part,dens,em,sp,step,disc)
          s1=sij4(ep(3),ep(4))
          s2=sij4(ep(5),ep(6))
          t3=sij4(ep(3),ep(4))-sij4(ep(1),ep(3))-sij4(ep(1),ep(4))
        endif
      endif

      call sprop(random(1),s1,dens,mass(ip(1)),width(ip(1)),nu(1)
     &     ,scutps(ep(3),ep(4),ngen)
     &     ,(roots-dsqrt(scutps(ep(5),ep(6),ngen)))**2,step,disc)
      call sprop(random(2),s2,dens,mass(ip(2)),width(ip(2)),nu(2)
     &     ,scutps(ep(5),ep(6),ngen),(roots-dsqrt(s1))**2,step,disc)
      call process(random(3),p(0,ep(2)),p(0,ep(1)),q2,q1,qt3,dens
     &     ,roots**2,s2,s1,t3,step,disc)
      call decay(random(5),q1,p(0,ep(3)),p(0,ep(4)),dens,s1,0d0,0d0,step
     &     ,disc)
      call decay(random(7),q2,p(0,ep(5)),p(0,ep(6)),dens,s2,0d0,0d0,step
     &     ,disc)

      if (em.ne.0) call map45(random(9),sij4,p,p5part,em,sp,step,disc)

      if (step.eq.1) then
        if (em.eq.0) nparticles=4
        if (em.ne.0) nparticles=5
        call dets(p,sij,nparticles,disc)
      endif

      end

*******************************************************************
*
*                                       derived from:
*
*                     xx ep3                                  xx ep3
*                   xx                            ip1   ip2 xx
*             ip2 xx               ep1 xxxxxxxxxxxxxxxxxxxxx
*               xx  xx                          x     xx    xx
*             xx      xx ep4                    x       xx    xx ep4
*       ip1 xx  xx                              x         xx
*         xx       xx                           x           xx
*   -----x           xx ep5                     x ip4         xx ep5
*         xx                                    x
*           xx                                  x
*             xx      xx ep6                    x             xx ep6
*          ip3  xx  xx                          x           xx
*                 xx               ep2 xxxxxxxxxxxxxxxxxxxxx
*                   xx                              ip3     xx
*                     xx ep7                                  xx ep7
*
*     step=1: generate momenta and invariants
*     step=2: calculate local density (=g) of given channel
*
*     no different values of 'nu' necessary since always nu=0
*     in subroutine 'process'
*******************************************************************
      subroutine topology_2(random,ep,ip,nu,p,sij,dens,ngen,step,disc)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/output.inc'
      integer ep(7),ip(3),ngen,step,disc,nparticles
      real*8 random(11),nu(3),p(0:3,7),sij(7,7),dens
      real*8 s1,s2,s3,t4,q1(0:3),q2(0:3),q3(0:3),qt4(0:3),roots
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width
c cutsps
      real*8 ccutps(7,7,maxgen),scutps(7,7,maxgen)
      common/cutsps/ccutps,scutps

      roots = dsqrt(2d0*(p(0,1)*p(0,2)-p(3,1)*p(3,2)))
      if(step.eq.2)then
        s1=sij(ep(3),ep(4))+sij(ep(3),ep(5))+sij(ep(4),ep(5))
        s2=sij(ep(3),ep(4))
        s3=sij(ep(6),ep(7))
        t4=sij(ep(6),ep(7))-sij(ep(2),ep(6))-sij(ep(2),ep(7))
      endif

      call sprop(random(1),s2,dens,mass(ip(2)),width(ip(2)),nu(1)
     &     ,scutps(ep(3),ep(4),ngen),(roots-dsqrt(scutps(ep(6),ep(7),
     &     ngen)))**2-scutps(ep(3),ep(5),ngen)-scutps(ep(4),ep(5),ngen)
     &     ,step,disc)
      call sprop(random(2),s1,dens,mass(ip(1)),width(ip(1)),nu(2)
     &     ,s2+scutps(ep(3),ep(5),ngen)+scutps(ep(4),ep(5),ngen),
     &     (roots-dsqrt(scutps(ep(6),ep(7),ngen)))**2,step,disc)
      call sprop(random(3),s3,dens,mass(ip(3)),width(ip(3)),nu(3)
     &     ,scutps(ep(6),ep(7),ngen),(roots-dsqrt(s1))**2,step,disc)
      call process(random(4),p(0,ep(2)),p(0,ep(1)),q3,q1,qt4,dens
     &     ,roots**2,s3,s1,t4,step,disc)
      call decay(random(6),q1,q2,p(0,ep(5)),dens,s1,s2,0d0,step,disc)
      call decay(random(8),q2,p(0,ep(3)),p(0,ep(4)),dens,s2,0d0,0d0,step
     &     ,disc)
      call decay(random(10),q3,p(0,ep(6)),p(0,ep(7)),dens,s3,0d0,0d0
     &     ,step,disc)

      if (step.eq.1.and.disc.ne.1) then
        nparticles=5
        call dets(p,sij,nparticles,disc)
      endif

      end

*******************************************************************
*
*     check if topology 1 exist
*
*       ep1 = Higgs
*       ep3,..,ep7 defined as outgoing, -ep3,..,-ep7 incoming
*       vertex v3() defined for incoming particles
*       convention for ip1,..ip3: going from left to right
*******************************************************************
      subroutine check_topology_1(ep1,ep2,ep3,ep4,ep5,ep6,ip1,ip2,exist)
      implicit none
      integer ep1,ep2,ep3,ep4,ep5,ep6,ip1,ip2,exist
      logical v3
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width

      exist=0
c symmetry (do not count symmetric channels twice)
      if(ep3.gt.ep4)return
      if(ep5.gt.ep6)return
      if(ep3.gt.ep5)return
c check if vertices exist
      if(.not.v3(ename(ep1),gname(-ip1),gname(-ip2)))return
      if(.not.v3(gname(ip1),ename(-ep3),ename(-ep4)))return
      if(.not.v3(gname(ip2),ename(-ep5),ename(-ep6)))return

      exist=1

      end

*******************************************************************
*
*     check if topology 2 exist
*
*******************************************************************
      subroutine check_topology_2(ep1,ep2,ep3,ep4,ep5,ep6,ep7
     &     ,ip1,ip2,ip3,exist)
      implicit none
      integer ep1,ep2,ep3,ep4,ep5,ep6,ep7,ip1,ip2,ip3,exist
      logical v3
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width

      exist=0
c symmetry (do not count symmetric channels twice)
      if(ep3.gt.ep4)return
      if(ep6.gt.ep7)return
c check if vertices exist
      if(.not.v3(ename(ep1),gname(-ip1),gname(-ip3)))return
      if(.not.v3(gname(ip1),gname(-ip2),ename(-ep5)))return
      if(.not.v3(gname(ip2),ename(-ep3),ename(-ep4)))return
      if(.not.v3(gname(ip3),ename(-ep6),ename(-ep7)))return

      exist=1


      end

*******************************************************************
*
*     s-channel propagator
*
*******************************************************************
      subroutine sprop(random,s,g,mass,width,nu,smin,smax,step,disc)
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/output.inc'
      real*8 random,s,g,mass,mass2,width,width2,nu
      real*8 smax,smin,omax,omin,h,jacobian,denom
      integer step,disc

      if(disc.eq.1)return
      mass2=mass*mass
      if(smax.le.smin)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a22,i5,i10,2d10.2)')' props: smin >= smax, '
     &          ,chan,nev,smin,smax
          numout=numout+1
        endif
        disc=1
        return
      endif
      if(smin.lt.0d0)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a18,i5,i10,d10.2)')' props: smin < 0, '
     &          ,chan,nev,smin
          numout=numout+1
        endif
        disc=1
        return
      endif
      if(width.gt.0d0)then
        width2=width*width
        omax=datan((smax-mass2)/mass/width)
        omin=datan((smin-mass2)/mass/width)
        if(step.eq.1)then
          if (nu.eq.0d0) then               ! no mapping applied
            s = smin + random*(smax-smin)
          else
            s=mass2+mass*width*dtan(random*(omax-omin)+omin)
          endif
        elseif(step.eq.2)then
          if (nu.eq.0d0) then               !no mapping applied
            denom = (smax-smin)
          else
            denom=(omax-omin)*((s-mass2)**2+mass2*width2)
            g=g*mass*width
          endif
          if(denom.gt.0d0)then
            g=g/denom
          else
            if(numout.lt.maxout)then
              write(nout,'(6x,a34,i5,i10,d10.2)')
     &          ' props: 1/density <= 0 (width>0), ',chan,nev,denom
              numout=numout+1
            endif
            disc=1
            return
          endif
        endif
      else
        if(step.eq.1)then
          s=h(random,nu,smin,smax,disc)
        elseif(step.eq.2)then
          denom=jacobian(nu,s,smin,smax,disc)
          if(denom.gt.0d0)then
            g=g/denom
          else
            if(nout.lt.maxout)then
              write(nout,'(6x,a34,i5,i10,d10.2)')
     *          ' props: 1/density <= 0 (width=0), ',chan,nev,denom
            endif
            numout=numout+1
            disc=1
            return
          endif
        endif
      endif
      end

*******************************************************************
*
*     2->2 particle process with t-channel propagator
*        q2 + q1 -> p2 + p1
*        qt = (q1 - p1) , s0 = (q2+q1)^2 , s1 =
*        adapted to 1->2 process (= no mapping: nu=0d0)
*                   (q2+q1) -> p2 + p1
*        step=1: generate momenta
*        step=2: calculate contribution to local density
*******************************************************************
      subroutine process(random,q1,q2,p1,p2,qt,g,s0,s1,s2,t,step,disc)
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      real*8 random(2),q1(0:3),q2(0:3),p1(0:3),p2(0:3),qt(0:3)
      real*8 g,mass,width,nu,s1,s2,t1,t2,tmin,tmax,ccut,qvec,q0
      real*8 lambdas,lambdat,mass2,t,phi,cost,s0,roots0
      real*8 omega,omin,omax,k1(0:3),q(0:3),h,jacobian,denom,cmax
      integer i1,step,disc

      if(disc.eq.1)return
c special to 1->2 process
      t1=0d0
      t2=0d0
      cmax=1d0
      nu=0d0
      mass=0d0
      width=0d0

      mass2=mass*mass
      lambdas=(s0-s1-s2)*(s0-s1-s2)-4d0*s1*s2
      if(lambdas.gt.0d0)then
        lambdas=dsqrt(lambdas)
      else
        if(numout.lt.maxout)then
          write(nout,'(6x,a24,i5,i10,d10.2)')' process: lambdas <= 0, '
     &          ,chan,nev,lambdas
          numout=numout+1
        endif
        disc=1
        return
      endif
      lambdat=(s0-t1-t2)*(s0-t1-t2)-4d0*t1*t2
      if(lambdat.gt.0d0)then
        lambdat=dsqrt(lambdat)
      else
        if(numout.lt.maxout)then
          write(nout,'(6x,a24,i5,i10,d10.2)')' process: lambdat <= 0, '
     &          ,chan,nev,lambdat
          numout=numout+1
        endif
        disc=1
        return
      endif
      tmin=s1+t1-((s0+s1-s2)*(s0+t1-t2)+lambdas*lambdat)/2d0/s0
      tmax=s1+t1-((s0+s1-s2)*(s0+t1-t2)-lambdas*lambdat*cmax)/2d0/s0
      if(dabs(tmax).lt.tequalzero)tmax=0d0
      if(tmax.le.tmin)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a24,i5,i10,2d10.2)')' process: tmin >= tmax, '
     &          ,chan,nev,tmin,tmax
          numout=numout+1
        endif
        disc=1
        return
      endif
      if(tmax.gt.0d0)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a20,i5,i10,d10.2)')' process: tmax > 0, '
     &          ,chan,nev,tmax
          numout=numout+1
        endif
        disc=1
        return
      endif
c generate momenta
      if (step.eq.1) then
        phi=2d0*pi*random(1)
        denom=lambdas*lambdat
        cost=((s0+s1-s2)*(s0+t1-t2)-2d0*s0*(t1+s1)
     *        -2d0*s0*h(random(2),nu,-tmax,-tmin,disc)
     *       )/lambdas/lambdat
        roots0=dsqrt(s0)
        p1(0)=(s0+s1-s2)*0.5d0/roots0
        p1(1)=0d0
        p1(2)=0d0
        p1(3)=lambdas*0.5d0/roots0
        phi=dsign(phi,q1(3))
        call rotation(p1,phi,cost,disc)
        do i1=0,3
          q(i1)=q1(i1)+q2(i1)
          k1(i1)=q1(i1)
        enddo
        call boost(q,k1,1d0,disc)
        if(k1(1).eq.0d0)then
          phi=dsign(pi*0.5d0,k1(2))
        else
          phi=datan(k1(2)/k1(1))
        endif
        if(k1(1).lt.0d0)phi=phi+pi
        cost=k1(3)/dsqrt(k1(1)*k1(1)+k1(2)*k1(2)+k1(3)*k1(3))
        call rotation(p1,-phi,cost,disc)
        call boost(q,p1,-1d0,disc)
        do i1=0,3
          p2(i1)=q(i1)-p1(i1)
          qt(i1)=q1(i1)-p1(i1)
        enddo
        t=qt(0)*qt(0)-qt(1)*qt(1)-qt(2)*qt(2)-qt(3)*qt(3)
c calculate contribution to local density
      elseif(step.eq.2)then
        denom=pi*jacobian(nu,mass2-t,-tmax,-tmin,disc)
        if(denom.ne.0d0)then
          g=g*2d0*lambdat/denom
        else
          if(numout.lt.maxout)then
            write(nout,'(6x,a35,i5,i10)')
     *        ' process: 1/density = 0 (width=0), ',chan,nev
          endif
          numout=numout+1
          disc=1
          return
        endif
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     1->2 particle decay without propagator                     c
c                                                                c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine decay(random,q,p1,p2,g,s0,s1,s2,step,disc)
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/output.inc'
      real*8 random(2),q(0:3),p1(0:3),p2(0:3)
      real*8 phi,cost,s0,roots0,lambda,denom,g,s1,s2
      integer i1,step,disc

      if(disc.eq.1)return
      lambda=(s0-s1-s2)*(s0-s1-s2)-4d0*s1*s2
      if(lambda.gt.0d0)then
        lambda=dsqrt(lambda)
      else
        if(numout.lt.maxout)then
          write(nout,'(6x,a21,i5,i10,d10.2)')' decay: lambda <= 0, '
     &          ,chan,nev,lambda
          numout=numout+1
        endif
        disc=1
        return
      endif
      if(step.eq.1)then
        roots0=dsqrt(s0)
        phi=2d0*pi*random(1)
        cost=2d0*random(2)-1d0
        p1(0)=(s0+s1-s2)*0.5d0/roots0
        p1(1)=0d0
        p1(2)=0d0
        p1(3)=lambda*0.5d0/roots0
        call rotation(p1,phi,cost,disc)
        call boost(q,p1,-1d0,disc)
        do i1=0,3
          p2(i1)=q(i1)-p1(i1)
        enddo
      elseif(step.eq.2)then
        denom=pi*lambda
        if(denom.gt.0d0)then
          g=g*2d0*s0/denom
        else
          if(numout.lt.maxout)then
            write(nout,'(6x,a22,i5,i10,d10.2)')' dec: 1/density <= 0, '
     &            ,chan,nev,denom
            numout=numout+1
          endif
          disc=1
          return
        endif
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     h function for importance sampling                         c
c
c      for propagators with vanishing width
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function h(random,nu,xmin,xmax,disc)
      implicit none
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      real*8 h,random,nu,xmin,xmax
      integer disc

      h=0d0
      if(disc.eq.1)return
      if(massmapping+xmax.lt.0d0.or.massmapping+xmin.lt.0d0)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a24,i5,i10,2d10.2)')
     &          ' h: massmapping+xmin, massmapping+xmax < 0, '
     &          ,chan,nev,massmapping+xmin,massmapping+xmax
          numout=numout+1
        endif
        disc=1
        return
      endif
      if(nu.eq.0d0)then
        h=random*xmax+(1d0-random)*xmin
      elseif(nu.eq.1d0)then
        h=dexp(random*dlog(massmapping+xmax)+(1d0-random)
     &        *dlog(massmapping+xmin))-massmapping
      else
        h=(random*(massmapping+xmax)**(1d0-nu)+(1d0-random)
     &       *(massmapping+xmin)**(1d0-nu))**(1d0/(1d0-nu))-massmapping
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     jacobian of the h function                                 c
c                                                                c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function jacobian(nu,x,xmin,xmax,disc)
      implicit none
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      real*8 jacobian,nu,x,xmin,xmax
      integer disc

      jacobian=0d0
      if(disc.eq.1)return
      if(massmapping+xmax.lt.0d0.or.massmapping+xmin.lt.0d0)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a30,i5,i10,2d10.2)')
     &      ' jacobian: massmapping+xmin, massmapping+xmax < 0, '
     &      ,chan,nev,massmapping+xmin,massmapping+xmax
          numout=numout+1
        endif
        disc=1
        return
      endif
      if(nu.eq.0)then
        jacobian=xmax-xmin
      elseif(nu.eq.1d0)then
        jacobian=(dlog(massmapping+xmax)-dlog(massmapping+xmin))
     &        *(massmapping+x)
      else
        jacobian=(((massmapping+xmax)**(1d0-nu)-(massmapping+xmin)
     &           **(1d0-nu))/(1d0-nu))*(massmapping+x)**nu
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     rotation                                                   c
c                                                                c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine rotation(p,phi,cost,disc)
      implicit none
      include 'COMMON/output.inc'
      real*8 p(0:3),phi,cost,sint,cosp,sinp,px,py,pz
      integer disc

      if(disc.eq.1)return
      cosp=dcos(phi)
      sinp=dsin(phi)
      if(dabs(cost).gt.1d0)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a38,i5,i10,d10.2)')
     *      ' rotation: |cost| > 1, event ignored, ',chan,nev,dabs(cost)
          numout=numout+1
        endif
        disc=1
        return
      endif
      sint=dsqrt((1d0-cost)*(1d0+cost))
      px=p(1)
      py=p(2)
      pz=p(3)
      p(1)=(pz*sint+px*cost)*cosp+py*sinp
      p(2)=-(pz*sint+px*cost)*sinp+py*cosp
      p(3)=pz*cost-px*sint
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     boost                                                      c
c                                                                c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine boost(q,p,dir,disc)
      implicit none
      include 'COMMON/output.inc'
      real*8 p(0:3),q(0:3),dir,m,m2,bx,by,bz,gamma
      real*8 p0,px,py,pz,bp,a
      integer disc

      if(disc.eq.1)return
      m2=q(0)*q(0)-q(1)*q(1)-q(2)*q(2)-q(3)*q(3)
      if(m2.le.0d0)then
        if(numout.lt.maxout)then
          write(nout,'(6x,a17,i5,i10,d10.2)')' boost: m2 <= 0, '
     &          ,chan,nev,m2
          numout=numout+1
        endif
        disc=1
        return
      endif
      m=dsqrt(m2)
      bx=-q(1)/m*dir
      by=-q(2)/m*dir
      bz=-q(3)/m*dir
      gamma=q(0)/m
      a=1d0/(1d0+gamma)
      p0=p(0)
      px=p(1)
      py=p(2)
      pz=p(3)
      bp=bx*px+by*py+bz*pz
      p(0)=gamma*p0+bp
      p(1)=px+bx*p0+a*bp*bx
      p(2)=py+by*p0+a*bp*by
      p(3)=pz+bz*p0+a*bp*bz
      end

*******************************************************************
*
*     calculation of invariants
*
*******************************************************************
      subroutine dets(p,sij,nparticles,disc)
      implicit none
      include 'COMMON/output.inc'
c local variables
      real*8 p(0:3,7),sij(7,7),q(0:3)
      integer i1,i2,i3,nparticles,disc

      if(disc.eq.1)return
      do i1=1,nparticles+2
      do i2=i1+1,nparticles+2
        do i3=0,3
          q(i3)=p(i3,i1)+p(i3,i2)
        enddo
        sij(i1,i2)=q(0)*q(0)-q(1)*q(1)-q(2)*q(2)-q(3)*q(3)
        if(sij(i1,i2).lt.0d0)then
          if(numout.lt.maxout)then
            write(nout,
     &        '("      dets: wrong sign of sij(",i1,","i2,")",i5,i9)')
     &        i1,i2,chan,nev
            numout=numout+1
          endif
          disc=1
          return
        endif
        sij(i2,i1)=sij(i1,i2)
      enddo
      enddo
      end

************************************************************************
*
*     mapping 5-particle phase space into 4-particle phase space
*        for step=2:
*          map p(5-part.phsp.) to p(4-part.phsp.),
*          store original 5-part.phsp. momenta in p5part,
*          generate invariants of 4-part.phsp. sij4
************************************************************************
      subroutine map54(sij,sij4,p,p5part,g,em,sp,step,disc)
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/output.inc'
      real*8 k(0:3,7),g,h,denum,jacobian
      real*8 z,y,v,p(0:3,7),p5part(0:3,7),sij(7,7),sij4(7,7)
      real*8 ck(0:3),ckt(0:3)
      integer nparticles,em,sp,ga,j,step,disc,i1,i2
c numap
      real*8 numap
      common/numap/numap

      if(disc.eq.1)return
      if(step.eq.2)then
        do i1=1,7
        do i2=0,3
          p5part(i2,i1)=p(i2,i1)  !save original 5-part. momenta
        enddo
        enddo
        ga=7
        z=sij(em,sp)/(sij(em,sp)+sij(sp,ga))
        y=sij(em,ga)/(sij(em,sp)+sij(em,ga)+sij(sp,ga))
        do i1=0,3
          p(i1,em)=p5part(i1,em)+p5part(i1,ga)-y/(1d0-y)*p5part(i1,sp)
          p(i1,sp)=1d0/(1d0-y)*p5part(i1,sp)
          p(i1,7)=0d0
        enddo
        denum=pi*(1d0-y)*jacobian(numap,1d0-y,0d0,1d0,disc)
     &                  *jacobian(numap,1d0-z,0d0,1d0,disc)

        nparticles=4
        call dets(p,sij4,nparticles,disc)
        if(denum.ge.0d0)then
          g=g*2d0/denum/sij4(em,sp)
        else
          if(numout.lt.maxout)then
            write(nout,'(a21,i5,i9)')' map54: 1/density <= 0',chan,nev
            numout=numout+1
          endif
          disc=1
          return
        endif
      endif

      end

************************************************************************
*
*     mapping 4-particle phase space into 5-particle phase space
*       step=1:
*         insert photon into 4-part.phsp., p(4-p.phsp.)->p(5-p.phsp.)
*       step=2:
*         restore p(5-part.phsp.) from p5part
************************************************************************
      subroutine map45(random,sij4,p,p5part,em,sp,step,disc)
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/output.inc'
      real*8 random(3),p(0:3,7),p5part(0:3,7),h,f,c,kt(0:3)
      real*8 z,y,v,k(0:3,7),sij4(7,7),ck(0:3),ckt(0:3)
      integer nparticles,em,sp,ga,j,i1,i2,step,disc
c numap
      real*8 numap
      common/numap/numap

      if(disc.eq.1)return
c setting momenta
      if(step.eq.1)then
        do i1=0,3
          p(i1,7)=0d0
          do i2=1,7
            k(i1,i2)=p(i1,i2)
          enddo
        enddo
        ga=7
        nparticles=4
        call dets(k,sij4,nparticles,disc)
        y=1d0-h(random(1),numap,0d0,1d0,disc)
        z=1d0-h(random(2),numap,0d0,1d0,disc)
        f=dsqrt(y*z/(1d0-z))
        kt(0)=-f*f*dsqrt(sij4(em,sp))
        kt(1)=f*dsqrt(sij4(em,sp))*dcos(2d0*pi*random(3))
        kt(2)=f*dsqrt(sij4(em,sp))*dsin(2d0*pi*random(3))
        kt(3)=-f*f*dsqrt(sij4(em,sp))
        call detkt(k,kt,sp,em,step,disc)
        do i1=0,3
          p(i1,em)=y*(1d0+z)*k(i1,sp)+z*k(i1,em)+(1d0-z)*kt(i1)
          p(i1,ga)=(1d0-z)*k(i1,em)-y*z*k(i1,sp)-(1d0-z)*kt(i1)
          p(i1,sp)=(1d0-y)*k(i1,sp)
        enddo
c restore original 5-particle momenta
      elseif(step.eq.2)then
        do i1=0,3
        do i2=1,7
          p(i1,i2)=p5part(i1,i2)
        enddo
        enddo
      endif
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     determination of kt                                         c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine detkt(p,kt,i,j,step,disc)
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/output.inc'
      real*8 p(0:3,7),k(0:3),kt(0:3),q(0:3),phi,cost,kvec
      integer i,j,i1,step,disc

      if(disc.eq.1)return
      do i1=0,3
        k(i1)=p(i1,i)
        q(i1)=p(i1,i)+p(i1,j)
      enddo
      call boost(q,k,1d0,disc)
      if(k(1).eq.0d0)then
        phi=dsign(pi*0.5d0,k(2))
      else
        phi=datan(k(2)/k(1))
      endif
      if(k(1).lt.0d0)phi=phi+pi
      if(k(1).gt.0d0.and.k(2).lt.0d0)phi=phi+2d0*pi
      kvec=dsqrt(k(1)*k(1)+k(2)*k(2)+k(3)*k(3))
      if(kvec.eq.0d0)then
        if(numout.lt.maxout)then
          write(nout,'(a18,i5,i9)')' detkt: kvec = 0, ',chan,nev
          numout=numout+1
        endif
        disc=1
        return
      endif
      cost=k(3)/kvec
      call rotation(kt,-phi,cost,disc)
      call boost(q,kt,-1d0,disc)
      end

************************************************************************
*
*     adaptive optimization
*
************************************************************************
      subroutine optimization(w,g,gsum,ngen,nchannels)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/output.inc'
      real*8 w,g(maxchan),gsum
      integer nchannels,i1,ngen
c adaptopt
      real*8 al(maxchan,maxopt,maxgen),be(0:maxchan,maxgen)
      real*8 wi(maxchan,maxgen),almin(maxgen)
      integer nopt(0:maxopt,maxgen),opt(maxgen)
      common/adaptopt/al,be,wi,almin,nopt,opt

      if(nev.le.nopt(opt(ngen),ngen))then
        if(w.ne.0d0.and.gsum.ne.0d0)then
          do i1=1,nchannels
            if(g(i1)*gsum.lt.0d0)then
              if(numout.lt.maxout)then
                write(nout,'(a29,i5,i9)')
     &            ' optimization: gi*gsum <= 0, ',chan,nev
                numout=numout+1
              endif
              return
            else
              wi(i1,ngen)=wi(i1,ngen)+dsqrt(g(i1)*w*w/gsum)
            endif
          enddo
        endif
        if(nev.eq.nopt(opt(ngen),ngen))then
          do i1=1,nchannels
            al(i1,opt(ngen)+1,ngen)=al(i1,opt(ngen),ngen)
     &            *dsqrt(wi(i1,ngen))
            be(i1,ngen)=be(i1-1,ngen)+al(i1,opt(ngen)+1,ngen)
            wi(i1,ngen)=0d0
          enddo
          opt(ngen)=opt(ngen)+1
          do i1=1,nchannels
            al(i1,opt(ngen),ngen)=max(al(i1,opt(ngen),ngen),
     &                            almin(ngen)*be(nchannels,ngen))
            be(i1,ngen)=be(i1-1,ngen)+al(i1,opt(ngen),ngen)
          enddo
          do i1=1,nchannels
            al(i1,opt(ngen),ngen)=al(i1,opt(ngen),ngen)
     &            /be(nchannels,ngen)
            be(i1,ngen)=be(i1,ngen)/be(nchannels,ngen)
          enddo
        endif
      endif
      end

************************************************************************
*
*     slicing cuts
*
************************************************************************
      function check_slicut(outmom)
      implicit none
      integer check_slicut,i
      real*8 outmom(0:3,5),pvec(5)
      include 'COMMON/processpar.inc'
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2

      check_slicut = 0
      if (outmom(0,5).lt.(deltaE*dsqrt(CMenergy2))) check_slicut = 1
      do i = 1,5
        pvec(i) = dsqrt(outmom(1,i)**2+outmom(2,i)**2+outmom(3,i)**2)
      enddo
      do i = 1,4
        if (q(i).ne.0.and.(outmom(1,i)*outmom(1,5)+outmom(2,i)
     &        *outmom(2,5)+outmom(3,i)*outmom(3,5))/pvec(i)/pvec(5)
     &        .gt.cos(deltaTh)) check_slicut = 1
      enddo

      end

*******************************************************************
*
*     pseudo-random-number generator
*
*******************************************************************
      function ran1()
      implicit none
      real*8 s1,s2,s3,ran1
      integer init
      save init,s1,s2,s3
      data init/0/,s1/0d0/,s2/0d0/,s3/0d0/
      if (init.eq.0) then
        init=1
        s1=dsqrt(2d0)-1d0
        s2=dsqrt(3d0)-1d0
        s3=dsqrt(5d0)-2d0
      endif
      s1=dmod(s1+s2+s3,1d0)
      s2=dmod(s1+s2+s3,1d0)
      s3=dmod(s1+s2+s3,1d0)
      ran1=s1

      end

*******************************************************************
*
*     pseudo-random-number generator RCARRY
*
*******************************************************************
      function ran_rcarry()
      implicit none
      real*8 B,ran_rcarry
      integer c,i,j,x(24),calls
      save x,c,B,calls
c B=2^30=maximum integer
      data x/1d6,3d6,2d5,2d6,4d6,3d5,7d6,1d7,2d6,1d7,1d5,1d7,2d6,
     &     2d6,3d5,1d6,5d6,7d6,3d6,1d7,3d6,6d6,8d6,7d6/,
     &     c/0/,B/1073741824d0/,calls/0/

 100  if (calls.eq.247) calls = 1
      calls = calls+1
      ran_rcarry = x(15)-x(1)-c
      c = 0
      if(ran_rcarry.lt.0d0)then
         ran_rcarry = ran_rcarry+B
         c = 1
      endif
      do i = 1,23
         x(i) = x(i+1)
      enddo
      x(24) = int(ran_rcarry)
      ran_rcarry = ran_rcarry/B
      if(calls.gt.24.or.ran_rcarry.eq.0) goto 100 !take 24, then skip 223 numb.

      end

*******************************************************************
*
*     3-particle vertex
*
*******************************************************************
      function v3(p1,p2,p3)
      implicit none
      character*8 p1,p2,p3
      logical v3,v3g
      if(p1.eq.'XXX'.or.p2.eq.'XXX'.or.p3.eq.'XXX')then
        v3=.false.
        return
      endif
      v3=.true.
      if(v3g(p1,p2,p3))return
      if(v3g(p1,p3,p2))return
      if(v3g(p2,p1,p3))return
      if(v3g(p2,p3,p1))return
      if(v3g(p3,p1,p2))return
      if(v3g(p3,p2,p1))return
      v3=.false.
      end

*******************************************************************
*
*     generic 3-particle vertex
*
*******************************************************************
      function v3g(p1,p2,p3)
      implicit none
      character*8 p1,p2,p3
      logical v3g
      include 'COMMON/options.inc'
      v3g=.true.
      if(p1.eq.'e'.and.p2.eq.'anti-nue'.and.p3.eq.'W+')return
      if(p1.eq.'anti-e'.and.p2.eq.'nue'.and.p3.eq.'W-')return
      if(p1.eq.'mu'.and.p2.eq.'anti-num'.and.p3.eq.'W+')return
      if(p1.eq.'anti-mu'.and.p2.eq.'num'.and.p3.eq.'W-')return
      if(p1.eq.'tau'.and.p2.eq.'anti-nut'.and.p3.eq.'W+')return
      if(p1.eq.'anti-tau'.and.p2.eq.'nut'.and.p3.eq.'W-')return
      if(p1.eq.'dq'.and.p2.eq.'anti-uq'.and.p3.eq.'W+')return
      if(p1.eq.'anti-dq'.and.p2.eq.'uq'.and.p3.eq.'W-')return
      if(p1.eq.'sq'.and.p2.eq.'anti-cq'.and.p3.eq.'W+')return
      if(p1.eq.'anti-sq'.and.p2.eq.'cq'.and.p3.eq.'W-')return
      if(p1.eq.'bq'.and.p2.eq.'anti-tq'.and.p3.eq.'W+')return
      if(p1.eq.'anti-bq'.and.p2.eq.'tq'.and.p3.eq.'W-')return
      if(p1.eq.'nue'.and.p2.eq.'anti-nue'.and.p3.eq.'Z0')return
      if(p1.eq.'e'.and.p2.eq.'anti-e'.and.p3.eq.'Z0')return
      if(p1.eq.'e'.and.p2.eq.'anti-e'.and.p3.eq.'ga')return
      if(p1.eq.'num'.and.p2.eq.'anti-num'.and.p3.eq.'Z0')return
      if(p1.eq.'mu'.and.p2.eq.'anti-mu'.and.p3.eq.'Z0')return
      if(p1.eq.'mu'.and.p2.eq.'anti-mu'.and.p3.eq.'ga')return
      if(p1.eq.'nut'.and.p2.eq.'anti-nut'.and.p3.eq.'Z0')return
      if(p1.eq.'tau'.and.p2.eq.'anti-tau'.and.p3.eq.'Z0')return
      if(p1.eq.'tau'.and.p2.eq.'anti-tau'.and.p3.eq.'ga')return
      if(p1.eq.'dq'.and.p2.eq.'anti-dq'.and.p3.eq.'Z0')return
      if(p1.eq.'dq'.and.p2.eq.'anti-dq'.and.p3.eq.'ga')return
      if(p1.eq.'uq'.and.p2.eq.'anti-uq'.and.p3.eq.'Z0')return
      if(p1.eq.'uq'.and.p2.eq.'anti-uq'.and.p3.eq.'ga')return
      if(p1.eq.'sq'.and.p2.eq.'anti-sq'.and.p3.eq.'Z0')return
      if(p1.eq.'sq'.and.p2.eq.'anti-sq'.and.p3.eq.'ga')return
      if(p1.eq.'cq'.and.p2.eq.'anti-cq'.and.p3.eq.'Z0')return
      if(p1.eq.'cq'.and.p2.eq.'anti-cq'.and.p3.eq.'ga')return
      if(p1.eq.'bq'.and.p2.eq.'anti-bq'.and.p3.eq.'Z0')return
      if(p1.eq.'bq'.and.p2.eq.'anti-bq'.and.p3.eq.'ga')return
      if(p1.eq.'tq'.and.p2.eq.'anti-tq'.and.p3.eq.'Z0')return
      if(p1.eq.'tq'.and.p2.eq.'anti-tq'.and.p3.eq.'ga')return
      if(p1.eq.'ga'.and.p2.eq.'W+'.and.p3.eq.'W-')return
      if(p1.eq.'Z0'.and.p2.eq.'W+'.and.p3.eq.'W-')return
      if(p1.eq.'H0'.and.p2.eq.'W+'.and.p3.eq.'W-')return
      if(p1.eq.'H0'.and.p2.eq.'Z0'.and.p3.eq.'Z0')return
      v3g=.false.
      end

************************************************************************
*
*   select events in singular regions (as check for subtraction)
*
*      singular=0: no singular event
*      singular=1: event lies in singular region,
*      flag=1: check whether photon energy < minen, out:photon energy
*      flag=2: check whether photon is closer to a fermion than mintheta
*              out: smallest angle with fermion 'ferm'
*      flag=3: check whether photon is closer to fermion 'ferm'
*              than mintheta, out: angle
************************************************************************
      function singular(outmom,ferm,minen,mintheta,out,flag)
      implicit none
      real*8 outmom(0:3,5),pvec(7),prod
      real*8 mintheta,minen,out,outmin
      integer i1,singular,ferm,flag
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'

      singular=1
      do i1=1,5
        pvec(i1)=dsqrt(outmom(1,i1)**2+outmom(2,i1)**2+outmom(3,i1)**2)
      enddo
c photon energy
      if(flag.eq.1)then
        if(outmom(0,5).lt.minen)then
           out=outmom(0,5)
           return
        endif
c minimal angle between photon and fermion
      elseif(flag.eq.2)then
        outmin=180d0
        do i1=1,4
          if(q(i1).ne.0d0)then
            prod=(outmom(1,i1)*outmom(1,5)+outmom(2,i1)*outmom(2,5)
     &            +outmom(3,i1)*outmom(3,5))
            if(prod.gt.dcos(mintheta*pi/180d0)*pvec(i1)*pvec(5))then
              out=dacos(prod/(pvec(i1)*pvec(5)))*180d0/pi
              if(out.lt.outmin) then
                outmin=out
                ferm=i1
              endif
            endif
          endif
        enddo
        if(outmin.lt.180d0)then
           out=outmin
           return
        endif
c angle between photon and fermion ferm
      elseif(flag.eq.3)then
        prod=(outmom(1,ferm)*outmom(1,5)+outmom(2,ferm)*outmom(2,5)
     &        +outmom(3,ferm)*outmom(3,5))
        if(prod.gt.dcos(mintheta*pi/180d0)*pvec(ferm)*pvec(5))then
          out=dacos(prod/(pvec(ferm)*pvec(5)))*180d0/pi
          return
        endif
      else
         write(*,*)'flag not defined'
         stop
      endif

      singular=0
      end

************************************************************************
*
*     lambda and mf dependent terms of virtual corrections
*
************************************************************************
      function lambdamf(sij,born)
      implicit none
      real*8 lambdamf,sij(7,7),born
      real*8 infrared
      real*8 sumq,sumqlm,sumql2m,sumqsij
      integer i,j
      include 'COMMON/processpar.inc'
      include 'COMMON/param.inc'
      include 'COMMON/xparam.inc'
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2

      infrared = 0d0
      lambdamf = 0d0
      do i=1,4
        if (q(i).ne.0d0)
     &    infrared = infrared + 2d0*log(lambda2)*q(i)**2*log(mf2(i))
     &        +2d0*log(lambda2)*q(i)**2 +q(i)**2*log(mf2(i))
     &        -q(i)**2*log(mf2(i))**2
        do j=1,4
          if (j.ne.i.and.q(j).ne.0d0)
     &      infrared = infrared + 2d0*(-1)**(i+j)*q(i)*q(j)
     &         *log(sij(i+2,j+2))*log(lambda2)
        enddo
      enddo

c$$$      sumq = 0d0
c$$$      sumqlm = 0d0
c$$$      sumql2m = 0d0
c$$$      sumqsij = 0d0
c$$$      do i = 1,4
c$$$        sumq = sumq + q(i)**2
c$$$        sumqlm = sumqlm + q(i)**2*log(mf2(i))
c$$$        sumql2m = sumql2m + q(i)**2*log(mf2(i))**2
c$$$        do j=i+1,4
c$$$          sumqsij = sumqsij + 2d0*(-1)**(i+j)*(q(i)*q(j)
c$$$     &                        *log(sij(i+2,j+2)))
c$$$        enddo
c$$$      enddo

c      infrared =
c     &     +2d0*log(lambda2)*sumq
ccc     &     -3d0*sumqlm
c     &     -2d0*dcmplx(0d0,1d0)*pi*log(lambda2)*sumq
c     &     +2d0*dcmplx(0d0,1d0)*pi*log(lambda2)*sumq
c     &     +2d0*log(lambda2)*sumqsij
c     &     +2d0*log(lambda2)*sumqlm
c     &     +sumqlm
c     &     -sumql2m

c     &     +2d0*log(lambda2)*sumqsij
c     &     +2d0*log(lambda2)*sumqlm
c     &     +4d0*sumqlm
c     &     -sumql2m

c counterterme (nur Massensingularitaeten, keine vollst. soft sing.)
c     &     +2d0*log(lambda2)*sumq
c     &     -3d0*sumqlm

c vertex
c     &     +2d0*q(1)*q(2)*(
c     &       -2d0*log(lambda2)*log(sij(3,4))
c     &       +2d0*log(lambda2)*log(mf2(1)*mf2(2))/2d0
c     &       -1d0/2d0*log(mf2(1))**2 - 1d0/2d0*log(mf2(2))**2
c     &       +2d0*log(mf2(1)) + 2d0*log(mf2(2)))
c     &     +2d0*q(3)*q(4)*(
c     &       -2d0*log(lambda2)*log(sij(5,6))
c     &       +2d0*log(lambda2)*log(mf2(3)*mf2(4))/2d0
c     &       -1d0/2d0*log(mf2(3))**2 - 1d0/2d0*log(mf2(4))**2
c     &       +2d0*log(mf2(3)) + 2d0*log(mf2(4)))
c     &     +q(1)
c     &     *4d0*log(mf2(1))*(1d0+xmw2/sij(3,4)*log(1d0-sij(3,4)/xmw2))
c     &     -q(2)
c     &     *4d0*log(mf2(2))*(1d0+xmw2/sij(3,4)*log(1d0-sij(3,4)/xmw2))
c     &     -q(3)
c     &     *4d0*log(mf2(3))*(1d0+xmw2/sij(5,6)*log(1d0-sij(5,6)/xmw2))
c     &     +q(4)
c     &     *4d0*log(mf2(4))*(1d0+xmw2/sij(5,6)*log(1d0-sij(5,6)/xmw2))


c boxes
c     &     -q(1)
c     &     *4d0*log(mf2(1))*(1d0+xmw2/sij(3,4)*log(1d0-sij(3,4)/xmw2))
c     &     +q(2)
c     &     *4d0*log(mf2(2))*(1d0+xmw2/sij(3,4)*log(1d0-sij(3,4)/xmw2))
c     &     +q(3)
c     &     *4d0*log(mf2(3))*(1d0+xmw2/sij(5,6)*log(1d0-sij(5,6)/xmw2))
c     &     -q(4)
c     &     *4d0*log(mf2(4))*(1d0+xmw2/sij(5,6)*log(1d0-sij(5,6)/xmw2))

c pentagons
c$$$     &     +4d0*log(lambda2)*(
c$$$     &        +q(1)*q(3)*log(sij(3,5))
c$$$     &        -q(1)*q(4)*log(sij(3,6))
c$$$     &        -q(2)*q(3)*log(sij(4,5))
c$$$     &        +q(2)*q(4)*log(sij(4,6)) )
c$$$     &     +2d0*log(lambda2)*(
c$$$     &        -q(1)*q(3)*log(mf2(1)*mf2(3))
c$$$     &        +q(1)*q(4)*log(mf2(1)*mf2(4))
c$$$     &        +q(2)*q(3)*log(mf2(2)*mf2(3))
c$$$     &        -q(2)*q(4)*log(mf2(2)*mf2(4)) )
c$$$     &     +(
c$$$     &        +q(1)*q(3)*(log(mf2(1))**2+log(mf2(3))**2)
c$$$     &        -q(1)*q(4)*(log(mf2(1))**2+log(mf2(4))**2)
c$$$     &        -q(2)*q(3)*(log(mf2(2))**2+log(mf2(3))**2)
c$$$     &        +q(2)*q(4)*(log(mf2(2))**2+log(mf2(4))**2) )
c$$$     &     +4d0*(
c$$$     &        -q(1)*q(3)*log(mf2(1)*mf2(3))
c$$$     &        +q(1)*q(4)*log(mf2(1)*mf2(4))
c$$$     &        +q(2)*q(3)*log(mf2(2)*mf2(3))
c$$$     &        -q(2)*q(4)*log(mf2(2)*mf2(4)) )

c$$$     &     -2d0*q(1)*q(3)*(
c$$$     &       -2d0*log(lambda2)*log(sij(3,5))
c$$$     &       +2d0*log(lambda2)*log(mf2(1)*mf2(3))/2d0
c$$$     &       -1d0/2d0*log(mf2(1))**2 - 1d0/2d0*log(mf2(3))**2
c$$$     &       +2d0*log(mf2(1)) + 2d0*log(mf2(3)))

c$$$     &     +2d0*q(1)*q(4)*(
c$$$     &       -2d0*log(lambda2)*log(sij(3,6))
c$$$     &       +2d0*log(lambda2)*log(mf2(1)*mf2(4))/2d0
c$$$     &       -1d0/2d0*log(mf2(1))**2 - 1d0/2d0*log(mf2(4))**2
c$$$     &       +2d0*log(mf2(1)) + 2d0*log(mf2(4)))

c$$$     &     +2d0*q(2)*q(3)*(
c$$$     &       -2d0*log(lambda2)*log(sij(4,5))
c$$$     &       +2d0*log(lambda2)*log(mf2(2)*mf2(3))/2d0
c$$$     &       -1d0/2d0*log(mf2(2))**2 - 1d0/2d0*log(mf2(3))**2
c$$$     &       +2d0*log(mf2(2)) + 2d0*log(mf2(3)))

c$$$     &     -2d0*q(2)*q(4)*(
c$$$     &       -2d0*log(lambda2)*log(sij(4,6))
c$$$     &       +2d0*log(lambda2)*log(mf2(2)*mf2(4))/2d0
c$$$     &       -1d0/2d0*log(mf2(2))**2 - 1d0/2d0*log(mf2(4))**2
c$$$     &       +2d0*log(mf2(2)) + 2d0*log(mf2(4)))

      lambdamf = lambdamf - alpha/4d0/pi *infrared*born

      end

************************************************************************
*
*     phase-space point for testing purpose
*       remember to set MH=200
************************************************************************
      subroutine test_phsp(p)
      implicit none
      real*8 p(0:3,7)

      p(0,3)=     35.737136961258450185d0
      p(1,3)=     12.903667496202508858d0
      p(2,3)=     32.958452445865162872d0
      p(3,3)=      4.937482729978506235d0
      p(0,4)=     64.694762138903740833d0
      p(1,4)=    -12.903667496202508858d0
      p(2,4)=    -32.958452445865162872d0
      p(3,4)=     54.153928996245753069d0
      p(0,5)=     22.856927684519696697d0
      p(1,5)=      2.523914283367739042d0
      p(2,5)=     16.297374249483024045d0
      p(3,5)=     15.826073184396159732d0
      p(0,6)=     76.711173215318098073d0
      p(1,6)=     -2.523914283367739042d0
      p(2,6)=    -16.297374249483024045d0
      p(3,6)=    -74.917484910620416372d0

      end

**********************************************************************
      subroutine llfsr(outmom,m2born,weight1,weightz,momz,nwabs
     &                ,nocut4f,random)
**********************************************************************
*     leading-log final-state radiation via structure function
*     subborn: subtract Born contribution
*     qllfsr = flag for contributions
*     qllfsr=1,2:dM2 = Gamma_FSR1*Gamma_FSR2*Gamma_FSR3*Gamma_FSR4
*                   -M2Born
*                   -Gamma_FSR1,1-Gamma_FSR2,1-Gamma_FSR3,1-Gamma_FSR4,1
*     qllfsr=1:beyond O(alpha) with exponentiation
*     qllfsr=2:beyond O(alpha) without exponentiation
*     qllfsr=3:O(alpha) up to O(alpha^3) with exponentiation
*     qllfsr=4:O(alpha) up to O(alpha^3) without exponentiation
*     qllfsr=5:O(alpha) up to O(alpha^2) without exponentiation
*     qllfsr=6:O(alpha) without exponentiation
*       weight1, outmom: all z=1
*       weightz(1-4), momz(1-4): 1 z.ne.1
*       weightz(5-10), momz(5-10): 2 z.ne.1
*       weightz(11-14), momz(11-14): 3 z.ne.1
*       weightz(15), momz(15): 4 z.ne.1
**********************************************************************
      implicit none
      integer i,j,k,l,m,imin,i2min,nocut4f,nocut,nfsr,nwabs,nparticles
      integer expon,subborn,suboa
      complex*16 cspen
      real*8 z(4),x,ran1,random(4),zz,yy,jacy(4),jacy2,beta,beta0,m2born
      real*8 weight1,weightz(15),outmom(0:3,5),momz(0:3,5,15)
      real*8 eps,GammaFSR,corrz(4),corr1(4),corroa(4)
      real*8 emin,e2min,dummy(0:3,5),zeta3
      parameter(zeta3=1.2020569031595942854d0)

      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/max.inc'
      include 'COMMON/options.inc'
      include 'COMMON/techpar.inc'
c factorization scale
      real*8 fsrfact,nfsrfact
      common/fsrfact/fsrfact,nfsrfact
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej

      subborn = 1 !subtract Born contribution
      if (qllfsr.eq.1.or.qllfsr.eq.3.or.qllfsr.eq.7) then
        expon = 1     !with exponentiation
      else
        expon = 0
      endif
      if (qllfsr.le.2) then
        suboa = 1     !subtract O(alpha) contribution
      else
        suboa = 0
      endif

      eps = 1d-16
      nfsr = 0
      emin = 0d0
      e2min = 0d0
      imin = 1
      i2min = 1
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c define Gamma_FSR = corrz-corr1
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      do i = 1,4
        corrz(i) = 0d0  !if no FSR: delta distribution
        corr1(i) = -1d0
        corroa(i) = 0d0
        jacy(i) = 1d0
c condition for including FSR
        if (q(i).ne.0d0.and.quark(i).eq.0) then
          nfsr = nfsr + 1     !number of fermions with FSR
          x = eps+(1d0-2d0*eps)*random(i)
          beta  = 2d0*alpha/pi*( log(fsrfact/mf2(i))-1d0 )
          beta0 = 2d0*alpha0/pi*( log(fsrfact/mf2(i))-1d0 )

          if (expon.eq.1) then
            x     = random(i)
            yy    = x**(2d0/beta0)
            jacy(i)  = 2d0/beta0*yy/x
            jacy2 = 2d0/beta0/x
c Jacobian jacy shifted to GammaFSR
          else
            yy    = x
          endif

          zz = 1d0-yy
          z(i) = zz

c O(alpha) corrections are subtracted in the end (->beta, not beta0)
          if (suboa.eq.1.and.expon.eq.1) then
            corroa(i) = beta/4d0*(1d0+zz**2)*jacy2
          elseif (suboa.eq.1.and.expon.eq.0) then
            corroa(i) = beta/4d0*(1d0+zz**2)/yy
          endif
c (beyond) O(alpha) up to O(alpha^3) with exponentiation
          if (expon.eq.1) then
            corrz(i) = GammaFSR(yy,beta0)
            corr1(i) = 0d0
c (beyond) O(alpha) up to O(alpha^3) without exponentiation
          elseif (qllfsr.eq.5) then
            corrz(i) = beta0/4d0*(1d0+zz**2)/yy
     &                +beta0**2/16d0*(3d0+4d0*log(yy))/yy
     &                -beta0**2/32d0*((1d0+3d0*zz**2)/yy*log(zz)
     &                               +4d0*(1d0+zz)*log(yy)+5d0+zz)
            corr1(i) = - 1d0
     &                +beta0/4d0*(1d0+zz**2)/yy
     &                +beta0**2/16d0*(3d0+4d0*log(yy))/yy
     &                -beta0**2*(9d0/128d0-pi**2/48d0)
          elseif (qllfsr.eq.6) then
            corrz(i) = beta0/4d0*(1d0+zz**2)/yy
            corr1(i) = - 1d0
     &                +beta0/4d0*(1d0+zz**2)/yy
          else
            corrz(i) = beta0/4d0*(1d0+zz**2)/yy
     &                +beta0**2/16d0*(3d0+4d0*log(yy))/yy
     &                -beta0**2/32d0*((1d0+3d0*zz**2)/yy*log(zz)
     &                               +4d0*(1d0+zz)*log(yy)+5d0+zz)
     &                +beta0**3/768d0*(27d0-8d0*pi**2+72d0*log(yy)
     &                                +48d0*log(yy)**2)/yy
     &                -beta0**3/384d0*(
     &                   (1d0+zz)*(6d0*cspen(dcmplx(zz))+12d0*log(yy)**2
     &                             -3d0*pi**2)
     &                   +(3d0/2d0*(1d0+8d0*zz+3d0*zz**2)*log(zz)
     &                     +6d0*(zz+5d0)*yy*log(yy)
     &                     +12d0*(1d0+zz**2)*log(zz)*log(yy)
     &                     -1d0/2d0*(1d0+7d0*zz**2)*log(zz)**2
     &                     +1d0/4d0*(39d0-24d0*zz-15d0*zz**2))/yy )
            corr1(i) = - 1d0
     &                +beta0/4d0*(1d0+zz**2)/yy
     &                +beta0**2/16d0*(3d0+4d0*log(yy))/yy
     &                -beta0**2*(9d0/128d0-pi**2/48d0)
     &                +beta0**3/768d0*(27d0-8d0*pi**2+72d0*log(yy)
     &                                +48d0*log(yy)**2)/yy
     &                -beta0**3*(9d0/1024d0-pi**2/128d0+zeta3/24d0)
          endif

c cutoff for z close to 1
          if ((1d0-z(i)).lt.sub4cut) then
            corroa(i) = 0d0
            corr1(i) = corr1(i) - corrz(i)
            corrz(i) = 0d0
          endif

c find photon with largest energy
          if ((1d0-z(i))*outmom(0,i).gt.emin) then
            emin = (1d0-z(i))*outmom(0,i)
            imin = i
          endif
        endif
      enddo

c find photon with next to largest energy
      do i = 1,4
        if (q(i).ne.0d0.and.quark(i).eq.0.and.i.ne.imin.and.
     &        (1d0-z(i))*outmom(0,i).gt.e2min) then
          e2min = (1d0-z(i))*outmom(0,i)
          i2min = i
        endif
      enddo

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c momenta for cut and histogram subroutines
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      do j = 0,3
      do i = 1,5
      do k = 1,15
        momz(j,i,k) = outmom(j,i)
      enddo
      enddo
      enddo

      if (qrecomb.ne.0) then
        nparticles = 4
      else
        nparticles = 5
        do j = 0,3
          l = 0
          do i = 1,4
            if (q(i).ne.0d0.and.quark(i).eq.0) then
c 4 sub-events with 1 z.ne.1
              momz(j,i,i) = outmom(j,i) *z(i)
              momz(j,5,i) = outmom(j,i) *(1d0-z(i))
            endif
c 6 sub-events with 2 z.ne.1
            do k = i+1,4
              l = l+1
              if (q(i).ne.0d0.and.quark(i).eq.0.and.
     &            q(k).ne.0d0.and.quark(k).eq.0) then
                momz(j,i,4+l) = outmom(j,i) *z(i)
                momz(j,k,4+l) = outmom(j,k) *z(k)
                if ((1d0-z(i))*outmom(0,i).gt.(1d0-z(k))*outmom(0,k))
     &               then                                  !photon with
                  momz(j,5,4+l) = outmom(j,i) *(1d0-z(i))  !highest energy
                else
                  momz(j,5,4+l) = outmom(j,k) *(1d0-z(k))
                endif
              endif
            enddo
c 4 sub-events with 3 z.ne.1
            if (nfsr.ge.3) then
              do k = 1,4
                if (k.ne.i.and.q(k).ne.0d0.and.quark(k).eq.0)
     &               momz(j,k,10+i) = outmom(j,k) *z(k)
              enddo
              if (imin.ne.i) then
                momz(j,5,10+i) = outmom(j,imin) *(1d0-z(imin))
              else
                momz(j,5,10+i) = outmom(j,i2min) *(1d0-z(i2min))
              endif
            endif
c 1 sub-event with 4 z.ne.1
          enddo
          if (nfsr.eq.4) then
            do i = 1,4
              momz(j,i,15) = outmom(j,i) *z(i)
            enddo
            momz(j,5,15) = outmom(j,imin) *(1d0-z(imin))
          endif
        enddo
      endif

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c calculate weights
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      do m = 1,15
        weightz(m) = 0d0
      enddo
c weight for all z.eq.1
      if (nocut4f.ne.0) then
        weight1 = weight1 + corr1(1)*corr1(2)*corr1(3)*corr1(4)*m2born
c subtract born and O(alpha) to avoid double counting
        if (subborn.eq.1) weight1 = weight1 - m2born
        if (suboa.eq.1) weight1 = weight1 + (corroa(1)+corroa(2)
     &          +corroa(3)+corroa(4))*m2born  !jac included in corroa
      endif

      l = 0
      do i = 1,4
c weight for 1 z.ne.1
        if (q(i).ne.0d0.and.quark(i).eq.0) then
          call cut(momz(0,1,i),dummy,nparticles,nocut,0)
          if (nocut.eq.0) then
            subrej(nwabs-1+i) = subrej(nwabs-1+i) + 1
          else
            weightz(i) = - m2born ! - from delta_+ prescription
            do m = 1,4
              if (m.eq.i) then
                weightz(i) = weightz(i) *corrz(m)
              else
                weightz(i) = weightz(i) *corr1(m)
              endif
            enddo
c subtract O(alpha) to avoid double counting
            if (suboa.eq.1) weightz(i) = weightz(i) - corroa(i)*m2born
          endif
        endif
c weight for 2 z.ne.1
        do k = i+1,4
          l = l+1
          if (q(i).ne.0d0.and.quark(i).eq.0.and.
     &        q(k).ne.0d0.and.quark(k).eq.0) then
            call cut(momz(0,1,4+l),dummy,nparticles,nocut,0)
            if (nocut.eq.0) then
              subrej(nwabs-1+4+l) = subrej(nwabs-1+4+l) + 1
            else
              weightz(4+l) = m2born
              do m = 1,4
                if (m.eq.i.or.m.eq.k) then
                  weightz(4+l) = weightz(4+l) *corrz(m)
                else
                  weightz(4+l) = weightz(4+l) *corr1(m)
                endif
              enddo
            endif
          endif
        enddo
c weight for 3 z.ne.1
        if(nfsr.eq.4.or.(nfsr.eq.3.and.(q(i).eq.0d0.or.quark(i).eq.1)))
     &       then
          call cut(momz(0,1,10+i),dummy,nparticles,nocut,0)
          if (nocut.eq.0) then
            subrej(nwabs-1+10+i) = subrej(nwabs-1+10+i) + 1
          else
            weightz(10+i) = - m2born ! - from delta_+ prescription
            do m = 1,4
              if (m.eq.i) then
                weightz(10+i) = weightz(10+i) *corr1(m)
              else
                weightz(10+i) = weightz(10+i) *corrz(m)
              endif
            enddo
          endif
        endif
      enddo
c weight for 4 z.ne.1
      if (nfsr.eq.4) then
        call cut(momz(0,1,15),dummy,nparticles,nocut,0)
        if (nocut.eq.0) then
          subrej(nwabs-1+15) = subrej(nwabs-1+15) + 1
        else
          weightz(15) = corrz(1)*corrz(2)*corrz(3)*corrz(4)*m2born
        endif
      endif

      end

*******************************************************************************
        function GammaFSR(y,beta)
*******************************************************************************
*       Leading-log structure function
*       NOTE:  x = 1-y
*
*       taken from Eq.(67) of Beenakker et al. in CERN 96-01, Vol.1, p.129
*       BETA choice:  beta = 2*alpha/pi*( log(Q^2/m^2)-1 )
*------------------------------------------------------------------------------
*       24.10.2001 Stefan Dittmaier
*******************************************************************************
        implicit real*8 (a-z)
        complex*16 cspen
        parameter(gammae=0.5772156649015328606d0)
        parameter(    pi=3.1415926535897931160d0)

        x = 1d0-y

c factor beta/2d0*y**(beta/2d0-1d0) extracted!

        GammaFSR= exp(-gammae*beta/2d0+3d0*beta/8d0)/Gamma(1d0+beta/2d0)

        if (y.ne.0d0) then
        GammaFSR = GammaFSR + 2d0/beta*y**(1d0-beta/2d0)*(
     &    -beta/4d0*(1d0+x)
     &    -beta**2/32d0
     &      *( (1d0+3d0*x**2)/y*log(x)+4d0*(1d0+x)*log(y)+5d0+x )
     &    -beta**3/384d0
     &      *( (1d0+x)*(6d0*cspen(dcmplx(x))+12d0*log(y)**2-3d0*pi**2)
     &        +( 3d0/2d0*(1d0+8d0*x+3d0*x**2)*log(x)
     &          +6d0*(x+5d0)*y*log(y)+12d0*(1d0+x**2)*log(x)*log(y)
     &          -(1d0+7d0*x**2)/2d0*log(x)**2
     &          +(39d0-24d0*x-15d0*x**2)/4d0 )/y  )
     &    )
        endif

        end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c     gamma function                                              c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function gamma(x)
      implicit none
c local variable
      integer max
      parameter(max=10)
      real*8 gamma,x,omx,a(max),prod,pi
      parameter(pi=3.1415926535897931160d0)
      integer i1
c a(1)=gammae
      data a/0.5772156649015328606d0,0d0,
     *       0.4006856343865314285d0,0d0,
     *       0.2073855510286739853d0,0d0,
     *       0.1440498967688461181d0,0d0,
     *       0.1113342658695646905d0,0d0/
c good convergence for small values of 1-x
      omx=x-1d0
      gamma=dexp(0.5d0*dlog(omx*pi/dsin(pi*omx)))
      do i1=1,max,2
        prod=dexp(-a(i1)*omx**i1)
        gamma=gamma*prod
        if(dabs(prod-1d0).lt.1d-10)return
      enddo
      write(*,'(a,g25.15)')' gamma: converges badly, x = ',x
      stop
      end
