**************************************************************
*
*     main program
*
**************************************************************
      implicit none

      call init_main    !some options, particle names/properties etc.
      call read_input   !from inputfile
      call fix_param    !derived parameters like alpha (from G_mu) and
                        !technical parameters like slicing cuts
      call check_parameters !check consistency of parameters
      call init_cuts    !phase-space cuts and photon recombination parameters
      call init_hepup   !initialize Les Houches accord common block

      call init_integration
      call init_fast    !fast evaluation of tensor integrals for qfast=1
      call begin_output

c      call checkamp   !switch off fast calc. of tensors (fastoff)
c      call checkuvir  !switch off fast calc. of tensors (fastoff)

      call integration

c      call write_multichannel   !output of multichannel integration

      end

**************************************************************
*
*     read input
*
*     convention: f1,f3:fermions; f2,f4: anti-fermions,
*                 f1+f2: W+/Z-boson, f3+f4: W-/Z-boson
**************************************************************
      subroutine read_input
      implicit none
      include 'COMMON/processpar.inc'
      include 'COMMON/param.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
      character*100 inputline,storeinput
      character*100 identifier,orgidentifier
      integer linputline,i1,isep1,isep2
      data linputline /100/
c output
      character*90 outputfile
      common/outputfile/outputfile
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c random generator
      integer rangen,ranshift
      common/random/rangen,ranshift
c factorization scale
      real*8 fsrfact,nfsrfact
      common/fsrfact/fsrfact,nfsrfact
c cuts
      real*8 ecut(5),scut(5,5),ccut(5,5),precomb(3)
      integer qcuts
      common/cuts/ecut,scut,ccut,precomb,qcuts

 50    read(*,'((a))',end=51,err=52) inputline
       if(inputline(1:1).eq.'!'.or.inputline(1:1).eq.'#'
     &      .or.inputline(1:1).eq.'') goto 50
       isep1=0
       isep2=linputline+1

       do i1=1,linputline
         if(inputline(i1:i1).eq.'=') then
           isep1=i1
           goto53
         elseif(inputline(i1:i1).eq.'!') then
           isep2=i1
           goto54
         endif
       enddo
 53    continue
       do i1=isep1+1,linputline
         if(inputline(i1:i1).eq.'!') then
           isep2=i1
           goto54
         endif
       enddo
 54    continue

       if(isep1.gt.0)then
         identifier= inputline(1:isep1-1)
         orgidentifier=identifier
         do i1=isep1-1,1,-1
           if (identifier(i1:i1).eq.' ') then
             identifier=identifier(1:i1-1)//identifier(i1+1:linputline)
           elseif(identifier(i1:i1).ge.'A'.and.identifier(i1:i1).le.'Z')
     &         then
             if(identifier(i1:i1).eq.'A') then
               identifier(i1:i1)='a'
             elseif(identifier(i1:i1).eq.'B') then
               identifier(i1:i1)='b'
             elseif(identifier(i1:i1).eq.'C') then
               identifier(i1:i1)='c'
             elseif(identifier(i1:i1).eq.'D') then
               identifier(i1:i1)='d'
             elseif(identifier(i1:i1).eq.'E') then
               identifier(i1:i1)='e'
             elseif(identifier(i1:i1).eq.'F') then
               identifier(i1:i1)='f'
             elseif(identifier(i1:i1).eq.'G') then
               identifier(i1:i1)='g'
             elseif(identifier(i1:i1).eq.'H') then
               identifier(i1:i1)='h'
             elseif(identifier(i1:i1).eq.'I') then
               identifier(i1:i1)='i'
             elseif(identifier(i1:i1).eq.'J') then
               identifier(i1:i1)='j'
             elseif(identifier(i1:i1).eq.'K') then
               identifier(i1:i1)='k'
             elseif(identifier(i1:i1).eq.'L') then
               identifier(i1:i1)='l'
             elseif(identifier(i1:i1).eq.'M') then
               identifier(i1:i1)='m'
             elseif(identifier(i1:i1).eq.'N') then
               identifier(i1:i1)='n'
             elseif(identifier(i1:i1).eq.'O') then
               identifier(i1:i1)='o'
             elseif(identifier(i1:i1).eq.'P') then
               identifier(i1:i1)='p'
             elseif(identifier(i1:i1).eq.'Q') then
               identifier(i1:i1)='q'
             elseif(identifier(i1:i1).eq.'R') then
               identifier(i1:i1)='r'
             elseif(identifier(i1:i1).eq.'S') then
               identifier(i1:i1)='s'
             elseif(identifier(i1:i1).eq.'T') then
               identifier(i1:i1)='t'
             elseif(identifier(i1:i1).eq.'U') then
               identifier(i1:i1)='u'
             elseif(identifier(i1:i1).eq.'V') then
               identifier(i1:i1)='v'
             elseif(identifier(i1:i1).eq.'W') then
               identifier(i1:i1)='w'
             elseif(identifier(i1:i1).eq.'X') then
               identifier(i1:i1)='x'
             elseif(identifier(i1:i1).eq.'Y') then
               identifier(i1:i1)='y'
             elseif(identifier(i1:i1).eq.'Z') then
               identifier(i1:i1)='z'
             endif
           endif
         enddo
         write(storeinput,*) inputline(isep1+1:isep2-1)

         if(identifier.eq.'outputfile') then
           read(storeinput,'(a100)',err=61) outputfile
           read(storeinput,*) outputfile
         elseif(identifier.eq.'nevents') then
           read(storeinput,'(i100)',err=61) nevents
           read(storeinput,*) nevents
         elseif(identifier.eq.'nunwevents') then
           read(storeinput,'(i100)',err=61) nunwevents
           read(storeinput,*) nunwevents
         elseif(identifier.eq.'contrib') then
           read(storeinput,'(i100)',err=61) contrib
           read(storeinput,*) contrib
         elseif(identifier.eq.'qqcd') then
           read(storeinput,'(i100)',err=61) qqcd
           read(storeinput,*) qqcd
         elseif(identifier.eq.'qsoftcoll') then
           read(storeinput,'(i100)',err=61) qsoftcoll
           read(storeinput,*) qsoftcoll
         elseif(identifier.eq.'qalp') then
           read(storeinput,'(i100)',err=61) qalp
           read(storeinput,*) qalp

         elseif(identifier.eq.'fname(1)') then
           read(storeinput,'(a100)',err=61) fname(1)
           read(storeinput,*) fname(1)
         elseif(identifier.eq.'fname(2)') then
           read(storeinput,'(a100)',err=61) fname(2)
           read(storeinput,*) fname(2)
         elseif(identifier.eq.'fname(3)') then
           read(storeinput,'(a100)',err=61) fname(3)
           read(storeinput,*) fname(3)
         elseif(identifier.eq.'fname(4)') then
           read(storeinput,'(a100)',err=61) fname(4)
           read(storeinput,*) fname(4)

         elseif(identifier.eq.'qrecomb') then
           read(storeinput,'(i100)',err=61) qrecomb
           read(storeinput,*) qrecomb
         elseif(identifier.eq.'invrecomb') then
           read(storeinput,'(f50.50)',err=61) precomb(1)
           read(storeinput,*) precomb(1)
         elseif(identifier.eq.'precomb(1)') then
           read(storeinput,'(f50.50)',err=61) precomb(1)
           read(storeinput,*) precomb(1)

         elseif(identifier.eq.'mh') then
           read(storeinput,'(f50.50)',err=61) mh
           read(storeinput,*) mh
         elseif(identifier.eq.'1/alpha0') then
           read(storeinput,'(f50.50)',err=61) alpha0
           read(storeinput,*) alpha0
           alpha0 = 1d0/alpha0
         elseif(identifier.eq.'alphas') then
           read(storeinput,'(f50.50)',err=61) alphas
           read(storeinput,*) alphas
         elseif(identifier.eq.'gf') then
           read(storeinput,'(f50.50)',err=61) gf
           read(storeinput,*) gf
         elseif(identifier.eq.'mz') then
           read(storeinput,'(f50.50)',err=61) mzi
           read(storeinput,*) mzi
         elseif(identifier.eq.'mw') then
           read(storeinput,'(f50.50)',err=61) mwi
           read(storeinput,*) mwi
         elseif(identifier.eq.'gammaz') then
           read(storeinput,'(f50.50)',err=61) gzi
           read(storeinput,*) gzi
         elseif(identifier.eq.'gammaw') then
           read(storeinput,'(f50.50)',err=61) gwi
           read(storeinput,*) gwi
         elseif(identifier.eq.'me') then
           read(storeinput,'(f50.50)',err=61) ml(1)
           read(storeinput,*) ml(1)
         elseif(identifier.eq.'mmu') then
           read(storeinput,'(f50.50)',err=61) ml(2)
           read(storeinput,*) ml(2)
         elseif(identifier.eq.'mtau') then
           read(storeinput,'(f50.50)',err=61) ml(3)
           read(storeinput,*) ml(3)
         elseif(identifier.eq.'md') then
           read(storeinput,'(f50.50)',err=61) mqm(1)
           read(storeinput,*) mqm(1)
         elseif(identifier.eq.'mu') then
           read(storeinput,'(f50.50)',err=61) mqp(1)
           read(storeinput,*) mqp(1)
         elseif(identifier.eq.'ms') then
           read(storeinput,'(f50.50)',err=61) mqm(2)
           read(storeinput,*) mqm(2)
         elseif(identifier.eq.'mc') then
           read(storeinput,'(f50.50)',err=61) mqp(2)
           read(storeinput,*) mqp(2)
         elseif(identifier.eq.'mb') then
           read(storeinput,'(f50.50)',err=61) mqm(3)
           read(storeinput,*) mqm(3)
         elseif(identifier.eq.'mt') then
           read(storeinput,'(f50.50)',err=61) mqp(3)
           read(storeinput,*) mqp(3)

         elseif(identifier.eq.'qllfsr') then
           read(storeinput,'(i100)',err=61) qllfsr
           read(storeinput,*) qllfsr
         elseif(identifier.eq.'deltae') then
           read(storeinput,'(f50.50)',err=61) deltae
           read(storeinput,*) deltae
         elseif(identifier.eq.'deltath') then
           read(storeinput,'(f50.50)',err=61) deltath
           read(storeinput,*) deltath
         elseif(identifier.eq.'ranshift') then
           read(storeinput,'(i100)',err=61) ranshift
           read(storeinput,*) ranshift
         elseif(identifier.eq.'qsoftcoll') then
           read(storeinput,'(i100)',err=61) qsoftcoll
           read(storeinput,*) qsoftcoll
         elseif(identifier.eq.'qhist') then
           read(storeinput,'(i100)',err=61) qhist
           read(storeinput,*) qhist
         elseif(identifier.eq.'moreout') then
           read(storeinput,'(i100)',err=61) moreout
           read(storeinput,*) moreout
         elseif(identifier.eq.'massmapping') then
           read(storeinput,'(f50.50)',err=61) massmapping
           read(storeinput,*) massmapping
         elseif(identifier.eq.'mininvmass') then
           read(storeinput,'(f50.50)',err=61) mininvmass
           read(storeinput,*) mininvmass
         elseif(identifier.eq.'nfsrfact') then
           read(storeinput,'(f50.50)',err=61) nfsrfact
           read(storeinput,*) nfsrfact
         elseif(identifier.ne.' ') then
           write(*,*) 'identifier not known:  ',identifier
           write(*,*) 'identifier not known:  ',orgidentifier
           stop
         endif
         goto 50
       endif

 52    write(*,*) 'input not known',inputline
       stop
 61    write(*,*) 'input not appropriate for identifier ''',
     &     identifier(1:isep1-1),''' : ',storeinput
       stop

 51    continue

      end

************************************************************************
*
*     initialize process etc.
*
************************************************************************
      subroutine init_main
      implicit none
      integer i,i1,i2
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      include 'COMMON/techpar.inc'
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width
c unit
      real*8 unit
      common/unit/unit
c random generator
      integer rangen,ranshift
      common/random/rangen,ranshift
c smearing
      integer smear
      real*8 smrat
      common/smearing/smrat,smear
c factorization scale
      real*8 fsrfact,nfsrfact
      common/fsrfact/fsrfact,nfsrfact
c output
      character*90 outputfile
      common/outputfile/outputfile
c cuts
      real*8 ecut(5),scut(5,5),ccut(5,5),precomb(3)
      integer qcuts
      common/cuts/ecut,scut,ccut,precomb,qcuts

c default parameters, can be reset by values given in inputfile
      mh       = 130d0                ! Higgs boson mass
      alpha0  = 1d0/137.03599911d0    ! alpha(0)
      alphas  = 0.1187d0              ! strong coupling constant
      gf      = 1.16637D-5            ! Fermi constant
      mzi     = 91.1876d0             ! Z boson mass in pdg scheme
      mwi     = 80.425d0              ! W boson mass in pdg scheme
      gzi     = 2.4952d0              ! Z boson width in pdg scheme
      gwi     = 2.124d0               ! W boson width in pdg scheme
      ml(1)   = 0.51099892d-3         ! electron mass
      ml(2)   = 105.658369d-3         ! muon mass
      ml(3)   = 1.77699d0             ! tau mass
      mqm(1)  = 0.066d0               ! d-quark mass
      mqp(1)  = 0.066d0               ! u-quark mass
      mqm(2)  = 0.066d0               ! s-quark mass
      mqp(2)  = 1.2d0                 ! c-quark mass
      mqm(3)  = 4.3d0                 ! b-quark mass
      mqp(3)  = 174.3d0               ! t-quark mass
c default options, can be reset via inputfile
      outputfile = ''  ! name of outputfile
      nevents = 10000000         ! number of weighted events
      nunwevents = 0             ! number of unweighted events
      contrib = 1          ! contrib: 1=best 2=IBA 3=Born
                           ! more options: 4=Brems 5=no virt.corr.&endp.
                           !              6=only virt.corr.&endp. 7=FSR
      qsoftcoll = 1        ! qsoftcoll: 1=subtraction, 2=slicing
c--- fermions: e,mu,nue,num,dq,uq,sq,cq,anti-e,...
c--- order: e.g. nue anti-e e anti-nue, uq anti-uq e anti-e, ...
      fname(1) = "e"         ! fermion 1
      fname(2) = "anti-e"    ! fermion 2
      fname(3) = "mu"        ! fermion 3
      fname(4) = "anti-mu"   ! fermion 4
      qrecomb = 1    ! qrecomb: 0= no recombination, 1=photon recombination
      precomb(1)=5d0 ! recombination condition (photon-fermion invariant mass)
      qqcd = 1       ! qqcd: 0=EW, 1=EW+QCD, 2=QCD, 3=a, 4=b, 5=c, 6=d
      qllfsr = 0     ! qllfsr: 0=no 1=expon 2=no expon 3/4=for IBA(exp/no exp)

c output
      nout=2
c more detailed output (for moreout=1)
      moreout=0
c maximum number of error messages
      maxout=0
c calculate virtual corr. only every nvirt'th time
      nvirt(1) = 100
      nvirt(2) = 20  !if nvirt(2).lt.nvirt(1), QCD has to be calculated first!
      if (qsoftcoll.eq.2) then
        nvirt(1) = 100
        nvirt(2) = 100
      endif
c ranshift: shift for random numbers (different seed)
      ranshift = 0
c fill histograms or don't (1 or 0)
      qhist = 1
c smearing of histograms
      smear = 0
c      smrat = 5d-1 !# on average 'smrat' of events are distributed over 2 bins
c alpha parametrization scheme
      qalp = 1
c factorization scale for FSR (x*MH)^2
      nfsrfact = 1d0
c fast calculation of loop integrals
      qfast = 1
c output in MeV
      unit = 1d3
c random generator (1:from RacoonWW, 2:RCarry)
      rangen = 2
c width scheme
      qprop = 1
c complex mass scheme
      qcomp = 1
c virtual part of corrections (qreal=0:with virtual part, qreal=1:neglect)
      qreal = 0
c bosonic and fermionc contributions to virtual corrections
c not applicable for wwidth and zwidth and corrections to decays in
c  NWA because all corrections needed to calculated W/Z width
      qbos  = 1
      qferm = 1
c parameter in h-function (mapping)
      massmapping = 1d-6
c cut for real corr.(min.*CMenergy)^2
      mininvmass = 1d-4
c cut for nlo/lo weight ratio, events with larger ratios are discarded
      weightcut = 1d6
c debugging output
      debug = 0

c generic particle names for phase-space generator
c different names than in RacoonWW and CofferGaGa
      gname(0)='XXX'
      gname(1)='e'
      gname(2)='mu'
      gname(3)='tau'
      gname(4)='nue'
      gname(5)='num'
      gname(6)='nut'
      gname(7)='dq'
      gname(8)='sq'
      gname(9)='bq'
      gname(10)='uq'
      gname(11)='cq'
      gname(12)='tq'
      do i = 1,12
        gname(-i) = gname(i)
        gname(-i) = 'anti-'//gname(i)
      enddo
      gname(13)='W-'
      gname(-13)='W+'
      gname(14)='Z0'
      gname(15)='ga'
      gname(16)='H0'
      do i = 14,15
        gname(-i) = gname(i)
      enddo
      gname(17)='   '

c polarizations of fermions and photon in Bremsstrahlungs process
      do i=1,5
        pol(i)=0
      enddo

      end

**************************************************************
*
*     check consistency of floating point behaviour
*
**************************************************************
      subroutine check_fp_testfunc(m2, m)
      implicit none
      real*8 m2,m
      real*8 mm2

      mm2 = m*m
      if (mm2 .ne. m2) then
         write(*,*) 'Inconsistent floating point behaviour (testfunc).'
         write(*,*) 'The floating point unit must be set to 64bit mode.'
         stop
      endif
      end

      subroutine check_fp_testfunc2(x)
      implicit none
      real*8 x

      if (1d0/x.ne.0.1d0) then
         write(*,*) 'Inconsistent floating point behaviour (testfunc2).'
         write(*,*) 'The floating point unit must be set to 64bit mode.'
         stop
      endif
      end

      subroutine check_fp_consistency
      implicit none
      real*8 ma,ma2

      ma = 80.41d0
      ma2 = ma*ma
      call check_fp_testfunc(ma2,ma)
      call check_fp_testfunc2(10d0)
      end

**************************************************************
*
*     check consistency of parameters
*
**************************************************************
      subroutine check_parameters
      implicit none
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      integer i1,i2,anti,cons
      logical exhist1,exhist2
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width

      call check_fp_consistency()
c inconsistent/unsupported options
      if (nunwevents.gt.0) then
         write(*,*)'Unweighting not supported in this version.'
         stop
      endif
      if (qsoftcoll.ne.1) then
         write(*,*)'Slicing not supported in this version.'
         stop
      endif
      if (qllfsr.ne.0) then
         write(*,*)'LL FSR not supported in this version.'
         stop
      endif
      if (nunwevents.gt.0.and.qsoftcoll.eq.1.and.contrib.eq.1) then
         write(*,*)'Unweighting only supported for slicing method',
     &        ' (qsoftcoll=2) in this version.'
         stop
      endif
      if (nevents.eq.0) then
         write(*,*)'The number of weighted events can never be 0'
         stop
      endif
      if (color.ne.1.and.qsoftcoll.eq.2.and.contrib.eq.1) then
         write(*,*)'For semi-leptonic and hadronic final states please',
     &        ' use the subtraction method (qsoftcoll=1).'
         stop
      endif
      if (nevents.eq.0) then
         write(*,*)'The number of weighted events can never be 0'
         stop
      endif
      if (nunwevents.gt.0.and.qllfsr.gt.0) then
         write(*,*)'LL FSR not supported in combination with',
     &        ' unweighted events in this version, qllfsr=',qllfsr,'.'
         stop
      endif
      if (contrib.eq.2.and.qllfsr.eq.1.or.qllfsr.eq.2) then
         write(*,*)'Please use qllfsr=3/4 in combination with IBA',
     &        'otherwise O(alpha) is subtracted.'
         stop
      endif
      if(qcomp.eq.1.and.qprop.ne.1)then
        write(*,*)'Incompatible options: qprop=',qprop,',qcomp=',qcomp
        stop
      endif

c consistency checks
c check charge conservation
      if (int(10d0*(q(1)-q(2)+q(3)-q(4))).ne.0d0) then
         write(*,*)'Check final state, charge is not conserved'
         stop
      endif
c check if fermion names exist and for correct (anti-)particles
      do i1 = 1,4
        cons = 0
        do i2 = 1,12
          anti = (-1)**(i1-1)  !particle(i1=1,3) or anti-particle(i1=2,4)
          if (fname(i1).eq.gname(anti*i2)) cons = 1
        enddo
        if (cons.eq.0) then
          write(*,*)'Wrong particle name or wrong order of'
     &          ,' (anti-)particles'
          stop
        endif
      enddo
c check correct order of quarks and generations
      if ((quark(1).ne.quark(2)).or.(quark(3).ne.quark(4)).or.
     &     (gen(1).ne.gen(2)).or.(gen(3).ne.gen(4))) then
        write(*,*)'Wrong order of particles or reaction not allowed'
        stop
      endif
c check isospin
      if (((iso(1).eq.iso(2)).and.(iso(3).ne.iso(4))).or.
     &    ((iso(1).ne.iso(2)).and.(iso(3).eq.iso(4)))) then
        write(*,*)'Reaction not allowed'
        stop
      endif
      if ((iso(1).ne.iso(2)).and.(iso(1).ne.1.or.iso(2).ne.-1
     &     .or.iso(3).ne.-1.or.iso(4).ne.1)) then
        write(*,*)'Wrong order of particles '
     &        ,'(CC order for mixed rections)'
        stop
      endif
      if (iso(1).lt.iso(3)) then
        write(*,*)'Wrong order of particles '
     &        ,'(Isospin: +1 +1 -1 -1)'
        stop
      endif

      end

************************************************************************
*
*     fill common blocks with values from input_param/inputfile
*     fix derived parameters
*         call init_process and input_param and input before
*
************************************************************************
      subroutine fix_param
      implicit none
      integer i,i1,i2,anti,temp_bos,temp_ferm
      real*8 wwidth,zwidth,dummy,ran1,ran_rcarry
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      include 'COMMON/xparam.inc'
c couplings
      complex*16 cvf(-1:1,3,4),cvs(3)
      common/couplings/cvf,cvs
c boson masses
      real*8 mb(4),mb2(4),gb(4)
      common/bosmass/mb,mb2,gb
c particle names
      real*8 mass(-17:17),width(-17:17)
      character*8 ename(-7:7),gname(-17:17)
      common/particles/ename,gname,mass,width
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
c renormalization scale
      real*8 mudim2
      common/uv/mudim2
c fermion masses as flag (i*1d-20)
      complex*16 cme2,cmf2(5)
      real*8 me2,rme2,fmass_mf2(5),rmf2(5)
      common/fmass/me2,rme2,cme2,fmass_mf2,rmf2,cmf2
c fermion charges of isospin duplets
      real*8 qf(4)
      common/fcharge/qf
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c factorization scale
      real*8 fsrfact,nfsrfact
      common/fsrfact/fsrfact,nfsrfact
c fitted constant for IBA
      real*8 chww,chzz
      common/IBA/chww,chzz
c random generator
      integer rangen,ranshift
      common/random/rangen,ranshift

c fixed quantities
      pi=4d0*datan(1d0)
c gauge boson width, 0=LO (used for Born preciction)
c   1=NLO (used for one-loop predictions, also for Born-part of one-loop)
      if (contrib.eq.3) then
         qwidth = 0
      else
         qwidth = 1
      endif
c conversion from input pdg W/Z masses to pole masses
      mz = mzi / dsqrt(1d0 + gzi**2/mzi**2)
      mw = mwi / dsqrt(1d0 + gwi**2/mwi**2)
c abbreviations
      mw2 = mw**2
      mz2 = mz**2
      mh2 = mh**2
      do i=1,3
        ml2(i)=ml(i)**2
        mqp2(i)=mqp(i)**2
        mqm2(i)=mqm(i)**2
      enddo
c weak mixing angle
      cw = mw/mz
      cw2=cw**2
      sw2=1d0-cw2
      sw=dsqrt(sw2)
c alpha parametrization scheme
      if(qalp.eq.0)then
	alpha=alpha0
      elseif(qalp.eq.1)then
        alpha=dsqrt(2d0)*GF/pi*mw2*sw2
      elseif(qalp.eq.2)then
	alpha=alphaz
      endif
      el=dsqrt(4d0*pi*alpha)
c center of mass energy
      CMenergy2 = mh2 ! Higgs mass
c factorization scale
      fsrfact = (nfsrfact*mh)**2
c photon mass
      lambda = 1d0
      lambda2 = lambda**2
c renormalization scale mu^2 in dimensional regularization
      mudim2 = 1d0
c slicing cuts (already in inputfile)
      deltaE  = 4d-3
      deltaTh = 3d-2
c define particle properties
      do i1 = 1,12
        do i2 = 1,4
          anti = (-1)**(i2-1)               !f1/f3:fermion, f2/f4:anti-fermion
          if (fname(i2).eq.gname(anti*i1)) then
            gen(i2) = mod(i1-1,3)+1         !generation of fermion
            iso(i2) = mod((i1-1)/3,2)*2-1   !isospin of fermion
            quark(i2) = (i1-1)/6            !quark:1, no quark:0
          endif
        enddo
      enddo
c elmg. charges of fermions
      do i=1,4
         q(i) = (iso(i)-1d0)/2d0+quark(i)*2d0/3d0
      enddo
c color factor
      color = 3**(quark(1)+quark(3))
c type of process
      if (iso(1).gt.iso(2)) then
        cc = 1
        if (gen(1).eq.gen(3).and.quark(1).eq.quark(3)) then
          nc = 1
        else
          nc = 0
        endif
      else
        cc = 0
        nc = 1
      endif
c calculate renormalization constants (real)
      call calcrenc
      call calcrenc_ferm
      call calcrenc_bgf
c vector boson width
      temp_bos = qbos
      temp_ferm = qferm
      qbos = 1
      qferm = 1
      if (qwidth.eq.0) then
         gw = wwidth(0,0)       !0:LO width
         gz = zwidth(0,0)
      else
         gw = wwidth(1,1)       !1:NLO width (including QCD corrections)
         gz = zwidth(1,1)
      endif
      qbos = temp_bos
      qferm = temp_ferm
      gt = GF*(mqp2(3)-mw2)**2*(mqp2(3)+2d0*mw2)
     &         /(8d0*pi*dsqrt(2d0)*mqp(3)**3)
c complex parameters
      xmz2 = dcmplx(mz2,-mz*gz)
      xmz = cdsqrt(xmz2)
      xmw2 = dcmplx(mw2,-mw*gw)
      xmw = cdsqrt(xmw2)
      xmh2 = dcmplx(mh2,0d0)
      xmh = cdsqrt(xmh2)
      do i=1,3
        xml2(i)  = ml2(i)
        xmqm2(i) = mqm2(i)
        xmqp2(i) = mqp2(i)
      enddo
      xmqp2(3) = dcmplx(mqp2(3),-mqp(3)*gt)
      null = dcmplx(0d0,0d0)
      xcw = cdsqrt(dcmplx(mw2,-mw*gw)
     &           /dcmplx(mz2,-mz*gz) )
      xcw2=xcw**2
      xsw2=1d0-xcw2
      xsw=cdsqrt(xsw2)
c boson masses and widths
      mb(1) = 0d0
      mb2(1) = 0d0
      gb(1) = 0d0
      mb(2) = mz
      mb2(2) = mz2
      gb(2) = gz
      mb(3) = mw
      mb2(3) = mw2
      gb(3) = gw
      mb(4) = mh
      mb2(4) = mh2
      gb(4) = 0d0

c external fermion masses and masses for loops connected to external fermions
      do i=1,4
        if (quark(i).eq.0) then
          if (iso(i).eq.-1) then
            mf(i) = ml(gen(i))
            mf2(i) = ml2(gen(i))
          else
            mf(i) = 0d0
            mf2(i) = 0d0
          endif
        else
          if (iso(i).eq.-1) then
            mf(i) = mqm(gen(i))
            mf2(i) = mqm2(gen(i))
          else
            mf(i) = mqp(gen(i))
            mf2(i) = mqp2(gen(i))
          endif
        endif
      enddo

c external particle names defined for subroutines check_topology
      ename(1)='H0'
      ename(2)='XXX'
      do i = 1,4
        ename(i+2) = fname(i)
      enddo
      ename(7) = 'ga'  ! photon
      do i=3,6
        if (ename(i)(1:4).eq.'anti') then
          ename(-i) = ename(i)(6:8)
        else
          ename(-i) = 'anti-'//ename(i)
        endif
      enddo
      ename(-7) = ename(7)

c electron mass as flag
      mxe = 1d-15
      mxe2 = mxe**2
c external fermion masses as flag (i*1d-20)
c convention:(f1,f2)/(f3,f4) form isospin duplet
c ->fmass_mf(1)/fmass_mf(4):up-type, fmass_mf(2)/fmass_mf(3):down-type
      me2  = ml2(1)
      rme2 = me2
      cme2 = me2
      fmass_mf2(5) = me2
      if (quark(1).eq.0) then
        fmass_mf2(1) = 0d0
        fmass_mf2(2) = ml2(gen(2))
      else
        fmass_mf2(1) = mqp2(gen(1))
        fmass_mf2(2) = mqm2(gen(2))
      endif
      if (quark(3).eq.0) then
        fmass_mf2(3) = ml2(gen(3))
        fmass_mf2(4) = 0d0
      else
        fmass_mf2(3) = mqm2(gen(3))
        fmass_mf2(4) = mqp2(gen(4))
      endif
      do i = 1,5
        rmf2(i) = i*1d-20
        cmf2(i) = rmf2(i)
      enddo
c fermion charges of isospin duplets
      qf(1) = quark(1)*2d0/3d0
      qf(2) = -1d0 + quark(2)*2d0/3d0
      qf(3) = -1d0 + quark(3)*2d0/3d0
      qf(4) = quark(4)*2d0/3d0
c couplings of fermions to gauge bosons
      do i=1,4
        cvf(-1,1,i) = -el*q(i)
        if (qcomp.eq.0) then
          cvf(+1,1,i) = -el*q(i)
          cvf(-1,2,i) = -el*(q(i)*sw/cw-iso(i)/2d0/sw/cw)
          cvf(+1,2,i) = -el* q(i)*sw/cw
          cvf(-1,3,i) =  el/dsqrt(2d0)/sw
        elseif (qcomp.eq.1) then
          cvf(+1,1,i) = -el*q(i)
          cvf(-1,2,i) = -el*(q(i)*xsw/xcw-iso(i)/2d0/xsw/xcw)
          cvf(+1,2,i) = -el* q(i)*xsw/xcw
          cvf(-1,3,i) =  el/dsqrt(2d0)/xsw
        endif
        cvf(+1,3,i) = 0d0
      enddo
c coupling of higgs to vector bosons
      if (qcomp.eq.0) then
        cvs(3)=el*mw/sw        ! coupling Higgs-W-W
        cvs(2)=el*mw/sw/cw2    ! coupling Higgs-Z-Z
      elseif (qcomp.eq.1)then
        cvs(3)=el*xmw/xsw          ! coupling Higgs-W-W
        cvs(2)=el*xmw/xsw/xcw2      ! coupling Higgs-Z-Z
      endif
c particle masses and widths for phase-space generator
      do i=-16,16
        mass(i)=0d0
        width(i)=0d0
      enddo
      mass(-13) = mw
      width(-13)= gw
      mass(-14) = mz
      width(-14)= gz
      do i=1,16
        mass(i)=mass(-i)
        width(i)=width(-i)
      enddo
c calculate renormalization constants (complex)
      call xcalcrenc
      call xcalcrenc_ferm
      call xcalcrenc_bgf
c fitted constant for IBA
      chww = 0.04
      chzz = 0.03
c shift of random numbers for uncorrelated results
      do i = 1,ranshift
        if (rangen.eq.1) then
          dummy = ran1()
        elseif (rangen.eq.2) then
          dummy = ran_rcarry()
        endif
      enddo

      end

**************************************************************
*
*     initialize separation cuts and photon recombination
*
**************************************************************
      subroutine init_cuts
      implicit none
      integer i1,i2
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
c inputcuts
      real*8 ecutp,ecutl,ecutq,scutqq,ccutpl,ccutpq,ccutll,ccutqq,ccutlq
      common/inputcuts/ecutp,ecutl,ecutq,scutqq,ccutpl,ccutpq,ccutll
     &     ,ccutqq,ccutlq
c cuts
      real*8 ecut(5),scut(5,5),ccut(5,5),precomb(3)
      integer qcuts
      common/cuts/ecut,scut,ccut,precomb,qcuts

      qcuts = 0   ! no separation cuts
      if (qcuts.eq.0)then
        ecutp = 0d0
        ecutl = 0d0
        ecutq = 0d0
        scutqq = 0d0
        ccutpl = 0d0
        ccutpq = 0d0
        ccutll = 0d0
        ccutqq = 0d0
        ccutlq = 0d0
c photon cuts for 1->5 Born cross section
      elseif (qcuts.eq.1) then
        ecutp=10d0             ! photon energy cut
        ecutl=0d0              ! charged-lepton energy cut
        ecutq=0d0              ! quark energy cut
        scutqq=0d0             ! quark-quark invariant mass cut
        ccutpl=5d0             ! angular cut between photon and charged lepton
        ccutpq=5d0             ! angular cut between photon and quark
        ccutll=0d0             ! angular cut between charged leptons
        ccutqq=0d0             ! angular cut between quarks
        ccutlq=0d0             ! angular cut between charged lepton and quark
c cuts for Born cross section (1->4, or 1->5)
      elseif (qcuts.eq.2) then
        ecutp=10d0             ! photon energy cut
        ecutl=10d0             ! charged-lepton energy cut
        ecutq=10d0             ! quark energy cut
        scutqq=10d0            ! quark-quark invariant mass cut
        ccutpl=5d0             ! angular cut between photon and charged lepton
        ccutpq=5d0             ! angular cut between photon and quark
        ccutll=5d0             ! angular cut between charged leptons
        ccutqq=0d0             ! angular cut between quarks
        ccutlq=5d0             ! angular cut between charged lepton and quark
c cuts for radiative corrections
      elseif(qcuts.eq.3)then
        ecutp=0d0              ! photon energy cut
        ecutl=50d0             ! charged-lepton energy cut
        ecutq=50d0             ! quark energy cut
        scutqq=50d0            ! quark-quark invariant mass cut
        ccutpl=0d0             ! angular cut between photon and charged lepton
        ccutpq=0d0             ! angular cut between photon and quark
        ccutll=55d0             ! angular cut between charged leptons
        ccutqq=0d0             ! angular cut between quarks
        ccutlq=55d0             ! angular cut between charged lepton and quark
      endif
c initialize arrays for cuts
      do i1=1,5
        ecut(i1)=0d0
        do i2=1,5
          scut(i1,i2)=0d0
          ccut(i1,i2)=0d0
        enddo
      enddo
c filling cut parameters, note that processpar:1..4, cuts:1..7
      if (qcuts.ne.0) then
        ecut(5)=ecutp
        do i1=1,4
          if (quark(i1).eq.0.and.iso(i1).lt.0) then     ! charged leptons
            ecut(i1)=ecutl
            ccut(i1,5)=ccutpl
          elseif (quark(i1).eq.1) then                  ! quarks
            ecut(i1)=ecutq
            ccut(i1,5)=ccutpq
          endif
        do i2=i1+1,4
          if (quark(i1).eq.1.and.quark(i2).eq.1) then
            scut(i1,i2)=scutqq
            ccut(i1,i2)=ccutqq
          elseif (quark(i1).eq.0.and.iso(i1).lt.0.and.
     &           quark(i2).eq.0.and.iso(i2).lt.0) then
            ccut(i1,i2)=ccutll
          elseif ((quark(i1).eq.0.and.iso(i1).lt.0.and.quark(i2)
     &        .eq.1).or.(quark(i1).eq.1.and.quark(i2).eq.0.and.
     &        iso(i2).lt.0)) then
            ccut(i1,i2)=ccutlq
          endif
        enddo
        enddo
      endif

c calculating cut parameters
c angle -> cos(angle)
      do i1=1,5
      do i2=1,5
        ccut(i1,i2)=dcos(ccut(i1,i2)*pi/180d0)
      enddo
      enddo
c mass -> mass squared
      do i1=1,5
      do i2=1,5
        scut(i1,i2)=scut(i1,i2)**2
      enddo
      enddo
c filling whole common block
      do i1=1,5
      do i2=i1+1,5
        ccut(i1,i2)=min(ccut(i1,i2),ccut(i2,i1))
        scut(i1,i2)=max(scut(i1,i2),scut(i2,i1))
      enddo
      enddo

      end

**************************************************************
*
*     initialize fast evaluation of tensor integrals
*       (calculate virtual corrections for 10 ph-sp points)
**************************************************************
      subroutine init_fast
      implicit none
      include 'COMMON/options.inc'
      include 'COMMON/max.inc'
      integer i,j,nocut,disc,ngen,nparticles
      real*8 random(18),inmom(0:3),outmom(0:3,5),distmom(0:3,5)
      real*8 dens(maxchan),denstotal,sij(7,7)
      real*8 dM2_virt_EW,dM2_virt_QCD,dummy,ran1,ran_rcarry
c center of mass energy
      real*8 CMenergy2
      common/CMenergy2/CMenergy2
c random generator
      integer rangen,ranshift
      common/random/rangen,ranshift

      if (qfast.eq.0) then
        call fastoff
        do j = 1,12 ! same random numbers independent of qfast
          do i = 1,18
            if (rangen.eq.1) then
              random(i) = ran1()
            elseif (rangen.eq.2) then
              random(i) = ran_rcarry()
            endif
          enddo
        enddo
      else
        ngen = 1
        nparticles = 4
        do j = 1,12 !2 more, in case of discarded event
          do i = 1,18
            if (rangen.eq.1) then
              random(i) = ran1()
            elseif (rangen.eq.2) then
              random(i) = ran_rcarry()
            endif
          enddo
c generate phase-space point
          inmom(0) = dsqrt(CMenergy2) ! Higgs mass
          do i = 1,3
            inmom(i) = 0d0
          enddo
          call phasespace(random,ngen,inmom,outmom,distmom
     &         ,sij,dens,denstotal,nparticles,nocut,disc)
          call setprods(outmom(0,1),nparticles)
          call calcsme()
          call fastinit(7)
          dummy = dM2_virt_QCD() !QCD before EW, keep same order!
          dummy = dM2_virt_EW()
        enddo
      endif

      end

**************************************************************
*
*     begin ouput, number of error messages, input parameters,...
*
**************************************************************
      subroutine begin_output
      implicit none
      include 'COMMON/output.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
      include 'COMMON/techpar.inc'
      integer i1,i2,length
      real*8 narrwidappr,nwa,nwaborn,nwacorr,br
      real*8 width_wff,width_zff,width_hww,width_hzz,wwidth,zwidth
      integer values(8)
      integer fehler
      character date*8,time*10,zone*5
c cuts
      real*8 ecut(5),scut(5,5),ccut(5,5),precomb(3)
      integer qcuts
      common/cuts/ecut,scut,ccut,precomb,qcuts
c slicing cuts
      real*8 deltaE,deltaTh
      common/slicuts/deltaE,deltaTh
c random generator
      integer rangen,ranshift
      common/random/rangen,ranshift
c IR
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
c renormalization scale
      real*8 mudim2
      common/uv/mudim2
c outputfile
      character*90 outputfile
      common/outputfile/outputfile
c smearing
      integer smear
      real*8 smrat
      common/smearing/smrat,smear
c fitted constant for IBA
      real*8 chww,chzz
      common/IBA/chww,chzz

      open(unit=nout,file=outputfile,status='unknown',iostat=fehler)
      if (fehler.ne.0) then
         close(nout)
	 outputfile='plot'
         nout=6
      endif
      write(nout,*)""
      write(nout,*)
     &      '======================================================='
      write(nout,*)' This is Prophecy4f, version 1.0                '
      write(nout,*)' authors: A. Denner, S. Dittmaier,              '
      write(nout,*)'          A. Mueck. M.M. Weber                  '
      write(nout,*)
     &      '======================================================='
      write(nout,*)""

      write(nout,*)' Input parameters:'
      write(nout,*)' -----------------'
      write(nout,'(a19,f12.8,a20,f8.4)')'    1/alpha(0) = 1/',1d0/alpha0
     &     ,'    alphas = ',alphas
      write(nout,'(a17,g12.7,a7)'),'GF = ',gf,' GeV^-2'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4,a10,f9.4,a4)')
     &     ,'MW = ',mwi,' GeV','     MZ = ',mzi,' GeV'
     &     ,'     MH = ',mh,' GeV'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4)')
     &     ,'GW = ',gwi,' GeV','     GZ = ',gzi,' GeV'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4,a10,f9.5,a4)')
     &     ,'me = ',ml(1)*1d3,' MeV','    mmu = ',ml(2),' GeV'
     &     ,'   mtau = ',ml(3),' GeV'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4,a10,f9.5,a4)')
     &     ,'mu = ',mqp(1),' GeV','     mc = ',mqp(2),' GeV'
     &     ,'     mt = ',mqp(3),' GeV'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4,a10,f9.5,a4)')
     &     ,'md = ',mqm(1),' GeV','     ms = ',mqm(2),' GeV'
     &     ,'     mb = ',mqm(3),' GeV'
      write(nout,*)
c derived parameters
      write(nout,*)' Derived parameters:'
      write(nout,*)' -----------------'
      write(nout,'(a19,f12.8)')'    1/alpha = 1/',1d0/alpha
      write(nout,*) '   W/Z pole masses and widths'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4)')
     &     ,'MW = ',mw,' GeV','     MZ = ',mz,' GeV'
      write(nout,'(a17,f9.5,a4,a10,f9.5,a4,a10,f9.5,a4,a10,f9.5,a4)')
     &     'GW = ',gw,' GeV','     GZ = ',gz,' GeV',
     &     '     gt = ',gt,' GeV'
      write(nout,*)
c options
      write(nout,*)' Options:'
      write(nout,*)' --------'
      write(nout,*)'  Process:  H ->   '
     &    ,fname(1)(1:length(fname(1))),' ',fname(2)(1:length(fname(2)))
     &    ,' ',fname(3)(1:length(fname(3))),' '
     &    ,fname(4)(1:length(fname(4))),' (+Photon/Gluon)'
      write(nout,*)
      write(nout,'(a33,i9)'),'   number of weighted events   = ',nevents
      write(nout,'(a33,i9)'),'   number of unweighted events = ',
     &     nunwevents
      write(nout,*)
      if(contrib.eq.1)then
         write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : Born cross section + radiative corrections'
      elseif(contrib.eq.2)then
        write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : Improved Born approximation'
      elseif(contrib.eq.3)then
        write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : Born cross section'
      elseif(contrib.eq.4)then
        write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : Bremsstrahlungs process H->4f+Photon'
      elseif(contrib.eq.5)then
        write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : slicing endpoint only'
      elseif(contrib.eq.6)then
        write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : virtual + subtraction endpoint only'
      elseif(contrib.eq.7)then
        write(nout,'(a16,i2,(a))'),'contrib =',contrib
     &        ,' : FSR only'
      endif
      if (qqcd.eq.0) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : ONLY EW corrections'
      elseif (qqcd.eq.1) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : EW+QCD corrections'
      elseif (qqcd.eq.2) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : ONLY QCD corrections'
      elseif (qqcd.eq.3) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : ONLY QCD corrections, category a'
      elseif (qqcd.eq.4) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : ONLY QCD corrections, category b'
      elseif (qqcd.eq.5) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : ONLY QCD corrections, category c'
      elseif (qqcd.eq.6) then
         write(nout,'(a16,i2,(a))'),'qqcd =',qqcd
     &        ,' : ONLY QCD corrections, category d'
      endif
      if(qsoftcoll.eq.1)then
         write(nout,'(a16,i2,(a))'),'qsoftcoll =',qsoftcoll
     &        ,' : subtraction method'
      elseif(qsoftcoll.eq.2)then
        write(nout,'(a16,i2,(a))'),'qsoftcoll =',qsoftcoll
     &        ,' : slicing method'
      endif
      if (qllfsr.eq.1.or.qllfsr.eq.3) then
         write(nout,'(a16,i2,2(a))'),'qllfsr =',qllfsr
     &        ,' : LL FSR up to O(alpha^3) included '
     &        ,'(with exponentiation)'
      elseif (qllfsr.eq.2.or.qllfsr.eq.4) then
         write(nout,'(a16,i2,2(a))'),'qllfsr =',qllfsr
     &        ,' : LL FSR up to O(alpha^3) included without '
     &        ,'exponentiation'
      elseif (qllfsr.eq.5) then
         write(nout,'(a16,i2,(a))'),'qllfsr =',qllfsr
     &        ,' : LL FSR up to O(alpha^2) included without '
     &        ,'exponentiation'
      elseif (qllfsr.eq.6) then
         write(nout,'(a16,i2,(a))'),'qllfsr =',qllfsr
     &        ,' : LL FSR up to O(alpha) included without '
     &        ,'exponentiation'
      endif
c photon recombination
      write(nout,*)
      if(qrecomb.eq.0)then
        write(nout,'(a16,i2,(a))')'qrecomb =',qrecomb,
     &        ' : no photon recombination applied'
      elseif(qrecomb.eq.1)then
        write(nout,'(a16,i2,(a))')'qrecomb =',qrecomb,
     &        ' : photon recombination applied'
        write(nout,'(20x,"invariant mass(f,ph) > ",f9.5," GeV")')
     &      precomb(1)
      endif
c separation cuts
      write(nout,*)
      if(qcuts.eq.0)then
        write(nout,'(a16,i2,(a))')'qcuts =',qcuts,
     &        ' : no separation cuts'
      else
        write(nout,'(a16,i2,(a))')'qcuts =',qcuts,
     &        ': with separation cuts:'
        do i1=1,5
          if(ecut(i1).gt.0d0)then
            write(nout,'(20x,"energy(",i1,")  > ",f9.5," GeV")')
     &        i1,ecut(i1)
          endif
        enddo
        do i1=1,5
        do i2=i1+1,5
          if(ccut(i1,i2).lt.1d0)then
            write(nout,'(20x,"angle(",i1,",",i1,") > ",f9.5," deg")')
     &        i1,i2,dacos(ccut(i1,i2))/pi*180d0
          endif
        enddo
        enddo
        do i1=1,5
        do i2=i1+1,5
          if(scut(i1,i2).gt.0d0)then
            write(nout,'(20x,"mass(",i1,",",i1,")  > ",f9.5," GeV")')
     &        i1,i2,dsqrt(scut(i1,i2))
          endif
        enddo
        enddo
      endif
c other options/parameters
      if (smear.eq.1) then
         write(nout,'(a16,i2,(a))'),'smear =',smear
     &        ,' : smearing of histograms'
      endif
      if(contrib.eq.2)then
        write(nout,'(a16,f6.2,3x,(a))')'chww =',chww,
     &         ' : fitted constant for IBA'
        write(nout,'(a16,f6.2,3x,(a))')'chzz =',chzz,
     &         ' : fitted constant for IBA'
      endif
      write(nout,*)
c technical parameters
      if (qsoftcoll.eq.2) then
        write(nout,'(a16,g9.1,(a))')'DeltaE =',deltae,
     &       ' :  Slicing energy cut'
        write(nout,'(a16,g9.1,(a))')'DeltaTh =',deltaTh,
     &       ' :  Slicing angular cut'
      endif
c more detailed output
      if (moreout.eq.1) then
         if(qalp.eq.0)then
            write(nout,'(a16,i2,(a))'),'qalp =',qalp,
     &           ' : alpha(0)-scheme'
         elseif(qalp.eq.1)then
            write(nout,'(a16,i2,(a))'),'qalp =',qalp,
     &           ' : alpha_G_mu-scheme'
         elseif(qalp.eq.2)then
            write(nout,'(a16,i2,(a))'),'qalp =',qalp,
     &           ' : alpha(MZ)-scheme'
         endif
         if(qcomp.eq.0)then
            write(nout,'(a16,i2,(a))'),'qcomp =',qcomp
     &           ,' : NO complex mass scheme'
         elseif(qcomp.eq.1)then
            write(nout,'(a16,i2,(a))'),'qcomp =',qcomp
     &           ,' : complex mass scheme'
         endif
         write(nout,*)
         write(nout,*)'  calculate virtual correction every '
     &        ,nvirt(1),'(EW) ',nvirt(2),'(QCD) ','th time'
         write(nout,*)
         write(nout,'(a16,g9.1,(a))')'massmapping =',massmapping,
     &        ' :  parameter in h-function (mapping)'
         if (qsoftcoll.eq.1)
     &        write(nout,'(a16,g9.1,(a))')'mininvmass =',mininvmass,
     &        ' :  invariant-mass cut for real corrections'
         if (qsoftcoll.eq.1)
     &        write(nout,'(a16,g9.1,(a))')'sub4cut =',sub4cut,
     &        ' :  same binning for 4-Phsp subtraction functions'
         if (qsoftcoll.eq.1)
     &        write(nout,'(a16,g9.1,(a))')'sub5cut =',sub5cut,
     &        ' :  same binning for 5-Phsp photon-subtr. functions'
         if (qsoftcoll.eq.1)
     &        write(nout,'(a16,g9.1,(a))')'sub5cutgl =',sub5cutgl,
     &        ' :  same binning for 5-Phsp gluon-subtr. functions'
         write(nout,*)
         if (qfast.eq.0) then
            write(nout,'(a16,i2,(a))')'qfast =',qfast,
     &           ' : NO fast evaluation of tensor integrals'
         elseif (qfast.eq.1) then
            write(nout,'(a16,i2,(a))')'qfast =',qfast,
     &           ' :  fast evaluation of tensor integrals'
         endif
         if (rangen.eq.1) then
            write(nout,'(a16,i2,(a))')'rangen =',rangen,
     &           ' :  old Random number generator'
         elseif (rangen.eq.2) then
            write(nout,'(a16,i2,(a))')'rangen =',rangen,
     &           ' :  RCarry random number generator'
         endif
         write(nout,'(a16,i2,(a))')'ranshift =',ranshift,
     &        ' :  different random variables'
         write(nout,*)
         write(nout,'(a16,g9.1,(a))')'lambda =',lambda,
     &        ' : Photon mass'
         write(nout,'(a16,g9.1,(a))')'mudim2 =',mudim2,
     &        ' : UV scale'
      endif
c narrow width approximation
      write(nout,*)
      write(nout,*)' Narrow width approximation (incl EW+QCD):'
      write(nout,*)' -----------------------------------------'

      nwa = narrwidappr(1)
      nwaborn = narrwidappr(0)
      nwacorr = nwa - nwaborn

 2000 format((a),g12.6,(a))
      write(nout,2000)'  Partial width: ',nwa,' MeV'
      if (qwidth.eq.0) then
         write(nout,2000)'           Born: ',nwaborn,
     &        ' MeV (LO gauge-boson width)'
      else
         write(nout,2000)'           Born: ',nwaborn,
     &        ' MeV (NLO gauge-boson width)'
      endif
      write(nout,2000)'     Correction: ',nwacorr,' MeV'
      write(nout,*)

c Branching ratio includes QCD corrections
c Born: corrections included in denominator
c      if (nc.eq.1) then
c        br = width_Zff(q(1),q(2),gen(1),1,1)/gz
c        write(nout,*)' Branching ratio (Z->',fname(1)(1:length
c     &       (fname(1))),' ',fname(2)(1:length(fname(2))),') =',br
c        br = width_Zff(q(1),q(2),gen(1),0,1)/gz
c        write(nout,*)'   Born:',br
c        br = width_Zff(q(3),q(4),gen(3),1,1)/gz
c        write(nout,*)' Branching ratio (Z->',fname(3)(1:length(
c     &       fname(3))),' ',fname(4)(1:length(fname(4))),') =',br
c        br = width_Zff(q(3),q(4),gen(3),0,1)/gz
c        write(nout,*)'   Born:',br
c      elseif (cc.eq.1) then
c        br = width_Wff(q(1),q(2),1,1)/gw
c        write(nout,*)' Branching ratio (W->',fname(1)(1:length
c     &       (fname(1))),' ',fname(2)(1:length(fname(2))),') =',br
c        br = width_Wff(q(1),q(2),0,1)/gw
c        write(nout,*)'   Born:',br
c        br = width_Wff(q(3),q(4),1,1)/gw
c        write(nout,*)' Branching ratio (W->',fname(3)(1:length
c     &       (fname(3))),' ',fname(4)(1:length(fname(4))),') =',br
c        br = width_Wff(q(3),q(4),0,1)/gw
c        write(nout,*)'   Born:',br
c      endif

c date
cacb
c      call date_and_time(date,time,zone,values)
c      write(nout,*)
c      write(nout,44)date(1:4),date(5:6),date(7:8),time(1:2),time(3:4)
c     &              ,zone
c 44    FORMAT('started integration:',1x,(a),'-',(a),'-',(a),1x,(a),
c     &     ':',(a),1x,(a))
c      write(nout,*)

      if (maxout.gt.0) then
        write(nout,*)' Rejected events:'
        write(nout,*)' -----------------'
        write(nout,*)
      endif

      end

**************************************************************
*
*     write final output
*
**************************************************************
      subroutine write_output(average,sigma)
      implicit none
      include 'COMMON/max.inc'
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      integer i,values(8)
      character date*8,time*10,zone*5
      real*8 average,sigma
c sub contributions
      real*8 subav(maxsubc),subsi(maxsubc),submax(maxsubc)
      integer subch(maxsubc),subn(maxsubc),subrej(maxw)
      common/subcontr/subav,subsi,submax,subch,subn,subrej


      if (moreout.eq.1) then
         if (qsoftcoll.eq.1) then
            write(nout,'(a27,i8,a17)')' inv.mass < mininvmass: ',nmininv
     &           ,' events discarded'
            write(nout,'(a27,i8,a17)')' 1-z < sub4cut: ',nsub4
     &           ,' events discarded'
            write(nout,'(a27,i8,a17)')' E_gamma,y < sub5cut: ',nsub5
     &           ,' events discarded (multiple counting)'
            write(nout,'(a27,i8,a17)')' E_gluon,y < sub5cutgl: ',nsub5gl
     &           ,' events discarded (multiple counting)'
         endif
         write(nout,'(a27,i8,a17)')' energy < min. energy: ',nminen
     &        ,' events discarded'
         write(nout,*)'   (double counting possible)'
         do i=1,maxw
            if (subrej(i).ne.0) then
               write(nout,'((a),i2,(a),f6.2,(a))')'    #',i,': '
     &              ,dfloat(subrej(i))/nevents*100d0,'%'
            endif
         enddo
      endif
 1000 format((a),f18.10," MeV")
 1005 format("    event number",i10,", channel",i3)
      write(nout,*)
      write(nout,*)' Results:'
      write(nout,*)' -------'
      write(nout,*)
      write(nout,1000)'  Average             = ',average
      write(nout,1000)'  Standard deviation  = ',sigma
      write(nout,*)
      write(nout,*)' Subcontributions:'
      write(nout,*)' -----------------'
      if (qwidth.eq.0) then
         write(nout,*)' Born cross section (LO gauge-',
     &        'boson width):'
      else
         write(nout,*)' Born cross section (NLO gauge-',
     &        'boson width):'
      endif
      write(nout,1000)'   Average            = ',subav(1)
      write(nout,1000)'   Standard deviation = ',subsi(1)
      if (debug.ne.0) then
        write(nout,1000)'   largest event      = ',submax(1)
        write(nout,1005)subn(1),subch(1)
      endif
      if (subav(2).ne.0d0 .and. subav(4).ne.0d0) then
        write(nout,*)
        write(nout,*)' EW corrections:'
        write(nout,1000)'   Average            = ',subav(2) + subav(4)
        write(nout,1000)'   Standard deviation = ',
     &       dsqrt(subsi(2)**2 + subsi(4)**2)
      endif
      if (subav(3).ne.0d0 .and. subav(5).ne.0d0) then
        write(nout,*)
        write(nout,*)' QCD corrections:'
        write(nout,1000)'   Average            = ',subav(3) + subav(5)
        write(nout,1000)'   Standard deviation = ',
     &       dsqrt(subsi(3)**2 + subsi(5)**2)
      endif
      if (subav(2).ne.0d0 .and. debug.ne.0) then
        write(nout,*)
        write(nout,*)' EW Virtual corrections:'
        write(nout,1000)'   Average            = ',subav(2)
        write(nout,1000)'   Standard deviation = ',subsi(2)
        write(nout,1000)'   largest event      = ',submax(2)
        write(nout,1005)subn(2),subch(2)
      endif
      if (subav(3).ne.0d0 .and. debug.ne.0) then
        write(nout,*)
        write(nout,*)' QCD Virtual corrections:'
        write(nout,1000)'   Average            = ',subav(3)
        write(nout,1000)'   Standard deviation = ',subsi(3)
        write(nout,1000)'   largest event      = ',submax(3)
        write(nout,1005)subn(3),subch(3)
      endif
      if (subav(4).ne.0d0 .and. debug.ne.0) then
        write(nout,*)
        write(nout,*)' Photon Bremsstrahlung:'
        write(nout,1000)'   Average            = ',subav(4)
        write(nout,1000)'   Standard deviation = ',subsi(4)
        write(nout,1000)'   largest event      = ',submax(4)
        write(nout,1005)subn(4),subch(4)
      endif
      if (subav(5).ne.0d0 .and. debug.ne.0) then
        write(nout,*)
        write(nout,*)' Gluon Bremsstrahlung:'
        write(nout,1000)'   Average            = ',subav(5)
        write(nout,1000)'   Standard deviation = ',subsi(5)
        write(nout,1000)'   largest event      = ',submax(5)
        write(nout,1005)subn(5),subch(5)
      endif
      if (debug .ne. 0) then
         write(nout,'(a,i10)') 'nancount = ', nancount
         write(nout,'(a,i10)') 'ewcount  = ', ewcount
         write(nout,'(a,i10)') 'qcdcount = ', qcdcount
      endif
c date
cacb
c      call date_and_time(date,time,zone,values)
c      write(nout,*)
c      write(nout,44)date(1:4),date(5:6),date(7:8),time(1:2),time(3:4)
c     &              ,zone
c 44    FORMAT('finished weighted events:',1x,(a),'-',(a),'-',(a),1x,(a),
c     &     ':',(a),1x,(a))

      end

************************************************************************
*
*     write output of unweighting procedure
*
************************************************************************
      subroutine write_unweighting(genunwev,genwev,novershoot,negev,
     &     maxweight,glomaxweight)
      implicit none
      include 'COMMON/options.inc'
      include 'COMMON/output.inc'
      include 'COMMON/max.inc'
      integer values(8)
      character date*8,time*10,zone*5
      integer genunwev,genwev,novershoot,negev
      real*8 maxweight(maxw),glomaxweight

      if (nunwevents.eq.0) return
      write(nout,*)
      write(nout,*)' Unweighted events:'
      write(nout,*)' ------------------'
      if (genunwev.lt.nunwevents) write(nout,*) ' only ', genunwev,
     &     ' unweighted events, unweighting procedure inefficient'

      write(nout,'((a),i7,(a))') ' generated ',genunwev,
     &     ' unweighted events (including negative events)'
      write(nout,'((a),i10,(a))')'   by generating weighted events ',
     &     genwev,' times'
      write(nout,'((a),(a),i4)') ' number of weights larger than ',
     &     'largest weight from 1st cycle: ',novershoot
      write(nout,'((a),i5)') ' number of negative unweighted events: ',
     &     negev
      if (moreout.eq.1) then
         write(nout,*) ' maximal weights: ',
     &        maxweight
      endif
      write(nout,'((a),(a),g10.5)') ' maximal weight used for',
     &     ' generating unweighted events: ',glomaxweight

      write(nout,*)

c date
cacb
c      call date_and_time(date,time,zone,values)
c      write(nout,*)
c      write(nout,44)date(1:4),date(5:6),date(7:8),time(1:2),time(3:4)
c     &              ,zone
c 44    FORMAT('finished unweighted events:',1x,(a),'-',(a),'-',(a),1x,
c     &     (a),':',(a),1x,(a))

      close(nout)

      end

******************************************************************
      INTEGER FUNCTION LENGTH(STRING)
*Returns length of string ignoring trailing blanks
******************************************************************
      CHARACTER*(*) STRING
      DO 35, I = LEN(STRING), 1, -1
         IF(STRING(I:I) .NE. ' ') GO TO 40
35    CONTINUE
40    LENGTH = I
      END

