************************************************************************
*
*     initialize Les Houches Accord common block (hep-ph/0109068)
*       for unweighted events
*
************************************************************************
      subroutine init_hepup()
      implicit none
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      integer i
c Les Houches accord common block
      integer MAXNUP
      parameter (MAXNUP=500)
      integer NUP,IDPRUP,IDUP,ISTUP,MOTHUP,ICOLUP
      double precision XWGTUP,SCALUP,AQEDUP,AQCDUP,
     &                 PUP,VTIMUP,SPINUP
      common/HEPEUP/NUP,IDPRUP,XWGTUP,SCALUP,AQEDUP,AQCDUP,
     &              IDUP(MAXNUP),ISTUP(MAXNUP),MOTHUP(2,MAXNUP),
     &              ICOLUP(2,MAXNUP),PUP(5,MAXNUP),VTIMUP(MAXNUP),
     &              SPINUP(MAXNUP)

      aqedup = alpha
      aqcdup = alphas

      do i = 1,4
         istup(i) = +1  !particles are outgoing
         idup(i) = 0
         if (fname(i).eq.'dq')       idup(i) = 1
         if (fname(i).eq.'anti-dq')  idup(i) = -1
         if (fname(i).eq.'uq')       idup(i) = 2
         if (fname(i).eq.'anti-uq')  idup(i) = -2
         if (fname(i).eq.'sq')       idup(i) = 3
         if (fname(i).eq.'anti-sq')  idup(i) = -3
         if (fname(i).eq.'cq')       idup(i) = 4
         if (fname(i).eq.'anti-cq')  idup(i) = -4
         if (fname(i).eq.'e')        idup(i) = 11
         if (fname(i).eq.'anti-e')   idup(i) = -11
         if (fname(i).eq.'nue')      idup(i) = 12
         if (fname(i).eq.'anti-nue') idup(i) = -12
         if (fname(i).eq.'mu')       idup(i) = 13
         if (fname(i).eq.'anti-mu')  idup(i) = -13
         if (fname(i).eq.'num')      idup(i) = 14
         if (fname(i).eq.'anti-num') idup(i) = -14
         if (idup(i).eq.0) then
            write(*,*) 'Process not supported.'
            stop
         endif
      enddo
      istup(5) = +1
      idup(5)  = 22  !photon (unweighted events for QCD not yet supported)

      end

**************************************************************
*
*     output of unweighted events
*
*         p:momentum(Lorentz index;fermion 1..4+photon/gluon)
*         evprop:0=only fermions, 1=photon event, 2=gluon event
*         weight: +/- integrated decay width
*
**************************************************************
      subroutine unweighted_output(p,evprop,weight)
      implicit none
      integer evprop,i
      real*8 p(0:3,5),weight
c Les Houches accord common block
      integer MAXNUP
      parameter (MAXNUP=500)
      integer NUP,IDPRUP,IDUP,ISTUP,MOTHUP,ICOLUP
      double precision XWGTUP,SCALUP,AQEDUP,AQCDUP,
     &                 PUP,VTIMUP,SPINUP
      common/HEPEUP/NUP,IDPRUP,XWGTUP,SCALUP,AQEDUP,AQCDUP,
     &              IDUP(MAXNUP),ISTUP(MAXNUP),MOTHUP(2,MAXNUP),
     &              ICOLUP(2,MAXNUP),PUP(5,MAXNUP),VTIMUP(MAXNUP),
     &              SPINUP(MAXNUP)


      xwgtup = weight

      do i = 1,5
         pup(1,i) = p(1,i)
         pup(2,i) = p(2,i)
         pup(3,i) = p(3,i)
         pup(4,i) = p(0,i)
         pup(5,i) = 0d0
      enddo

      if (evprop.eq.0) then
         nup = 4      !4 final state particles
      else
         nup = 5      !4 fermions + photon(gluon)
      endif

      end

************************************************************************
*
*     definition of histograms
*
*  input:
*  ------
*  weight        : weight of subevent
*  outmom(0:3,5) : four-momenta of outgoing particles
*  steps=1: initialization
*  steps=2: start filling histograms
*  steps=3: finish filling of histograms
*  steps=4: print data for histograms
*
************************************************************************
      subroutine create_histo(weight,outmom,step,unweighted)
      implicit none
      real*8 weight,outmom(0:3,5),boostmom(0:3,4)
      integer step,i1,i2,i3,length,num,unweighted,nev
      integer coltype
      real*8 s(5,5),kp(0:3),km(0:3),kpvec,kmvec,pvec(5),boostvec(4)
      real*8 energy,x,q1(0:3),q2(0:3),q3(0:3),denum,en(4),entr(4)
      character*100 name
      character*25 dir
      include 'COMMON/options.inc'
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
c outputfile
      character*90 outputfile
      common/outputfile/outputfile

      if (qhist.eq.0) return
      if (weight.eq.0d0.and.step.eq.2) return
      if (unweighted.eq.0) then
         dir='HISTOGRAMS/'
         nev=nevents
      else
         dir='HISTUNWEIGHTED/'
         nev=nunwevents
      endif
      energy = mh     !Higgs mass
      num = 0         !number of distribution
c calculation of sum of pairs of momenta and the
c according invariant masses squared:
      do i1=1,5
      do i2=i1+1,5
        do i3=0,3
          q1(i3)=outmom(i3,i1)+outmom(i3,i2)
        enddo
        s(i1,i2)=q1(0)*q1(0)-q1(1)*q1(1)-q1(2)*q1(2)-q1(3)*q1(3)
        s(i2,i1)=s(i1,i2)
      enddo
      enddo
c calculation of the W+ and W- momenta:
      do i1=0,3
        kp(i1)=outmom(i1,1)+outmom(i1,2)
        km(i1)=outmom(i1,3)+outmom(i1,4)
      enddo
      kpvec=dsqrt(kp(1)*kp(1)+kp(2)*kp(2)+kp(3)*kp(3))
      kmvec=dsqrt(km(1)*km(1)+km(2)*km(2)+km(3)*km(3))
      do i1=1,5
        pvec(i1)=dsqrt(outmom(1,i1)**2+outmom(2,i1)**2+outmom(3,i1)**2)
      enddo
c type of process, used to selectively switch of histograms
c 0 = leptonic,  1 = semileptonic,  2 = hadronic
      coltype = quark(1) + quark(3)

c no histograms for hadronic processes
      if (coltype.eq.2 .and. debug.eq.0) then
         return
      endif

c invariant mass distributions
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv12.7585'
      num = num+1
      call histogram(75d0,85d0,dsqrt(s(1,2)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv34.7585'
      num = num+1
      call histogram(75d0,85d0,dsqrt(s(3,4)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv12.5090'
      num = num+1
      call histogram(50d0,90d0,dsqrt(s(1,2)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv34.5090'
      num = num+1
      call histogram(50d0,90d0,dsqrt(s(3,4)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv12.8595'
      num = num+1
      call histogram(85d0,95d0,dsqrt(s(1,2)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv34.8595'
      num = num+1
      call histogram(85d0,95d0,dsqrt(s(3,4)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv12.60100'
      num = num+1
      call histogram(60d0,100d0,dsqrt(s(1,2)),weight,num,nev,name,50
     &     ,step)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.inv34.60100'
      num = num+1
      call histogram(60d0,100d0,dsqrt(s(3,4)),weight,num,nev,name,50
     &     ,step)

c decay angle distribution of f3 in rest frame of km(f3+f4)
      if (coltype .eq. 0) then
      call boostrestframe(km,outmom(0,3),boostmom(0,3))
      boostvec(3)=dsqrt( boostmom(1,3)**2+boostmom(2,3)**2
     &                  +boostmom(3,3)**2)
      x = 2d0 ! events that are screwed up are sorted in non-existing bin
      if (kmvec*boostvec(3).gt.0d0)
     &  x = (km(1)*boostmom(1,3)+km(2)*boostmom(2,3)
     &       +km(3)*boostmom(3,3))/kmvec/boostvec(3)
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.cthv2f3'
      num = num+1
      call histogram(-1d0,1d0,x,weight,num,nev,name,50,step)
      endif

c angle between decay planes (called in phi' in publication)
c cos(phi) = (k_12 x k_1) (k_12 x k_3)/   k_12 -> k_1 + k_2
c            |k_12 x k_1| |k_12 x k_3)|
c sgn(sin(phi)) = sgn(k_12*((k_12 x k_1) x (k_12 x k_3)))
      if (coltype .eq. 0) then
      q1(1)=kp(2)*outmom(3,1)-kp(3)*outmom(2,1)
      q1(2)=kp(3)*outmom(1,1)-kp(1)*outmom(3,1)
      q1(3)=kp(1)*outmom(2,1)-kp(2)*outmom(1,1)
      q2(1)=kp(2)*outmom(3,3)-kp(3)*outmom(2,3)
      q2(2)=kp(3)*outmom(1,3)-kp(1)*outmom(3,3)
      q2(3)=kp(1)*outmom(2,3)-kp(2)*outmom(1,3)
      denum=dsqrt(q1(1)**2+q1(2)**2+q1(3)**2)
     &     *dsqrt(q2(1)**2+q2(2)**2+q2(3)**2)
      x=-1d0
      if(denum.ne.0d0)then
        x=(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/denum
        if(dabs(x).le.1d0)then
          x=dacos(x)*180d0/pi
          q3(1)=q1(2)*q2(3)-q1(3)*q2(2)
          q3(2)=q1(3)*q2(1)-q1(1)*q2(3)
          q3(3)=q1(1)*q2(2)-q1(2)*q2(1)
          if((kp(1)*q3(1)+kp(2)*q3(2)+kp(3)*q3(3)).lt.0d0)x=360d0-x
        else
          x=-1d0
        endif
      endif
      name=dir(1:length(dir))//outputfile(1:length(outputfile))//'.phi'
      num = num+1
      call histogram(0d0,360d0,x,weight,num,nev,name,50,step)
      endif

c angle between decay planes (H->uu(dd)ee)
c |cos phi| = |(k_12 x k_1) (k_12 x k_3)|/    (k_12->qq)
c              |k_12 x k_1| |k_12 x k_3)|
      if (coltype .eq. 1) then
      q1(1)=kp(2)*outmom(3,1)-kp(3)*outmom(2,1)
      q1(2)=kp(3)*outmom(1,1)-kp(1)*outmom(3,1)
      q1(3)=kp(1)*outmom(2,1)-kp(2)*outmom(1,1)
      q2(1)=kp(2)*outmom(3,3)-kp(3)*outmom(2,3)
      q2(2)=kp(3)*outmom(1,3)-kp(1)*outmom(3,3)
      q2(3)=kp(1)*outmom(2,3)-kp(2)*outmom(1,3)
      denum=dsqrt(q1(1)**2+q1(2)**2+q1(3)**2)
     &     *dsqrt(q2(1)**2+q2(2)**2+q2(3)**2)
      x=-2d0
      if(denum.ne.0d0)then
        x=dabs(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/denum
        if (dabs(x).gt.1d0) x=-2d0
      endif
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.cphihad'
      num = num+1
      call histogram(0d0,1d0,x,weight,num,nev,name,50,step)
      endif

c angle between W(34) and e(2) for H -> nu e q q
      denum = kmvec*pvec(2)
      if (denum.ne.0d0) then
        x = (km(1)*outmom(1,2)+km(2)*outmom(2,2)+km(3)*outmom(3,2))
     &        /denum
      endif
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.cthv2f2'
      num = num+1
      call histogram(-1d0,1d0,x,weight,num,nev,name,50,step)

c angle between fermion 1 and 3
      if (coltype .eq. 0) then
      en(1) = dsqrt(outmom(1,1)**2+outmom(2,1)**2+outmom(3,1)**2)
      en(3) = dsqrt(outmom(1,3)**2+outmom(2,3)**2+outmom(3,3)**2)
      denum = en(1)*en(3)
      x = -10d0
      if (denum.ne.0d0) then
        x = (outmom(1,1)*outmom(1,3)+outmom(2,1)*outmom(2,3)
     &      +outmom(3,1)*outmom(3,3))/denum
      endif
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.cthf1f3'
      num = num+1
      call histogram(-1d0,1d0,x,weight,num,nev,name,50,step)
      endif

c angle between transverse part of fermion 2 and 3
      if (coltype .eq. 0) then
      entr(2) = dsqrt(outmom(1,2)**2+outmom(2,2)**2)
      entr(3) = dsqrt(outmom(1,3)**2+outmom(2,3)**2)
      denum = entr(2)*entr(3)
      x = -1d0
      if (denum.ne.0d0) then
        x = (outmom(1,2)*outmom(1,3)+outmom(2,2)*outmom(2,3))/denum
        if(dabs(x).le.1d0)then
          x = dacos(x)*180d0/pi
          if ((outmom(1,2)*outmom(2,3)-outmom(2,2)*outmom(1,3)).lt.0d0)
     &         x = 360d0-x
        else
          x = -1d0
        endif
      endif
      name=dir(1:length(dir))//outputfile(1:length(outputfile))
     &     //'.phitrf2f3'
      num = num+1
      call histogram(0d0,360d0,x,weight,num,nev,name,50,step)
      endif

      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     histogram routine                                           c
c     Markus Roth, 11. August 2000, modified by Axel              c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine histogram(min,max,x,y,histo,n,name,parts,step)
      implicit none
      include 'COMMON/max.inc'
      real*8 average(maxh,maxp),sigma(maxh,maxp),ysum(maxh,maxp)
      real*8 avborn(maxh,maxp),siborn(maxh,maxp)
      real*8 min,max,x,y,av,si,avb,sib,shift,pos
      integer i,isum(maxh,maxp),imax(maxh),rhs
      integer i1,i2,histo,parts,n,hiout,step,fehler
      character*100 name
      save average,sigma,ysum,isum,imax,avborn,siborn
c smearing
      integer smear
      real*8 smrat
      common/smearing/smrat,smear
c initialization
      if(step.eq.1)then
        hiout=25
        open(unit=hiout,file=name,status='unknown',iostat=fehler)
        if (fehler.ne.0) then
           write(*,*)'Cannot open ',name
           write(*,*)'Please create directory.'
           stop
        endif
        close(hiout)
        do i1=1,maxh
          imax(i1)=0
          do i2=1,maxp
            ysum(i1,i2)=0d0
            average(i1,i2)=0d0
            sigma(i1,i2)=0d0
            avborn(i1,i2)=0d0
            siborn(i1,i2)=0d0
            isum(i1,i2)=0
          enddo
        enddo
      endif
c input of single subevents
      if(step.eq.2)then
        if(x.ge.min.and.x.lt.max)then
          i=int((x-min)/(max-min)*parts)+1
          imax(histo)=imax(histo)+1   !# of contributing subevents
          isum(histo,imax(histo))=i   !# of bin for subevent
          if (smear.eq.0) then
            ysum(histo,i)=y+ysum(histo,i)
          elseif (smear.eq.1) then
            pos = (x-min)/(max-min)*parts - (i-1)
            if (pos.lt.smrat/2d0.and.i.ne.1) then
              ysum(histo,i)=(0.5d0+pos/smrat)*y+ysum(histo,i)
              ysum(histo,i-1)=(0.5d0-pos/smrat)*y+ysum(histo,i-1)
            elseif (pos.gt.(1d0-smrat/2d0).and.i.ne.parts) then
              ysum(histo,i)=(0.5d0+(1d0-pos)/smrat)*y+ysum(histo,i)
              ysum(histo,i+1)=(0.5d0-(1d0-pos)/smrat)*y+ysum(histo,i+1)
            else
              ysum(histo,i)=y+ysum(histo,i)
            endif
          endif
        endif
c input born event and final calculation of whole event
      elseif(step.eq.3)then
        if(x.ge.min.and.x.lt.max)then
          i=int((x-min)/(max-min)*parts)+1
          if (smear.eq.0) then
            avborn(histo,i)=y+avborn(histo,i)
            siborn(histo,i)=y*y+siborn(histo,i)
          elseif (smear.eq.1) then
            pos = (x-min)/(max-min)*parts - (i-1)
            if (pos.lt.smrat/2d0.and.i.ne.1) then
              avborn(histo,i)=(0.5d0+pos/smrat)*y+avborn(histo,i)
              avborn(histo,i-1)=(0.5d0-pos/smrat)*y+avborn(histo,i-1)
              siborn(histo,i)=(0.5d0+pos/smrat)*y*y+siborn(histo,i)
              siborn(histo,i-1)=(0.5d0-pos/smrat)*y*y+siborn(histo,i-1)
            elseif (pos.gt.(1d0-smrat/2d0).and.i.ne.parts) then
              avborn(histo,i)=(0.5d0+(1d0-pos)/smrat)*y
     &                          +avborn(histo,i)
              avborn(histo,i+1)=(0.5d0-(1d0-pos)/smrat)*y
     &                          +avborn(histo,i+1)
              siborn(histo,i)=(0.5d0+(1d0-pos)/smrat)*y*y
     &                          +siborn(histo,i)
              siborn(histo,i+1)=(0.5d0-(1d0-pos)/smrat)*y*y
     &                          +siborn(histo,i+1)
            else
              avborn(histo,i)=y+avborn(histo,i)
              siborn(histo,i)=y*y+siborn(histo,i)
            endif
          endif
        endif
        if(imax(histo).ge.1)then
          do i1=1,imax(histo)
            i=isum(histo,i1)
            average(histo,i)=ysum(histo,i)+average(histo,i)
            sigma(histo,i)=ysum(histo,i)*ysum(histo,i)+sigma(histo,i)
            ysum(histo,i)=0d0
            if (smear.eq.1) then
              if (i.ne.1) then
                average(histo,i-1)=ysum(histo,i-1)+average(histo,i-1)
                sigma(histo,i-1)=ysum(histo,i-1)**2+sigma(histo,i-1)
                ysum(histo,i-1)=0d0
              endif
              if (i.ne.parts) then
                average(histo,i+1)=ysum(histo,i+1)+average(histo,i+1)
                sigma(histo,i+1)=ysum(histo,i+1)**2+sigma(histo,i+1)
                ysum(histo,i+1)=0d0
              endif
            endif
          enddo
        endif
        imax(histo)=0
c output
      elseif(step.eq.4)then
        hiout=25
        open(unit=hiout,file=name,status='unknown')
        shift=1d0
        rhs=1
        write(hiout,'("# histogram:",i3,", bins:",i4)'),histo,parts
        write(hiout,'(a)')
     &       '# columns: x, NLO result, corresponding MC error,',
     &       '#              LO result, corresponding MC error'
        if(shift.eq.0d0)write(hiout,'(a)')
     &       '# x = right corner of histogram bin'
        if(shift.eq.0.5d0)write(hiout,'(a)')
     &       '# x = middle of histogram bin'
        if(shift.eq.1d0)write(hiout,'(a)')
     &       '# x = left corner of histogram bin'
        write(hiout,'(a)')'# '
        do i=1,parts
          av=average(histo,i)/dble(n)/(max-min)*dble(parts)
          si=sigma(histo,i)/(dble(n)*(max-min)/dble(parts))**2
          si=dsqrt(dabs(si-av*av/dble(n)))
          avb=avborn(histo,i)/dble(n)/(max-min)*dble(parts)
          sib=siborn(histo,i)/(dble(n)*(max-min)/dble(parts))**2
          sib=dsqrt(dabs(sib-avb*avb/dble(n)))
          write(hiout,'(5(e12.6,2x))')
     *      min+(max-min)/dble(parts)*(dble(i)-shift),av,si,avb,sib
        enddo
        if(rhs.eq.1)write(hiout,'(5(e12.6,2x))')
     *      max,av,si,avb,sib
        write(hiout,'(a)') '# '
        close(hiout)
      endif

      end

************************************************************************
*     cut subroutine
*
* photon recombination for qrecomb=1 (qrecomb=0: no recombination)
* jet recombination for gluon=1
* phase-space cuts for qcuts.ne.0
*
* gluon=0: 4-particle phsp or photon radiation
* gluon=1: gluon radiation
* qrecomb=1: invariant-mass condition (precomb(1))

*
* input:
* ------
* four-momenta of external particles: outmom(0:3,5)
* outmom(0:3,1-4): fermions
* outmom(0:3,5): photon/gluon
* p(0,n): energy, p(1-3,n): x,y,z component
* nparticles: number of particles in the final state (4 or 5)
*
* output:
* -------
* nocut: 1: event passes cuts, 0: event is rejected
* recmom(0:3,5): momenta (after possible recombination)
*
************************************************************************
      subroutine cut(outmom,recmom,nparticles,nocut,gluon)
      implicit none
      real*8 outmom(0:3,5),recmom(0:3,5),p(0:3),pvec(5),s,prod
      real*8 ctheta(4),cthetamax,si(4),simin,sij(5,5),sijmin
      integer nparticles,npart,nocut,comb,fcomb,i1,i2,i3,jet1,jet2
      integer smalljet,bigjet,gluon
      include 'COMMON/param.inc'
      include 'COMMON/processpar.inc'
      include 'COMMON/options.inc'
c cuts
      real*8 ecut(5),scut(5,5),ccut(5,5),precomb(3)
      integer qcuts
      common/cuts/ecut,scut,ccut,precomb,qcuts

      npart = nparticles
      do i1=1,5
      do i2=0,3 !momentum after recombination and for histograms
        recmom(i2,i1) = outmom(i2,i1)
      enddo
      enddo
c photon recombination
      if (nparticles.eq.5.and.gluon.eq.0) then
        comb=0
        if (qrecomb.eq.1) then
           do i1=1,4
             if(q(i1).ne.0d0)then
               do i2 = 0,3
                 p(i2) = outmom(i2,i1) + outmom(i2,5)
               enddo
               si(i1) = p(0)**2-p(1)**2-p(2)**2-p(3)**2
               if(si(i1).lt.precomb(1)**2) comb=1
             endif
           enddo
           simin = 1d10
           do i1=1,4
              if(q(i1).ne.0d0.and.simin.gt.si(i1))then
                simin = si(i1)
                fcomb=i1
              endif
           enddo
        endif
c photon is recombined with fermion fcomb if comb=1
        if(comb.eq.1)then
          do i1=0,3
            recmom(i1,fcomb)=outmom(i1,fcomb)+outmom(i1,5)
            recmom(i1,5)=0d0
          enddo
          npart = 4     ! apply cuts only to fermions (after recombination)
        endif
      endif
c jet recombination
      sijmin = 1d10
      if (nparticles.eq.5.and.gluon.eq.1) then
        do i1=1,4
        do i2=i1+1,5
          if ((i2.eq.5.and.quark(i1).eq.1).or.(quark(i1).eq.1
     &          .and.quark(i2).eq.1)) then
            do i3=0,3
              p(i3) = outmom(i3,i1) + outmom(i3,i2)
            enddo
            sij(i1,i2) = p(0)**2 - p(1)**2 - p(2)**2 - p(3)**2
            if (sij(i1,i2).lt.sijmin) then
              sijmin = sij(i1,i2)
              jet1 = i1
              jet2 = i2
            endif
          endif
        enddo
        enddo
        if (jet2.eq.5) then !gluon and quark are recombined
          do i3=0,3
            recmom(i3,jet1) = outmom(i3,jet1) + outmom(i3,5)
            recmom(i3,5) = 0d0
          enddo
        else    !2 quarks are recombined, smaller one is replaced with gluon
          if (outmom(0,jet1).gt.outmom(0,jet2)) then
            bigjet = jet1
            smalljet = jet2
          else
            bigjet = jet2
            smalljet = jet1
          endif
          do i3=0,3
            recmom(i3,bigjet) = outmom(i3,bigjet) + outmom(i3,smalljet)
            recmom(i3,smalljet) = outmom(i3,5)
            recmom(i3,5) = 0d0
          enddo
        endif
        npart = 4     ! apply cuts only to fermions (after recombination)
      endif

c apply cuts
      if (qcuts.ne.0) then
        nocut=0
c energy cuts
        do i1=1,npart
          if(recmom(0,i1).lt.ecut(i1))return
        enddo
c invariant-mass cuts
        do i1=1,npart
        do i2=i1+1,npart
          do i3=0,3
            p(i3)=recmom(i3,i1)+recmom(i3,i2)
          enddo
          s=p(0)*p(0)-p(1)*p(1)-p(2)*p(2)-p(3)*p(3)
          if(s.lt.scut(i1,i2))return
        enddo
        enddo
c angular cuts
        do i1=1,npart
          pvec(i1)=dsqrt(recmom(1,i1)**2+recmom(2,i1)**2
     &          +recmom(3,i1)**2)
        enddo
        do i1=1,npart
        do i2=i1+1,npart
          prod=(recmom(1,i1)*recmom(1,i2)+recmom(2,i1)*recmom(2,i2)
     &          +recmom(3,i1)*recmom(3,i2))
          if(prod.gt.ccut(i1,i2)*pvec(i1)*pvec(i2))return
        enddo
        enddo
      endif

      nocut=1

      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     boost particle with momentum pold into restframe           c
c     of particle with momentum q -> pnew                        c
c                                                                c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine boostrestframe(q,pold,pnew)
      implicit none
c local variable
      real*8 pold(0:3),pnew(0:3),q(0:3),m,m2,bx,by,bz,gamma
      real*8 bp,a
      m2=q(0)*q(0)-q(1)*q(1)-q(2)*q(2)-q(3)*q(3)
      if(m2.le.0d0)return
      m=dsqrt(dabs(m2))
      bx=-q(1)/m
      by=-q(2)/m
      bz=-q(3)/m
      gamma=q(0)/m
      a=1d0/(1d0+gamma)
      bp=bx*pold(1)+by*pold(2)+bz*pold(3)
      pnew(0)=gamma*pold(0)+bp
      pnew(1)=pold(1)+bx*pold(0)+a*bp*bx
      pnew(2)=pold(2)+by*pold(0)+a*bp*by
      pnew(3)=pold(3)+bz*pold(0)+a*bp*bz
      end
