************************************************************************
        function xA0_(m2)
************************************************************************
*	scalar 1-point function with complex mass
*-----------------------------------------------------------------------
*	3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xA0_,m2
        complex*16 fct(110),x(16)
        integer nrank
        logical nocalc
        character*11 name

	COMMON/UV/MUDIM2

        x(1)=m2
        nrank=0
        name='xA0_'
        call fastin(fct,x,1,1,9,nrank,name,nocalc)
        if(nocalc)then
           xA0_=fct(1)
           return
        endif

        if (m2.ne.(0d0,0d0)) then
	  xA0_ = m2*(1d0+log(mudim2/m2))
	else
	  xA0_ = 0d0
	endif

        fct(1)=xA0_
        call fastout(fct,1,9)

	end

************************************************************************
        subroutine Bargs(xp,xm02,xm12,p,m02,m12,qsing)
************************************************************************
*	change arguments for 2-point functions with complex masses
*-----------------------------------------------------------------------
*	9.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 m02,m12,xm02,xm12,cmp2,cm2
        complex*16 cme2,cmf2(5)
        real*8 mf2(5),rmf2(5)
	integer ext,qsing
	logical regp,regm02,regm12

        common/fmass/me2,rme2,cme2,mf2,rmf2,cmf2

	rmp2(rm2) = mf2(nint(rm2*1d20))
	cmp2(cm2) = mf2(nint(dreal(cm2*1d20)))

	regp   = (abs(xp)  .gt.1d-15)
        regm02 = (abs(xm02).gt.1d-15)
        regm12 = (abs(xm12).gt.1d-15)
	p   = xp
        m02 = xm02
        m12 = xm12
	if (regp.or.regm02.or.regm12) then
c regular case
	  if (.not.regp)   p   = 0d0
	  if (.not.regm02) m02 = 0d0
	  if (.not.regm12) m12 = 0d0
	  qsing = 0
	else
c singular case
	  if ((.not.regp).and.(p.ne.0d0))           p   = rmp2(p)
	  if ((.not.regm02).and.(m02.ne.(0d0,0d0))) m02 = cmp2(m02)
	  if ((.not.regm12).and.(m12.ne.(0d0,0d0))) m12 = cmp2(m12)
	  qsing = 1
	endif

        end

************************************************************************
        function xB0_(xp,xm02,xm12,ext)
************************************************************************
*	scalar 2-point function with complex masses
*-----------------------------------------------------------------------
*	3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 m02,m12,xm02,xm12,xB0_,ieps,sqe,x1,x2
	integer ext,qsing
        complex*16 fct(110),x(16)
        integer nrank
        logical nocalc
        character*11 name

	COMMON/UV/MUDIM2

        x(1)=xp
        x(2)=xm02
        x(3)=xm12
        x(4)=ext
        nrank=0
        name='xB0_'
        call fastin(fct,x,1,4,8,nrank,name,nocalc)
        if(nocalc)then
           xB0_=fct(1)
           return
        endif

	call Bargs(xp,xm02,xm12,p,m02,m12,qsing)

	eps  = 1d-20
	ieps = dcmplx(0d0,eps)

c*** cases with p=0
	if (p.eq.0d0) then
	  if (m02.eq.(0d0,0d0)) then
	    xB0_ = 1d0+log(mudim2/m12)
	  elseif (m12.eq.(0d0,0d0)) then
	    xB0_ = 1d0+log(mudim2/m02)
	  elseif (m02.eq.m12) then
	    xB0_ = log(mudim2/m02)
	  else
	    xB0_ = 1d0-(m02*log(m02/mudim2)-m12*log(m12/mudim2))
     &			/(m02-m12)
	  endif
c*** cases with p=/=0 and mi^2=0
	elseif (m02*m12.eq.(0d0,0d0)) then
	  if ((m02.eq.(0d0,0d0)).and.(m12.eq.(0d0,0d0))) then
	    xB0_ = 2d0-log(-p/mudim2-ieps)
	  elseif (dcmplx(p).eq.m02) then
	    xB0_ = 2d0+log(mudim2/m02)
	  elseif (dcmplx(p).eq.m12) then
	    xB0_ = 2d0+log(mudim2/m12)
	  elseif (m02.eq.(0d0,0d0)) then
	    xB0_ = 2d0-log(m12/mudim2)+(m12/p-1d0)*log(1d0-p/m12-ieps)
	  elseif (m12.eq.(0d0,0d0)) then
	    xB0_ = 2d0-log(m02/mudim2)+(m02/p-1d0)*log(1d0-p/m02-ieps)
	  endif
c*** general case
	else
	  x1 = sqe(dcmplx(p),-p+m02-m12,m12-abs(p)*ieps)
	  x2 = (m12-abs(p)*ieps)/p/x1
	  xB0_ = 2d0-log(m12/mudim2)+(x1-1d0)*log(1d0-1d0/x1)
     &			+(x2-1d0)*log(1d0-1d0/x2)
	endif

        fct(1)=xB0_
        call fastout(fct,1,8)

	end

************************************************************************
        function xDB0_(xp,xm02,xm12,ext)
************************************************************************
*	momentum derivative of scalar 2-point function with complex masses
*-----------------------------------------------------------------------
*	3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,xDB0_,ieps,sqe,x1,x2
	integer ext,qsing

	COMMON/IR/LAMBDA,LAMBDA2

	call Bargs(xp,xm02,xm12,p,m02,m12,qsing)

	eps  = 1d-20
	ieps = dcmplx(0d0,eps)

c*** cases with p=0
	if (p.eq.0d0) then
	  if (m02.eq.(0d0,0d0)) then
	    xDB0_ = 1d0/2d0/m12
	  elseif (m12.eq.(0d0,0d0)) then
	    xDB0_ = 1d0/2d0/m02
	  elseif (m02.eq.m12) then
	    xDB0_ = 1d0/6d0/m02
	  else
	    xDB0_ = (m02**2-m12**2-2d0*m02*m12*log(m02/m12))
     &		  /2d0/(m02-m12)**3
	  endif

c*** cases with p=/=0 and mi=0
	elseif (m02*m12.eq.(0d0,0d0)) then
	  if ((m02.eq.(0d0,0d0)).and.(m12.eq.(0d0,0d0))) then
	    xDB0_ = -1d0/p
	  elseif (dcmplx(p).eq.m02) then
	    xDB0_ = -1d0/2d0/m02*(2d0+log(lambda2/m02))
	  elseif (dcmplx(p).eq.m12) then
	    xDB0_ = -1d0/2d0/m12*(2d0+log(lambda2/m12))
	  elseif (m02.eq.(0d0,0d0)) then
	    xDB0_ = -1d0/p**2*(p+m12*log(1d0-p/m12-ieps))
	  elseif (m12.eq.(0d0,0d0)) then
	    xDB0_ = -1d0/p**2*(p+m02*log(1d0-p/m02-ieps))
	  endif
c*** general case
	else
	  x1 = sqe(dcmplx(p),-p+m02-m12,m12-abs(p)*ieps)
	  x2 = (m12-abs(p)*ieps)/p/x1
	  xDB0_ = (-1d0 + x1*(1d0-x1)/(x1-x2)*log(1d0-1d0/x1)
     &	                - x2*(1d0-x2)/(x1-x2)*log(1d0-1d0/x2) )/p
	endif

	end

************************************************************************
        function xB1_(xp,xm02,xm12,ext)
************************************************************************
*	coefficient for vector 2-point function with complex masses
*-----------------------------------------------------------------------
*	3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,xB1_,xB0_,xDB0_,xA0_
	integer ext,qsing
        complex*16 fct(110),x(16)
        integer nrank
        logical nocalc
        character*11 name

        x(1)=xp
        x(2)=xm02
        x(3)=xm12
        x(4)=ext
        nrank=0
        name='xB1_'
        call fastin(fct,x,1,4,7,nrank,name,nocalc)
        if(nocalc)then
           xB1_=fct(1)
           return
        endif

	call Bargs(xp,xm02,xm12,p,m02,m12,qsing)

	if (p.eq.0d0) then
	  xB1_ = -xB0_(0d0,m02,m12,ext)/2d0
     &		+(m12-m02)/2d0*xDB0_(0d0,m02,m12,ext)
	else
	  xB1_ = ( xA0_(m02)-xA0_(m12)-(p-m12+m02)*xB0_(p,m02,m12,ext) )
     &		/2d0/p
	endif

        fct(1)=xB1_
        call fastout(fct,1,7)

	end

************************************************************************
        function xDB1_(xp,xm02,xm12,ext)
************************************************************************
*	momentum derivative of coefficient for vector 2-point function
*	with complex masses
*-----------------------------------------------------------------------
*       3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,m04,m14
	complex*16 xDB1_,xB1_,xB0_,xDB0_,xA0_
	integer ext,qsing

	call Bargs(xp,xm02,xm12,p,m02,m12,qsing)

	m04 = m02**2
	m14 = m12**2

	if (p.eq.0d0) then
	  if (m02.eq.(0d0,0d0)) then
	    xDB1_ = -1d0/6d0/m12
	  elseif (m12.eq.(0d0,0d0)) then
	    xDB1_ = -1d0/3d0/m02
	  elseif (m02.eq.m12) then
	    xDB1_ = -1d0/12d0/m02
	  else
	    xDB1_ = ( 2d0*m04+5d0*m02*m12-m14
     &		  -6d0*m02*(xA0_(m12)*m02-xA0_(m02)*m12)/(m02-m12))
     &		 /6d0/(m12-m02)**3
	  endif
	else
	  xDB1_ = -( 2d0*xB1_(p,m02,m12,ext)+xB0_(p,m02,m12,ext)
     &		   +(p-m12+m02)*xDB0_(p,m02,m12,ext) )/2d0/p
	endif

	end

************************************************************************
        function xB20_(xp,xm02,xm12,ext)
************************************************************************
*	tensor coefficient B00 for 2nd-rank tensor 2-point function
*	with complex masses
*-----------------------------------------------------------------------
*       3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,xB20_,xB1_,xB0_,xA0_
	integer ext

        p   = xp
        m02 = xm02
        m12 = xm12
        if (abs(xp).lt.1d-15)     p = 0d0
        if (abs(xm02).lt.1d-15) m02 = 0d0
        if (abs(xm12).lt.1d-15) m12 = 0d0

	if (p.eq.0d0) then
	  if (m02.eq.m12) then
	    xB20_ = xA0_(m12)/2d0
	  else
	    xB20_ = (m02+m12)/8D0+(xA0_(m02)*m02-xA0_(m12)*m12)
     &			/4D0/(m02-m12)
	  endif
	else
	  xB20_ = ( xA0_(m12)+2d0*m02*xB0_(p,m02,m12,ext)
     &		   +(p-m12+m02)*xB1_(p,m02,m12,ext)+m02+m12-p/3d0 )/6d0
	endif

	end

************************************************************************
        function xB21_(xp,xm02,xm12,ext)
************************************************************************
*	tensor coefficient B11 for 2nd-rank tensor 2-point function
*	with complex masses
*-----------------------------------------------------------------------
*       3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,m04,m14,xB21_,xB1_,xB0_,xA0_
	integer ext,qsing

	call Bargs(xp,xm02,xm12,p,m02,m12,qsing)

	m04 = m02**2
	m14 = m12**2

	if (p.eq.0d0) then
	  if (m02.eq.m12) then
	    xB21_ = xB0_(0d0,m02,m12,ext)/3d0
	  else
	    xB21_ = ( 5d0*m04+5d0*m02*m12-4d0*m14
     &		   +6d0/(m02-m12)*xA0_(m02)*m04
     &		   -6d0/(m02-m12)*xA0_(m12)*(3d0*m04-3d0*m02*m12+m14) )
     &		  /18d0/(m02-m12)**2
	  endif
	else
	  xB21_ = ( 2d0*xA0_(m12)-2d0*m02*xB0_(p,m02,m12,ext)
     &	    -4d0*(p-m12+m02)*xB1_(p,m02,m12,ext)-m02-m12+p/3d0 )/6d0/p
	endif

	end

************************************************************************
        function xB001_(xp,xm02,xm12,ext)
************************************************************************
*	tensor coefficient B001 for 3rd-rank tensor 2-point function
*	with complex masses
*-----------------------------------------------------------------------
*       3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,m04,m14,xB001_,xA0_,xB1_,xB21_
	integer ext

        p   = xp
        m02 = xm02
        m12 = xm12
        if (abs(xp).lt.1d-15)     p = 0d0
        if (abs(xm02).lt.1d-15) m02 = 0d0
        if (abs(xm12).lt.1d-15) m12 = 0d0

	m04 = m02**2
	m14 = m12**2

	if (p.eq.0d0) then
	  if (m02.eq.m12) then
	    xB001_ = -xA0_(m02)/4d0
	  else
	    xB001_ = -(5d0*m04+5d0*m02*m12-4d0*m14)/72d0/(m02-m12)
     &		  -m04*xA0_(m02)/12d0/(m02-m12)**2
     &		  +m12*(3d0*m02-2d0*m12)*xA0_(m12)/12d0/(m02-m12)**2
	  endif
	else
	  xB001_ = ( -xA0_(m12)+2d0*m02*xB1_(p,m02,m12,ext)
     &		    +(p-m12+m02)*xB21_(p,m02,m12,ext)
     &		    +(p-2d0*m02-4d0*m12)/6d0 )/8d0
	endif

	end

************************************************************************
        function xB111_(xp,xm02,xm12,ext)
************************************************************************
*	tensor coefficient B111 for 3rd-rank tensor 2-point function
*	with complex masses
*-----------------------------------------------------------------------
*       3.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 xm02,xm12,m02,m12,m04,m14,m06,m16
	complex*16 xB111_,xB0_,xB20_,xB21_,xA0_
	integer ext,qsing

	call Bargs(xp,xm02,xm12,p,m02,m12,qsing)

	m04 = m02**2
	m14 = m12**2

	if (p.eq.0d0) then
	  if (m02.eq.m12) then
	    xB111_ = (1d0-xB0_(0d0,(0d0,0d0),m02,ext))/4d0
	  else
	  m06 = m02**3
	  m16 = m12**3
	    xB111_ = -(13D0*M06+13D0*M04*M12-23D0*M02*M14
     &		      +9D0*M16)/(M02-M12)**3/48D0
     &              -M06/(M02-M12)**4*xA0_(M02)/4D0
     &              +(2D0*M02-M12)*(2D0*M04-2D0*M02*M12+M14)
     &		      /(M02-M12)**4*xA0_(M12)/4D0
	  endif
	else
	  xB111_ = -( (M04-M14)/P + (2D0*M02*xA0_(M02))/P
     &               -2D0*(M12-2D0*P)*xA0_(M12)/P
     &               -(8D0*(M02-M12+P)*xB20_(p,m02,m12,ext))/P
     &               +4D0*(M02-M12+P)*xB21_(p,m02,m12,ext) )/(8D0*P)
	endif

	end

***********************************************************************
	SUBROUTINE xB012345_(xp2,xm02,xm12,B0,B1,B2,B3,B4,B5,rank,ext)
***********************************************************************
*	2-POINT TENSOR COEFFICIENTS with complex masses
*-----------------------------------------------------------------------
*       11.11.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
        COMPLEX*16 xm02,xm12,m02,m12,B0,B1(1),B2(0:2,0:2),B3(0:2,0:2,2)
        complex*16 B4(0:2,0:2,0:2,0:2),B5(0:2,0:2,0:2,0:2,2)
	complex*16 ieps,m2ieps,sqe,y(2),f(0:10,2),xA0_,luv,A0m0,A0m1
	integer rank,ext,k,l,n,qsing
        complex*16 fct(110),x(16)
        integer nrank
        logical nocalc
        character*11 name
        complex*16 xB0_,xB1_,xB21_,xB20_,xB001_,xB111_

	COMMON/UV/MUDIM2

        x(1)=xp2
        x(2)=xm02
        x(3)=xm12
        x(4)=ext
        nrank=rank
        name='xB012345_'
        call fastin(fct,x,12,4,6,nrank,name,nocalc)
        if(nocalc)then
           B0=fct(1)
           B1(1)=fct(2)
           B2(0,0)=fct(3)
           B2(1,1)=fct(4)
           B3(0,0,1)=fct(5)
           B3(1,1,1)=fct(6)
           B4(0,0,0,0)=fct(7)
           B4(0,0,1,1)=fct(8)
           B4(1,1,1,1)=fct(9)
           B5(0,0,0,0,1)=fct(10)
           B5(0,0,1,1,1)=fct(11)
           B5(1,1,1,1,1)=fct(12)
           return
        endif

	call Bargs(xp2,xm02,xm12,p2,m02,m12,qsing)
	ieps   = dcmplx(0d0,1d-18)
	m2ieps = (abs(p2)+abs(m02)+abs(m12))*ieps

c*** m0 =/= 0
	if (m02.ne.(0d0,0d0)) then
	  luv = log(mudim2/m02)

	  if (p2.eq.0d0) then
	    y(1) = 1d20
	    if (m02.eq.m12) then
	      y(2) = 1d20
	    else
	      y(2) = (m02-abs(m02)*ieps)/(m02-m12)
	    endif
	  else
	    if (m12.eq.(0d0,0d0)) then
	      y(1) = sqe(dcmplx(p2),-p2-m02+m2ieps,m02-m2ieps)
	    else
	      y(1) = sqe(dcmplx(p2),-p2-m02+m12,m02-m2ieps)
	    endif
	    y(2) = (m02-m2ieps)/p2/y(1)
	  endif
	  do 100 k=1,2
	  do 100 n=0,nrank
	    if (abs(y(k)).lt.10d0) then
	      if (1d0-1d0/y(k).ne.(0d0,0d0)) then
	        f(n,k) = (1d0-y(k)**(n+1))*log(1d0-1d0/y(k))
	      else
		f(n,k) = 0d0
	      endif
	      do l=0,n
	        f(n,k) = f(n,k) - y(k)**(n-l)/(l+1d0)
	      enddo
	    else
	      f(n,k) = log(1d0-1d0/y(k))
	      if (abs(y(k)).lt.1d20) then
	        do l=n+1,n+20
	          f(n,k) = f(n,k) + y(k)**(n-l)/(l+1d0)
	        enddo
	      endif
	    endif
100	  continue

c*** m0 = 0
	else
	  luv   = 0d0
	  if (p2.ne.0d0) then
	    y(1)  = 1d0-m12/p2+ieps*sign(1d0,p2)
	  else
	    y(1) = 1d20
	  endif
	  do 200 n=0,nrank
	    if (dcmplx(p2).eq.m12) then
	      f(n,1) = -log(mudim2/m12) - 1d0/(n+1d0)
	      f(n,2) = - 1d0/(n+1d0)
	    else
	      if (abs(y(1)).lt.10d0) then
	        if (1d0-1d0/y(1).ne.(0d0,0d0)) then
	          f(n,1) = (1d0-y(1)**(n+1))*log(1d0-1d0/y(1))
	        else
  		  f(n,1) = 0d0
	        endif
	        do l=0,n
	          f(n,1) = f(n,1) - y(1)**(n-l)/(l+1d0)
	        enddo
	      else
	        f(n,1) = log(1d0-1d0/y(1))
		if (abs(y(1)).lt.1d20) then
	          do l=n+1,n+20
	            f(n,1) = f(n,1) + y(1)**(n-l)/(l+1d0)
	          enddo
	        endif
	      endif
	      f(n,2) = log((m12-p2)/mudim2-ieps) - 1d0/(n+1d0)
	    endif
200	  continue
	endif

	B0 = luv - f(0,1) - f(0,2)

	if (nrank.eq.0) goto 1002

	B1(1) = -( luv - f(1,1) - f(1,2) )/2d0

	if (nrank.eq.1) goto 1002

	A0m0 = xA0_(m02)
	A0m1 = xA0_(m12)


	if (qsing.eq.1) then
	  B2(0,0) = 0d0
	elseif (p2.eq.0d0) then
	  if (m02.eq.m12) then
	    B2(0,0) = A0m1/2d0
	  else
	    B2(0,0) = (m02+m12)/8D0+(A0m0*m02-A0m1*m12)
     &			/4D0/(m02-m12)
	  endif
	else
	  B2(0,0) = ( A0m1+2d0*m02*B0
     &		     +(p2-m12+m02)*B1(1)+m02+m12-p2/3d0 )/6d0
	endif
	B2(1,1) = ( luv - f(2,1) - f(2,2) )/3d0

	if (nrank.eq.2) goto 1002

	if (qsing.eq.1) then
	  B3(0,0,1) = 0d0
	elseif (p2.eq.0d0) then
	  if (m02.eq.m12) then
	    B3(0,0,1) = -A0m0/4d0
	  else
	    B3(0,0,1) = -(5d0*m02**2+5d0*m02*m12-4d0*m12**2)
     &		  /72d0/(m02-m12) -m02**2*A0m0/12d0/(m02-m12)**2
     &		  +m12*(3d0*m02-2d0*m12)*A0m1/12d0/(m02-m12)**2
	  endif
	else
	  B3(0,0,1) = ( -A0m1+2d0*m02*B1(1)+(p2-m12+m02)*B2(1,1)
     &	  	      +(p2-2d0*m02-4d0*m12)/6d0 )/8d0
	endif
	B3(1,1,1) = -( luv - f(3,1) - f(3,2) )/4d0

	if (nrank.eq.3) goto 1002

	B4(1,1,1,1) = ( luv - f(4,1) - f(4,2) )/5d0
	if (qsing.eq.1) then
	  B4(0,0,0,0) = 0d0
	  B4(0,0,1,1) = 0d0
	elseif (p2.eq.0d0) then
	  if (m02.eq.m12) then
	    B4(0,0,0,0) = m02*(m02+2d0*A0m0)/16d0
	    B4(0,0,1,1) = A0m0/6d0
	  else
	    B4(0,0,0,0) = 5d0*(m02**2+m02*m12+m12**2)/144d0
     &			 +(m02**2*A0m0-m12**2*A0m1)/(24d0*(m02-m12))
	    B4(0,0,1,1) = (A0m1+(m12-m02)*B3(1,1,1))/6d0
	  endif
	else
	  B4(0,0,1,1) = (A0m1-(p2-m12+m02)*B3(1,1,1)
     &			-2d0*p2*B4(1,1,1,1))/6d0
	  B4(0,0,0,0) = m12**2/16d0 + m12*A0m1/8d0
     &			-(m02-m12+p2)*B3(0,0,1)/2d0-p2*B4(0,0,1,1)
	endif

	if (nrank.eq.4) goto 1002

	B5(1,1,1,1,1) = -( luv - f(5,1) - f(5,2) )/6d0
	if (qsing.eq.1) then
	  B5(0,0,0,0,1) = 0d0
	  B5(0,0,1,1,1) = 0d0
	elseif (p2.eq.0d0) then
	  if (m02.eq.m12) then
	    B5(0,0,0,0,1) = -m02*(2d0*A0m0+m02)/32d0
	    B5(0,0,1,1,1) = -A0m0/8d0
	  else
	    B5(0,0,1,1,1) = -A0m1/8d0+(-m02+m12)*B4(1,1,1,1)/8d0
	    B5(0,0,0,0,1) = -A0m1*m12/16d0-m12**2/32d0
     &			  +(-m02+m12)*B4(0,0,1,1)/4d0
	  endif
	else
	  B5(0,0,1,1,1) = -A0m1/8d0+(-m02+m12-p2)*B4(1,1,1,1)/8d0
     &			  -p2*B5(1,1,1,1,1)/4d0
	  B5(0,0,0,0,1) = -A0m1*m12/16d0-m12**2/32d0
     &			  +(-m02+m12-p2)*B4(0,0,1,1)/4d0
     &			  -p2*B5(0,0,1,1,1)/2d0
	endif

 1002   fct(1)=B0
        fct(2)=B1(1)
        fct(3)=B2(0,0)
        fct(4)=B2(1,1)
        fct(5)=B3(0,0,1)
        fct(6)=B3(1,1,1)
        fct(7)=B4(0,0,0,0)
        fct(8)=B4(0,0,1,1)
        fct(9)=B4(1,1,1,1)
        fct(10)=B5(0,0,0,0,1)
        fct(11)=B5(0,0,1,1,1)
        fct(12)=B5(1,1,1,1,1)
        call fastout(fct,12,6)

	end

***********************************************************************
        function xdet(A,n)
************************************************************************
*	determinant of complex nxn-matrix A
*-----------------------------------------------------------------------
*       10.6.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer n
	complex*16 A(n,n),Q(n,n),R(n,n),detQ,xdet
	integer i

	call xQRdecomp(A,Q,R,detQ,n)

	xdet = detQ*r(n,n)
	do i=1,n-1
	  xdet = xdet*r(i,i)
	enddo

	end

***********************************************************************
        subroutine xinverse(A,iA,n)
************************************************************************
*	inverse of complex nxn-matrix A
*-----------------------------------------------------------------------
*       8.1.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer n
	complex*16 A(n,n),Q(n,n),R(n,n),iA(n,n),iR(n,n),detQ,w
	complex*16 del(n,n),iA2(n,n),aux
	integer i,j,k

	call xQRdecomp(A,Q,R,detQ,n)

c invert R
	do i=n,1,-1
	  do j=1,n
	    if (j.lt.i) then
	      iR(i,j) = 0d0
	    else
	      w = 0d0
	      do k=i+1,n
	        w = w + R(i,k)*iR(k,j)
	      enddo
	      if (i.eq.j) then
	        iR(i,j) = (1d0-w)/R(i,i)
	      else
	        iR(i,j) = -w/R(i,i)
	      endif
	    endif
	  enddo
	enddo

c inverse of A
	do i=1,n
	  do j=1,n
	    iA(i,j) = 0d0
	    do k=1,n
	      iA(i,j) = iA(i,j) + iR(i,k)*dconjg(Q(j,k))
	    enddo
	  enddo
	enddo


	end

***********************************************************************
        subroutine xQRdecomp(A,Q,R,detQ,n)
************************************************************************
*	QR-decomposition of complex nxn-matrix A using Householder matrices
*	Q = orthogonal matrix, R = upper triangle matrix
*-----------------------------------------------------------------------
*       8.1.04 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer n
	complex*16 A(n,n),Q(n,n),R(n,n),w(n),wr(n),qw(n),detQ,aux,kap
	complex*16 rmax
	integer i,j,k,imax

	detQ = 1d0

c*** start with R=A and Q=unit matrix
	do i=1,n
	  do j=1,n
	    r(i,j) = a(i,j)
	    if (i.eq.j) then
	      q(i,j) = 1d0
	    else
	      q(i,j) = 0d0
	    endif
	  enddo
	enddo

c*** n-1 steps of decomposition
	do 200 i=1,n-1

c exchange rows such that |r(i,i)| is maximal
	  imax = i
	  rmax = r(i,i)
	  do j=i+1,n
	    if (abs(r(j,i)).gt.abs(rmax)) then
	      imax = j
	      rmax = r(j,i)
	    endif
	  enddo
	  if (imax.ne.i) then
	    do j=1,n
	      aux       = r(i,j)
	      r(i,j)    = r(imax,j)
	      r(imax,j) = aux
	      aux       = q(j,i)
	      q(j,i)    = q(j,imax)
	      q(j,imax) = aux
	    enddo
	    detQ = -detQ
	  endif
c calculate R and Q if r(i,i) is non-zero
	  if (r(i,i).eq.(0d0,0d0)) goto 200
	  sig2 = 0d0
	  do j=i,n
	    sig2 = sig2 + abs(r(j,i))**2
	  enddo
	  sig = sqrt(sig2)
	  kap   = -r(i,i)/abs(r(i,i))*sig
	  wnorm = sqrt(2d0*sig*(sig+abs(r(i,i))))
	  do j=i,n
	    w(j) = r(j,i)/wnorm
	  enddo
	  w(i) = w(i) - kap/wnorm

	  do j=i,n
	    wr(j) = 0d0
	    do k=i,n
	      wr(j) = wr(j) + dconjg(w(k))*r(k,j)
	    enddo
	  enddo
	  do j=i,n
	    if (j.eq.i) then
	      r(j,i) = r(j,i) - 2d0*w(j)*wr(i)
	    else
	      r(j,i) = 0d0
	    endif
	    do k=i+1,n
	      r(j,k) = r(j,k) - 2d0*w(j)*wr(k)
	    enddo
	  enddo

c calculate matrix Q
	  do j=1,n
	    qw(j) = 0d0
	    do k=i,n
	      qw(j) = qw(j) + q(j,k)*w(k)
	    enddo
	  enddo
	  do j=1,n
	    do k=i,n
	      q(j,k) = q(j,k) - 2d0*dconjg(w(k))*qw(j)
	    enddo
	  enddo
	  detQ = -detQ

200	continue

	end

***********************************************************************
	SUBROUTINE xC01234_(p1,p2,p3,M12,M22,M32,
     &			  C0,C1,C2,C3,C4,B0,B1,B2,B3,rank,ext)
***********************************************************************
*	3-POINT TENSOR COEFFICIENTS with complex masses
*-----------------------------------------------------------------------
*       11.6.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
        COMPLEX*16 M12,M22,M32,F1,F2,y12,y13,y23,xdet
        COMPLEX*16 xB001_,xB111_
        COMPLEX*16   B0,  B1(1),  B2(0:2,0:2),  B3(0:2,0:2,2)
        COMPLEX*16 B0_2,B1_2(1),B2_2(0:2,0:2),B3_2(0:2,0:2,2)
        COMPLEX*16 B0_3,B1_3(1),B2_3(0:2,0:2),B3_3(0:2,0:2,2)
        COMPLEX*16   B4(0:2,0:2,0:2,0:2),  B5(0:2,0:2,0:2,0:2,2)
        COMPLEX*16 B4_2(0:2,0:2,0:2,0:2),B5_2(0:2,0:2,0:2,0:2,2)
        COMPLEX*16 B4_3(0:2,0:2,0:2,0:2),B5_3(0:2,0:2,0:2,0:2,2)
        COMPLEX*16 BB,R1(2),R2(2,2),R3X0(2),R3(2,2),R3X12(2),R3X21(2)
        COMPLEX*16 xC0_,C0,C1(2),C2(0:2,0:2),C3(0:2,0:2,2)
	COMPLEX*16 C4(0:2,0:2,0:2,0:2)
        COMPLEX*16 cC0,cC1(2),cC2(0:2,0:2),cC3(0:2,0:2,2)
	COMPLEX*16 cC4(0:2,0:2,0:2,0:2)
        COMPLEX*16 eC0,eC1(2),eC2(0:2,0:2),eC3(0:2,0:2,2)
	COMPLEX*16 eC4(0:2,0:2,0:2,0:2)
	COMPLEX*16 R4X0(2,2),R4(2,2),R4X112(2),R4X121(2),R4X211(2)
	COMPLEX*16 R4X122(2),R4X212(2),R4X221(2)
	REAL*8 XI(2,2),xx(2,2)
	INTEGER I,J,K,rank,ext
        complex*16 fct(110),x(16)
        integer nrank,i1,i2,i3
        logical nocalc,qsoft1,qsoft2,qsoft3,qcoll12,qcoll23,qcoll31
        character*11 name

	common/cproblem/crelworst
	common/Bsub/B0_2,B1_2,B2_2,B3_2,B4_2,B5_2,
     &	            B0_3,B1_3,B2_3,B3_3,B4_3,B5_3

        x(1)=p1
        x(2)=p2
        x(3)=p3
        x(4)=m12
        x(5)=m22
        x(6)=m32
        x(7)=ext
        nrank=rank
c at present only rank=3 is supported !
	if (rank.gt.3) then
	  write(*,*) 'rank>3 not yet supported in xC01234_!'
	  stop
	endif
        nrank=3
        name='xC01234_'
        call fastin(fct,x,29,7,4,nrank,name,nocalc)
        if(nocalc)then
           B0=fct(1)
           B1(1)=fct(2)
           B2(0,0)=fct(3)
           B2(1,1)=fct(4)
           B3(0,0,1)=fct(5)
           B3(1,1,1)=fct(6)
           B4(0,0,0,0)=fct(7)
           B4(0,0,1,1)=fct(8)
           B4(1,1,1,1)=fct(9)
           B5(0,0,0,0,1)=fct(10)
           B5(0,0,1,1,1)=fct(11)
           B5(1,1,1,1,1)=fct(12)
           C0=fct(13)
           C1(1)=fct(14)
           C1(2)=fct(15)
           C2(0,0)=fct(16)
           C2(1,1)=fct(17)
           C2(1,2)=fct(18)
           C2(2,2)=fct(19)
           C3(0,0,1)=fct(20)
           C3(0,0,2)=fct(21)
           C3(1,1,1)=fct(22)
           C3(1,1,2)=fct(23)
           C3(1,2,2)=fct(24)
           C3(2,2,2)=fct(25)
           C4(0,0,0,0)=fct(26)
           C4(0,0,1,1)=fct(27)
           C4(0,0,1,2)=fct(28)
           C4(0,0,2,2)=fct(29)
           C2(2,1)=C2(1,2)
           do i1=1,2
           do i2=i1,2
           do i3=i2,2
             C3(i1,i3,i2)=C3(i1,i2,i3)
             C3(i2,i1,i3)=C3(i1,i2,i3)
             C3(i2,i3,i1)=C3(i1,i2,i3)
             C3(i3,i1,i2)=C3(i1,i2,i3)
             C3(i3,i2,i1)=C3(i1,i2,i3)
           enddo
           enddo
           enddo
           C4(0,0,2,1)=C4(0,0,1,2)
           return
        endif

	C0 = xC0_(P1,P2,P3,M12,M22,M32,ext)

	if (nrank.eq.0) goto 1005

	Q1  = P1
	Q2  = P3
	Q12 = (P1+P3-P2)/2D0

	DX  = Q1*Q2-Q12**2
	XI(1,1) = Q2/DX
	XI(1,2) = -Q12/DX
	XI(2,1) = XI(1,2)
	XI(2,2) = Q1/DX

	F1 = Q1-M22+M12
	F2 = Q2-M32+M12

c*** higher tensors may be needed in xC012345_cayley and xC01234_sing2
	CALL xB012345_(P2,M22,M32,B0  ,B1  ,B2  ,  B3,  B4,  B5,5,ext)
	CALL xB012345_(P3,M12,M32,B0_2,B1_2,B2_2,B3_2,B4_2,B5_2,5,ext)
	CALL xB012345_(P1,M12,M22,B0_3,B1_3,B2_3,B3_3,B4_3,B5_3,5,ext)

	R1(1) = (B0_2-B0-F1*C0)/2D0
	R1(2) = (B0_3-B0-F2*C0)/2D0

	DO 1 I=1,2
	C1(I) = 0D0
	DO 1 K=1,2
	C1(I) = C1(I) + XI(I,K)*R1(K)
1	CONTINUE

	C2(0,0) = M12*C0/2D0+(B0+F1*C1(1)+F2*C1(2)+1D0)/4D0

	if (nrank.eq.1) goto 1005

	R2(1,1) = (        B1(1)+B0-F1*C1(1))/2D0 - C2(0,0)
	R2(1,2) = (B1_2(1)-B1(1)   -F1*C1(2))/2D0
	R2(2,1) = (B1_3(1)+B1(1)+B0-F2*C1(1))/2D0
	R2(2,2) = (       -B1(1)   -F2*C1(2))/2D0 - C2(0,0)

	DO 2 I=1,2
	DO 2 J=1,2
	C2(I,J) = 0D0
	DO 2 K=1,2
	C2(I,J) = C2(I,J) + XI(I,K)*R2(K,J)
2	CONTINUE

	crelacc = abs( (4d0*C2(0,0)-1d0/2d0+p1*C2(1,1)+p3*C2(2,2)
     &		        +(p1+p3-p2)*C2(1,2)-m12*C0)/B0-1d0 )
      	crelacc = max( crelacc, abs(C2(1,2)/C2(2,1)-1d0) )

	R3X0(1) = (B2_2(0,0)-B2(0,0)-F1*C2(0,0))/2D0
	R3X0(2) = (B2_3(0,0)-B2(0,0)-F2*C2(0,0))/2D0

	DO 30 I=1,2
	C3(0,0,I) = 0D0
	DO 30 K=1,2
	C3(0,0,I) = C3(0,0,I) + XI(I,K)*R3X0(K)
30	CONTINUE

	if (nrank.eq.2) goto 1005

	BB = B2(1,1)+2D0*B1(1)+B0
	R3(1,1) = (         -BB     -F1*C2(1,1))/2D0
     *					- 2D0*C3(0,0,1)
	R3(1,2) = (B2_2(1,1)-B2(1,1)-F1*C2(2,2))/2D0
	R3(2,1) = (B2_3(1,1)-BB     -F2*C2(1,1))/2D0
	R3(2,2) = (         -B2(1,1)-F2*C2(2,2))/2D0
     *					- 2D0*C3(0,0,2)

	DO 3 I=1,2
	DO 3 J=1,2
	C3(I,J,J) = 0D0
	DO 3 K=1,2
	C3(I,J,J) = C3(I,J,J) + XI(I,K)*R3(K,J)
3	CONTINUE

c redundant relations
	BB = -B2(1,1)-B1(1)
	R3X12(1) = ( -BB -F1*C2(1,2))/2D0 - C3(0,0,2)
	R3X12(2) = ( -BB -F2*C2(1,2))/2D0 - C3(0,0,1)
	R3X21(1) = ( -BB -F1*C2(2,1))/2D0 - C3(0,0,2)
	R3X21(2) = ( -BB -F2*C2(2,1))/2D0 - C3(0,0,1)

	DO 300 I=1,2
	C3(I,1,2) = 0D0
	C3(I,2,1) = 0D0
	DO 300 K=1,2
	C3(I,1,2) = C3(I,1,2) + XI(I,K)*R3X12(K)
	C3(I,2,1) = C3(I,2,1) + XI(I,K)*R3X21(K)
300	CONTINUE

      	crelacc = max( crelacc, abs(C3(1,1,2)/C3(1,2,1)-1d0) )
      	crelacc = max( crelacc, abs(C3(1,1,2)/C3(2,1,1)-1d0) )
      	crelacc = max( crelacc, abs(C3(1,2,2)/C3(2,1,2)-1d0) )
      	crelacc = max( crelacc, abs(C3(1,2,2)/C3(2,2,1)-1d0) )

	C4(0,0,0,0) = B2(0,0)/8d0 + m12*C2(0,0)/4d0
     &		      + f1*C3(0,0,1)/8d0 + f2*C3(0,0,2)/8d0
     &		      + (m12+m22+m32)/48d0 -(p1+p2+p3)/192d0

	R4X0(1,1) = (            B3(0,0,1)+B2(0,0)-F1*C3(0,0,1))/2D0
     &			- C4(0,0,0,0)
	R4X0(1,2) = (B3_2(0,0,1)-B3(0,0,1)        -F1*C3(0,0,2))/2D0
	R4X0(2,1) = (B3_3(0,0,1)+B3(0,0,1)+B2(0,0)-F2*C3(0,0,1))/2D0
	R4X0(2,2) = (           -B3(0,0,1)        -F2*C3(0,0,2))/2D0
     &			- C4(0,0,0,0)

	DO 4 I=1,2
	DO 4 J=1,2
	C4(0,0,I,J) = 0D0
	DO 4 K=1,2
	C4(0,0,I,J) = C4(0,0,I,J) + XI(I,K)*R4X0(K,J)
4	CONTINUE

	if (B2(0,0).ne.(0d0,0d0)) then
	  crelacc = max( crelacc,
     &	  abs( (6d0*C4(0,0,0,0)+(p1+p2+p3)/48d0-(m12+m22+m32)/12d0
     &		+p1*C4(0,0,1,1)+p3*C4(0,0,2,2)+(p1+p3-p2)*C4(0,0,1,2)
     &		-m12*C2(0,0))/B2(0,0)-1d0 ) )
	endif
      	crelacc = max( crelacc, abs(C4(0,0,1,2)/C4(0,0,2,1)-1d0) )

	if (nrank.eq.3) goto 1005

	R4(1,1) = (             + B3(1,1,1)+3d0*B2(1,1)+3d0*B1(1)+B0
     &			    - f1*C3(1,1,1) )/2d0 - 3d0*C4(0,0,1,1)
	R4(1,2) = ( B3_2(1,1,1) - B3(1,1,1) - f1*C3(2,2,2) )/2d0
	R4(2,1) = ( B3_3(1,1,1) + B3(1,1,1)+3d0*B2(1,1)+3d0*B1(1)+B0
     &			    - f2*C3(1,1,1) )/2d0
	R4(2,2) = (             - B3(1,1,1) - f2*C3(2,2,2) )/2d0
     &			                         - 3d0*C4(0,0,2,2)

	DO 5 I=1,2
	DO 5 J=1,2
	C4(I,J,J,J) = 0D0
	DO 5 K=1,2
	C4(I,J,J,J) = C4(I,J,J,J) + XI(I,K)*R4(K,J)
5	CONTINUE

	BB = -B3(1,1,1)-B2(1,1)
	R4X112(1) = (-BB-f1*C3(1,1,2))/2d0-2d0*C4(0,0,1,2)
	R4X112(2) = (-BB-f2*C3(1,1,2))/2d0-C4(0,0,1,1)
	R4X121(1) = (-BB-f1*C3(1,2,1))/2d0-2d0*C4(0,0,1,2)
	R4X121(2) = (-BB-f2*C3(1,2,1))/2d0-C4(0,0,1,1)
	R4X211(1) = (-BB-f1*C3(2,1,1))/2d0-2d0*C4(0,0,1,2)
	R4X211(2) = (-BB-f2*C3(2,1,1))/2d0-C4(0,0,1,1)
	R4X122(1) = (-BB-f1*C3(1,2,2))/2d0-C4(0,0,2,2)
	R4X122(2) = (-BB-f2*C3(1,2,2))/2d0-2d0*C4(0,0,1,2)
	R4X212(1) = (-BB-f1*C3(2,1,2))/2d0-C4(0,0,2,2)
	R4X212(2) = (-BB-f2*C3(2,1,2))/2d0-2d0*C4(0,0,1,2)
	R4X221(1) = (-BB-f1*C3(2,2,1))/2d0-C4(0,0,2,2)
	R4X221(2) = (-BB-f2*C3(2,2,1))/2d0-2d0*C4(0,0,1,2)


	DO 6 I=1,2
	C4(I,1,1,2) = 0D0
	C4(I,1,2,1) = 0D0
	C4(I,2,1,1) = 0D0
	C4(I,1,2,2) = 0D0
	C4(I,2,1,2) = 0D0
	C4(I,2,2,1) = 0D0
	DO 6 K=1,2
	C4(I,1,1,2) = C4(I,1,1,2) + XI(I,K)*R4X112(K)
	C4(I,1,2,1) = C4(I,1,2,1) + XI(I,K)*R4X121(K)
	C4(I,2,1,1) = C4(I,2,1,1) + XI(I,K)*R4X211(K)
	C4(I,1,2,2) = C4(I,1,2,2) + XI(I,K)*R4X122(K)
	C4(I,2,1,2) = C4(I,2,1,2) + XI(I,K)*R4X212(K)
	C4(I,2,2,1) = C4(I,2,2,1) + XI(I,K)*R4X221(K)
6	CONTINUE

	do i=1,2
      	crelacc = max( crelacc, abs(C4(i,1,1,2)/C4(i,1,2,1)-1d0) )
      	crelacc = max( crelacc, abs(C4(i,1,1,2)/C4(i,2,1,1)-1d0) )
      	crelacc = max( crelacc, abs(C4(i,1,2,2)/C4(i,2,1,2)-1d0) )
      	crelacc = max( crelacc, abs(C4(i,1,2,2)/C4(i,2,2,1)-1d0) )
	enddo

1005   	continue

c safety factor
	crelacc = crelacc * 1d3

	rcrit = 1d-4

c PaVe reduction insufficient
	if (crelacc.gt.rcrit) then
	  ref  = (abs(p1)+abs(p2)+abs(p3))/3d0
          gdet = abs(q1*q2-q12**2)/ref**2
	  y12  = m12+m22-p1
	  y23  = m22+m32-p2
	  y13  = m12+m32-p3
	  ydet = abs(-y13**2*2d0*m22+2d0*y12*y13*y23-2d0*m12*y23**2
     &		     -y12**2*2d0*m32+8d0*m12*m22*m32)/ref**3

	  qsoft1  = (m12.eq.(0d0,0d0)).and.
     &		    (dcmplx(p1).eq.m22).and.(dcmplx(p3).eq.m32)
	  qsoft2  = (m22.eq.(0d0,0d0)).and.
     &		    (dcmplx(p2).eq.m32).and.(dcmplx(p1).eq.m12)
	  qsoft3  = (m32.eq.(0d0,0d0)).and.
     &		    (dcmplx(p3).eq.m12).and.(dcmplx(p2).eq.m22)
	  qcoll12 = (abs(m12+m22)+abs(p1).lt.1d-15)
	  qcoll23 = (abs(m22+m32)+abs(p2).lt.1d-15)
	  qcoll31 = (abs(m32+m12)+abs(p3).lt.1d-15)

	  if (qsoft1.or.qsoft2.or.qsoft3.or.
     &		qcoll12.or.qcoll23.or.qcoll31) ydet = 0d0

	  crelacc2 = 1d20
	  crelacc3 = 1d20

	  if (ydet.gt.1d-6) then
c Try Cayley reduction
	    cC0 = C0
cacb comment out 3 lines for faster calculation
      	    crelacc2 = 1d20
c	    call xC01234_cayley(P1,P2,P3,M12,M22,M32,
c     &               cC0,cC1,cC2,cC3,cC4,B0,B1,B2,B3,B4,B5,nrank,ext)
c      	    crelacc2 = abs(cC0/C0-1d0)
	    if (crelacc2.lt.rcrit) then
c Cayley reduction sufficient
	      crelmin = crelacc2
	      goto 400
	    endif
	  endif

	  if (qsoft1.or.qsoft2.or.qsoft3) then
c Soft-singular case (always ydet=0)
c improvement not yet implemented

	  elseif (qcoll12.or.qcoll23.or.qcoll31) then
c Collinear-singular case (always ydet=0)
	    eC0 = C0
	    call xC_coll(P1,P2,P3,M12,M22,M32,
     &                     eC0,eC1,eC2,eC3,eC4,crelacc3)
	    if (crelacc3.lt.rcrit) then
c specific treatment of colinear-singular case sufficient
	      crelmin = crelacc3
	      goto 400
	    endif

	  elseif ((gdet.lt.1d-2).and.(ydet.lt.1d-2)) then
c Try combined Gram/Cayley expansion
            call xC01234_sing2(P1,P2,P3,M12,M22,M32,
     &              eC0,eC1,eC2,eC3,eC4,B0,B1,B2,B3,B4,B5,4,ext)
      	    crelacc3 = abs(eC0/C0-1d0)
	    if (crelacc3.lt.rcrit) then
c Gram/Cayley expansion sufficient
	      crelmin = crelacc3
	      goto 400
	    endif
	  endif

	  crelmin = min(min(crelacc,crelacc2),crelacc3)

c Output for new worst case
c	  if (crelmin.gt.crelworst) then
c	    crelworst = crelmin
c	    write(*,*)
c	    write(*,*) 'Insufficient accuracy in C-tensors',rank
c	    write(*,*) 'crelacc  = ',crelacc
c	    write(*,*) 'crelacc2 = ',crelacc2
c	    write(*,*) 'crelacc3 = ',crelacc3
c	    write(*,*) 'gdet = ',gdet
c	    write(*,*) 'ydet = ',ydet
c	    write(99,*)
c	    write(99,*) 'q01 = ',p1
c	    write(99,*) 'q12 = ',p2
c	    write(99,*) 'q20 = ',p3
c	    write(99,*) 'm02 = ',m12
c	    write(99,*) 'm12 = ',m22
c	    write(99,*) 'm22 = ',m32
c	    write(99,*) 'rank = ',rank
c	    write(99,*) 'gdet = ',gdet
c	    write(99,*) 'ydet = ',ydet
c	    write(99,*) 'crelacc  = ',crelacc
c	    write(99,*) 'crelacc2 = ',crelacc2
c	    write(99,*) 'crelacc3 = ',crelacc3
c	  endif

400	  continue
	  if (crelmin.eq.crelacc2) then
	      C0            = cC0
	      C2(0,0)       = cC2(0,0)
	      C4(0,0,0,0)   = cC4(0,0,0,0)
	    do 500 i1=1,2
	      C1(i1)        = cC1(i1)
	      C3(0,0,i1)    = cC3(0,0,i1)
	    do 500 i2=1,2
	      C2(i1,i2)     = cC2(i1,i2)
	      C4(0,0,i1,i2) = cC4(0,0,i1,i2)
	    do 500 i3=1,2
	      C3(i1,i2,i3)  = cC3(i1,i2,i3)
500	    continue
	  elseif (crelmin.eq.crelacc3) then
	      C0            = eC0
	      C2(0,0)       = eC2(0,0)
	      C4(0,0,0,0)   = eC4(0,0,0,0)
	    do 600 i1=1,2
	      C1(i1)        = eC1(i1)
	      C3(0,0,i1)    = eC3(0,0,i1)
	    do 600 i2=1,2
	      C2(i1,i2)     = eC2(i1,i2)
	      C4(0,0,i1,i2) = eC4(0,0,i1,i2)
	    do 600 i3=1,2
	      C3(i1,i2,i3)  = eC3(i1,i2,i3)
600	    continue
	  endif

	endif

 1003   fct(1)=B0
        fct(2)=B1(1)
        fct(3)=B2(0,0)
        fct(4)=B2(1,1)
        fct(5)=B3(0,0,1)
        fct(6)=B3(1,1,1)
        fct(7)=B4(0,0,0,0)
        fct(8)=B4(0,0,1,1)
        fct(9)=B4(1,1,1,1)
        fct(10)=B5(0,0,0,0,1)
        fct(11)=B5(0,0,1,1,1)
        fct(12)=B5(1,1,1,1,1)
        fct(13)=C0
        fct(14)=C1(1)
        fct(15)=C1(2)
        fct(16)=C2(0,0)
        fct(17)=C2(1,1)
        fct(18)=C2(1,2)
        fct(19)=C2(2,2)
        fct(20)=C3(0,0,1)
        fct(21)=C3(0,0,2)
        fct(22)=C3(1,1,1)
        fct(23)=C3(1,1,2)
        fct(24)=C3(1,2,2)
        fct(25)=C3(2,2,2)
        fct(26)=C4(0,0,0,0)
        fct(27)=C4(0,0,1,1)
        fct(28)=C4(0,0,1,2)
        fct(29)=C4(0,0,2,2)
        call fastout(fct,29,4)

	END

***********************************************************************
	SUBROUTINE xD012345_(P1,P2,P3,P4,P23,P12,M12,M22,M32,M42,
     *			   D0,D1,D2,D3,D4,D5,C0,C1,C2,C3,C4,rank,ext)
***********************************************************************
*	4-POINT TENSOR COEFFICIENTS with complex masses
*-----------------------------------------------------------------------
*       11.6.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
	complex*16 xD0000_,xD000000_,aux
        COMPLEX*16 y12,y13,y14,y23,y24,y34
        COMPLEX*16 M12,M22,M32,M42,F1,F2,F3
        COMPLEX*16   B0,  B1(1),  B2(0:2,0:2),  B3(0:2,0:2,2)
        COMPLEX*16   C0,  C1(2),  C2(0:2,0:2),  C3(0:2,0:2,2)
	COMPLEX*16   C4(0:2,0:2,0:2,0:2)
        COMPLEX*16 C0_2,C1_2(2),C2_2(0:2,0:2),C3_2(0:2,0:2,2)
	COMPLEX*16 C4_2(0:2,0:2,0:2,0:2)
        COMPLEX*16 C0_3,C1_3(2),C2_3(0:2,0:2),C3_3(0:2,0:2,2)
	COMPLEX*16 C4_3(0:2,0:2,0:2,0:2)
        COMPLEX*16 C0_4,C1_4(2),C2_4(0:2,0:2),C3_4(0:2,0:2,2)
	COMPLEX*16 C4_4(0:2,0:2,0:2,0:2)
        COMPLEX*16 CC,S1(3),S2(3,3),S3X0(3),S3(3,3)
	COMPLEX*16 S3X12(3),S3X13(3),S3X23(3)
	COMPLEX*16 S4X0(3,3),S4(3,3),S4X233(3),S4X113(3),S4X122(3)
        COMPLEX*16 xD0_,D0,D1(3),D2(0:3,0:3),D3(0:3,0:3,3)
	COMPLEX*16 D4(0:3,0:3,0:3,0:3),D5(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 cD0,cD1(3),cD2(0:3,0:3),cD3(0:3,0:3,3)
	COMPLEX*16 cD4(0:3,0:3,0:3,0:3),cD5(0:3,0:3,0:3,0:3,3)
	COMPLEX*16 S5X00(3),S5X0IJJ(3,3),S5X023(3)
	COMPLEX*16 C00mn_sum1(2),C00mn_sum,C00m_sum,C00_sum
	REAL*8 YI(3,3),YY(3,3)
	INTEGER I,J,K,rank,ext
        complex*16 fct(110),x(16)
        integer nrank,i1,i2,i3,i4
        logical nocalc
        character*11 name

	common/dproblem/drelworst
	common/Csub/C0_2,C1_2,C2_2,C3_2,C4_2,
     &	            C0_3,C1_3,C2_3,C3_3,C4_3,
     &		    C0_4,C1_4,C2_4,C3_4,C4_4

        x(1)=p1
        x(2)=p2
        x(3)=p3
        x(4)=p4
        x(5)=p23
        x(6)=p12
        x(7)=m12
        x(8)=m22
        x(9)=m32
        x(10)=m42
        x(11)=ext
        nrank=rank
c at present only rank=3 is supported !
        nrank=3
        name='xD012345_'
        call fastin(fct,x,76,11,2,nrank,name,nocalc)
        if(nocalc)then
           C0=fct(1)
           C1(1)=fct(2)
           C1(2)=fct(3)
           C2(0,0)=fct(4)
           C2(1,1)=fct(5)
           C2(1,2)=fct(6)
           C2(2,2)=fct(7)
           C3(0,0,1)=fct(8)
           C3(0,0,2)=fct(9)
           C3(1,1,1)=fct(10)
           C3(1,1,2)=fct(11)
           C3(1,2,2)=fct(12)
           C3(2,2,2)=fct(13)
           C4(0,0,0,0)=fct(14)
           C4(0,0,1,1)=fct(15)
           C4(0,0,1,2)=fct(16)
           C4(0,0,2,2)=fct(17)
           D0=fct(18)
           D1(1)=fct(19)
           D1(2)=fct(20)
           D1(3)=fct(21)
           D2(0,0)=fct(22)
           D2(1,1)=fct(23)
           D2(1,2)=fct(24)
           D2(1,3)=fct(25)
           D2(2,2)=fct(26)
           D2(2,3)=fct(27)
           D2(3,3)=fct(28)
           D3(0,0,1)=fct(29)
           D3(0,0,2)=fct(30)
           D3(0,0,3)=fct(31)
           D3(1,1,1)=fct(32)
           D3(1,1,2)=fct(33)
           D3(1,1,3)=fct(34)
           D3(1,2,2)=fct(35)
           D3(1,2,3)=fct(36)
           D3(1,3,3)=fct(37)
           D3(2,2,2)=fct(38)
           D3(2,2,3)=fct(39)
           D3(2,3,3)=fct(40)
           D3(3,3,3)=fct(41)
           D4(0,0,0,0)=fct(42)
           D4(0,0,1,1)=fct(43)
           D4(0,0,1,2)=fct(44)
           D4(0,0,1,3)=fct(45)
           D4(0,0,2,2)=fct(46)
           D4(0,0,2,3)=fct(47)
           D4(0,0,3,3)=fct(48)
           D4(1,1,1,1)=fct(49)
           D4(1,1,1,2)=fct(50)
           D4(1,1,1,3)=fct(51)
           D4(1,1,2,2)=fct(52)
           D4(1,1,2,3)=fct(53)
           D4(1,1,3,3)=fct(54)
           D4(1,2,2,2)=fct(55)
           D4(1,2,2,3)=fct(56)
           D4(1,2,3,3)=fct(57)
           D4(1,3,3,3)=fct(58)
           D4(2,2,2,2)=fct(59)
           D4(2,2,2,3)=fct(60)
           D4(2,2,3,3)=fct(61)
           D4(2,3,3,3)=fct(62)
           D4(3,3,3,3)=fct(63)
           D5(0,0,0,0,1)=fct(64)
           D5(0,0,0,0,2)=fct(65)
           D5(0,0,0,0,3)=fct(66)
           D5(0,0,1,1,1)=fct(67)
           D5(0,0,1,1,2)=fct(68)
           D5(0,0,1,1,3)=fct(69)
           D5(0,0,1,2,2)=fct(70)
           D5(0,0,1,2,3)=fct(71)
           D5(0,0,1,3,3)=fct(72)
           D5(0,0,2,2,2)=fct(73)
           D5(0,0,2,2,3)=fct(74)
           D5(0,0,2,3,3)=fct(75)
           D5(0,0,3,3,3)=fct(76)
           C2(1,2)=C2(2,1)
           do i1=1,2
           do i2=i1,2
           do i3=i2,2
             C3(i1,i3,i2)=C3(i1,i2,i3)
             C3(i2,i1,i3)=C3(i1,i2,i3)
             C3(i2,i3,i1)=C3(i1,i2,i3)
             C3(i3,i1,i2)=C3(i1,i2,i3)
             C3(i3,i2,i1)=C3(i1,i2,i3)
           enddo
           enddo
           enddo
           C4(0,0,2,1)=C4(0,0,1,2)
           do i1=1,3
           do i2=i1+1,3
             D2(i2,i1)=D2(i1,i2)
           enddo
           enddo
           do i1=1,3
           do i2=i1,3
           do i3=i2,3
             D3(i1,i3,i2)=D3(i1,i2,i3)
             D3(i2,i1,i3)=D3(i1,i2,i3)
             D3(i2,i3,i1)=D3(i1,i2,i3)
             D3(i3,i1,i2)=D3(i1,i2,i3)
             D3(i3,i2,i1)=D3(i1,i2,i3)
           enddo
           enddo
           enddo
           do i1=1,3
           do i2=i1,3
           do i3=i2,3
           do i4=i3,3
             D4(i1,i2,i4,i3)=D4(i1,i2,i3,i4)
             D4(i1,i3,i2,i4)=D4(i1,i2,i3,i4)
             D4(i1,i3,i4,i2)=D4(i1,i2,i3,i4)
             D4(i1,i4,i2,i3)=D4(i1,i2,i3,i4)
             D4(i1,i4,i3,i2)=D4(i1,i2,i3,i4)
             D4(i2,i1,i3,i4)=D4(i1,i2,i3,i4)
             D4(i2,i1,i4,i3)=D4(i1,i2,i3,i4)
             D4(i2,i3,i1,i4)=D4(i1,i2,i3,i4)
             D4(i2,i3,i4,i1)=D4(i1,i2,i3,i4)
             D4(i2,i4,i1,i3)=D4(i1,i2,i3,i4)
             D4(i2,i4,i3,i1)=D4(i1,i2,i3,i4)
             D4(i3,i1,i2,i4)=D4(i1,i2,i3,i4)
             D4(i3,i1,i4,i2)=D4(i1,i2,i3,i4)
             D4(i3,i2,i1,i4)=D4(i1,i2,i3,i4)
             D4(i3,i2,i4,i1)=D4(i1,i2,i3,i4)
             D4(i3,i4,i1,i2)=D4(i1,i2,i3,i4)
             D4(i3,i4,i2,i1)=D4(i1,i2,i3,i4)
             D4(i4,i1,i2,i3)=D4(i1,i2,i3,i4)
             D4(i4,i1,i3,i2)=D4(i1,i2,i3,i4)
             D4(i4,i2,i1,i3)=D4(i1,i2,i3,i4)
             D4(i4,i2,i3,i1)=D4(i1,i2,i3,i4)
             D4(i4,i3,i1,i2)=D4(i1,i2,i3,i4)
             D4(i4,i3,i2,i1)=D4(i1,i2,i3,i4)
           enddo
           enddo
           enddo
           enddo
           do i1=1,3
           do i2=i1+1,3
             D4(0,0,i2,i1)=D4(0,0,i1,i2)
           enddo
           enddo
           do i1=1,3
           do i2=i1,3
           do i3=i2,3
             D5(0,0,i1,i3,i2)=D5(0,0,i1,i2,i3)
             D5(0,0,i2,i1,i3)=D5(0,0,i1,i2,i3)
             D5(0,0,i2,i3,i1)=D5(0,0,i1,i2,i3)
             D5(0,0,i3,i1,i2)=D5(0,0,i1,i2,i3)
             D5(0,0,i3,i2,i1)=D5(0,0,i1,i2,i3)
           enddo
           enddo
           enddo
           return
        endif

	D0 = xD0_(P1,P2,P3,P4,P12,P23,M12,M22,M32,M42,ext)

	if (nrank.eq.0) goto 1006

	Q1  = P1
	Q2  = P12
	Q3  = P4
	Q12 = (P1+P12-P2)/2D0
	Q23 = (P4+P12-P3)/2D0
	Q13 = (P1+P4-P23)/2D0

	DY  = Q1*Q2*Q3+2D0*Q12*Q23*Q13
     *		-Q1*Q23**2-Q2*Q13**2-Q3*Q12**2

	YI(1,1) = (Q2*Q3-Q23**2)/DY
	YI(1,2) = (Q13*Q23-Q3*Q12)/DY
	YI(1,3) = (Q12*Q23-Q2*Q13)/DY
	YI(2,1) = YI(1,2)
	YI(2,2) = (Q1*Q3-Q13**2)/DY
	YI(2,3) = (Q12*Q13-Q1*Q23)/DY
	YI(3,1) = YI(1,3)
	YI(3,2) = YI(2,3)
	YI(3,3) = (Q1*Q2-Q12**2)/DY

c	yy(1,1) = q1
c	yy(1,2) = q12
c	yy(1,3) = q13
c	yy(2,1) = yy(1,2)
c	yy(2,2) = q2
c	yy(2,3) = q23
c	yy(3,1) = yy(1,3)
c	yy(3,2) = yy(2,3)
c	yy(3,3) = q3
c	call inverse(yy,yi,3)

	F1 = Q1-M22+M12
	F2 = Q2-M32+M12
	F3 = Q3-M42+M12

c*** higher tensors may be needed in xD012345_cayley
	CALL xC01234_(P2,P3,P23,M22,M32,M42,
     &			C0  ,C1  ,C2  ,C3  ,C4  ,B0,B1,B2,B3,3,ext)
	CALL xC01234_(P12,P3,P4,M12,M32,M42,
     &			C0_2,C1_2,C2_2,C3_2,C4_2,B0,B1,B2,B3,3,ext)
	CALL xC01234_(P1,P23,P4,M12,M22,M42,
     &			C0_3,C1_3,C2_3,C3_3,C4_3,B0,B1,B2,B3,3,ext)
	CALL xC01234_(P1,P2,P12,M12,M22,M32,
     &			C0_4,C1_4,C2_4,C3_4,C4_4,B0,B1,B2,B3,3,ext)

	S1(1) = (C0_2-C0-F1*D0)/2D0
	S1(2) = (C0_3-C0-F2*D0)/2D0
	S1(3) = (C0_4-C0-F3*D0)/2D0

	DO 1 I=1,3
	D1(I) = 0D0
	DO 1 K=1,3
	D1(I) = D1(I) + YI(I,K)*S1(K)
1	CONTINUE

	D2(0,0) = M12*D0+(C0
     *			+F1*D1(1)+F2*D1(2)+F3*D1(3))/2D0

	if (nrank.eq.1) goto 1006

	CC = C1(1)+C1(2)+C0
	S2(1,1) = (        CC    - F1*D1(1))/2D0 - D2(0,0)
	S2(1,2) = (C1_2(1)-C1(1) - F1*D1(2))/2D0
	S2(1,3) = (C1_2(2)-C1(2) - F1*D1(3))/2D0
	S2(2,1) = (C1_3(1)+CC    - F2*D1(1))/2D0
	S2(2,2) = (       -C1(1) - F2*D1(2))/2D0 - D2(0,0)
	S2(2,3) = (C1_3(2)-C1(2) - F2*D1(3))/2D0
	S2(3,1) = (C1_4(1)+CC    - F3*D1(1))/2D0
	S2(3,2) = (C1_4(2)-C1(1) - F3*D1(2))/2D0
	S2(3,3) = (       -C1(2) - F3*D1(3))/2D0 - D2(0,0)

	DO 2 I=1,3
	DO 2 J=1,3
	D2(I,J) = 0D0
	DO 2 K=1,3
	D2(I,J) = D2(I,J) + YI(I,K)*S2(K,J)
2	CONTINUE

	drelacc = abs( (4d0*D2(0,0)+p1*D2(1,1)+p12*D2(2,2)+p4*D2(3,3)
     &		        +(p1+p12-p2)*D2(1,2)+(p4+p12-p3)*D2(2,3)
     &		        +(p1+p4-p23)*D2(1,3)-m12*D0)/C0-1d0 )

      	drelacc = max( drelacc, abs(D2(1,2)/D2(2,1)-1d0) )
      	drelacc = max( drelacc, abs(D2(1,3)/D2(3,1)-1d0) )
      	drelacc = max( drelacc, abs(D2(2,3)/D2(3,2)-1d0) )

	S3X0(1) = (C2_2(0,0)-C2(0,0)-F1*D2(0,0))/2D0
	S3X0(2) = (C2_3(0,0)-C2(0,0)-F2*D2(0,0))/2D0
	S3X0(3) = (C2_4(0,0)-C2(0,0)-F3*D2(0,0))/2D0

	DO 30 I=1,3
	D3(0,0,I) = 0D0
	DO 30 K=1,3
	D3(0,0,I) = D3(0,0,I) + YI(I,K)*S3X0(K)
30	CONTINUE

	if (nrank.eq.2) goto 1006

	CC = C2(1,1)+2D0*C2(1,2)+C2(2,2)
     *		+2D0*C1(1)+2D0*C1(2)+C0
	S3(1,1) = (         -CC      - F1*D2(1,1))/2D0
     *			- 2D0*D3(0,0,1)
	S3(1,2) = (C2_2(1,1)-C2(1,1) - F1*D2(2,2))/2D0
	S3(1,3) = (C2_2(2,2)-C2(2,2) - F1*D2(3,3))/2D0
	S3(2,1) = (C2_3(1,1)-CC      - F2*D2(1,1))/2D0
	S3(2,2) = (         -C2(1,1) - F2*D2(2,2))/2D0
     *			- 2D0*D3(0,0,2)
	S3(2,3) = (C2_3(2,2)-C2(2,2) - F2*D2(3,3))/2D0
	S3(3,1) = (C2_4(1,1)-CC      - F3*D2(1,1))/2D0
	S3(3,2) = (C2_4(2,2)-C2(1,1) - F3*D2(2,2))/2D0
	S3(3,3) = (         -C2(2,2) - F3*D2(3,3))/2D0
     *			- 2D0*D3(0,0,3)

	DO 3 I=1,3
	DO 3 J=1,3
	D3(I,J,J) = 0D0
	DO 3 K=1,3
	D3(I,J,J) = D3(I,J,J) + YI(I,K)*S3(K,J)
3	CONTINUE

	S3X23(1) = (C2_2(1,2)-C2(1,2)-F1*D2(2,3))/2D0
	S3X23(2) = (         -C2(1,2)-F2*D2(2,3))/2D0 - D3(0,0,3)
	S3X23(3) = (         -C2(1,2)-F3*D2(2,3))/2D0 - D3(0,0,2)

	DO 31 I=1,3
	D3(I,2,3) = 0D0
	DO 31 K=1,3
	D3(I,2,3) = D3(I,2,3) + YI(I,K)*S3X23(K)
31	CONTINUE

c*** redundant relations
	CC = -C2(1,1)-C2(1,2)-C1(1)
	S3X12(1) = (         -CC     -F1*D2(1,2))/2D0 - D3(0,0,2)
	S3X12(2) = (         -CC     -F2*D2(1,2))/2D0 - D3(0,0,1)
	S3X12(3) = (C2_4(1,2)-CC     -F3*D2(1,2))/2D0
	CC = -C2(1,2)-C2(2,2)-C1(2)
	S3X13(1) = (         -CC     -F1*D2(1,3))/2D0 - D3(0,0,3)
	S3X13(2) = (C2_3(1,2)-CC     -F2*D2(1,3))/2D0
	S3X13(3) = (         -CC     -F3*D2(1,3))/2D0 - D3(0,0,1)

	DO 310 I=1,3
	D3(I,1,2) = 0D0
	D3(I,1,3) = 0D0
	DO 310 K=1,3
	D3(I,1,2) = D3(I,1,2) + YI(I,K)*S3X12(K)
	D3(I,1,3) = D3(I,1,3) + YI(I,K)*S3X13(K)
310	CONTINUE

	drelacc = max( drelacc, abs(D3(1,1,2)/D3(2,1,1)-1d0) )
	drelacc = max( drelacc, abs(D3(2,1,2)/D3(1,2,2)-1d0) )
	drelacc = max( drelacc, abs(D3(3,1,2)/D3(1,2,3)-1d0) )
	drelacc = max( drelacc, abs(D3(1,1,3)/D3(3,1,1)-1d0) )
	drelacc = max( drelacc, abs(D3(2,1,3)/D3(1,2,3)-1d0) )
	drelacc = max( drelacc, abs(D3(3,1,3)/D3(1,3,3)-1d0) )

c	D3(1,1,2) = D3(2,1,1)
c	D3(1,1,3) = D3(3,1,1)

c*** remaining non-ordered coefficients
	D3(1,3,2) = D3(1,2,3)
c	D3(2,1,3) = D3(1,2,3)
	D3(2,3,1) = D3(1,2,3)
c	D3(3,1,2) = D3(1,2,3)
	D3(3,2,1) = D3(1,2,3)
	D3(2,2,1) = D3(1,2,2)
	D3(3,3,1) = D3(1,3,3)
	D3(3,3,2) = D3(2,3,3)
	D3(1,2,1) = D3(2,1,1)
	D3(1,3,1) = D3(3,1,1)
c	D3(2,1,2) = D3(1,2,2)
	D3(2,3,2) = D3(3,2,2)
c	D3(3,1,3) = D3(1,3,3)

	D4(0,0,0,0) = M12*D2(0,0)/3D0 + (C2(0,0)
     *			+F1*D3(0,0,1)+F2*D3(0,0,2)
     *			+F3*D3(0,0,3)+1D0/6D0)/6D0

	CC = C3(0,0,1)+C3(0,0,2)+C2(0,0)
	S4X0(1,1) = (            CC
     *			- F1*D3(0,0,1))/2D0 - D4(0,0,0,0)
	S4X0(1,2) = (C3_2(0,0,1)-C3(0,0,1)
     *			- F1*D3(0,0,2))/2D0
	S4X0(1,3) = (C3_2(0,0,2)-C3(0,0,2)
     *			- F1*D3(0,0,3))/2D0
	S4X0(2,1) = (C3_3(0,0,1)+CC
     *			- F2*D3(0,0,1))/2D0
	S4X0(2,2) = (           -C3(0,0,1)
     *			- F2*D3(0,0,2))/2D0 - D4(0,0,0,0)
	S4X0(2,3) = (C3_3(0,0,2)-C3(0,0,2)
     *			- F2*D3(0,0,3))/2D0
	S4X0(3,1) = (C3_4(0,0,1)+CC
     *			- F3*D3(0,0,1))/2D0
	S4X0(3,2) = (C3_4(0,0,2)-C3(0,0,1)
     *			- F3*D3(0,0,2))/2D0
	S4X0(3,3) = (           -C3(0,0,2)
     *			- F3*D3(0,0,3))/2D0 - D4(0,0,0,0)

	DO 40 I=1,3
	DO 40 J=1,3
	D4(0,0,I,J) = 0D0
	DO 40 K=1,3
	D4(0,0,I,J) = D4(0,0,I,J) + YI(I,K)*S4X0(K,J)
40	CONTINUE

	drelacc = max( drelacc, abs( (6d0*D4(0,0,0,0)-1d0/12d0
     &		   +p1*D4(0,0,1,1)+p12*D4(0,0,2,2)+p4*D4(0,0,3,3)
     &		   +(p1+p12-p2)*D4(0,0,1,2)+(p4+p12-p3)*D4(0,0,2,3)
     &		   +(p1+p4-p23)*D4(0,0,1,3)-m12*D2(0,0))/C2(0,0)-1d0 ) )

      	drelacc = max( drelacc, abs(D4(0,0,1,2)/D4(0,0,2,1)-1d0) )
      	drelacc = max( drelacc, abs(D4(0,0,1,3)/D4(0,0,3,1)-1d0) )
      	drelacc = max( drelacc, abs(D4(0,0,2,3)/D4(0,0,3,2)-1d0) )

	if (nrank.eq.3) goto 1006

	CC = C3(1,1,1)+3D0*C3(1,1,2)+3D0*C3(1,2,2)+C3(2,2,2)
     *		+3D0*C2(1,1)+6D0*C2(1,2)+3D0*C2(2,2)
     *		+3D0*C1(1)+3D0*C1(2)+C0

	S4(1,1) = (            CC        - F1*D3(1,1,1))/2D0
     *			-3D0*D4(0,0,1,1)
	S4(1,2) = (C3_2(1,1,1)-C3(1,1,1) - F1*D3(2,2,2))/2D0
	S4(1,3) = (C3_2(2,2,2)-C3(2,2,2) - F1*D3(3,3,3))/2D0
	S4(2,1) = (C3_3(1,1,1)+CC        - F2*D3(1,1,1))/2D0
	S4(2,2) = (           -C3(1,1,1) - F2*D3(2,2,2))/2D0
     *			-3D0*D4(0,0,2,2)
	S4(2,3) = (C3_3(2,2,2)-C3(2,2,2) - F2*D3(3,3,3))/2D0
	S4(3,1) = (C3_4(1,1,1)+CC        - F3*D3(1,1,1))/2D0
	S4(3,2) = (C3_4(2,2,2)-C3(1,1,1) - F3*D3(2,2,2))/2D0
	S4(3,3) = (           -C3(2,2,2) - F3*D3(3,3,3))/2D0
     *			-3D0*D4(0,0,3,3)

	DO 4 I=1,3
	DO 4 J=1,3
	D4(I,J,J,J) = 0D0
	DO 4 K=1,3
	D4(I,J,J,J) = D4(I,J,J,J) + YI(I,K)*S4(K,J)
4	CONTINUE

	D4(1,1,1,2) = D4(2,1,1,1)
	D4(1,1,1,3) = D4(3,1,1,1)
	D4(2,2,2,3) = D4(3,2,2,2)
	D4(2,2,2,1) = D4(1,2,2,2)

	S4X233(1) = (C3_2(1,2,2)-C3(1,2,2)-F1*D3(2,3,3))/2D0
	S4X233(2) = (           -C3(1,2,2)-F2*D3(2,3,3))/2D0
     *						-D4(0,0,3,3)
	S4X233(3) = (           -C3(1,2,2)-F3*D3(2,3,3))/2D0
     *						-2D0*D4(0,0,2,3)

	DO 41 I=1,3
	D4(I,2,3,3) = 0D0
	DO 41 K=1,3
	D4(I,2,3,3) = D4(I,2,3,3) + YI(I,K)*S4X233(K)
41	CONTINUE

	CC = C3(1,1,2)+2D0*C3(1,2,2)+C3(2,2,2)
     *			+2D0*C2(1,2)+2D0*C2(2,2)+C1(2)
	S4X113(1) = (           -CC-F1*D3(1,1,3))/2D0
     *						-2D0*D4(0,0,1,3)
	S4X113(2) = (C3_3(1,1,2)-CC-F2*D3(1,1,3))/2D0
	S4X113(3) = (           -CC-F3*D3(1,1,3))/2D0
     *						-D4(0,0,1,1)

	DO 42 I=1,3
	D4(1,1,I,3) = 0D0
	DO 42 K=1,3
	D4(1,1,I,3) = D4(1,1,I,3) + YI(I,K)*S4X113(K)
42	CONTINUE

	CC = C3(1,1,1)+C3(1,1,2)+C2(1,1)
	S4X122(1) = (            CC-F1*D3(1,2,2))/2D0
     *						-D4(0,0,2,2)
	S4X122(2) = (            CC-F2*D3(1,2,2))/2D0
     *						-2D0*D4(0,0,1,2)
	S4X122(3) = (C3_4(1,2,2)+CC-F3*D3(1,2,2))/2D0

	DO 43 I=1,3
	D4(I,1,2,2) = 0D0
	DO 43 K=1,3
	D4(I,1,2,2) = D4(I,1,2,2) + YI(I,K)*S4X122(K)
43	CONTINUE

	D4(1,2,2,3) = D4(3,1,2,2)

c*** remaining non-ordered coefficients NOT set !

      	drelacc = max( drelacc, abs(D4(3,2,3,3)/D4(2,3,3,3)-1d0) )
      	drelacc = max( drelacc, abs(D4(1,1,1,3)/D4(3,1,1,1)-1d0) )
      	drelacc = max( drelacc, abs(D4(2,1,2,2)/D4(1,2,2,2)-1d0) )

	S5X00(1) = (C4_2(0,0,0,0)-C4(0,0,0,0)-f1*D4(0,0,0,0))/2d0
	S5X00(2) = (C4_3(0,0,0,0)-C4(0,0,0,0)-f2*D4(0,0,0,0))/2d0
	S5X00(3) = (C4_4(0,0,0,0)-C4(0,0,0,0)-f3*D4(0,0,0,0))/2d0

	DO 50 I=1,3
	D5(0,0,0,0,I) = 0D0
	DO 50 J=1,3
	D5(0,0,0,0,I) = D5(0,0,0,0,I) + YI(I,J)*S5X00(J)
50	CONTINUE

	C00mn_sum1(1)= C4(0,0,1,1) + C4(0,0,1,2)
	C00mn_sum1(2)= C4(0,0,1,2) + C4(0,0,2,2)
	C00mn_sum    = C00mn_sum1(1) + C00mn_sum1(2)
	C00m_sum     = C3(0,0,1) + C3(0,0,2)
	C00_sum      = C00mn_sum + 2d0*C00m_sum + C2(0,0)

	S5X0IJJ(1,1) = (             -C00_sum    -F1*D4(0,0,1,1))/2D0
     *			- 2D0*D5(0,0,0,0,1)
	S5X0IJJ(1,2) = (C4_2(0,0,1,1)-C4(0,0,1,1)-F1*D4(0,0,2,2))/2D0
	S5X0IJJ(1,3) = (C4_2(0,0,2,2)-C4(0,0,2,2)-F1*D4(0,0,3,3))/2D0
	S5X0IJJ(2,1) = (C4_3(0,0,1,1)-C00_sum    -F2*D4(0,0,1,1))/2D0
	S5X0IJJ(2,2) = (             -C4(0,0,1,1)-F2*D4(0,0,2,2))/2D0
     *			- 2D0*D5(0,0,0,0,2)
	S5X0IJJ(2,3) = (C4_3(0,0,2,2)-C4(0,0,2,2)-F2*D4(0,0,3,3))/2D0
	S5X0IJJ(3,1) = (C4_4(0,0,1,1)-C00_sum    -F3*D4(0,0,1,1))/2D0
	S5X0IJJ(3,2) = (C4_4(0,0,2,2)-C4(0,0,1,1)-F3*D4(0,0,2,2))/2D0
	S5X0IJJ(3,3) = (             -C4(0,0,2,2)-F3*D4(0,0,3,3))/2D0
     *			- 2D0*D5(0,0,0,0,3)

	DO 51 I=1,3
	DO 51 J=1,3
	D5(0,0,I,J,J) = 0D0
	DO 51 K=1,3
	D5(0,0,I,J,J) = D5(0,0,I,J,J) + YI(I,K)*S5X0IJJ(K,J)
51	CONTINUE

	S5X023(1) = (C4_2(0,0,1,2)-C4(0,0,1,2)-F1*D4(0,0,2,3))/2D0
	S5X023(2) = (             -C4(0,0,1,2)-F2*D4(0,0,2,3))/2D0
     *			- D5(0,0,0,0,3)
	S5X023(3) = (             -C4(0,0,1,2)-F3*D4(0,0,2,3))/2D0
     *			- D5(0,0,0,0,2)

	DO 52 I=1,3
	D5(0,0,I,2,3) = 0D0
	DO 52 K=1,3
	D5(0,0,I,2,3) = D5(0,0,I,2,3) + YI(I,K)*S5X023(K)
52	CONTINUE

	D5(0,0,1,1,2) = D5(0,0,2,1,1)
	D5(0,0,1,1,3) = D5(0,0,3,1,1)

c*** remaining non-ordered coefficients
	D5(0,0,1,3,2) = D5(0,0,1,2,3)
	D5(0,0,2,1,3) = D5(0,0,1,2,3)
	D5(0,0,2,3,1) = D5(0,0,1,2,3)
	D5(0,0,3,1,2) = D5(0,0,1,2,3)
	D5(0,0,3,2,1) = D5(0,0,1,2,3)
	D5(0,0,2,2,1) = D5(0,0,1,2,2)
	D5(0,0,3,3,1) = D5(0,0,1,3,3)
	D5(0,0,3,3,2) = D5(0,0,2,3,3)
	D5(0,0,1,2,1) = D5(0,0,2,1,1)
	D5(0,0,1,3,1) = D5(0,0,3,1,1)
	D5(0,0,2,1,2) = D5(0,0,1,2,2)
	D5(0,0,2,3,2) = D5(0,0,3,2,2)
	D5(0,0,3,1,3) = D5(0,0,1,3,3)

1006   	continue

c safety factor
	drelacc = drelacc * 1d3

	rcrit = 1d-3
	if (drelacc.gt.rcrit) then

c PaVe reduction insufficient
	  ref  = (abs(p1)+abs(p2)+abs(p3)+abs(p4)+abs(p12)+abs(p23))/6d0
          gdet = dy/ref**3
	  y12  = m12+m22-p1
	  y13  = m12+m32-p12
	  y14  = m12+m42-p4
	  y23  = m22+m32-p2
	  y24  = m22+m42-p23
	  y34  = m32+m42-p3
	  ydet = abs(
     &	  16D0*M12*M22*M32*M42-4D0*M32*M42*Y12**2-4D0*M22*M42*Y13**2-
     &    4D0*M22*M32*Y14**2+4D0*M42*Y12*Y13*Y23-4D0*M12*M42*Y23**2+
     &    Y14**2*Y23**2+4D0*M32*Y12*Y14*Y24-2D0*Y13*Y14*Y23*Y24-4D0*
     &    M12*M32*Y24**2+Y13**2*Y24**2+4D0*M22*Y13*Y14*Y34-2D0*Y12*
     &    Y14*Y23*Y34-2D0*Y12*Y13*Y24*Y34+4D0*M12*Y23*Y24*Y34-4D0*M12*
     &    M22*Y34**2+Y12**2*Y34**2)/ref**4

          drelacc2 = 1d20
	  if (ydet.gt.1d-15) then
c Try Cayley reduction
c            call xD012345_cayley(P1,P2,P3,P4,P23,P12,M12,M22,M32,M42,
c     *             cD0,cD1,cD2,cD3,cD4,cD5,C0,C1,C2,C3,C4,nrank,ext)
c      	    drelacc2 = abs(cD0/D0-1d0)
	    if (drelacc2.lt.rcrit) then
c Cayley reduction sufficient
              drelmin = drelacc2
              goto 400
            endif
          endif
          drelmin = min(drelacc,drelacc2)

c	  if (drelmin.gt.drelworst) then
c            drelworst = drelmin
c	    write(*,*)
c	    write(*,*) 'Insufficient accuracy in D-tensors',rank
c	    write(*,*) 'drelacc = ',drelacc
c	    write(*,*) 'drelacc2 = ',drelacc2
c	    write(*,*) 'gdet = ',gdet
c	    write(*,*) 'ydet = ',ydet
c	    write(99,*)
c	    write(99,*) 'p1  = ',p1
c	    write(99,*) 'p2  = ',p2
c	    write(99,*) 'p3  = ',p3
c	    write(99,*) 'p4  = ',p4
c	    write(99,*) 'p23 = ',p23
c	    write(99,*) 'p12 = ',p12
c	    write(99,*) 'm02 = ',m12
c	    write(99,*) 'm12 = ',m22
c	    write(99,*) 'm22 = ',m32
c	    write(99,*) 'm32 = ',m42
c	    write(99,*) 'rank = ',rank
c	    write(99,*) 'gdet = ',gdet
c	    write(99,*) 'ydet = ',ydet
c	    write(99,*) 'drelacc  = ',drelacc
c	    write(99,*) 'drelacc2 = ',drelacc2
c	  endif

400       continue
          if (drelmin.eq.drelacc2) then
              D0            = cD0
              D2(0,0)       = cD2(0,0)
              D4(0,0,0,0)   = cD4(0,0,0,0)
            do 500 i1=1,3
              D1(i1)        = cD1(i1)
              D3(0,0,i1)    = cD3(0,0,i1)
            do 500 i2=1,3
              D2(i1,i2)     = cD2(i1,i2)
              D4(0,0,i1,i2) = cD4(0,0,i1,i2)
            do 500 i3=1,3
              D3(i1,i2,i3)  = cD3(i1,i2,i3)
500         continue
          endif
	endif

 1001   fct(1)=C0
        fct(2)=C1(1)
        fct(3)=C1(2)
        fct(4)=C2(0,0)
        fct(5)=C2(1,1)
        fct(6)=C2(1,2)
        fct(7)=C2(2,2)
        fct(8)=C3(0,0,1)
        fct(9)=C3(0,0,2)
        fct(10)=C3(1,1,1)
        fct(11)=C3(1,1,2)
        fct(12)=C3(1,2,2)
        fct(13)=C3(2,2,2)
        fct(14)=C4(0,0,0,0)
        fct(15)=C4(0,0,1,1)
        fct(16)=C4(0,0,1,2)
        fct(17)=C4(0,0,2,2)
        fct(18)=D0
        fct(19)=D1(1)
        fct(20)=D1(2)
        fct(21)=D1(3)
        fct(22)=D2(0,0)
        fct(23)=D2(1,1)
        fct(24)=D2(1,2)
        fct(25)=D2(1,3)
        fct(26)=D2(2,2)
        fct(27)=D2(2,3)
        fct(28)=D2(3,3)
        fct(29)=D3(0,0,1)
        fct(30)=D3(0,0,2)
        fct(31)=D3(0,0,3)
        fct(32)=D3(1,1,1)
        fct(33)=D3(1,1,2)
        fct(34)=D3(1,1,3)
        fct(35)=D3(1,2,2)
        fct(36)=D3(1,2,3)
        fct(37)=D3(1,3,3)
        fct(38)=D3(2,2,2)
        fct(39)=D3(2,2,3)
        fct(40)=D3(2,3,3)
        fct(41)=D3(3,3,3)
        fct(42)=D4(0,0,0,0)
        fct(43)=D4(0,0,1,1)
        fct(44)=D4(0,0,1,2)
        fct(45)=D4(0,0,1,3)
        fct(46)=D4(0,0,2,2)
        fct(47)=D4(0,0,2,3)
        fct(48)=D4(0,0,3,3)
        fct(49)=D4(1,1,1,1)
        fct(50)=D4(1,1,1,2)
        fct(51)=D4(1,1,1,3)
        fct(52)=D4(1,1,2,2)
        fct(53)=D4(1,1,2,3)
        fct(54)=D4(1,1,3,3)
        fct(55)=D4(1,2,2,2)
        fct(56)=D4(1,2,2,3)
        fct(57)=D4(1,2,3,3)
        fct(58)=D4(1,3,3,3)
        fct(59)=D4(2,2,2,2)
        fct(60)=D4(2,2,2,3)
        fct(61)=D4(2,2,3,3)
        fct(62)=D4(2,3,3,3)
        fct(63)=D4(3,3,3,3)
        fct(64)=D5(0,0,0,0,1)
        fct(65)=D5(0,0,0,0,2)
        fct(66)=D5(0,0,0,0,3)
        fct(67)=D5(0,0,1,1,1)
        fct(68)=D5(0,0,1,1,2)
        fct(69)=D5(0,0,1,1,3)
        fct(70)=D5(0,0,1,2,2)
        fct(71)=D5(0,0,1,2,3)
        fct(72)=D5(0,0,1,3,3)
        fct(73)=D5(0,0,2,2,2)
        fct(74)=D5(0,0,2,2,3)
        fct(75)=D5(0,0,2,3,3)
        fct(76)=D5(0,0,3,3,3)
        call fastout(fct,76,2)

	END

***********************************************************************
	SUBROUTINE xE01234_(p1,p2,p3,p4,p5,s12,s23,s34,s45,s15,
     &		m02,m12,m22,m32,m42,E0,E1,E2,E3,E4,D0,D1,D2,D3,rank,ext)
***********************************************************************
*	5-POINT TENSOR COEFFICIENTS with complex masses
*-----------------------------------------------------------------------
*       11.6.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
        COMPLEX*16 m02,m12,m22,m32,m42,f(4)
        COMPLEX*16 C0,C1(2),C2(0:2,0:2),C3(0:2,0:2,2)
	COMPLEX*16 C4(0:2,0:2,0:2,0:2)
        COMPLEX*16 D0,D1(3),D2(0:3,0:3),D3(0:3,0:3,3)
	COMPLEX*16 D4(0:3,0:3,0:3,0:3),D5(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_1,D1_1(3),D2_1(0:3,0:3),D3_1(0:3,0:3,3)
	COMPLEX*16 D4_1(0:3,0:3,0:3,0:3),D5_1(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_2,D1_2(3),D2_2(0:3,0:3),D3_2(0:3,0:3,3)
	COMPLEX*16 D4_2(0:3,0:3,0:3,0:3),D5_2(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_3,D1_3(3),D2_3(0:3,0:3),D3_3(0:3,0:3,3)
	COMPLEX*16 D4_3(0:3,0:3,0:3,0:3),D5_3(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_4,D1_4(3),D2_4(0:3,0:3),D3_4(0:3,0:3,3)
	COMPLEX*16 D4_4(0:3,0:3,0:3,0:3),D5_4(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 E0_,E0,E1(4),E2(4,4),E3(4,4,4),E4(4,4,4,4)
        COMPLEX*16 T1(4),T2(4,4),T3(4,4,4),T4(4,4,4,4)
        COMPLEX*16 D1sum,D2sum0,D2sum(3),D3sum00,D3sum0(3),D3sum(3,3)
        COMPLEX*16 D00diff(4),D00kdiff(4,4)
	REAL*8 ZI(4,4)
	INTEGER j,k,l,m,n,rank,ext

        E0 = xE0_(p1,p2,p3,p4,p5,s12,s23,s34,s45,s15,
     &		  m02,m12,m22,m32,m42,ext)
	if (rank.eq.0) return

	p1p1 = 2d0* p1
	p2p2 = 2d0*s12
	p3p3 = 2d0*s45
	p4p4 = 2d0* p5
	p1p2 = s12+ p1- p2
	p2p3 = s45+s12- p3
	p3p4 = s45+ p5- p4
	p1p3 =  p1+s45-s23
	p2p4 =  p5+s12-s34
	p1p4 =  p1+ p5-s15

	f(1) =  p1-m12+m02
	f(2) = s12-m22+m02
	f(3) = s45-m32+m02
	f(4) =  p5-m42+m02

        ZDET = P1P4**2*P2P3**2-2D0*P1P3*P1P4*P2P3*P2P4+P1P3**2*
     &  P2P4**2-P1P4**2*P2P2*P3P3+2D0*P1P2*P1P4*P2P4*P3P3-P1P1*
     &  P2P4**2*P3P3+2D0*P1P3*P1P4*P2P2*P3P4-2D0*P1P2*P1P4*P2P3*
     &  P3P4-2D0*P1P2*P1P3*P2P4*P3P4+2D0*P1P1*P2P3*P2P4*P3P4+P1P2**
     &  2*P3P4**2-P1P1*P2P2*P3P4**2-P1P3**2*P2P2*P4P4+2D0*P1P2*P1P3*
     &  P2P3*P4P4-P1P1*P2P3**2*P4P4-P1P2**2*P3P3*P4P4+P1P1*P2P2*
     &  P3P3*P4P4
        ZI(1,1) = (-P2P4**2*P3P3+2D0*P2P3*P2P4*P3P4-P2P2*P3P4**2-
     &  	   P2P3**2*P4P4+P2P2*P3P3*P4P4)/ZDET
        ZI(1,2) = (P1P4*P2P4*P3P3-P1P4*P2P3*P3P4-P1P3*P2P4*P3P4+
     &  	   P1P2*P3P4**2+P1P3*P2P3*P4P4-P1P2*P3P3*P4P4)/ZDET
        ZI(1,3) = (-P1P4*P2P3*P2P4+P1P3*P2P4**2+P1P4*P2P2*P3P4-
     &  	   P1P2*P2P4*P3P4-P1P3*P2P2*P4P4+P1P2*P2P3*P4P4)/ZDET
        ZI(1,4) = (P1P4*P2P3**2-P1P3*P2P3*P2P4-P1P4*P2P2*P3P3+P1P2*
     &  	   P2P4*P3P3+P1P3*P2P2*P3P4-P1P2*P2P3*P3P4)/ZDET
        ZI(2,2) = (-P1P4**2*P3P3+2D0*P1P3*P1P4*P3P4-P1P1*P3P4**2-
     &  	   P1P3**2*P4P4+P1P1*P3P3*P4P4)/ZDET
        ZI(2,3) = (P1P4**2*P2P3-P1P3*P1P4*P2P4-P1P2*P1P4*P3P4+P1P1*
     &  	   P2P4*P3P4+P1P2*P1P3*P4P4-P1P1*P2P3*P4P4)/ZDET
        ZI(2,4) = (-P1P3*P1P4*P2P3+P1P3**2*P2P4+P1P2*P1P4*P3P3-
     &  	   P1P1*P2P4*P3P3-P1P2*P1P3*P3P4+P1P1*P2P3*P3P4)/ZDET
        ZI(3,3) = (-P1P4**2*P2P2+2D0*P1P2*P1P4*P2P4-P1P1*P2P4**2-
     &  	   P1P2**2*P4P4+P1P1*P2P2*P4P4)/ZDET
        ZI(3,4) = (P1P3*P1P4*P2P2-P1P2*P1P4*P2P3-P1P2*P1P3*P2P4+
     &  	   P1P1*P2P3*P2P4+P1P2**2*P3P4-P1P1*P2P2*P3P4)/ZDET
        ZI(4,4) = (-P1P3**2*P2P2+2D0*P1P2*P1P3*P2P3-P1P1*P2P3**2-
     &  	   P1P2**2*P3P3+P1P1*P2P2*P3P3)/ZDET
        ZI(2,1) = ZI(1,2)
        ZI(3,1) = ZI(1,3)
        ZI(3,2) = ZI(2,3)
        ZI(4,1) = ZI(1,4)
        ZI(4,2) = ZI(2,4)
        ZI(4,3) = ZI(3,4)

c*** Note order of arguments 5 and 6 in D01234_ !
	call xD012345_( p2, p3, p4,s15,s34,s23,m12,m22,m32,m42,
     *		D0,D1,D2,D3,D4,D5,C0,C1,C2,C3,C4,rank-1,ext)
 	call xD012345_(s12, p3, p4, p5,s34,s45,m02,m22,m32,m42,
     *		D0_1,D1_1,D2_1,D3_1,D4_1,D5_1,C0,C1,C2,C3,C4,rank-1,ext)
	call xD012345_( p1,s23, p4, p5,s15,s45,m02,m12,m32,m42,
     *		D0_2,D1_2,D2_2,D3_2,D4_2,D5_2,C0,C1,C2,C3,C4,rank-1,ext)
	call xD012345_( p1, p2,s34, p5,s15,s12,m02,m12,m22,m42,
     *		D0_3,D1_3,D2_3,D3_3,D4_3,D5_3,C0,C1,C2,C3,C4,rank-1,ext)
	call xD012345_( p1, p2, p3,s45,s23,s12,m02,m12,m22,m32,
     *		D0_4,D1_4,D2_4,D3_4,D4_4,D5_4,C0,C1,C2,C3,C4,rank-1,ext)

	T1(1) = D0_1-D0
	T1(2) = D0_2-D0
	T1(3) = D0_3-D0
	T1(4) = D0_4-D0

	do 1 k=1,4
	E1(k) = 0d0
	do 1 j=1,4
	E1(k) = E1(k) + ZI(k,j)*(T1(j)-f(j)*E0)
1	continue

	if (rank.eq.1) return

	D1sum = D0 + D1(1) + D1(2) + D1(3)

	T2(1,1) =         + D1sum
	T2(1,2) = D1_1(1) - D1(1)
	T2(1,3) = D1_1(2) - D1(2)
	T2(1,4) = D1_1(3) - D1(3)

	T2(2,1) = D1_2(1) + D1sum
	T2(2,2) =         - D1(1)
	T2(2,3) = D1_2(2) - D1(2)
	T2(2,4) = D1_2(3) - D1(3)

	T2(3,1) = D1_3(1) + D1sum
	T2(3,2) = D1_3(2) - D1(1)
	T2(3,3) =         - D1(2)
	T2(3,4) = D1_3(3) - D1(3)

	T2(4,1) = D1_4(1) + D1sum
	T2(4,2) = D1_4(2) - D1(1)
	T2(4,3) = D1_4(3) - D1(2)
	T2(4,4) =         - D1(3)

	do 2 k=1,4
	do 2 l=1,4
	E2(k,l) = 0d0
	do 2 j=1,4
	E2(k,l) = E2(k,l) + ZI(k,j)*(T2(j,l)-f(j)*E1(l))
2	continue

	if (rank.eq.2) return

	D2sum0   = D2(1,1)+D2(2,2)+D2(3,3)+2d0*(D2(1,2)+D2(1,3)+D2(2,3))
     &		   +2d0*(D1(1)+D1(2)+D1(3))+D0
	D2sum(1) = D2(1,1)+D2(1,2)+D2(1,3)+D1(1)
	D2sum(2) = D2(1,2)+D2(2,2)+D2(2,3)+D1(2)
	D2sum(3) = D2(1,3)+D2(2,3)+D2(3,3)+D1(3)

	D00diff(1) = D2_1(0,0) - D2(0,0)
	D00diff(2) = D2_2(0,0) - D2(0,0)
	D00diff(3) = D2_3(0,0) - D2(0,0)
	D00diff(4) = D2_4(0,0) - D2(0,0)

	T3(1,1,1) =           - D2sum0
	T3(1,1,2) =             D2sum(1)
	T3(1,1,3) =             D2sum(2)
	T3(1,1,4) =             D2sum(3)
	T3(1,2,2) = D2_1(1,1) - D2(1,1)
	T3(1,2,3) = D2_1(1,2) - D2(1,2)
	T3(1,2,4) = D2_1(1,3) - D2(1,3)
	T3(1,3,3) = D2_1(2,2) - D2(2,2)
	T3(1,3,4) = D2_1(2,3) - D2(2,3)
	T3(1,4,4) = D2_1(3,3) - D2(3,3)

	T3(2,1,1) = D2_2(1,1) - D2sum0
	T3(2,1,2) =             D2sum(1)
	T3(2,1,3) = D2_2(1,2) + D2sum(2)
	T3(2,1,4) = D2_2(1,3) + D2sum(3)
	T3(2,2,2) =           - D2(1,1)
	T3(2,2,3) =           - D2(1,2)
	T3(2,2,4) =           - D2(1,3)
	T3(2,3,3) = D2_2(2,2) - D2(2,2)
	T3(2,3,4) = D2_2(2,3) - D2(2,3)
	T3(2,4,4) = D2_2(3,3) - D2(3,3)

	T3(3,1,1) = D2_3(1,1) - D2sum0
	T3(3,1,2) = D2_3(1,2) + D2sum(1)
	T3(3,1,3) =             D2sum(2)
	T3(3,1,4) = D2_3(1,3) + D2sum(3)
	T3(3,2,2) = D2_3(2,2) - D2(1,1)
	T3(3,2,3) =           - D2(1,2)
	T3(3,2,4) = D2_3(2,3) - D2(1,3)
	T3(3,3,3) =           - D2(2,2)
	T3(3,3,4) =           - D2(2,3)
	T3(3,4,4) = D2_3(3,3) - D2(3,3)

	T3(4,1,1) = D2_4(1,1) - D2sum0
	T3(4,1,2) = D2_4(1,2) + D2sum(1)
	T3(4,1,3) = D2_4(1,3) + D2sum(2)
	T3(4,1,4) =             D2sum(3)
	T3(4,2,2) = D2_4(2,2) - D2(1,1)
	T3(4,2,3) = D2_4(2,3) - D2(1,2)
	T3(4,2,4) =           - D2(1,3)
	T3(4,3,3) = D2_4(3,3) - D2(2,2)
	T3(4,3,4) =           - D2(2,3)
	T3(4,4,4) =           - D2(3,3)

	do 30 j=1,4
	do 30 l=1,4
	do 30 m=1,l-1
	T3(j,l,m) = T3(j,m,l)
30	continue

	do 3 k=1,4
	do 3 l=1,4
	do 3 m=1,4
	E3(k,l,m) = 0d0
	do 3 j=1,4
	E3(k,l,m) = E3(k,l,m) + ZI(k,j)*( T3(j,l,m)
     &		+ 2d0*ZI(l,m)*D00diff(j) - f(j)*E2(l,m) )
3	continue

	if (rank.eq.3) return

	D3sum00    =              D0          + 6d0*ZI(1,1)*D2(0,0)
	do 41 l=1,3
	D3sum00    = D3sum00    + 3d0*D1(l)   + 6d0*ZI(1,1)*D3(0,0,l)
	D3sum0(l)  =              D1(l)       - 4d0*ZI(1,l+1)*D2(0,0)
	do 41 m=1,3
	D3sum00    = D3sum00    + 3d0*D2(l,m)
	D3sum0(l)  = D3sum0(l)  + 2d0*D2(l,m) - 4d0*ZI(1,l+1)*D3(0,0,m)
	D3sum(l,m) =              D2(l,m)     + 2d0*ZI(l+1,m+1)*D2(0,0)
	do 41 n=1,3
	D3sum00    = D3sum00    + D3(l,m,n)
	D3sum0(l)  = D3sum0(l)  + D3(l,m,n)
	D3sum(l,m) = D3sum(l,m) + D3(l,m,n) + 2d0*ZI(l+1,m+1)*D3(0,0,n)
41	continue

	D00kdiff(1,1) = 0d0
	D00kdiff(1,2) = D3_1(0,0,1) - D3(0,0,1)
	D00kdiff(1,3) = D3_1(0,0,2) - D3(0,0,2)
	D00kdiff(1,4) = D3_1(0,0,3) - D3(0,0,3)

	D00kdiff(2,1) = D3_2(0,0,1)
	D00kdiff(2,2) =             - D3(0,0,1)
	D00kdiff(2,3) = D3_2(0,0,2) - D3(0,0,2)
	D00kdiff(2,4) = D3_2(0,0,3) - D3(0,0,3)

	D00kdiff(3,1) = D3_3(0,0,1)
	D00kdiff(3,2) = D3_3(0,0,2) - D3(0,0,1)
	D00kdiff(3,3) =             - D3(0,0,2)
	D00kdiff(3,4) = D3_3(0,0,3) - D3(0,0,3)

	D00kdiff(4,1) = D3_4(0,0,1)
	D00kdiff(4,2) = D3_4(0,0,2) - D3(0,0,1)
	D00kdiff(4,3) = D3_4(0,0,3) - D3(0,0,2)
	D00kdiff(4,4) =             - D3(0,0,3)

	T4(1,1,1,1) =               D3sum00
	T4(1,1,1,2) =             - D3sum0(1)
	T4(1,1,1,3) =             - D3sum0(2)
	T4(1,1,1,4) =             - D3sum0(3)
	T4(1,1,2,2) =               D3sum(1,1)
	T4(1,1,2,3) =               D3sum(1,2)
	T4(1,1,2,4) =               D3sum(1,3)
	T4(1,1,3,3) =               D3sum(2,2)
	T4(1,1,3,4) =               D3sum(2,3)
	T4(1,1,4,4) =               D3sum(3,3)
	T4(1,2,2,2) = D3_1(1,1,1) - D3(1,1,1)
	T4(1,2,2,3) = D3_1(1,1,2) - D3(1,1,2)
	T4(1,2,2,4) = D3_1(1,1,3) - D3(1,1,3)
	T4(1,2,3,3) = D3_1(1,2,2) - D3(1,2,2)
	T4(1,2,3,4) = D3_1(1,2,3) - D3(1,2,3)
	T4(1,2,4,4) = D3_1(1,3,3) - D3(1,3,3)
	T4(1,3,3,3) = D3_1(2,2,2) - D3(2,2,2)
	T4(1,3,3,4) = D3_1(2,2,3) - D3(2,2,3)
	T4(1,3,4,4) = D3_1(2,3,3) - D3(2,3,3)
	T4(1,4,4,4) = D3_1(3,3,3) - D3(3,3,3)

	T4(2,1,1,1) = D3_2(1,1,1) + D3sum00
	T4(2,1,1,2) =             - D3sum0(1)
	T4(2,1,1,3) = D3_2(1,1,2) - D3sum0(2)
	T4(2,1,1,4) = D3_2(1,1,3) - D3sum0(3)
	T4(2,1,2,2) =               D3sum(1,1)
	T4(2,1,2,3) =               D3sum(1,2)
	T4(2,1,2,4) =               D3sum(1,3)
	T4(2,1,3,3) = D3_2(1,2,2) + D3sum(2,2)
	T4(2,1,3,4) = D3_2(1,2,3) + D3sum(2,3)
	T4(2,1,4,4) = D3_2(1,3,3) + D3sum(3,3)
	T4(2,2,2,2) =             - D3(1,1,1)
	T4(2,2,2,3) =             - D3(1,1,2)
	T4(2,2,2,4) =             - D3(1,1,3)
	T4(2,2,3,3) =             - D3(1,2,2)
	T4(2,2,3,4) =             - D3(1,2,3)
	T4(2,2,4,4) =             - D3(1,3,3)
	T4(2,3,3,3) = D3_2(2,2,2) - D3(2,2,2)
	T4(2,3,3,4) = D3_2(2,2,3) - D3(2,2,3)
	T4(2,3,4,4) = D3_2(2,3,3) - D3(2,3,3)
	T4(2,4,4,4) = D3_2(3,3,3) - D3(3,3,3)

	T4(3,1,1,1) = D3_3(1,1,1) + D3sum00
	T4(3,1,1,2) = D3_3(1,1,2) - D3sum0(1)
	T4(3,1,1,3) =             - D3sum0(2)
	T4(3,1,1,4) = D3_3(1,1,3) - D3sum0(3)
	T4(3,1,2,2) = D3_3(1,2,2) + D3sum(1,1)
	T4(3,1,2,3) =               D3sum(1,2)
	T4(3,1,2,4) = D3_3(1,2,3) + D3sum(1,3)
	T4(3,1,3,3) =               D3sum(2,2)
	T4(3,1,3,4) =               D3sum(2,3)
	T4(3,1,4,4) = D3_3(1,3,3) + D3sum(3,3)
	T4(3,2,2,2) = D3_3(2,2,2) - D3(1,1,1)
	T4(3,2,2,3) =             - D3(1,1,2)
	T4(3,2,2,4) = D3_3(2,2,3) - D3(1,1,3)
	T4(3,2,3,3) =             - D3(1,2,2)
	T4(3,2,3,4) =             - D3(1,2,3)
	T4(3,2,4,4) = D3_3(2,3,3) - D3(1,3,3)
	T4(3,3,3,3) =             - D3(2,2,2)
	T4(3,3,3,4) =             - D3(2,2,3)
	T4(3,3,4,4) =             - D3(2,3,3)
	T4(3,4,4,4) = D3_3(3,3,3) - D3(3,3,3)

	T4(4,1,1,1) = D3_4(1,1,1) + D3sum00
	T4(4,1,1,2) = D3_4(1,1,2) - D3sum0(1)
	T4(4,1,1,3) = D3_4(1,1,3) - D3sum0(2)
	T4(4,1,1,4) =             - D3sum0(3)
	T4(4,1,2,2) = D3_4(1,2,2) + D3sum(1,1)
	T4(4,1,2,3) = D3_4(1,2,3) + D3sum(1,2)
	T4(4,1,2,4) =               D3sum(1,3)
	T4(4,1,3,3) = D3_4(1,3,3) + D3sum(2,2)
	T4(4,1,3,4) =               D3sum(2,3)
	T4(4,1,4,4) =               D3sum(3,3)
	T4(4,2,2,2) = D3_4(2,2,2) - D3(1,1,1)
	T4(4,2,2,3) = D3_4(2,2,3) - D3(1,1,2)
	T4(4,2,2,4) =             - D3(1,1,3)
	T4(4,2,3,3) = D3_4(2,3,3) - D3(1,2,2)
	T4(4,2,3,4) =             - D3(1,2,3)
	T4(4,2,4,4) =             - D3(1,3,3)
	T4(4,3,3,3) = D3_4(3,3,3) - D3(2,2,2)
	T4(4,3,3,4) =             - D3(2,2,3)
	T4(4,3,4,4) =             - D3(2,3,3)
	T4(4,4,4,4) =             - D3(3,3,3)

	do 40 j=1,4
	do 40 l=1,4
	do 40 m=1,l-1
	T4(j,l,l,m) = T4(j,m,l,l)
	T4(j,l,m,l) = T4(j,m,l,l)
	T4(j,m,l,m) = T4(j,m,m,l)
	T4(j,l,m,m) = T4(j,m,m,l)
	do 40 n=1,m-1
	T4(j,n,l,m) = T4(j,n,m,l)
	T4(j,l,m,n) = T4(j,n,m,l)
	T4(j,l,n,m) = T4(j,n,m,l)
	T4(j,m,l,n) = T4(j,n,m,l)
	T4(j,m,n,l) = T4(j,n,m,l)
40	continue

	do 4 k=1,4
	do 4 l=1,4
	do 4 m=1,4
	do 4 n=1,4
	E4(k,l,m,n) = 0d0
	do 4 j=1,4
	E4(k,l,m,n) = E4(k,l,m,n) + ZI(k,j)*( T4(j,l,m,n)
     &		+ 2d0*ZI(m,n)*D00kdiff(j,l) + 2d0*ZI(l,n)*D00kdiff(j,m)
     &		+ 2d0*ZI(l,m)*D00kdiff(j,n) - f(j)*E3(l,m,n)	)
4	continue

	END

***********************************************************************
	SUBROUTINE xE01234_new(p1,p2,p3,p4,p5,s12,s23,s34,s45,s15,
     &	  m02,m12,m22,m32,m42,E0n,E1n,E2n,E3n,E4n,D0,D1,D2,D3,rank,ext)
***********************************************************************
*	5-POINT TENSOR COEFFICIENTS with complex masses
*	+++ new decomposition +++
*	*** new evaluation of determinants etc. ***
*-----------------------------------------------------------------------
*       11.6.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
        COMPLEX*16 m02,m12,m22,m32,m42,xdet
        COMPLEX*16 C0,C1(2),C2(0:2,0:2),C3(0:2,0:2,2)
	COMPLEX*16 C4(0:2,0:2,0:2,0:2)
        COMPLEX*16 D0,D1(3),D2(0:3,0:3),D3(0:3,0:3,3)
	COMPLEX*16 D4(0:3,0:3,0:3,0:3),D5(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_1,D1_1(3),D2_1(0:3,0:3),D3_1(0:3,0:3,3)
	COMPLEX*16 D4_1(0:3,0:3,0:3,0:3),D5_1(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_2,D1_2(3),D2_2(0:3,0:3),D3_2(0:3,0:3,3)
	COMPLEX*16 D4_2(0:3,0:3,0:3,0:3),D5_2(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_3,D1_3(3),D2_3(0:3,0:3),D3_3(0:3,0:3,3)
	COMPLEX*16 D4_3(0:3,0:3,0:3,0:3),D5_3(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 D0_4,D1_4(3),D2_4(0:3,0:3),D3_4(0:3,0:3,3)
	COMPLEX*16 D4_4(0:3,0:3,0:3,0:3),D5_4(0:3,0:3,0:3,0:3,3)
        COMPLEX*16 E0n,E1n(4),E2n(0:4,0:4),E3n(0:4,0:4,4)
        COMPLEX*16 E4n(0:4,0:4,0:4,0:4)
        COMPLEX*16 Dmnl_sum1(3,3),Dmnl_sum2(3),Dmn_sum1(3),D00mn_sum1(3)
        COMPLEX*16 Dmnl_sum,Dmn_sum,Dm_sum,D00mn_sum,D00m_sum
	COMPLEX*16 D0000m_sum,D00mnl_sum1(3,3),D00mnl_sum2(3),D00mnl_sum
	COMPLEX*16 Dmnlk_sum1(3,3,3),Dmnlk_sum2(3,3)
	COMPLEX*16 Dmnlk_sum3(3),Dmnlk_sum
	complex*16 y(0:4,0:4),dety(0:4),detz(4,4),detyy
	complex*16 yy(5,5),y0(5,5),y1(5,5),y2(5,5),y3(5,5),y4(5,5)
	INTEGER j,k,l,m,n,rank,ext
        complex*16 fct(110),x(16)
        integer nrank,i1,i2,i3,i4
        logical nocalc
        character*11 name

	det3(a11,a12,a13,a21,a22,a23,a31,a32,a33) =
     &		  a11*a22*a33+a12*a23*a31+a13*a21*a32
     &		 -a11*a23*a32-a12*a21*a33-a13*a22*a31

        x(1)=p1
        x(2)=p2
        x(3)=p3
        x(4)=p4
        x(5)=p5
        x(6)=s12
        x(7)=s23
        x(8)=s34
        x(9)=s45
        x(10)=s15
        x(11)=m02
        x(12)=m12
        x(13)=m22
        x(14)=m32
        x(15)=m42
        x(16)=ext
        nrank=rank
        name='xE01234_new'
        call fastin(fct,x,110,16,1,nrank,name,nocalc)
        if(nocalc)then
           D0=fct(1)
           D1(1)=fct(2)
           D1(2)=fct(3)
           D1(3)=fct(4)
           D2(0,0)=fct(5)
           D2(1,1)=fct(6)
           D2(1,2)=fct(7)
           D2(1,3)=fct(8)
           D2(2,2)=fct(9)
           D2(2,3)=fct(10)
           D2(3,3)=fct(11)
           D3(0,0,1)=fct(12)
           D3(0,0,2)=fct(13)
           D3(0,0,3)=fct(14)
           D3(1,1,1)=fct(15)
           D3(1,1,2)=fct(16)
           D3(1,1,3)=fct(17)
           D3(1,2,2)=fct(18)
           D3(1,2,3)=fct(19)
           D3(1,3,3)=fct(20)
           D3(2,2,2)=fct(21)
           D3(2,2,3)=fct(22)
           D3(2,3,3)=fct(23)
           D3(3,3,3)=fct(24)
           E0n=fct(25)
           E1n(1)=fct(26)
           E1n(2)=fct(27)
           E1n(3)=fct(28)
           E1n(4)=fct(29)
           E2n(0,0)=fct(30)
           E2n(1,1)=fct(31)
           E2n(1,2)=fct(32)
           E2n(1,3)=fct(33)
           E2n(1,4)=fct(34)
           E2n(2,2)=fct(35)
           E2n(2,3)=fct(36)
           E2n(2,4)=fct(37)
           E2n(3,3)=fct(38)
           E2n(3,4)=fct(39)
           E2n(4,4)=fct(40)
           E3n(0,0,1)=fct(41)
           E3n(0,0,2)=fct(42)
           E3n(0,0,3)=fct(43)
           E3n(0,0,4)=fct(44)
           E3n(1,1,1)=fct(45)
           E3n(1,1,2)=fct(46)
           E3n(1,1,3)=fct(47)
           E3n(1,1,4)=fct(48)
           E3n(1,2,2)=fct(49)
           E3n(1,2,3)=fct(50)
           E3n(1,2,4)=fct(51)
           E3n(1,3,3)=fct(52)
           E3n(1,3,4)=fct(53)
           E3n(1,4,4)=fct(54)
           E3n(2,2,2)=fct(55)
           E3n(2,2,3)=fct(56)
           E3n(2,2,4)=fct(57)
           E3n(2,3,3)=fct(58)
           E3n(2,3,4)=fct(59)
           E3n(2,4,4)=fct(60)
           E3n(3,3,3)=fct(61)
           E3n(3,3,4)=fct(62)
           E3n(3,4,4)=fct(63)
           E3n(4,4,4)=fct(64)
           E4n(0,0,0,0)=fct(65)
           E4n(0,0,1,1)=fct(66)
           E4n(0,0,1,2)=fct(67)
           E4n(0,0,1,3)=fct(68)
           E4n(0,0,1,4)=fct(69)
           E4n(0,0,2,2)=fct(70)
           E4n(0,0,2,3)=fct(71)
           E4n(0,0,2,4)=fct(72)
           E4n(0,0,3,3)=fct(73)
           E4n(0,0,3,4)=fct(74)
           E4n(0,0,4,4)=fct(75)
           E4n(1,1,1,1)=fct(76)
           E4n(1,1,1,2)=fct(77)
           E4n(1,1,1,3)=fct(78)
           E4n(1,1,1,4)=fct(79)
           E4n(1,1,2,2)=fct(80)
           E4n(1,1,2,3)=fct(81)
           E4n(1,1,2,4)=fct(82)
           E4n(1,1,3,3)=fct(83)
           E4n(1,1,3,4)=fct(84)
           E4n(1,1,4,4)=fct(85)
           E4n(1,2,2,2)=fct(86)
           E4n(1,2,2,3)=fct(87)
           E4n(1,2,2,4)=fct(88)
           E4n(1,2,3,3)=fct(89)
           E4n(1,2,3,4)=fct(90)
           E4n(1,2,4,4)=fct(91)
           E4n(1,3,3,3)=fct(92)
           E4n(1,3,3,4)=fct(93)
           E4n(1,3,4,4)=fct(94)
           E4n(1,4,4,4)=fct(95)
           E4n(2,2,2,2)=fct(96)
           E4n(2,2,2,3)=fct(97)
           E4n(2,2,2,4)=fct(98)
           E4n(2,2,3,3)=fct(99)
           E4n(2,2,3,4)=fct(100)
           E4n(2,2,4,4)=fct(101)
           E4n(2,3,3,3)=fct(102)
           E4n(2,3,3,4)=fct(103)
           E4n(2,3,4,4)=fct(104)
           E4n(2,4,4,4)=fct(105)
           E4n(3,3,3,3)=fct(106)
           E4n(3,3,3,4)=fct(107)
           E4n(3,3,4,4)=fct(108)
           E4n(3,4,4,4)=fct(109)
           E4n(4,4,4,4)=fct(110)
           do i1=1,3
           do i2=i1+1,3
             D2(i2,i1)=D2(i1,i2)
           enddo
           enddo
           do i1=1,3
           do i2=i1,3
           do i3=i2,3
             D3(i1,i3,i2)=D3(i1,i2,i3)
             D3(i2,i1,i3)=D3(i1,i2,i3)
             D3(i2,i3,i1)=D3(i1,i2,i3)
             D3(i3,i1,i2)=D3(i1,i2,i3)
             D3(i3,i2,i1)=D3(i1,i2,i3)
           enddo
           enddo
           enddo
           do i1=1,4
           do i2=i1+1,4
             E2n(i2,i1)=E2n(i1,i2)
           enddo
           enddo
           do i1=1,4
           do i2=i1,4
           do i3=i2,4
             E3n(i1,i3,i2)=E3n(i1,i2,i3)
             E3n(i2,i1,i3)=E3n(i1,i2,i3)
             E3n(i2,i3,i1)=E3n(i1,i2,i3)
             E3n(i3,i1,i2)=E3n(i1,i2,i3)
             E3n(i3,i2,i1)=E3n(i1,i2,i3)
           enddo
           enddo
           enddo
           do i1=1,4
           do i2=i1,4
           do i3=i2,4
           do i4=i3,4
             E4n(i1,i2,i4,i3)=E4n(i1,i2,i3,i4)
             E4n(i1,i3,i2,i4)=E4n(i1,i2,i3,i4)
             E4n(i1,i3,i4,i2)=E4n(i1,i2,i3,i4)
             E4n(i1,i4,i2,i3)=E4n(i1,i2,i3,i4)
             E4n(i1,i4,i3,i2)=E4n(i1,i2,i3,i4)
             E4n(i2,i1,i3,i4)=E4n(i1,i2,i3,i4)
             E4n(i2,i1,i4,i3)=E4n(i1,i2,i3,i4)
             E4n(i2,i3,i1,i4)=E4n(i1,i2,i3,i4)
             E4n(i2,i3,i4,i1)=E4n(i1,i2,i3,i4)
             E4n(i2,i4,i1,i3)=E4n(i1,i2,i3,i4)
             E4n(i2,i4,i3,i1)=E4n(i1,i2,i3,i4)
             E4n(i3,i1,i2,i4)=E4n(i1,i2,i3,i4)
             E4n(i3,i1,i4,i2)=E4n(i1,i2,i3,i4)
             E4n(i3,i2,i1,i4)=E4n(i1,i2,i3,i4)
             E4n(i3,i2,i4,i1)=E4n(i1,i2,i3,i4)
             E4n(i3,i4,i1,i2)=E4n(i1,i2,i3,i4)
             E4n(i3,i4,i2,i1)=E4n(i1,i2,i3,i4)
             E4n(i4,i1,i2,i3)=E4n(i1,i2,i3,i4)
             E4n(i4,i1,i3,i2)=E4n(i1,i2,i3,i4)
             E4n(i4,i2,i1,i3)=E4n(i1,i2,i3,i4)
             E4n(i4,i2,i3,i1)=E4n(i1,i2,i3,i4)
             E4n(i4,i3,i1,i2)=E4n(i1,i2,i3,i4)
             E4n(i4,i3,i2,i1)=E4n(i1,i2,i3,i4)
           enddo
           enddo
           enddo
           enddo
           return
        endif

c*** Note order of arguments 5 and 6 in D01234_ !
	call xD012345_( p2, p3, p4,s15,s34,s23,m12,m22,m32,m42,
     *	       D0,D1,D2,D3,D4,D5,C0,C1,C2,C3,C4,nrank,ext)
 	call xD012345_(s12, p3, p4, p5,s34,s45,m02,m22,m32,m42,
     *	       D0_1,D1_1,D2_1,D3_1,D4_1,D5_1,C0,C1,C2,C3,C4,nrank,ext)
	call xD012345_( p1,s23, p4, p5,s15,s45,m02,m12,m32,m42,
     *	       D0_2,D1_2,D2_2,D3_2,D4_2,D5_2,C0,C1,C2,C3,C4,nrank,ext)
	call xD012345_( p1, p2,s34, p5,s15,s12,m02,m12,m22,m42,
     *	       D0_3,D1_3,D2_3,D3_3,D4_3,D5_3,C0,C1,C2,C3,C4,nrank,ext)
	call xD012345_( p1, p2, p3,s45,s23,s12,m02,m12,m22,m32,
     *	       D0_4,D1_4,D2_4,D3_4,D4_4,D5_4,C0,C1,C2,C3,C4,nrank,ext)

	y(0,0) = 2d0*m02
	y(1,1) = 2d0*m12
	y(2,2) = 2d0*m22
	y(3,3) = 2d0*m32
	y(4,4) = 2d0*m42
	y(0,1) = m02+m12-p1
	y(1,2) = m12+m22-p2
	y(2,3) = m22+m32-p3
	y(3,4) = m32+m42-p4
	y(0,4) = m42+m02-p5
	y(0,2) = m02+m22-s12
	y(1,3) = m12+m32-s23
	y(2,4) = m22+m42-s34
	y(0,3) = m32+m02-s45
	y(1,4) = m42+m12-s15
	do j=0,3
	  do k=j+1,4
	    y(k,j) = y(j,k)
	  enddo
	enddo

	do j=1,5
	  do k=1,5
	    yy(j,k) = y(j-1,k-1)
	    if (k.ne.1) then
	      y0(j,k) = y(j-1,k-1)
	    else
	      y0(j,k) = 1d0
	    endif
	    if (k.ne.2) then
	      y1(j,k) = y(j-1,k-1)
	    else
	      y1(j,k) = 1d0
	    endif
	    if (k.ne.3) then
	      y2(j,k) = y(j-1,k-1)
	    else
	      y2(j,k) = 1d0
	    endif
	    if (k.ne.4) then
	      y3(j,k) = y(j-1,k-1)
	    else
	      y3(j,k) = 1d0
	    endif
	    if (k.ne.5) then
	      y4(j,k) = y(j-1,k-1)
	    else
	      y4(j,k) = 1d0
	    endif
	  enddo
	enddo

	detyy   =  xdet(yy,5)
	dety(0) = -xdet(y0,5)/detyy
	dety(1) = -xdet(y1,5)/detyy
	dety(2) = -xdet(y2,5)/detyy
	dety(3) = -xdet(y3,5)/detyy
	dety(4) = -xdet(y4,5)/detyy

	E0n = D0*dety(0)+D0_1*dety(1)+D0_2*dety(2)
     &	                +D0_3*dety(3)+D0_4*dety(4)

	if (nrank.eq.0) goto 1007

	p1p1 = 2d0* p1
	p2p2 = 2d0*s12
	p3p3 = 2d0*s45
	p4p4 = 2d0* p5
	p1p2 = s12+ p1- p2
	p2p3 = s45+s12- p3
	p3p4 = s45+ p5- p4
	p1p3 =  p1+s45-s23
	p2p4 =  p5+s12-s34
	p1p4 =  p1+ p5-s15

	detz(1,1) = det3(p2p2,p2p3,p2p4,
     &		         p2p3,p3p3,p3p4,
     &		         p2p4,p3p4,p4p4)/detyy
	detz(1,2) = det3(p1p2,p2p3,p2p4,
     &		 	 p1p3,p3p3,p3p4,
     &		 	 p1p4,p3p4,p4p4)/detyy
	detz(1,3) = det3(p1p2,p2p2,p2p4,
     &		 	 p1p3,p2p3,p3p4,
     &		 	 p1p4,p2p4,p4p4)/detyy
	detz(1,4) = det3(p1p2,p2p2,p2p3,
     &		 	 p1p3,p2p3,p3p3,
     &		 	 p1p4,p2p4,p3p4)/detyy
	detz(2,2) = det3(p1p1,p1p3,p1p4,
     &		 	 p1p3,p3p3,p3p4,
     &		 	 p1p4,p3p4,p4p4)/detyy
	detz(2,3) = det3(p1p1,p1p2,p1p4,
     &		 	 p1p3,p2p3,p3p4,
     &		 	 p1p4,p2p4,p4p4)/detyy
	detz(2,4) = det3(p1p1,p1p2,p1p3,
     &		 	 p1p3,p2p3,p3p3,
     &		 	 p1p4,p2p4,p3p4)/detyy
	detz(3,3) = det3(p1p1,p1p2,p1p4,
     &		 	 p1p2,p2p2,p2p4,
     &		 	 p1p4,p2p4,p4p4)/detyy
	detz(3,4) = det3(p1p1,p1p2,p1p3,
     &		 	 p1p2,p2p2,p2p3,
     &		 	 p1p4,p2p4,p3p4)/detyy
	detz(4,4) = det3(p1p1,p1p2,p1p3,
     &		 	 p1p2,p2p2,p2p3,
     &		 	 p1p3,p2p3,p3p3)/detyy

        detzz = (P1P4**2*P2P3**2-2D0*P1P3*P1P4*P2P3*P2P4+P1P3**2*
     &  P2P4**2-P1P4**2*P2P2*P3P3+2D0*P1P2*P1P4*P2P4*P3P3-P1P1*
     &  P2P4**2*P3P3+2D0*P1P3*P1P4*P2P2*P3P4-2D0*P1P2*P1P4*P2P3*
     &  P3P4-2D0*P1P2*P1P3*P2P4*P3P4+2D0*P1P1*P2P3*P2P4*P3P4+P1P2**
     &  2*P3P4**2-P1P1*P2P2*P3P4**2-P1P3**2*P2P2*P4P4+2D0*P1P2*P1P3*
     &  P2P3*P4P4-P1P1*P2P3**2*P4P4-P1P2**2*P3P3*P4P4+P1P1*P2P2*
     &  P3P3*P4P4 )/detyy

	do j=1,3
	do k=j+1,4
	  detz(k,j) = detz(j,k)
	enddo
	enddo

	Dm_sum = D1(1)+D1(2)+D1(3) + D0

	E1n(1) =   D1_2(1)*dety(2)+D1_3(1)*dety(3)+D1_4(1)*dety(4)
     &		  -Dm_sum*dety(0)
     &		  -2d0*(-(D2_1(0,0)-D2(0,0))*detz(1,1)
     &		        +(D2_2(0,0)-D2(0,0))*detz(2,1)
     &		        -(D2_3(0,0)-D2(0,0))*detz(3,1)
     &		        +(D2_4(0,0)-D2(0,0))*detz(4,1))

	E1n(2) =   D1_1(1)*dety(1)+D1_3(2)*dety(3)+D1_4(2)*dety(4)
     &		  +D1(1)*dety(0)
     &		  +2d0*(-(D2_1(0,0)-D2(0,0))*detz(1,2)
     &		        +(D2_2(0,0)-D2(0,0))*detz(2,2)
     &		        -(D2_3(0,0)-D2(0,0))*detz(3,2)
     &		        +(D2_4(0,0)-D2(0,0))*detz(4,2))

	E1n(3) =   D1_1(2)*dety(1)+D1_2(2)*dety(2)+D1_4(3)*dety(4)
     &		  +D1(2)*dety(0)
     &		  -2d0*(-(D2_1(0,0)-D2(0,0))*detz(1,3)
     &		        +(D2_2(0,0)-D2(0,0))*detz(2,3)
     &		        -(D2_3(0,0)-D2(0,0))*detz(3,3)
     &		        +(D2_4(0,0)-D2(0,0))*detz(4,3))

	E1n(4) =   D1_1(3)*dety(1)+D1_2(3)*dety(2)+D1_3(3)*dety(3)
     &		  +D1(3)*dety(0)
     &		  +2d0*(-(D2_1(0,0)-D2(0,0))*detz(1,4)
     &		        +(D2_2(0,0)-D2(0,0))*detz(2,4)
     &		        -(D2_3(0,0)-D2(0,0))*detz(3,4)
     &		        +(D2_4(0,0)-D2(0,0))*detz(4,4))

	if (nrank.eq.1) goto 1007

	D00m_sum    = D3(0,0,1)+D3(0,0,2)+D3(0,0,3) + D2(0,0)

	Dmn_sum1(1) = D2(1,1)+D2(1,2)+D2(1,3) + D1(1)
	Dmn_sum1(2) = D2(1,2)+D2(2,2)+D2(2,3) + D1(2)
	Dmn_sum1(3) = D2(1,3)+D2(2,3)+D2(3,3) + D1(3)
	Dmn_sum     = Dmn_sum1(1)+Dmn_sum1(2)+Dmn_sum1(3) + Dm_sum


	E2n(0,0) =  D2(0,0)*dety(0)+D2_1(0,0)*dety(1)+D2_2(0,0)*dety(2)
     &	           +D2_3(0,0)*dety(3)+D2_4(0,0)*dety(4)

	E2n(1,1) =   D2_2(1,1)*dety(2)+D2_3(1,1)*dety(3)
     &	            +D2_4(1,1)*dety(4)+Dmn_sum*dety(0)
     &		    +4d0*(             D00m_sum )*detz(1,1)
     &		    -4d0*( D3_2(0,0,1)+D00m_sum )*detz(2,1)
     &		    +4d0*( D3_3(0,0,1)+D00m_sum )*detz(3,1)
     &		    -4d0*( D3_4(0,0,1)+D00m_sum )*detz(4,1)

	E2n(1,2) =   D2_3(1,2)*dety(3)+D2_4(1,2)*dety(4)
     &	            -Dmn_sum1(1)*dety(0)
     &		    -2d0*(             D00m_sum  )*detz(1,2)
     &		    +2d0*( D3_2(0,0,1)+D00m_sum  )*detz(2,2)
     &		    -2d0*( D3_3(0,0,1)+D00m_sum  )*detz(3,2)
     &		    +2d0*( D3_4(0,0,1)+D00m_sum  )*detz(4,2)
     &		    +2d0*( D3_1(0,0,1)-D3(0,0,1) )*detz(1,1)
     &		    -2d0*(            -D3(0,0,1) )*detz(2,1)
     &		    +2d0*( D3_3(0,0,2)-D3(0,0,1) )*detz(3,1)
     &		    -2d0*( D3_4(0,0,2)-D3(0,0,1) )*detz(4,1)

	E2n(1,3) =   D2_2(1,2)*dety(2)+D2_4(1,3)*dety(4)
     &	            -Dmn_sum1(2)*dety(0)
     &		    +2d0*(             D00m_sum  )*detz(1,3)
     &		    -2d0*( D3_2(0,0,1)+D00m_sum  )*detz(2,3)
     &		    +2d0*( D3_3(0,0,1)+D00m_sum  )*detz(3,3)
     &		    -2d0*( D3_4(0,0,1)+D00m_sum  )*detz(4,3)
     &		    +2d0*( D3_1(0,0,2)-D3(0,0,2) )*detz(1,1)
     &		    -2d0*( D3_2(0,0,2)-D3(0,0,2) )*detz(2,1)
     &		    +2d0*(            -D3(0,0,2) )*detz(3,1)
     &		    -2d0*( D3_4(0,0,3)-D3(0,0,2) )*detz(4,1)

	E2n(1,4) =   D2_2(1,3)*dety(2)+D2_3(1,3)*dety(3)
     &	            -Dmn_sum1(3)*dety(0)
     &		    -2d0*(             D00m_sum  )*detz(1,4)
     &		    +2d0*( D3_2(0,0,1)+D00m_sum  )*detz(2,4)
     &		    -2d0*( D3_3(0,0,1)+D00m_sum  )*detz(3,4)
     &		    +2d0*( D3_4(0,0,1)+D00m_sum  )*detz(4,4)
     &		    +2d0*( D3_1(0,0,3)-D3(0,0,3) )*detz(1,1)
     &		    -2d0*( D3_2(0,0,3)-D3(0,0,3) )*detz(2,1)
     &		    +2d0*( D3_3(0,0,3)-D3(0,0,3) )*detz(3,1)
     &		    -2d0*(            -D3(0,0,3) )*detz(4,1)

	E2n(2,2) =   D2_1(1,1)*dety(1)+D2_3(2,2)*dety(3)
     &	            +D2_4(2,2)*dety(4)+D2(1,1)*dety(0)
     &		    -4d0*( D3_1(0,0,1)-D3(0,0,1) )*detz(1,2)
     &		    +4d0*(            -D3(0,0,1) )*detz(2,2)
     &		    -4d0*( D3_3(0,0,2)-D3(0,0,1) )*detz(3,2)
     &		    +4d0*( D3_4(0,0,2)-D3(0,0,1) )*detz(4,2)

	E2n(2,3) =   D2_1(1,2)*dety(1)
     &	            +D2_4(2,3)*dety(4)+D2(1,2)*dety(0)
     &		    +2d0*( D3_1(0,0,1)-D3(0,0,1) )*detz(1,3)
     &		    -2d0*(            -D3(0,0,1) )*detz(2,3)
     &		    +2d0*( D3_3(0,0,2)-D3(0,0,1) )*detz(3,3)
     &		    -2d0*( D3_4(0,0,2)-D3(0,0,1) )*detz(4,3)
     &		    -2d0*( D3_1(0,0,2)-D3(0,0,2) )*detz(1,2)
     &		    +2d0*( D3_2(0,0,2)-D3(0,0,2) )*detz(2,2)
     &		    -2d0*(            -D3(0,0,2) )*detz(3,2)
     &		    +2d0*( D3_4(0,0,3)-D3(0,0,2) )*detz(4,2)

	E2n(2,4) =   D2_1(1,3)*dety(1)
     &	            +D2_3(2,3)*dety(3)+D2(1,3)*dety(0)
     &		    -2d0*( D3_1(0,0,1)-D3(0,0,1) )*detz(1,4)
     &		    +2d0*(            -D3(0,0,1) )*detz(2,4)
     &		    -2d0*( D3_3(0,0,2)-D3(0,0,1) )*detz(3,4)
     &		    +2d0*( D3_4(0,0,2)-D3(0,0,1) )*detz(4,4)
     &		    -2d0*( D3_1(0,0,3)-D3(0,0,3) )*detz(1,2)
     &		    +2d0*( D3_2(0,0,3)-D3(0,0,3) )*detz(2,2)
     &		    -2d0*( D3_3(0,0,3)-D3(0,0,3) )*detz(3,2)
     &		    +2d0*(            -D3(0,0,3) )*detz(4,2)

	E2n(3,3) =   D2_1(2,2)*dety(1)+D2_2(2,2)*dety(2)
     &	            +D2_4(3,3)*dety(4)+D2(2,2)*dety(0)
     &		    +4d0*( D3_1(0,0,2)-D3(0,0,2) )*detz(1,3)
     &		    -4d0*( D3_2(0,0,2)-D3(0,0,2) )*detz(2,3)
     &		    +4d0*(            -D3(0,0,2) )*detz(3,3)
     &		    -4d0*( D3_4(0,0,3)-D3(0,0,2) )*detz(4,3)

	E2n(3,4) =   D2_1(2,3)*dety(1)
     &	            +D2_2(2,3)*dety(2)+D2(2,3)*dety(0)
     &		    -2d0*( D3_1(0,0,2)-D3(0,0,2) )*detz(1,4)
     &		    +2d0*( D3_2(0,0,2)-D3(0,0,2) )*detz(2,4)
     &		    -2d0*(            -D3(0,0,2) )*detz(3,4)
     &		    +2d0*( D3_4(0,0,3)-D3(0,0,2) )*detz(4,4)
     &		    +2d0*( D3_1(0,0,3)-D3(0,0,3) )*detz(1,3)
     &		    -2d0*( D3_2(0,0,3)-D3(0,0,3) )*detz(2,3)
     &		    +2d0*( D3_3(0,0,3)-D3(0,0,3) )*detz(3,3)
     &		    -2d0*(            -D3(0,0,3) )*detz(4,3)

	E2n(4,4) =   D2_1(3,3)*dety(1)+D2_2(3,3)*dety(2)
     &	            +D2_3(3,3)*dety(3)+D2(3,3)*dety(0)
     &		    -4d0*( D3_1(0,0,3)-D3(0,0,3) )*detz(1,4)
     &		    +4d0*( D3_2(0,0,3)-D3(0,0,3) )*detz(2,4)
     &		    -4d0*( D3_3(0,0,3)-D3(0,0,3) )*detz(3,4)
     &		    +4d0*(            -D3(0,0,3) )*detz(4,4)

	do j=1,3
	do k=j+1,4
	  E2n(k,j) = E2n(j,k)
	enddo
	enddo

	if (nrank.eq.2) goto 1007

	D00mn_sum1(1)  = D4(0,0,1,1)+D4(0,0,1,2)+D4(0,0,1,3) + D3(0,0,1)
	D00mn_sum1(2)  = D4(0,0,1,2)+D4(0,0,2,2)+D4(0,0,2,3) + D3(0,0,2)
	D00mn_sum1(3)  = D4(0,0,1,3)+D4(0,0,2,3)+D4(0,0,3,3) + D3(0,0,3)
	D00mn_sum      = D00mn_sum1(1)+D00mn_sum1(2)+D00mn_sum1(3)
     &			 + D00m_sum

        Dmnl_sum1(1,1) = D3(1,1,1)+D3(1,1,2)+D3(1,1,3) + D2(1,1)
        Dmnl_sum1(1,2) = D3(1,1,2)+D3(1,2,2)+D3(1,2,3) + D2(1,2)
        Dmnl_sum1(1,3) = D3(1,1,3)+D3(1,2,3)+D3(1,3,3) + D2(1,3)
        Dmnl_sum1(2,2) = D3(1,2,2)+D3(2,2,2)+D3(2,2,3) + D2(2,2)
        Dmnl_sum1(2,3) = D3(1,2,3)+D3(2,2,3)+D3(2,3,3) + D2(2,3)
        Dmnl_sum1(3,3) = D3(1,3,3)+D3(2,3,3)+D3(3,3,3) + D2(3,3)

        Dmnl_sum2(1)   = Dmnl_sum1(1,1)+Dmnl_sum1(1,2)+Dmnl_sum1(1,3)
     &			 + Dmn_sum1(1)
        Dmnl_sum2(2)   = Dmnl_sum1(1,2)+Dmnl_sum1(2,2)+Dmnl_sum1(2,3)
     &			 + Dmn_sum1(2)
        Dmnl_sum2(3)   = Dmnl_sum1(1,3)+Dmnl_sum1(2,3)+Dmnl_sum1(3,3)
     &			 + Dmn_sum1(3)
        Dmnl_sum       = Dmnl_sum2(1)+Dmnl_sum2(2)+Dmnl_sum2(3)
     &			 + Dmn_sum

	E3n(0,0,1) =   D3_2(0,0,1)*dety(2)+D3_3(0,0,1)*dety(3)
     &		      +D3_4(0,0,1)*dety(4)-D00m_sum*dety(0)
     &		   +2d0*(D4_1(0,0,0,0)-D4(0,0,0,0))*detz(1,1)
     &		   -2d0*(D4_2(0,0,0,0)-D4(0,0,0,0))*detz(2,1)
     &		   +2d0*(D4_3(0,0,0,0)-D4(0,0,0,0))*detz(3,1)
     &		   -2d0*(D4_4(0,0,0,0)-D4(0,0,0,0))*detz(4,1)
	E3n(0,0,2) =   D3_1(0,0,1)*dety(1)+D3_3(0,0,2)*dety(3)
     &                +D3_4(0,0,2)*dety(4)+D3(0,0,1)*dety(0)
     &		   -2d0*(D4_1(0,0,0,0)-D4(0,0,0,0))*detz(1,2)
     &		   +2d0*(D4_2(0,0,0,0)-D4(0,0,0,0))*detz(2,2)
     &		   -2d0*(D4_3(0,0,0,0)-D4(0,0,0,0))*detz(3,2)
     &		   +2d0*(D4_4(0,0,0,0)-D4(0,0,0,0))*detz(4,2)
	E3n(0,0,3) =   D3_1(0,0,2)*dety(1)+D3_2(0,0,2)*dety(2)
     &                +D3_4(0,0,3)*dety(4)+D3(0,0,2)*dety(0)
     &		   +2d0*(D4_1(0,0,0,0)-D4(0,0,0,0))*detz(1,3)
     &		   -2d0*(D4_2(0,0,0,0)-D4(0,0,0,0))*detz(2,3)
     &		   +2d0*(D4_3(0,0,0,0)-D4(0,0,0,0))*detz(3,3)
     &		   -2d0*(D4_4(0,0,0,0)-D4(0,0,0,0))*detz(4,3)
	E3n(0,0,4) =   D3_1(0,0,3)*dety(1)+D3_2(0,0,3)*dety(2)
     &                +D3_3(0,0,3)*dety(3)+D3(0,0,3)*dety(0)
     &		   -2d0*(D4_1(0,0,0,0)-D4(0,0,0,0))*detz(1,4)
     &		   +2d0*(D4_2(0,0,0,0)-D4(0,0,0,0))*detz(2,4)
     &		   -2d0*(D4_3(0,0,0,0)-D4(0,0,0,0))*detz(3,4)
     &		   +2d0*(D4_4(0,0,0,0)-D4(0,0,0,0))*detz(4,4)

	E3n(1,1,1) =   D3_2(1,1,1)*dety(2)+D3_3(1,1,1)*dety(3)
     &		      +D3_4(1,1,1)*dety(4)-Dmnl_sum*dety(0)
     &		   +6d0*(             -D00mn_sum)*detz(1,1)
     &		   -6d0*(D4_2(0,0,1,1)-D00mn_sum)*detz(2,1)
     &		   +6d0*(D4_3(0,0,1,1)-D00mn_sum)*detz(3,1)
     &		   -6d0*(D4_4(0,0,1,1)-D00mn_sum)*detz(4,1)
	E3n(1,1,2) =   D3_3(1,1,2)*dety(3)+D3_4(1,1,2)*dety(4)
     &		      +Dmnl_sum2(1)*dety(0)
     &		   +4d0*(             +D00mn_sum1(1))*detz(1,1)
     &		   -4d0*(             +D00mn_sum1(1))*detz(2,1)
     &		   +4d0*(D4_3(0,0,1,2)+D00mn_sum1(1))*detz(3,1)
     &		   -4d0*(D4_4(0,0,1,2)+D00mn_sum1(1))*detz(4,1)
     &		   -2d0*(             -D00mn_sum    )*detz(1,2)
     &		   +2d0*(D4_2(0,0,1,1)-D00mn_sum    )*detz(2,2)
     &		   -2d0*(D4_3(0,0,1,1)-D00mn_sum    )*detz(3,2)
     &		   +2d0*(D4_4(0,0,1,1)-D00mn_sum    )*detz(4,2)
	E3n(1,1,3) =   D3_2(1,1,2)*dety(2)+D3_4(1,1,3)*dety(4)
     &		      +Dmnl_sum2(2)*dety(0)
     &		   +4d0*(             +D00mn_sum1(2))*detz(1,1)
     &		   -4d0*(D4_2(0,0,1,2)+D00mn_sum1(2))*detz(2,1)
     &		   +4d0*(             +D00mn_sum1(2))*detz(3,1)
     &		   -4d0*(D4_4(0,0,1,3)+D00mn_sum1(2))*detz(4,1)
     &		   +2d0*(             -D00mn_sum    )*detz(1,3)
     &		   -2d0*(D4_2(0,0,1,1)-D00mn_sum    )*detz(2,3)
     &		   +2d0*(D4_3(0,0,1,1)-D00mn_sum    )*detz(3,3)
     &		   -2d0*(D4_4(0,0,1,1)-D00mn_sum    )*detz(4,3)
	E3n(1,1,4) =   D3_2(1,1,3)*dety(2)+D3_3(1,1,3)*dety(3)
     &		      +Dmnl_sum2(3)*dety(0)
     &		   +4d0*(             +D00mn_sum1(3))*detz(1,1)
     &		   -4d0*(D4_2(0,0,1,3)+D00mn_sum1(3))*detz(2,1)
     &		   +4d0*(D4_3(0,0,1,3)+D00mn_sum1(3))*detz(3,1)
     &		   -4d0*(             +D00mn_sum1(3))*detz(4,1)
     &		   -2d0*(             -D00mn_sum    )*detz(1,4)
     &		   +2d0*(D4_2(0,0,1,1)-D00mn_sum    )*detz(2,4)
     &		   -2d0*(D4_3(0,0,1,1)-D00mn_sum    )*detz(3,4)
     &		   +2d0*(D4_4(0,0,1,1)-D00mn_sum    )*detz(4,4)
	E3n(1,2,2) =   D3_3(1,2,2)*dety(3)+D3_4(1,2,2)*dety(4)
     &		      -Dmnl_sum1(1,1)*dety(0)
     &		   +2d0*(D4_1(0,0,1,1)-D4(0,0,1,1)  )*detz(1,1)
     &		   -2d0*(             -D4(0,0,1,1)  )*detz(2,1)
     &		   +2d0*(D4_3(0,0,2,2)-D4(0,0,1,1)  )*detz(3,1)
     &		   -2d0*(D4_4(0,0,2,2)-D4(0,0,1,1)  )*detz(4,1)
     &		   -4d0*(             +D00mn_sum1(1))*detz(1,2)
     &		   +4d0*(             +D00mn_sum1(1))*detz(2,2)
     &		   -4d0*(D4_3(0,0,1,2)+D00mn_sum1(1))*detz(3,2)
     &		   +4d0*(D4_4(0,0,1,2)+D00mn_sum1(1))*detz(4,2)
	E3n(1,2,3) =   D3_4(1,2,3)*dety(4)
     &		      -Dmnl_sum1(1,2)*dety(0)
     &		   +2d0*(D4_1(0,0,1,2)-D4(0,0,1,2)  )*detz(1,1)
     &		   -2d0*(             -D4(0,0,1,2)  )*detz(2,1)
     &		   +2d0*(             -D4(0,0,1,2)  )*detz(3,1)
     &		   -2d0*(D4_4(0,0,2,3)-D4(0,0,1,2)  )*detz(4,1)
     &		   -2d0*(             +D00mn_sum1(2))*detz(1,2)
     &		   +2d0*(D4_2(0,0,1,2)+D00mn_sum1(2))*detz(2,2)
     &		   -2d0*(             +D00mn_sum1(2))*detz(3,2)
     &		   +2d0*(D4_4(0,0,1,3)+D00mn_sum1(2))*detz(4,2)
     &		   +2d0*(             +D00mn_sum1(1))*detz(1,3)
     &		   -2d0*(             +D00mn_sum1(1))*detz(2,3)
     &		   +2d0*(D4_3(0,0,1,2)+D00mn_sum1(1))*detz(3,3)
     &		   -2d0*(D4_4(0,0,1,2)+D00mn_sum1(1))*detz(4,3)
	E3n(1,2,4) =   D3_3(1,2,3)*dety(3)
     &		      -Dmnl_sum1(1,3)*dety(0)
     &		   +2d0*(D4_1(0,0,1,3)-D4(0,0,1,3)  )*detz(1,1)
     &		   -2d0*(             -D4(0,0,1,3)  )*detz(2,1)
     &		   +2d0*(D4_3(0,0,2,3)-D4(0,0,1,3)  )*detz(3,1)
     &		   -2d0*(             -D4(0,0,1,3)  )*detz(4,1)
     &		   -2d0*(             +D00mn_sum1(3))*detz(1,2)
     &		   +2d0*(D4_2(0,0,1,3)+D00mn_sum1(3))*detz(2,2)
     &		   -2d0*(D4_3(0,0,1,3)+D00mn_sum1(3))*detz(3,2)
     &		   +2d0*(             +D00mn_sum1(3))*detz(4,2)
     &		   -2d0*(             +D00mn_sum1(1))*detz(1,4)
     &		   +2d0*(             +D00mn_sum1(1))*detz(2,4)
     &		   -2d0*(D4_3(0,0,1,2)+D00mn_sum1(1))*detz(3,4)
     &		   +2d0*(D4_4(0,0,1,2)+D00mn_sum1(1))*detz(4,4)
	E3n(1,3,3) =   D3_2(1,2,2)*dety(2)+D3_4(1,3,3)*dety(4)
     &		      -Dmnl_sum1(2,2)*dety(0)
     &		   +2d0*(D4_1(0,0,2,2)-D4(0,0,2,2)  )*detz(1,1)
     &		   -2d0*(D4_2(0,0,2,2)-D4(0,0,2,2)  )*detz(2,1)
     &		   +2d0*(             -D4(0,0,2,2)  )*detz(3,1)
     &		   -2d0*(D4_4(0,0,3,3)-D4(0,0,2,2)  )*detz(4,1)
     &		   +4d0*(             +D00mn_sum1(2))*detz(1,3)
     &		   -4d0*(D4_2(0,0,1,2)+D00mn_sum1(2))*detz(2,3)
     &		   +4d0*(             +D00mn_sum1(2))*detz(3,3)
     &		   -4d0*(D4_4(0,0,1,3)+D00mn_sum1(2))*detz(4,3)
	E3n(1,3,4) =   D3_2(1,2,3)*dety(2)
     &		      -Dmnl_sum1(2,3)*dety(0)
     &		   +2d0*(D4_1(0,0,2,3)-D4(0,0,2,3)  )*detz(1,1)
     &		   -2d0*(D4_2(0,0,2,3)-D4(0,0,2,3)  )*detz(2,1)
     &		   +2d0*(             -D4(0,0,2,3)  )*detz(3,1)
     &		   -2d0*(             -D4(0,0,2,3)  )*detz(4,1)
     &		   +2d0*(             +D00mn_sum1(3))*detz(1,3)
     &		   -2d0*(D4_2(0,0,1,3)+D00mn_sum1(3))*detz(2,3)
     &		   +2d0*(D4_3(0,0,1,3)+D00mn_sum1(3))*detz(3,3)
     &		   -2d0*(             +D00mn_sum1(3))*detz(4,3)
     &		   -2d0*(             +D00mn_sum1(2))*detz(1,4)
     &		   +2d0*(D4_2(0,0,1,2)+D00mn_sum1(2))*detz(2,4)
     &		   -2d0*(             +D00mn_sum1(2))*detz(3,4)
     &		   +2d0*(D4_4(0,0,1,3)+D00mn_sum1(2))*detz(4,4)
	E3n(1,4,4) =   D3_2(1,3,3)*dety(2)+D3_3(1,3,3)*dety(3)
     &		      -Dmnl_sum1(3,3)*dety(0)
     &		   +2d0*(D4_1(0,0,3,3)-D4(0,0,3,3)  )*detz(1,1)
     &		   -2d0*(D4_2(0,0,3,3)-D4(0,0,3,3)  )*detz(2,1)
     &		   +2d0*(D4_3(0,0,3,3)-D4(0,0,3,3)  )*detz(3,1)
     &		   -2d0*(             -D4(0,0,3,3)  )*detz(4,1)
     &		   -4d0*(             +D00mn_sum1(3))*detz(1,4)
     &		   +4d0*(D4_2(0,0,1,3)+D00mn_sum1(3))*detz(2,4)
     &		   -4d0*(D4_3(0,0,1,3)+D00mn_sum1(3))*detz(3,4)
     &		   +4d0*(             +D00mn_sum1(3))*detz(4,4)
	E3n(2,2,2) =   D3_1(1,1,1)*dety(1)+D3_3(2,2,2)*dety(3)
     &		      +D3_4(2,2,2)*dety(4)+D3(1,1,1)*dety(0)
     &		   -6d0*(D4_1(0,0,1,1)-D4(0,0,1,1))*detz(1,2)
     &		   +6d0*(             -D4(0,0,1,1))*detz(2,2)
     &		   -6d0*(D4_3(0,0,2,2)-D4(0,0,1,1))*detz(3,2)
     &		   +6d0*(D4_4(0,0,2,2)-D4(0,0,1,1))*detz(4,2)
	E3n(2,2,3) =   D3_1(1,1,2)*dety(1)+D3_4(2,2,3)*dety(4)
     &		      +D3(1,1,2)*dety(0)
     &		   -4d0*(D4_1(0,0,1,2)-D4(0,0,1,2))*detz(1,2)
     &		   +4d0*(             -D4(0,0,1,2))*detz(2,2)
     &		   -4d0*(             -D4(0,0,1,2))*detz(3,2)
     &		   +4d0*(D4_4(0,0,2,3)-D4(0,0,1,2))*detz(4,2)
     &		   +2d0*(D4_1(0,0,1,1)-D4(0,0,1,1))*detz(1,3)
     &		   -2d0*(             -D4(0,0,1,1))*detz(2,3)
     &		   +2d0*(D4_3(0,0,2,2)-D4(0,0,1,1))*detz(3,3)
     &		   -2d0*(D4_4(0,0,2,2)-D4(0,0,1,1))*detz(4,3)
	E3n(2,2,4) =   D3_1(1,1,3)*dety(1)+D3_3(2,2,3)*dety(3)
     &		      +D3(1,1,3)*dety(0)
     &		   -4d0*(D4_1(0,0,1,3)-D4(0,0,1,3))*detz(1,2)
     &		   +4d0*(             -D4(0,0,1,3))*detz(2,2)
     &		   -4d0*(D4_3(0,0,2,3)-D4(0,0,1,3))*detz(3,2)
     &		   +4d0*(             -D4(0,0,1,3))*detz(4,2)
     &		   -2d0*(D4_1(0,0,1,1)-D4(0,0,1,1))*detz(1,4)
     &		   +2d0*(             -D4(0,0,1,1))*detz(2,4)
     &		   -2d0*(D4_3(0,0,2,2)-D4(0,0,1,1))*detz(3,4)
     &		   +2d0*(D4_4(0,0,2,2)-D4(0,0,1,1))*detz(4,4)
	E3n(2,3,3) =   D3_1(1,2,2)*dety(1)+D3_4(2,3,3)*dety(4)
     &		      +D3(1,2,2)*dety(0)
     &		   -2d0*(D4_1(0,0,2,2)-D4(0,0,2,2))*detz(1,2)
     &		   +2d0*(D4_2(0,0,2,2)-D4(0,0,2,2))*detz(2,2)
     &		   -2d0*(             -D4(0,0,2,2))*detz(3,2)
     &		   +2d0*(D4_4(0,0,3,3)-D4(0,0,2,2))*detz(4,2)
     &		   +4d0*(D4_1(0,0,1,2)-D4(0,0,1,2))*detz(1,3)
     &		   -4d0*(             -D4(0,0,1,2))*detz(2,3)
     &		   +4d0*(             -D4(0,0,1,2))*detz(3,3)
     &		   -4d0*(D4_4(0,0,2,3)-D4(0,0,1,2))*detz(4,3)
	E3n(2,3,4) =   D3_1(1,2,3)*dety(1)+D3(1,2,3)*dety(0)
     &		   -2d0*(D4_1(0,0,2,3)-D4(0,0,2,3))*detz(1,2)
     &		   +2d0*(D4_2(0,0,2,3)-D4(0,0,2,3))*detz(2,2)
     &		   -2d0*(             -D4(0,0,2,3))*detz(3,2)
     &		   +2d0*(             -D4(0,0,2,3))*detz(4,2)
     &		   +2d0*(D4_1(0,0,1,3)-D4(0,0,1,3))*detz(1,3)
     &		   -2d0*(             -D4(0,0,1,3))*detz(2,3)
     &		   +2d0*(D4_3(0,0,2,3)-D4(0,0,1,3))*detz(3,3)
     &		   -2d0*(             -D4(0,0,1,3))*detz(4,3)
     &		   -2d0*(D4_1(0,0,1,2)-D4(0,0,1,2))*detz(1,4)
     &		   +2d0*(             -D4(0,0,1,2))*detz(2,4)
     &		   -2d0*(             -D4(0,0,1,2))*detz(3,4)
     &		   +2d0*(D4_4(0,0,2,3)-D4(0,0,1,2))*detz(4,4)
	E3n(2,4,4) =   D3_1(1,3,3)*dety(1)+D3_3(2,3,3)*dety(3)
     &		      +D3(1,3,3)*dety(0)
     &		   -2d0*(D4_1(0,0,3,3)-D4(0,0,3,3))*detz(1,2)
     &		   +2d0*(D4_2(0,0,3,3)-D4(0,0,3,3))*detz(2,2)
     &		   -2d0*(D4_3(0,0,3,3)-D4(0,0,3,3))*detz(3,2)
     &		   +2d0*(             -D4(0,0,3,3))*detz(4,2)
     &		   -4d0*(D4_1(0,0,1,3)-D4(0,0,1,3))*detz(1,4)
     &		   +4d0*(             -D4(0,0,1,3))*detz(2,4)
     &		   -4d0*(D4_3(0,0,2,3)-D4(0,0,1,3))*detz(3,4)
     &		   +4d0*(             -D4(0,0,1,3))*detz(4,4)
	E3n(3,3,3) =   D3_1(2,2,2)*dety(1)+D3_2(2,2,2)*dety(2)
     &		      +D3_4(3,3,3)*dety(4)+D3(2,2,2)*dety(0)
     &		   +6d0*(D4_1(0,0,2,2)-D4(0,0,2,2))*detz(1,3)
     &		   -6d0*(D4_2(0,0,2,2)-D4(0,0,2,2))*detz(2,3)
     &		   +6d0*(             -D4(0,0,2,2))*detz(3,3)
     &		   -6d0*(D4_4(0,0,3,3)-D4(0,0,2,2))*detz(4,3)
	E3n(3,3,4) =   D3_1(2,2,3)*dety(1)+D3_2(2,2,3)*dety(2)
     &		      +D3(2,2,3)*dety(0)
     &		   +4d0*(D4_1(0,0,2,3)-D4(0,0,2,3))*detz(1,3)
     &		   -4d0*(D4_2(0,0,2,3)-D4(0,0,2,3))*detz(2,3)
     &		   +4d0*(             -D4(0,0,2,3))*detz(3,3)
     &		   -4d0*(             -D4(0,0,2,3))*detz(4,3)
     &		   -2d0*(D4_1(0,0,2,2)-D4(0,0,2,2))*detz(1,4)
     &		   +2d0*(D4_2(0,0,2,2)-D4(0,0,2,2))*detz(2,4)
     &		   -2d0*(             -D4(0,0,2,2))*detz(3,4)
     &		   +2d0*(D4_4(0,0,3,3)-D4(0,0,2,2))*detz(4,4)
	E3n(3,4,4) =   D3_1(2,3,3)*dety(1)+D3_2(2,3,3)*dety(2)
     &		      +D3(2,3,3)*dety(0)
     &		   +2d0*(D4_1(0,0,3,3)-D4(0,0,3,3))*detz(1,3)
     &		   -2d0*(D4_2(0,0,3,3)-D4(0,0,3,3))*detz(2,3)
     &		   +2d0*(D4_3(0,0,3,3)-D4(0,0,3,3))*detz(3,3)
     &		   -2d0*(             -D4(0,0,3,3))*detz(4,3)
     &		   -4d0*(D4_1(0,0,2,3)-D4(0,0,2,3))*detz(1,4)
     &		   +4d0*(D4_2(0,0,2,3)-D4(0,0,2,3))*detz(2,4)
     &		   -4d0*(             -D4(0,0,2,3))*detz(3,4)
     &		   +4d0*(             -D4(0,0,2,3))*detz(4,4)
	E3n(4,4,4) =   D3_1(3,3,3)*dety(1)+D3_2(3,3,3)*dety(2)
     &		      +D3_3(3,3,3)*dety(3)+D3(3,3,3)*dety(0)
     &		   -6d0*(D4_1(0,0,3,3)-D4(0,0,3,3))*detz(1,4)
     &		   +6d0*(D4_2(0,0,3,3)-D4(0,0,3,3))*detz(2,4)
     &		   -6d0*(D4_3(0,0,3,3)-D4(0,0,3,3))*detz(3,4)
     &		   +6d0*(             -D4(0,0,3,3))*detz(4,4)

	do j=1,3
	do k=j+1,4
	  E3n(k,j,k) = E3n(j,k,k)
	  E3n(k,k,j) = E3n(j,k,k)
          E3n(k,j,j) = E3n(j,j,k)
          E3n(j,k,j) = E3n(j,j,k)
	enddo
	enddo

	do j=1,2
	do k=j+1,3
	do l=k+1,4
	  E3n(k,l,j) = E3n(j,k,l)
	  E3n(l,j,k) = E3n(j,k,l)
	  E3n(l,k,j) = E3n(j,k,l)
	  E3n(k,j,l) = E3n(j,k,l)
	  E3n(j,l,k) = E3n(j,k,l)
	enddo
	enddo
	enddo

	if (nrank.eq.3) goto 1007

	D0000m_sum        = D5(0,0,0,0,1)+D5(0,0,0,0,2)+D5(0,0,0,0,3)
     &			    + D4(0,0,0,0)

        D00mnl_sum1(1,1)  = D5(0,0,1,1,1)+D5(0,0,1,1,2)+D5(0,0,1,1,3)
     &			    + D4(0,0,1,1)
        D00mnl_sum1(1,2)  = D5(0,0,1,1,2)+D5(0,0,1,2,2)+D5(0,0,1,2,3)
     &			    + D4(0,0,1,2)
        D00mnl_sum1(1,3)  = D5(0,0,1,1,3)+D5(0,0,1,2,3)+D5(0,0,1,3,3)
     &			    + D4(0,0,1,3)
        D00mnl_sum1(2,2)  = D5(0,0,1,2,2)+D5(0,0,2,2,2)+D5(0,0,2,2,3)
     &			    + D4(0,0,2,2)
        D00mnl_sum1(2,3)  = D5(0,0,1,2,3)+D5(0,0,2,2,3)+D5(0,0,2,3,3)
     &			    + D4(0,0,2,3)
        D00mnl_sum1(3,3)  = D5(0,0,1,3,3)+D5(0,0,2,3,3)+D5(0,0,3,3,3)
     &			    + D4(0,0,3,3)

        D00mnl_sum2(1)    = D00mnl_sum1(1,1)+D00mnl_sum1(1,2)
     &			   +D00mnl_sum1(1,3) + D00mn_sum1(1)
        D00mnl_sum2(2)    = D00mnl_sum1(1,2)+D00mnl_sum1(2,2)
     &			   +D00mnl_sum1(2,3) + D00mn_sum1(2)
        D00mnl_sum2(3)    = D00mnl_sum1(1,3)+D00mnl_sum1(2,3)
     &			   +D00mnl_sum1(3,3) + D00mn_sum1(3)

        D00mnl_sum        = D00mnl_sum2(1)+D00mnl_sum2(2)+D00mnl_sum2(3)
     &			    + D00mn_sum

        Dmnlk_sum1(1,1,1)= D4(1,1,1,1)+D4(1,1,1,2)+D4(1,1,1,3)+D3(1,1,1)
        Dmnlk_sum1(1,1,2)= D4(1,1,1,2)+D4(1,1,2,2)+D4(1,1,2,3)+D3(1,1,2)
        Dmnlk_sum1(1,1,3)= D4(1,1,1,3)+D4(1,1,2,3)+D4(1,1,3,3)+D3(1,1,3)
        Dmnlk_sum1(1,2,2)= D4(1,1,2,2)+D4(1,2,2,2)+D4(1,2,2,3)+D3(1,2,2)
        Dmnlk_sum1(1,2,3)= D4(1,1,2,3)+D4(1,2,2,3)+D4(1,2,3,3)+D3(1,2,3)
        Dmnlk_sum1(1,3,3)= D4(1,1,3,3)+D4(1,2,3,3)+D4(1,3,3,3)+D3(1,3,3)
        Dmnlk_sum1(2,2,2)= D4(1,2,2,2)+D4(2,2,2,2)+D4(2,2,2,3)+D3(2,2,2)
        Dmnlk_sum1(2,2,3)= D4(1,2,2,3)+D4(2,2,2,3)+D4(2,2,3,3)+D3(2,2,3)
        Dmnlk_sum1(2,3,3)= D4(1,2,3,3)+D4(2,2,3,3)+D4(2,3,3,3)+D3(2,3,3)
        Dmnlk_sum1(3,3,3)= D4(1,3,3,3)+D4(2,3,3,3)+D4(3,3,3,3)+D3(3,3,3)

        Dmnlk_sum2(1,1)   = Dmnlk_sum1(1,1,1)+Dmnlk_sum1(1,1,2)
     &			   +Dmnlk_sum1(1,1,3) + Dmnl_sum1(1,1)
        Dmnlk_sum2(1,2)   = Dmnlk_sum1(1,1,2)+Dmnlk_sum1(1,2,2)
     &			   +Dmnlk_sum1(1,2,3) + Dmnl_sum1(1,2)
        Dmnlk_sum2(1,3)   = Dmnlk_sum1(1,1,3)+Dmnlk_sum1(1,2,3)
     &			   +Dmnlk_sum1(1,3,3) + Dmnl_sum1(1,3)
        Dmnlk_sum2(2,2)   = Dmnlk_sum1(1,2,2)+Dmnlk_sum1(2,2,2)
     &			   +Dmnlk_sum1(2,2,3) + Dmnl_sum1(2,2)
        Dmnlk_sum2(2,3)   = Dmnlk_sum1(1,2,3)+Dmnlk_sum1(2,2,3)
     &			   +Dmnlk_sum1(2,3,3) + Dmnl_sum1(2,3)
        Dmnlk_sum2(3,3)   = Dmnlk_sum1(1,3,3)+Dmnlk_sum1(2,3,3)
     &			   +Dmnlk_sum1(3,3,3) + Dmnl_sum1(3,3)

        Dmnlk_sum3(1) = Dmnlk_sum2(1,1)+Dmnlk_sum2(1,2)+Dmnlk_sum2(1,3)
     &			+ Dmnl_sum2(1)
        Dmnlk_sum3(2) = Dmnlk_sum2(1,2)+Dmnlk_sum2(2,2)+Dmnlk_sum2(2,3)
     &			+ Dmnl_sum2(2)
        Dmnlk_sum3(3) = Dmnlk_sum2(1,3)+Dmnlk_sum2(2,3)+Dmnlk_sum2(3,3)
     &			+ Dmnl_sum2(3)

        Dmnlk_sum     = Dmnlk_sum3(1)+Dmnlk_sum3(2)+Dmnlk_sum3(3)
     &			+ Dmnl_sum

	E4n(0,0,0,0) = -1d0/48d0*detzz
     &		       +   D4(0,0,0,0)*dety(0)
     &		       + D4_1(0,0,0,0)*dety(1)
     &		       + D4_2(0,0,0,0)*dety(2)
     &		       + D4_3(0,0,0,0)*dety(3)
     &		       + D4_4(0,0,0,0)*dety(4)

	E4n(0,0,1,1) =   D4_2(0,0,1,1)*dety(2)+D4_3(0,0,1,1)*dety(3)
     &	                +D4_4(0,0,1,1)*dety(4)+D00mn_sum*dety(0)
     &	+4d0*(                 D0000m_sum)*detz(1,1)
     &	-4d0*( D5_2(0,0,0,0,1)+D0000m_sum)*detz(2,1)
     &	+4d0*( D5_3(0,0,0,0,1)+D0000m_sum)*detz(3,1)
     &	-4d0*( D5_4(0,0,0,0,1)+D0000m_sum)*detz(4,1)

	E4n(0,0,1,2) =   D4_3(0,0,1,2)*dety(3)+D4_4(0,0,1,2)*dety(4)
     &	                -D00mn_sum1(1)*dety(0)
     &	-2d0*(                 D0000m_sum    )*detz(1,2)
     &	+2d0*( D5_2(0,0,0,0,1)+D0000m_sum    )*detz(2,2)
     &	-2d0*( D5_3(0,0,0,0,1)+D0000m_sum    )*detz(3,2)
     &	+2d0*( D5_4(0,0,0,0,1)+D0000m_sum    )*detz(4,2)
     &	+2d0*( D5_1(0,0,0,0,1)-D5(0,0,0,0,1) )*detz(1,1)
     &	-2d0*(                -D5(0,0,0,0,1) )*detz(2,1)
     &	+2d0*( D5_3(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(3,1)
     &	-2d0*( D5_4(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(4,1)

	E4n(0,0,1,3) =   D4_2(0,0,1,2)*dety(2)+D4_4(0,0,1,3)*dety(4)
     &	                -D00mn_sum1(2)*dety(0)
     &	+2d0*(                 D0000m_sum    )*detz(1,3)
     &	-2d0*( D5_2(0,0,0,0,1)+D0000m_sum    )*detz(2,3)
     &	+2d0*( D5_3(0,0,0,0,1)+D0000m_sum    )*detz(3,3)
     &	-2d0*( D5_4(0,0,0,0,1)+D0000m_sum    )*detz(4,3)
     &	+2d0*( D5_1(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(1,1)
     &	-2d0*( D5_2(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(2,1)
     &	+2d0*(                -D5(0,0,0,0,2) )*detz(3,1)
     &	-2d0*( D5_4(0,0,0,0,3)-D5(0,0,0,0,2) )*detz(4,1)

	E4n(0,0,1,4) =   D4_2(0,0,1,3)*dety(2)+D4_3(0,0,1,3)*dety(3)
     &	                -D00mn_sum1(3)*dety(0)
     &	-2d0*(                 D0000m_sum    )*detz(1,4)
     &	+2d0*( D5_2(0,0,0,0,1)+D0000m_sum    )*detz(2,4)
     &	-2d0*( D5_3(0,0,0,0,1)+D0000m_sum    )*detz(3,4)
     &	+2d0*( D5_4(0,0,0,0,1)+D0000m_sum    )*detz(4,4)
     &	+2d0*( D5_1(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(1,1)
     &	-2d0*( D5_2(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(2,1)
     &	+2d0*( D5_3(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(3,1)
     &	-2d0*(                -D5(0,0,0,0,3) )*detz(4,1)

	E4n(0,0,2,2) =   D4_1(0,0,1,1)*dety(1)+D4_3(0,0,2,2)*dety(3)
     &                  +D4_4(0,0,2,2)*dety(4)+D4(0,0,1,1)*dety(0)
     &	-4d0*( D5_1(0,0,0,0,1)-D5(0,0,0,0,1) )*detz(1,2)
     &	+4d0*(                -D5(0,0,0,0,1) )*detz(2,2)
     &	-4d0*( D5_3(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(3,2)
     &	+4d0*( D5_4(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(4,2)

	E4n(0,0,2,3) =   D4_1(0,0,1,2)*dety(1)
     &	                +D4_4(0,0,2,3)*dety(4)+D4(0,0,1,2)*dety(0)
     &	+2d0*( D5_1(0,0,0,0,1)-D5(0,0,0,0,1) )*detz(1,3)
     &	-2d0*(                -D5(0,0,0,0,1) )*detz(2,3)
     &	+2d0*( D5_3(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(3,3)
     &	-2d0*( D5_4(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(4,3)
     &	-2d0*( D5_1(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(1,2)
     &	+2d0*( D5_2(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(2,2)
     &	-2d0*(                -D5(0,0,0,0,2) )*detz(3,2)
     &	+2d0*( D5_4(0,0,0,0,3)-D5(0,0,0,0,2) )*detz(4,2)

	E4n(0,0,2,4) =   D4_1(0,0,1,3)*dety(1)
     &            +D4_3(0,0,2,3)*dety(3)+D4(0,0,1,3)*dety(0)
     &	-2d0*( D5_1(0,0,0,0,1)-D5(0,0,0,0,1) )*detz(1,4)
     &	+2d0*(                -D5(0,0,0,0,1) )*detz(2,4)
     &	-2d0*( D5_3(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(3,4)
     &	+2d0*( D5_4(0,0,0,0,2)-D5(0,0,0,0,1) )*detz(4,4)
     &	-2d0*( D5_1(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(1,2)
     &	+2d0*( D5_2(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(2,2)
     &	-2d0*( D5_3(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(3,2)
     &	+2d0*(                -D5(0,0,0,0,3) )*detz(4,2)

	E4n(0,0,3,3) =   D4_1(0,0,2,2)*dety(1)+D4_2(0,0,2,2)*dety(2)
     &            +D4_4(0,0,3,3)*dety(4)+D4(0,0,2,2)*dety(0)
     &	+4d0*( D5_1(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(1,3)
     &	-4d0*( D5_2(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(2,3)
     &	+4d0*(                -D5(0,0,0,0,2) )*detz(3,3)
     &	-4d0*( D5_4(0,0,0,0,3)-D5(0,0,0,0,2) )*detz(4,3)

	E4n(0,0,3,4) =   D4_1(0,0,2,3)*dety(1)
     &            +D4_2(0,0,2,3)*dety(2)+D4(0,0,2,3)*dety(0)
     &	-2d0*( D5_1(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(1,4)
     &	+2d0*( D5_2(0,0,0,0,2)-D5(0,0,0,0,2) )*detz(2,4)
     &	-2d0*(                -D5(0,0,0,0,2) )*detz(3,4)
     &	+2d0*( D5_4(0,0,0,0,3)-D5(0,0,0,0,2) )*detz(4,4)
     &	+2d0*( D5_1(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(1,3)
     &	-2d0*( D5_2(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(2,3)
     &	+2d0*( D5_3(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(3,3)
     &	-2d0*(                -D5(0,0,0,0,3) )*detz(4,3)

	E4n(0,0,4,4) =   D4_1(0,0,3,3)*dety(1)+D4_2(0,0,3,3)*dety(2)
     &            +D4_3(0,0,3,3)*dety(3)+D4(0,0,3,3)*dety(0)
     &	-4d0*( D5_1(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(1,4)
     &	+4d0*( D5_2(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(2,4)
     &	-4d0*( D5_3(0,0,0,0,3)-D5(0,0,0,0,3) )*detz(3,4)
     &	+4d0*(                -D5(0,0,0,0,3) )*detz(4,4)

        E4n(1,1,1,1) =   D4_2(1,1,1,1)*dety(2)+D4_3(1,1,1,1)*dety(3)
     &	                +D4_4(1,1,1,1)*dety(4)+Dmnlk_sum*dety(0)
     &	+8d0*(                 D00mnl_sum)*detz(1,1)
     &	-8d0*( D5_2(0,0,1,1,1)+D00mnl_sum)*detz(2,1)
     &	+8d0*( D5_3(0,0,1,1,1)+D00mnl_sum)*detz(3,1)
     &	-8d0*( D5_4(0,0,1,1,1)+D00mnl_sum)*detz(4,1)

        E4n(1,1,1,2) =   D4_3(1,1,1,2)*dety(3)+D4_4(1,1,1,2)*dety(4)
     &		        -Dmnlk_sum3(1)*dety(0)
     &	+6d0*(               -D00mnl_sum2(1))*detz(1,1)
     &	-6d0*(               -D00mnl_sum2(1))*detz(2,1)
     &	+6d0*(D5_3(0,0,1,1,2)-D00mnl_sum2(1))*detz(3,1)
     &	-6d0*(D5_4(0,0,1,1,2)-D00mnl_sum2(1))*detz(4,1)
     &	-2d0*(               +D00mnl_sum    )*detz(1,2)
     &	+2d0*(D5_2(0,0,1,1,1)+D00mnl_sum    )*detz(2,2)
     &	-2d0*(D5_3(0,0,1,1,1)+D00mnl_sum    )*detz(3,2)
     &	+2d0*(D5_4(0,0,1,1,1)+D00mnl_sum    )*detz(4,2)

        E4n(1,1,1,3) =   D4_2(1,1,1,2)*dety(2)+D4_4(1,1,1,3)*dety(4)
     &		        -Dmnlk_sum3(2)*dety(0)
     &	+6d0*(               -D00mnl_sum2(2))*detz(1,1)
     &	-6d0*(D5_2(0,0,1,1,2)-D00mnl_sum2(2))*detz(2,1)
     &	+6d0*(               -D00mnl_sum2(2))*detz(3,1)
     &	-6d0*(D5_4(0,0,1,1,3)-D00mnl_sum2(2))*detz(4,1)
     &	+2d0*(               +D00mnl_sum    )*detz(1,3)
     &	-2d0*(D5_2(0,0,1,1,1)+D00mnl_sum    )*detz(2,3)
     &	+2d0*(D5_3(0,0,1,1,1)+D00mnl_sum    )*detz(3,3)
     &	-2d0*(D5_4(0,0,1,1,1)+D00mnl_sum    )*detz(4,3)

        E4n(1,1,1,4) =   D4_2(1,1,1,3)*dety(2)+D4_3(1,1,1,3)*dety(3)
     &		        -Dmnlk_sum3(3)*dety(0)
     &	+6d0*(               -D00mnl_sum2(3))*detz(1,1)
     &	-6d0*(D5_2(0,0,1,1,3)-D00mnl_sum2(3))*detz(2,1)
     &	+6d0*(D5_3(0,0,1,1,3)-D00mnl_sum2(3))*detz(3,1)
     &	-6d0*(               -D00mnl_sum2(3))*detz(4,1)
     &	-2d0*(               +D00mnl_sum    )*detz(1,4)
     &	+2d0*(D5_2(0,0,1,1,1)+D00mnl_sum    )*detz(2,4)
     &	-2d0*(D5_3(0,0,1,1,1)+D00mnl_sum    )*detz(3,4)
     &	+2d0*(D5_4(0,0,1,1,1)+D00mnl_sum    )*detz(4,4)

        E4n(1,1,2,2) =   D4_3(1,1,2,2)*dety(3)+D4_4(1,1,2,2)*dety(4)
     &		        +Dmnlk_sum2(1,1)*dety(0)
     &	+4d0*(               +D00mnl_sum1(1,1))*detz(1,1)
     &	-4d0*(               +D00mnl_sum1(1,1))*detz(2,1)
     &	+4d0*(D5_3(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(3,1)
     &	-4d0*(D5_4(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(4,1)
     &	-4d0*(               -D00mnl_sum2(1)  )*detz(1,2)
     &	+4d0*(               -D00mnl_sum2(1)  )*detz(2,2)
     &	-4d0*(D5_3(0,0,1,1,2)-D00mnl_sum2(1)  )*detz(3,2)
     &	+4d0*(D5_4(0,0,1,1,2)-D00mnl_sum2(1)  )*detz(4,2)

        E4n(1,1,2,3) =   D4_4(1,1,2,3)*dety(4)
     &		        +Dmnlk_sum2(1,2)*dety(0)
     &	+4d0*(               +D00mnl_sum1(1,2))*detz(1,1)
     &	-4d0*(               +D00mnl_sum1(1,2))*detz(2,1)
     &	+4d0*(               +D00mnl_sum1(1,2))*detz(3,1)
     &	-4d0*(D5_4(0,0,1,2,3)+D00mnl_sum1(1,2))*detz(4,1)
     &	-2d0*(               -D00mnl_sum2(2)  )*detz(1,2)
     &	+2d0*(D5_2(0,0,1,1,2)-D00mnl_sum2(2)  )*detz(2,2)
     &	-2d0*(               -D00mnl_sum2(2)  )*detz(3,2)
     &	+2d0*(D5_4(0,0,1,1,3)-D00mnl_sum2(2)  )*detz(4,2)
     &	+2d0*(               -D00mnl_sum2(1)  )*detz(1,3)
     &	-2d0*(               -D00mnl_sum2(1)  )*detz(2,3)
     &	+2d0*(D5_3(0,0,1,1,2)-D00mnl_sum2(1)  )*detz(3,3)
     &	-2d0*(D5_4(0,0,1,1,2)-D00mnl_sum2(1)  )*detz(4,3)


        E4n(1,1,2,4) =   D4_3(1,1,2,3)*dety(3)
     &		        +Dmnlk_sum2(1,3)*dety(0)
     &	+4d0*(               +D00mnl_sum1(1,3))*detz(1,1)
     &	-4d0*(               +D00mnl_sum1(1,3))*detz(2,1)
     &	+4d0*(D5_3(0,0,1,2,3)+D00mnl_sum1(1,3))*detz(3,1)
     &	-4d0*(               +D00mnl_sum1(1,3))*detz(4,1)
     &	-2d0*(               -D00mnl_sum2(3)  )*detz(1,2)
     &	+2d0*(D5_2(0,0,1,1,3)-D00mnl_sum2(3)  )*detz(2,2)
     &	-2d0*(D5_3(0,0,1,1,3)-D00mnl_sum2(3)  )*detz(3,2)
     &	+2d0*(               -D00mnl_sum2(3)  )*detz(4,2)
     &	-2d0*(               -D00mnl_sum2(1)  )*detz(1,4)
     &	+2d0*(               -D00mnl_sum2(1)  )*detz(2,4)
     &	-2d0*(D5_3(0,0,1,1,2)-D00mnl_sum2(1)  )*detz(3,4)
     &	+2d0*(D5_4(0,0,1,1,2)-D00mnl_sum2(1)  )*detz(4,4)

        E4n(1,1,3,3) =   D4_2(1,1,2,2)*dety(2)+D4_4(1,1,3,3)*dety(4)
     &		        +Dmnlk_sum2(2,2)*dety(0)
     &	+4d0*(               +D00mnl_sum1(2,2))*detz(1,1)
     &	-4d0*(D5_2(0,0,1,2,2)+D00mnl_sum1(2,2))*detz(2,1)
     &	+4d0*(               +D00mnl_sum1(2,2))*detz(3,1)
     &	-4d0*(D5_4(0,0,1,3,3)+D00mnl_sum1(2,2))*detz(4,1)
     &	+4d0*(               -D00mnl_sum2(2)  )*detz(1,3)
     &	-4d0*(D5_2(0,0,1,1,2)-D00mnl_sum2(2)  )*detz(2,3)
     &	+4d0*(               -D00mnl_sum2(2)  )*detz(3,3)
     &	-4d0*(D5_4(0,0,1,1,3)-D00mnl_sum2(2)  )*detz(4,3)

        E4n(1,1,3,4) =   D4_2(1,1,2,3)*dety(2)
     &		        +Dmnlk_sum2(2,3)*dety(0)
     &	+4d0*(               +D00mnl_sum1(2,3))*detz(1,1)
     &	-4d0*(D5_2(0,0,1,2,3)+D00mnl_sum1(2,3))*detz(2,1)
     &	+4d0*(               +D00mnl_sum1(2,3))*detz(3,1)
     &	-4d0*(               +D00mnl_sum1(2,3))*detz(4,1)
     &	+2d0*(               -D00mnl_sum2(3)  )*detz(1,3)
     &	-2d0*(D5_2(0,0,1,1,3)-D00mnl_sum2(3)  )*detz(2,3)
     &	+2d0*(D5_3(0,0,1,1,3)-D00mnl_sum2(3)  )*detz(3,3)
     &	-2d0*(               -D00mnl_sum2(3)  )*detz(4,3)
     &	-2d0*(               -D00mnl_sum2(2)  )*detz(1,4)
     &	+2d0*(D5_2(0,0,1,1,2)-D00mnl_sum2(2)  )*detz(2,4)
     &	-2d0*(               -D00mnl_sum2(2)  )*detz(3,4)
     &	+2d0*(D5_4(0,0,1,1,3)-D00mnl_sum2(2)  )*detz(4,4)

        E4n(1,1,4,4) =   D4_2(1,1,3,3)*dety(2)+D4_3(1,1,3,3)*dety(3)
     &		        +Dmnlk_sum2(3,3)*dety(0)
     &	+4d0*(               +D00mnl_sum1(3,3))*detz(1,1)
     &	-4d0*(D5_2(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(2,1)
     &	+4d0*(D5_3(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(3,1)
     &	-4d0*(               +D00mnl_sum1(3,3))*detz(4,1)
     &	-4d0*(               -D00mnl_sum2(3)  )*detz(1,4)
     &	+4d0*(D5_2(0,0,1,1,3)-D00mnl_sum2(3)  )*detz(2,4)
     &	-4d0*(D5_3(0,0,1,1,3)-D00mnl_sum2(3)  )*detz(3,4)
     &	+4d0*(               -D00mnl_sum2(3)  )*detz(4,4)

        E4n(1,2,2,2) =   D4_3(1,2,2,2)*dety(3)+D4_4(1,2,2,2)*dety(4)
     &		        -Dmnlk_sum1(1,1,1)*dety(0)
     &	+2d0*(D5_1(0,0,1,1,1)-D5(0,0,1,1,1)   )*detz(1,1)
     &	-2d0*(               -D5(0,0,1,1,1)   )*detz(2,1)
     &	+2d0*(D5_3(0,0,2,2,2)-D5(0,0,1,1,1)   )*detz(3,1)
     &	-2d0*(D5_4(0,0,2,2,2)-D5(0,0,1,1,1)   )*detz(4,1)
     &	-6d0*(               +D00mnl_sum1(1,1))*detz(1,2)
     &	+6d0*(               +D00mnl_sum1(1,1))*detz(2,2)
     &	-6d0*(D5_3(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(3,2)
     &	+6d0*(D5_4(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(4,2)

        E4n(1,2,2,3) =   D4_4(1,2,2,3)*dety(4)
     &		        -Dmnlk_sum1(1,1,2)*dety(0)
     &	+2d0*(D5_1(0,0,1,1,2)-D5(0,0,1,1,2)   )*detz(1,1)
     &	-2d0*(               -D5(0,0,1,1,2)   )*detz(2,1)
     &	+2d0*(               -D5(0,0,1,1,2)   )*detz(3,1)
     &	-2d0*(D5_4(0,0,2,2,3)-D5(0,0,1,1,2)   )*detz(4,1)
     &	-4d0*(               +D00mnl_sum1(1,2))*detz(1,2)
     &	+4d0*(               +D00mnl_sum1(1,2))*detz(2,2)
     &	-4d0*(               +D00mnl_sum1(1,2))*detz(3,2)
     &	+4d0*(D5_4(0,0,1,2,3)+D00mnl_sum1(1,2))*detz(4,2)
     &	+2d0*(               +D00mnl_sum1(1,1))*detz(1,3)
     &	-2d0*(               +D00mnl_sum1(1,1))*detz(2,3)
     &	+2d0*(D5_3(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(3,3)
     &	-2d0*(D5_4(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(4,3)

        E4n(1,2,2,4) =   D4_3(1,2,2,3)*dety(3)
     &		        -Dmnlk_sum1(1,1,3)*dety(0)
     &	+2d0*(D5_1(0,0,1,1,3)-D5(0,0,1,1,3)   )*detz(1,1)
     &	-2d0*(               -D5(0,0,1,1,3)   )*detz(2,1)
     &	+2d0*(D5_3(0,0,2,2,3)-D5(0,0,1,1,3)   )*detz(3,1)
     &	-2d0*(               -D5(0,0,1,1,3)   )*detz(4,1)
     &	-4d0*(               +D00mnl_sum1(1,3))*detz(1,2)
     &	+4d0*(               +D00mnl_sum1(1,3))*detz(2,2)
     &	-4d0*(D5_3(0,0,1,2,3)+D00mnl_sum1(1,3))*detz(3,2)
     &	+4d0*(               +D00mnl_sum1(1,3))*detz(4,2)
     &	-2d0*(               +D00mnl_sum1(1,1))*detz(1,4)
     &	+2d0*(               +D00mnl_sum1(1,1))*detz(2,4)
     &	-2d0*(D5_3(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(3,4)
     &	+2d0*(D5_4(0,0,1,2,2)+D00mnl_sum1(1,1))*detz(4,4)

        E4n(1,2,3,3) =   D4_4(1,2,3,3)*dety(4)
     &		        -Dmnlk_sum1(1,2,2)*dety(0)
     &	+2d0*(D5_1(0,0,1,2,2)-D5(0,0,1,2,2)   )*detz(1,1)
     &	-2d0*(               -D5(0,0,1,2,2)   )*detz(2,1)
     &	+2d0*(               -D5(0,0,1,2,2)   )*detz(3,1)
     &	-2d0*(D5_4(0,0,2,3,3)-D5(0,0,1,2,2)   )*detz(4,1)
     &	-2d0*(               +D00mnl_sum1(2,2))*detz(1,2)
     &	+2d0*(D5_2(0,0,1,2,2)+D00mnl_sum1(2,2))*detz(2,2)
     &	-2d0*(               +D00mnl_sum1(2,2))*detz(3,2)
     &	+2d0*(D5_4(0,0,1,3,3)+D00mnl_sum1(2,2))*detz(4,2)
     &	+4d0*(               +D00mnl_sum1(1,2))*detz(1,3)
     &	-4d0*(               +D00mnl_sum1(1,2))*detz(2,3)
     &	+4d0*(               +D00mnl_sum1(1,2))*detz(3,3)
     &	-4d0*(D5_4(0,0,1,2,3)+D00mnl_sum1(1,2))*detz(4,3)

        E4n(1,2,3,4) =   -Dmnlk_sum1(1,2,3)*dety(0)
     &	+2d0*(D5_1(0,0,1,2,3)-D5(0,0,1,2,3)   )*detz(1,1)
     &	-2d0*(               -D5(0,0,1,2,3)   )*detz(2,1)
     &	+2d0*(               -D5(0,0,1,2,3)   )*detz(3,1)
     &	-2d0*(               -D5(0,0,1,2,3)   )*detz(4,1)
     &	-2d0*(               +D00mnl_sum1(2,3))*detz(1,2)
     &	+2d0*(D5_2(0,0,1,2,3)+D00mnl_sum1(2,3))*detz(2,2)
     &	-2d0*(               +D00mnl_sum1(2,3))*detz(3,2)
     &	+2d0*(               +D00mnl_sum1(2,3))*detz(4,2)
     &	+2d0*(               +D00mnl_sum1(1,3))*detz(1,3)
     &	-2d0*(               +D00mnl_sum1(1,3))*detz(2,3)
     &	+2d0*(D5_3(0,0,1,2,3)+D00mnl_sum1(1,3))*detz(3,3)
     &	-2d0*(               +D00mnl_sum1(1,3))*detz(4,3)
     &	-2d0*(               +D00mnl_sum1(1,2))*detz(1,4)
     &	+2d0*(               +D00mnl_sum1(1,2))*detz(2,4)
     &	-2d0*(               +D00mnl_sum1(1,2))*detz(3,4)
     &	+2d0*(D5_4(0,0,1,2,3)+D00mnl_sum1(1,2))*detz(4,4)

        E4n(1,2,4,4) =   D4_3(1,2,3,3)*dety(3)
     &		        -Dmnlk_sum1(1,3,3)*dety(0)
     &	+2d0*(D5_1(0,0,1,3,3)-D5(0,0,1,3,3)   )*detz(1,1)
     &	-2d0*(               -D5(0,0,1,3,3)   )*detz(2,1)
     &	+2d0*(D5_3(0,0,2,3,3)-D5(0,0,1,3,3)   )*detz(3,1)
     &	-2d0*(               -D5(0,0,1,3,3)   )*detz(4,1)
     &	-2d0*(               +D00mnl_sum1(3,3))*detz(1,2)
     &	+2d0*(D5_2(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(2,2)
     &	-2d0*(D5_3(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(3,2)
     &	+2d0*(               +D00mnl_sum1(3,3))*detz(4,2)
     &	-4d0*(               +D00mnl_sum1(1,3))*detz(1,4)
     &	+4d0*(               +D00mnl_sum1(1,3))*detz(2,4)
     &	-4d0*(D5_3(0,0,1,2,3)+D00mnl_sum1(1,3))*detz(3,4)
     &	+4d0*(               +D00mnl_sum1(1,3))*detz(4,4)

        E4n(1,3,3,3) =   D4_2(1,2,2,2)*dety(2)+D4_4(1,3,3,3)*dety(4)
     &		        -Dmnlk_sum1(2,2,2)*dety(0)
     &	+2d0*(D5_1(0,0,2,2,2)-D5(0,0,2,2,2)   )*detz(1,1)
     &	-2d0*(D5_2(0,0,2,2,2)-D5(0,0,2,2,2)   )*detz(2,1)
     &	+2d0*(               -D5(0,0,2,2,2)   )*detz(3,1)
     &	-2d0*(D5_4(0,0,3,3,3)-D5(0,0,2,2,2)   )*detz(4,1)
     &	+6d0*(               +D00mnl_sum1(2,2))*detz(1,3)
     &	-6d0*(D5_2(0,0,1,2,2)+D00mnl_sum1(2,2))*detz(2,3)
     &	+6d0*(               +D00mnl_sum1(2,2))*detz(3,3)
     &	-6d0*(D5_4(0,0,1,3,3)+D00mnl_sum1(2,2))*detz(4,3)

        E4n(1,3,3,4) =   D4_2(1,2,2,3)*dety(2)
     &		        -Dmnlk_sum1(2,2,3)*dety(0)
     &	+2d0*(D5_1(0,0,2,2,3)-D5(0,0,2,2,3)   )*detz(1,1)
     &	-2d0*(D5_2(0,0,2,2,3)-D5(0,0,2,2,3)   )*detz(2,1)
     &	+2d0*(               -D5(0,0,2,2,3)   )*detz(3,1)
     &	-2d0*(               -D5(0,0,2,2,3)   )*detz(4,1)
     &	+4d0*(               +D00mnl_sum1(2,3))*detz(1,3)
     &	-4d0*(D5_2(0,0,1,2,3)+D00mnl_sum1(2,3))*detz(2,3)
     &	+4d0*(               +D00mnl_sum1(2,3))*detz(3,3)
     &	-4d0*(               +D00mnl_sum1(2,3))*detz(4,3)
     &	-2d0*(               +D00mnl_sum1(2,2))*detz(1,4)
     &	+2d0*(D5_2(0,0,1,2,2)+D00mnl_sum1(2,2))*detz(2,4)
     &	-2d0*(               +D00mnl_sum1(2,2))*detz(3,4)
     &	+2d0*(D5_4(0,0,1,3,3)+D00mnl_sum1(2,2))*detz(4,4)

        E4n(1,3,4,4) =   D4_2(1,2,3,3)*dety(2)
     &		        -Dmnlk_sum1(2,3,3)*dety(0)
     &	+2d0*(D5_1(0,0,2,3,3)-D5(0,0,2,3,3)   )*detz(1,1)
     &	-2d0*(D5_2(0,0,2,3,3)-D5(0,0,2,3,3)   )*detz(2,1)
     &	+2d0*(               -D5(0,0,2,3,3)   )*detz(3,1)
     &	-2d0*(               -D5(0,0,2,3,3)   )*detz(4,1)
     &	+2d0*(               +D00mnl_sum1(3,3))*detz(1,3)
     &	-2d0*(D5_2(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(2,3)
     &	+2d0*(D5_3(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(3,3)
     &	-2d0*(               +D00mnl_sum1(3,3))*detz(4,3)
     &	-4d0*(               +D00mnl_sum1(2,3))*detz(1,4)
     &	+4d0*(D5_2(0,0,1,2,3)+D00mnl_sum1(2,3))*detz(2,4)
     &	-4d0*(               +D00mnl_sum1(2,3))*detz(3,4)
     &	+4d0*(               +D00mnl_sum1(2,3))*detz(4,4)

        E4n(1,4,4,4) =   D4_2(1,3,3,3)*dety(2)+D4_3(1,3,3,3)*dety(3)
     &		        -Dmnlk_sum1(3,3,3)*dety(0)
     &	+2d0*(D5_1(0,0,3,3,3)-D5(0,0,3,3,3)   )*detz(1,1)
     &	-2d0*(D5_2(0,0,3,3,3)-D5(0,0,3,3,3)   )*detz(2,1)
     &	+2d0*(D5_3(0,0,3,3,3)-D5(0,0,3,3,3)   )*detz(3,1)
     &	-2d0*(               -D5(0,0,3,3,3)   )*detz(4,1)
     &	-6d0*(               +D00mnl_sum1(3,3))*detz(1,4)
     &	+6d0*(D5_2(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(2,4)
     &	-6d0*(D5_3(0,0,1,3,3)+D00mnl_sum1(3,3))*detz(3,4)
     &	+6d0*(               +D00mnl_sum1(3,3))*detz(4,4)

        E4n(2,2,2,2) =   D4_1(1,1,1,1)*dety(1)+D4_3(2,2,2,2)*dety(3)
     &		        +D4_4(2,2,2,2)*dety(4)+D4(1,1,1,1)*dety(0)
     &	-8d0*(D5_1(0,0,1,1,1)-D5(0,0,1,1,1))*detz(1,2)
     &	+8d0*(               -D5(0,0,1,1,1))*detz(2,2)
     &	-8d0*(D5_3(0,0,2,2,2)-D5(0,0,1,1,1))*detz(3,2)
     &	+8d0*(D5_4(0,0,2,2,2)-D5(0,0,1,1,1))*detz(4,2)

        E4n(3,3,3,3) =   D4_1(2,2,2,2)*dety(1)+D4_2(2,2,2,2)*dety(2)
     &		        +D4_4(3,3,3,3)*dety(4)+D4(2,2,2,2)*dety(0)
     &	+8d0*(D5_1(0,0,2,2,2)-D5(0,0,2,2,2))*detz(1,3)
     &	-8d0*(D5_2(0,0,2,2,2)-D5(0,0,2,2,2))*detz(2,3)
     &	+8d0*(               -D5(0,0,2,2,2))*detz(3,3)
     &	-8d0*(D5_4(0,0,3,3,3)-D5(0,0,2,2,2))*detz(4,3)

        E4n(2,2,2,3) =   D4_1(1,1,1,2)*dety(1)+D4_4(2,2,2,3)*dety(4)
     &		        +D4(1,1,1,2)*dety(0)
     &	-6d0*(D5_1(0,0,1,1,2)-D5(0,0,1,1,2))*detz(1,2)
     &	+6d0*(               -D5(0,0,1,1,2))*detz(2,2)
     &	-6d0*(               -D5(0,0,1,1,2))*detz(3,2)
     &	+6d0*(D5_4(0,0,2,2,3)-D5(0,0,1,1,2))*detz(4,2)
     &	+2d0*(D5_1(0,0,1,1,1)-D5(0,0,1,1,1))*detz(1,3)
     &	-2d0*(               -D5(0,0,1,1,1))*detz(2,3)
     &	+2d0*(D5_3(0,0,2,2,2)-D5(0,0,1,1,1))*detz(3,3)
     &	-2d0*(D5_4(0,0,2,2,2)-D5(0,0,1,1,1))*detz(4,3)

        E4n(2,2,2,4) =   D4_1(1,1,1,3)*dety(1)+D4_3(2,2,2,3)*dety(3)
     &		        +D4(1,1,1,3)*dety(0)
     &	-6d0*(D5_1(0,0,1,1,3)-D5(0,0,1,1,3))*detz(1,2)
     &	+6d0*(               -D5(0,0,1,1,3))*detz(2,2)
     &	-6d0*(D5_3(0,0,2,2,3)-D5(0,0,1,1,3))*detz(3,2)
     &	+6d0*(               -D5(0,0,1,1,3))*detz(4,2)
     &	-2d0*(D5_1(0,0,1,1,1)-D5(0,0,1,1,1))*detz(1,4)
     &	+2d0*(               -D5(0,0,1,1,1))*detz(2,4)
     &	-2d0*(D5_3(0,0,2,2,2)-D5(0,0,1,1,1))*detz(3,4)
     &	+2d0*(D5_4(0,0,2,2,2)-D5(0,0,1,1,1))*detz(4,4)

        E4n(2,2,3,3) =   D4_1(1,1,2,2)*dety(1)+D4_4(2,2,3,3)*dety(4)
     &		        +D4(1,1,2,2)*dety(0)
     &	-4d0*(D5_1(0,0,1,2,2)-D5(0,0,1,2,2))*detz(1,2)
     &	+4d0*(               -D5(0,0,1,2,2))*detz(2,2)
     &	-4d0*(               -D5(0,0,1,2,2))*detz(3,2)
     &	+4d0*(D5_4(0,0,2,3,3)-D5(0,0,1,2,2))*detz(4,2)
     &	+4d0*(D5_1(0,0,1,1,2)-D5(0,0,1,1,2))*detz(1,3)
     &	-4d0*(               -D5(0,0,1,1,2))*detz(2,3)
     &	+4d0*(               -D5(0,0,1,1,2))*detz(3,3)
     &	-4d0*(D5_4(0,0,2,2,3)-D5(0,0,1,1,2))*detz(4,3)

        E4n(2,2,3,4) =   D4_1(1,1,2,3)*dety(1)
     &		        +D4(1,1,2,3)*dety(0)
     &	-4d0*(D5_1(0,0,1,2,3)-D5(0,0,1,2,3))*detz(1,2)
     &	+4d0*(               -D5(0,0,1,2,3))*detz(2,2)
     &	-4d0*(               -D5(0,0,1,2,3))*detz(3,2)
     &	+4d0*(               -D5(0,0,1,2,3))*detz(4,2)
     &	+2d0*(D5_1(0,0,1,1,3)-D5(0,0,1,1,3))*detz(1,3)
     &	-2d0*(               -D5(0,0,1,1,3))*detz(2,3)
     &	+2d0*(D5_3(0,0,2,2,3)-D5(0,0,1,1,3))*detz(3,3)
     &	-2d0*(               -D5(0,0,1,1,3))*detz(4,3)
     &	-2d0*(D5_1(0,0,1,1,2)-D5(0,0,1,1,2))*detz(1,4)
     &	+2d0*(               -D5(0,0,1,1,2))*detz(2,4)
     &	-2d0*(               -D5(0,0,1,1,2))*detz(3,4)
     &	+2d0*(D5_4(0,0,2,2,3)-D5(0,0,1,1,2))*detz(4,4)

        E4n(2,2,4,4) =   D4_1(1,1,3,3)*dety(1)+D4_3(2,2,3,3)*dety(3)
     &		        +D4(1,1,3,3)*dety(0)
     &	-4d0*(D5_1(0,0,1,3,3)-D5(0,0,1,3,3))*detz(1,2)
     &	+4d0*(               -D5(0,0,1,3,3))*detz(2,2)
     &	-4d0*(D5_3(0,0,2,3,3)-D5(0,0,1,3,3))*detz(3,2)
     &	+4d0*(               -D5(0,0,1,3,3))*detz(4,2)
     &	-4d0*(D5_1(0,0,1,1,3)-D5(0,0,1,1,3))*detz(1,4)
     &	+4d0*(               -D5(0,0,1,1,3))*detz(2,4)
     &	-4d0*(D5_3(0,0,2,2,3)-D5(0,0,1,1,3))*detz(3,4)
     &	+4d0*(               -D5(0,0,1,1,3))*detz(4,4)

        E4n(2,3,3,3) =   D4_1(1,2,2,2)*dety(1)+D4_4(2,3,3,3)*dety(4)
     &		        +D4(1,2,2,2)*dety(0)
     &	-2d0*(D5_1(0,0,2,2,2)-D5(0,0,2,2,2))*detz(1,2)
     &	+2d0*(D5_2(0,0,2,2,2)-D5(0,0,2,2,2))*detz(2,2)
     &	-2d0*(               -D5(0,0,2,2,2))*detz(3,2)
     &	+2d0*(D5_4(0,0,3,3,3)-D5(0,0,2,2,2))*detz(4,2)
     &	+6d0*(D5_1(0,0,1,2,2)-D5(0,0,1,2,2))*detz(1,3)
     &	-6d0*(               -D5(0,0,1,2,2))*detz(2,3)
     &	+6d0*(               -D5(0,0,1,2,2))*detz(3,3)
     &	-6d0*(D5_4(0,0,2,3,3)-D5(0,0,1,2,2))*detz(4,3)

        E4n(2,3,3,4) =   D4_1(1,2,2,3)*dety(1)
     &		        +D4(1,2,2,3)*dety(0)
     &	-2d0*(D5_1(0,0,2,2,3)-D5(0,0,2,2,3))*detz(1,2)
     &	+2d0*(D5_2(0,0,2,2,3)-D5(0,0,2,2,3))*detz(2,2)
     &	-2d0*(               -D5(0,0,2,2,3))*detz(3,2)
     &	+2d0*(               -D5(0,0,2,2,3))*detz(4,2)
     &	+4d0*(D5_1(0,0,1,2,3)-D5(0,0,1,2,3))*detz(1,3)
     &	-4d0*(               -D5(0,0,1,2,3))*detz(2,3)
     &	+4d0*(               -D5(0,0,1,2,3))*detz(3,3)
     &	-4d0*(               -D5(0,0,1,2,3))*detz(4,3)
     &	-2d0*(D5_1(0,0,1,2,2)-D5(0,0,1,2,2))*detz(1,4)
     &	+2d0*(               -D5(0,0,1,2,2))*detz(2,4)
     &	-2d0*(               -D5(0,0,1,2,2))*detz(3,4)
     &	+2d0*(D5_4(0,0,2,3,3)-D5(0,0,1,2,2))*detz(4,4)

        E4n(2,3,4,4) =   D4_1(1,2,3,3)*dety(1)
     &		        +D4(1,2,3,3)*dety(0)
     &	-2d0*(D5_1(0,0,2,3,3)-D5(0,0,2,3,3))*detz(1,2)
     &	+2d0*(D5_2(0,0,2,3,3)-D5(0,0,2,3,3))*detz(2,2)
     &	-2d0*(               -D5(0,0,2,3,3))*detz(3,2)
     &	+2d0*(               -D5(0,0,2,3,3))*detz(4,2)
     &	+2d0*(D5_1(0,0,1,3,3)-D5(0,0,1,3,3))*detz(1,3)
     &	-2d0*(               -D5(0,0,1,3,3))*detz(2,3)
     &	+2d0*(D5_3(0,0,2,3,3)-D5(0,0,1,3,3))*detz(3,3)
     &	-2d0*(               -D5(0,0,1,3,3))*detz(4,3)
     &	-4d0*(D5_1(0,0,1,2,3)-D5(0,0,1,2,3))*detz(1,4)
     &	+4d0*(               -D5(0,0,1,2,3))*detz(2,4)
     &	-4d0*(               -D5(0,0,1,2,3))*detz(3,4)
     &	+4d0*(               -D5(0,0,1,2,3))*detz(4,4)

        E4n(2,4,4,4) =   D4_1(1,3,3,3)*dety(1)+D4_3(2,3,3,3)*dety(3)
     &		        +D4(1,3,3,3)*dety(0)
     &	-2d0*(D5_1(0,0,3,3,3)-D5(0,0,3,3,3))*detz(1,2)
     &	+2d0*(D5_2(0,0,3,3,3)-D5(0,0,3,3,3))*detz(2,2)
     &	-2d0*(D5_3(0,0,3,3,3)-D5(0,0,3,3,3))*detz(3,2)
     &	+2d0*(               -D5(0,0,3,3,3))*detz(4,2)
     &	-6d0*(D5_1(0,0,1,3,3)-D5(0,0,1,3,3))*detz(1,4)
     &	+6d0*(               -D5(0,0,1,3,3))*detz(2,4)
     &	-6d0*(D5_3(0,0,2,3,3)-D5(0,0,1,3,3))*detz(3,4)
     &	+6d0*(               -D5(0,0,1,3,3))*detz(4,4)

        E4n(3,3,3,4) =   D4_1(2,2,2,3)*dety(1)+D4_2(2,2,2,3)*dety(2)
     &		        +D4(2,2,2,3)*dety(0)
     &	+6d0*(D5_1(0,0,2,2,3)-D5(0,0,2,2,3))*detz(1,3)
     &	-6d0*(D5_2(0,0,2,2,3)-D5(0,0,2,2,3))*detz(2,3)
     &	+6d0*(               -D5(0,0,2,2,3))*detz(3,3)
     &	-6d0*(               -D5(0,0,2,2,3))*detz(4,3)
     &	-2d0*(D5_1(0,0,2,2,2)-D5(0,0,2,2,2))*detz(1,4)
     &	+2d0*(D5_2(0,0,2,2,2)-D5(0,0,2,2,2))*detz(2,4)
     &	-2d0*(               -D5(0,0,2,2,2))*detz(3,4)
     &	+2d0*(D5_4(0,0,3,3,3)-D5(0,0,2,2,2))*detz(4,4)

        E4n(3,3,4,4) =   D4_1(2,2,3,3)*dety(1)+D4_2(2,2,3,3)*dety(2)
     &		        +D4(2,2,3,3)*dety(0)
     &	+4d0*(D5_1(0,0,2,3,3)-D5(0,0,2,3,3))*detz(1,3)
     &	-4d0*(D5_2(0,0,2,3,3)-D5(0,0,2,3,3))*detz(2,3)
     &	+4d0*(               -D5(0,0,2,3,3))*detz(3,3)
     &	-4d0*(               -D5(0,0,2,3,3))*detz(4,3)
     &	-4d0*(D5_1(0,0,2,2,3)-D5(0,0,2,2,3))*detz(1,4)
     &	+4d0*(D5_2(0,0,2,2,3)-D5(0,0,2,2,3))*detz(2,4)
     &	-4d0*(               -D5(0,0,2,2,3))*detz(3,4)
     &	+4d0*(               -D5(0,0,2,2,3))*detz(4,4)

        E4n(3,4,4,4) =   D4_1(2,3,3,3)*dety(1)+D4_2(2,3,3,3)*dety(2)
     &		        +D4(2,3,3,3)*dety(0)
     &	+2d0*(D5_1(0,0,3,3,3)-D5(0,0,3,3,3))*detz(1,3)
     &	-2d0*(D5_2(0,0,3,3,3)-D5(0,0,3,3,3))*detz(2,3)
     &	+2d0*(D5_3(0,0,3,3,3)-D5(0,0,3,3,3))*detz(3,3)
     &	-2d0*(               -D5(0,0,3,3,3))*detz(4,3)
     &	-6d0*(D5_1(0,0,2,3,3)-D5(0,0,2,3,3))*detz(1,4)
     &	+6d0*(D5_2(0,0,2,3,3)-D5(0,0,2,3,3))*detz(2,4)
     &	-6d0*(               -D5(0,0,2,3,3))*detz(3,4)
     &	+6d0*(               -D5(0,0,2,3,3))*detz(4,4)

        E4n(4,4,4,4) =   D4_1(3,3,3,3)*dety(1)+D4_2(3,3,3,3)*dety(2)
     &		        +D4_3(3,3,3,3)*dety(3)+D4(3,3,3,3)*dety(0)
     &	-8d0*(D5_1(0,0,3,3,3)-D5(0,0,3,3,3))*detz(1,4)
     &	+8d0*(D5_2(0,0,3,3,3)-D5(0,0,3,3,3))*detz(2,4)
     &	-8d0*(D5_3(0,0,3,3,3)-D5(0,0,3,3,3))*detz(3,4)
     &	+8d0*(               -D5(0,0,3,3,3))*detz(4,4)

	do j=1,3
	do k=j+1,4
	  E4n(0,0,k,j) = E4n(0,0,j,k)
	enddo
	enddo

c*** NOTE: E4n(j,k,l,m) only set for j >= k >= l >= m

 1007   fct(1)=D0
        fct(2)=D1(1)
        fct(3)=D1(2)
        fct(4)=D1(3)
        fct(5)=D2(0,0)
        fct(6)=D2(1,1)
        fct(7)=D2(1,2)
        fct(8)=D2(1,3)
        fct(9)=D2(2,2)
        fct(10)=D2(2,3)
        fct(11)=D2(3,3)
        fct(12)=D3(0,0,1)
        fct(13)=D3(0,0,2)
        fct(14)=D3(0,0,3)
        fct(15)=D3(1,1,1)
        fct(16)=D3(1,1,2)
        fct(17)=D3(1,1,3)
        fct(18)=D3(1,2,2)
        fct(19)=D3(1,2,3)
        fct(20)=D3(1,3,3)
        fct(21)=D3(2,2,2)
        fct(22)=D3(2,2,3)
        fct(23)=D3(2,3,3)
        fct(24)=D3(3,3,3)
        fct(25)=E0n
        fct(26)=E1n(1)
        fct(27)=E1n(2)
        fct(28)=E1n(3)
        fct(29)=E1n(4)
        fct(30)=E2n(0,0)
        fct(31)=E2n(1,1)
        fct(32)=E2n(1,2)
        fct(33)=E2n(1,3)
        fct(34)=E2n(1,4)
        fct(35)=E2n(2,2)
        fct(36)=E2n(2,3)
        fct(37)=E2n(2,4)
        fct(38)=E2n(3,3)
        fct(39)=E2n(3,4)
        fct(40)=E2n(4,4)
        fct(41)=E3n(0,0,1)
        fct(42)=E3n(0,0,2)
        fct(43)=E3n(0,0,3)
        fct(44)=E3n(0,0,4)
        fct(45)=E3n(1,1,1)
        fct(46)=E3n(1,1,2)
        fct(47)=E3n(1,1,3)
        fct(48)=E3n(1,1,4)
        fct(49)=E3n(1,2,2)
        fct(50)=E3n(1,2,3)
        fct(51)=E3n(1,2,4)
        fct(52)=E3n(1,3,3)
        fct(53)=E3n(1,3,4)
        fct(54)=E3n(1,4,4)
        fct(55)=E3n(2,2,2)
        fct(56)=E3n(2,2,3)
        fct(57)=E3n(2,2,4)
        fct(58)=E3n(2,3,3)
        fct(59)=E3n(2,3,4)
        fct(60)=E3n(2,4,4)
        fct(61)=E3n(3,3,3)
        fct(62)=E3n(3,3,4)
        fct(63)=E3n(3,4,4)
        fct(64)=E3n(4,4,4)
        fct(65)=E4n(0,0,0,0)
        fct(66)=E4n(0,0,1,1)
        fct(67)=E4n(0,0,1,2)
        fct(68)=E4n(0,0,1,3)
        fct(69)=E4n(0,0,1,4)
        fct(70)=E4n(0,0,2,2)
        fct(71)=E4n(0,0,2,3)
        fct(72)=E4n(0,0,2,4)
        fct(73)=E4n(0,0,3,3)
        fct(74)=E4n(0,0,3,4)
        fct(75)=E4n(0,0,4,4)
        fct(76)=E4n(1,1,1,1)
        fct(77)=E4n(1,1,1,2)
        fct(78)=E4n(1,1,1,3)
        fct(79)=E4n(1,1,1,4)
        fct(80)=E4n(1,1,2,2)
        fct(81)=E4n(1,1,2,3)
        fct(82)=E4n(1,1,2,4)
        fct(83)=E4n(1,1,3,3)
        fct(84)=E4n(1,1,3,4)
        fct(85)=E4n(1,1,4,4)
        fct(86)=E4n(1,2,2,2)
        fct(87)=E4n(1,2,2,3)
        fct(88)=E4n(1,2,2,4)
        fct(89)=E4n(1,2,3,3)
        fct(90)=E4n(1,2,3,4)
        fct(91)=E4n(1,2,4,4)
        fct(92)=E4n(1,3,3,3)
        fct(93)=E4n(1,3,3,4)
        fct(94)=E4n(1,3,4,4)
        fct(95)=E4n(1,4,4,4)
        fct(96)=E4n(2,2,2,2)
        fct(97)=E4n(2,2,2,3)
        fct(98)=E4n(2,2,2,4)
        fct(99)=E4n(2,2,3,3)
        fct(100)=E4n(2,2,3,4)
        fct(101)=E4n(2,2,4,4)
        fct(102)=E4n(2,3,3,3)
        fct(103)=E4n(2,3,3,4)
        fct(104)=E4n(2,3,4,4)
        fct(105)=E4n(2,4,4,4)
        fct(106)=E4n(3,3,3,3)
        fct(107)=E4n(3,3,3,4)
        fct(108)=E4n(3,3,4,4)
        fct(109)=E4n(3,4,4,4)
        fct(110)=E4n(4,4,4,4)
        call fastout(fct,110,1)
	END

***********************************************************************
	SUBROUTINE xF0123(p1,p2,p3,p4,p5,p6,s12,s23,s34,s45,s56,s16,
     &	 s123,s234,s345,m02,m12,m22,m32,m42,m52,F0,F1,F2,F3,E0,E1,E2,
     &   rank,ext)
***********************************************************************
*	6-POINT TENSOR COEFFICIENTS with complex masses
*	conventional Passarino-Veltman reduction
*-----------------------------------------------------------------------
*       11.6.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
        COMPLEX*16 m02,m12,m22,m32,m42,m52,f(4),xdet
        COMPLEX*16 D0,D1(3),D2(0:3,0:3),D3(0:3,0:3,3)
        COMPLEX*16 E0,  E1(4),  E2(4,4),  E3(4,4,4),E4(4,4,4,4)
        COMPLEX*16 E0_1,E1_1(4),E2_1(4,4),E3_1(4,4,4),E4_1(4,4,4,4)
        COMPLEX*16 E0_2,E1_2(4),E2_2(4,4),E3_2(4,4,4),E4_2(4,4,4,4)
        COMPLEX*16 E0_3,E1_3(4),E2_3(4,4),E3_3(4,4,4),E4_3(4,4,4,4)
        COMPLEX*16 E0_4,E1_4(4),E2_4(4,4),E3_4(4,4,4),E4_4(4,4,4,4)
        COMPLEX*16 E0_5,E1_5(4),E2_5(4,4),E3_5(4,4,4),E4_5(4,4,4,4)
        COMPLEX*16 F0,F1(4),F2(4,4),F3(4,4,4)
        COMPLEX*16 T1(4),T2(4,4),T3(4,4,4),T4(4,4,4,4)
        COMPLEX*16 E1sum,E2sum0,E2sum(4)
	complex*16 y(0:5,0:5),dety(0:5),detyy
	complex*16 yy(6,6),y0(6,6),y1(6,6),y2(6,6),y3(6,6),y4(6,6),y5(6,6)
	real*8 pp(4,4),zi(4,4),p5p(4),xx(4)
	INTEGER j,k,l,m,n,rank,ext

	call xE01234_(p2,p3,p4,p5,s16, s23,s34,s45,s234,s345, m12,m22,
     *	   m32,m42,m52,E0,E1,E2,E3,E4,D0,D1,D2,D3,rank-1,ext)
 	call xE01234_(s12,p3,p4,p5,p6, s123,s34,s45,s56,s345, m02,m22,
     *	   m32,m42,m52,E0_1,E1_1,E2_1,E3_1,E4_1,D0,D1,D2,D3,rank-1,ext)
	call xE01234_(p1,s23,p4,p5,p6, s123,s234,s45,s56,s16, m02,m12,
     *	   m32,m42,m52,E0_2,E1_2,E2_2,E3_2,E4_2,D0,D1,D2,D3,rank-1,ext)
	call xE01234_(p1,p2,s34,p5,p6, s12,s234,s345,s56,s16, m02,m12,
     *	   m22,m42,m52,E0_3,E1_3,E2_3,E3_3,E4_3,D0,D1,D2,D3,rank-1,ext)
	call xE01234_(p1,p2,p3,s45,p6, s12,s23,s345,s123,s16, m02,m12,
     *	   m22,m32,m52,E0_4,E1_4,E2_4,E3_4,E4_4,D0,D1,D2,D3,rank-1,ext)
	call xE01234_(p1,p2,p3,p4,s56, s12,s23,s34,s123,s234, m02,m12,
     *	   m22,m32,m42,E0_5,E1_5,E2_5,E3_5,E4_5,D0,D1,D2,D3,rank-1,ext)

	y(0,0) = 2d0*m02
	y(1,1) = 2d0*m12
	y(2,2) = 2d0*m22
	y(3,3) = 2d0*m32
	y(4,4) = 2d0*m42
	y(5,5) = 2d0*m52
	y(0,1) = m02+m12-p1
	y(1,2) = m12+m22-p2
	y(2,3) = m22+m32-p3
	y(3,4) = m32+m42-p4
	y(4,5) = m42+m52-p5
	y(0,5) = m52+m02-p6
	y(0,2) = m02+m22-s12
	y(1,3) = m12+m32-s23
	y(2,4) = m22+m42-s34
	y(3,5) = m32+m52-s45
	y(0,4) = m42+m02-s56
	y(1,5) = m52+m12-s16
	y(0,3) = m02+m32-s123
	y(1,4) = m12+m42-s234
	y(2,5) = m22+m52-s345

	do j=0,4
	  do k=j+1,5
	    y(k,j) = y(j,k)
	  enddo
	enddo

	do j=1,6
	  do k=1,6
	    yy(j,k) = y(j-1,k-1)
	    if (k.ne.1) then
	      y0(j,k) = y(j-1,k-1)
	    else
	      y0(j,k) = 1d0
	    endif
	    if (k.ne.2) then
	      y1(j,k) = y(j-1,k-1)
	    else
	      y1(j,k) = 1d0
	    endif
	    if (k.ne.3) then
	      y2(j,k) = y(j-1,k-1)
	    else
	      y2(j,k) = 1d0
	    endif
	    if (k.ne.4) then
	      y3(j,k) = y(j-1,k-1)
	    else
	      y3(j,k) = 1d0
	    endif
	    if (k.ne.5) then
	      y4(j,k) = y(j-1,k-1)
	    else
	      y4(j,k) = 1d0
	    endif
	    if (k.ne.6) then
	      y5(j,k) = y(j-1,k-1)
	    else
	      y5(j,k) = 1d0
	    endif
	  enddo
	enddo

	detyy   =  xdet(yy,6)
	dety(0) = -xdet(y0,6)/detyy
	dety(1) = -xdet(y1,6)/detyy
	dety(2) = -xdet(y2,6)/detyy
	dety(3) = -xdet(y3,6)/detyy
	dety(4) = -xdet(y4,6)/detyy
	dety(5) = -xdet(y5,6)/detyy

	F0 =  E0*dety(0)  +E0_1*dety(1)+E0_2*dety(2)
     &	     +E0_3*dety(3)+E0_4*dety(4)+E0_5*dety(5)

	if (rank.eq.0) return

	pp(1,1) = 2d0* p1
	pp(2,2) = 2d0*s12
	pp(3,3) = 2d0*s123
	pp(4,4) = 2d0*s56
	pp(1,2) = s12 +  p1-p2
	pp(2,3) = s12 +s123-p3
	pp(3,4) = s123+s56 -p4
	pp(1,3) = p1  +s123-s23
	pp(2,4) = s12 +s56 -s34
	pp(1,4) = p1  +s56 -s234

	do j=1,3
	  do k=j+1,4
	    pp(k,j) = pp(j,k)
	  enddo
	enddo

        call inverse(pp,zi,4)

	f(1) =  p1 -m12+m02
	f(2) = s12 -m22+m02
	f(3) = s123-m32+m02
	f(4) = s56 -m42+m02

	T1(1) = E0_1-E0
	T1(2) = E0_2-E0
	T1(3) = E0_3-E0
	T1(4) = E0_4-E0

	do 1 k=1,4
	F1(k) = 0d0
	do 1 j=1,4
	F1(k) = F1(k) + ZI(k,j)*(T1(j)-f(j)*F0)
1	continue

	if (rank.eq.1) return

	p5p(1) = p1  +p6-s16
	p5p(2) = s12 +p6-s345
	p5p(3) = s123+p6-s45
	p5p(4) = s56 +p6-p5

	do j=1,4
	  xx(j) = 0d0
	  do k=1,4
	    xx(j) = xx(j) + zi(j,k)*p5p(k)
	  enddo
	enddo

	E1sum = E0 + E1(1) + E1(2) + E1(3) + E1(4)

	T2(1,1) =         + E1sum + xx(1)*(E1_1(4)-E1(4))
	T2(1,2) = E1_1(1) - E1(1) + xx(2)*(E1_1(4)-E1(4))
	T2(1,3) = E1_1(2) - E1(2) + xx(3)*(E1_1(4)-E1(4))
	T2(1,4) = E1_1(3) - E1(3) + xx(4)*(E1_1(4)-E1(4))

	T2(2,1) = E1_2(1) + E1sum + xx(1)*(E1_2(4)-E1(4))
	T2(2,2) =         - E1(1) + xx(2)*(E1_2(4)-E1(4))
	T2(2,3) = E1_2(2) - E1(2) + xx(3)*(E1_2(4)-E1(4))
	T2(2,4) = E1_2(3) - E1(3) + xx(4)*(E1_2(4)-E1(4))

	T2(3,1) = E1_3(1) + E1sum + xx(1)*(E1_3(4)-E1(4))
	T2(3,2) = E1_3(2) - E1(1) + xx(2)*(E1_3(4)-E1(4))
	T2(3,3) =         - E1(2) + xx(3)*(E1_3(4)-E1(4))
	T2(3,4) = E1_3(3) - E1(3) + xx(4)*(E1_3(4)-E1(4))

	T2(4,1) = E1_4(1) + E1sum + xx(1)*(E1_4(4)-E1(4))
	T2(4,2) = E1_4(2) - E1(1) + xx(2)*(E1_4(4)-E1(4))
	T2(4,3) = E1_4(3) - E1(2) + xx(3)*(E1_4(4)-E1(4))
	T2(4,4) =         - E1(3) + xx(4)*(E1_4(4)-E1(4))

	do 2 k=1,4
	do 2 l=1,4
	F2(k,l) = 0d0
	do 2 j=1,4
	F2(k,l) = F2(k,l) + ZI(k,j)*(T2(j,l)-f(j)*F1(l))
2	continue

	if (rank.eq.2) return

	E2sum0   = E2(1,1)+E2(2,2)+E2(3,3)+E2(4,4)
     &	  +2d0*(E2(1,2)+E2(1,3)+E2(1,4)+E2(2,3)+E2(2,4)+E2(3,4))
     &	  +2d0*(E1(1)+E1(2)+E1(3)+E1(4))+E0
	E2sum(1) = E2(1,1)+E2(1,2)+E2(1,3)+E2(1,4)+E1(1)
	E2sum(2) = E2(1,2)+E2(2,2)+E2(2,3)+E2(2,4)+E1(2)
	E2sum(3) = E2(1,3)+E2(2,3)+E2(3,3)+E2(3,4)+E1(3)
	E2sum(4) = E2(1,4)+E2(2,4)+E2(3,4)+E2(4,4)+E1(4)

	T3(1,1,1) =           - E2sum0
     &		+ xx(1)**2*(E2_1(4,4)-E2(4,4))
     &		+ 2d0*xx(1)*E2sum(4)
	T3(1,1,2) =             E2sum(1)
     &		+ xx(1)*xx(2)*(E2_1(4,4)-E2(4,4))   + xx(2)*E2sum(4)
     &		+ xx(1)*(E2_1(1,4)-E2(1,4))
	T3(1,1,3) =             E2sum(2)
     &		+ xx(1)*xx(3)*(E2_1(4,4)-E2(4,4))   + xx(3)*E2sum(4)
     &		+ xx(1)*(E2_1(2,4)-E2(2,4))
	T3(1,1,4) =             E2sum(3)
     &		+ xx(1)*xx(4)*(E2_1(4,4)-E2(4,4))   + xx(4)*E2sum(4)
     &		+ xx(1)*(E2_1(3,4)-E2(3,4))
	T3(1,2,2) = E2_1(1,1) - E2(1,1)
     &		+ xx(2)**2*(E2_1(4,4)-E2(4,4))
     &		+ 2d0*xx(2)*(E2_1(1,4)-E2(1,4))
	T3(1,2,3) = E2_1(1,2) - E2(1,2)
     &		+ xx(2)*xx(3)*(E2_1(4,4)-E2(4,4))
     &		+ xx(2)*(E2_1(2,4)-E2(2,4)) + xx(3)*(E2_1(1,4)-E2(1,4))
	T3(1,2,4) = E2_1(1,3) - E2(1,3)
     &		+ xx(2)*xx(4)*(E2_1(4,4)-E2(4,4))
     &		+ xx(2)*(E2_1(3,4)-E2(3,4)) + xx(4)*(E2_1(1,4)-E2(1,4))
	T3(1,3,3) = E2_1(2,2) - E2(2,2)
     &		+ xx(3)**2*(E2_1(4,4)-E2(4,4))
     &		+ 2d0*xx(3)*(E2_1(2,4)-E2(2,4))
	T3(1,3,4) = E2_1(2,3) - E2(2,3)
     &		+ xx(3)*xx(4)*(E2_1(4,4)-E2(4,4))
     &		+ xx(3)*(E2_1(3,4)-E2(3,4)) + xx(4)*(E2_1(2,4)-E2(2,4))
	T3(1,4,4) = E2_1(3,3) - E2(3,3)
     &		+ xx(4)**2*(E2_1(4,4)-E2(4,4))
     &		+ 2d0*xx(4)*(E2_1(3,4)-E2(3,4))

	T3(2,1,1) = E2_2(1,1) - E2sum0
     &		+ xx(1)**2*(E2_2(4,4)-E2(4,4))+ 2d0*xx(1)*E2sum(4)
     &		+ 2d0*xx(1)*E2_2(1,4)
	T3(2,1,2) =             E2sum(1)
     &		+ xx(1)*xx(2)*(E2_2(4,4)-E2(4,4)) + xx(2)*E2sum(4)
     &		+ xx(2)*E2_2(1,4) - xx(1)*E2(1,4)
	T3(2,1,3) = E2_2(1,2) + E2sum(2)
     &		+ xx(1)*xx(3)*(E2_2(4,4)-E2(4,4)) + xx(3)*E2sum(4)
     &		+ xx(3)*E2_2(1,4) + xx(1)*(E2_2(2,4)-E2(2,4))
	T3(2,1,4) = E2_2(1,3) + E2sum(3)
     &		+ xx(1)*xx(4)*(E2_2(4,4)-E2(4,4)) + xx(4)*E2sum(4)
     &		+ xx(4)*E2_2(1,4) + xx(1)*(E2_2(3,4)-E2(3,4))
	T3(2,2,2) =           - E2(1,1)
     &		+ xx(2)**2*(E2_2(4,4)-E2(4,4))
     &		- 2d0*xx(2)*E2(1,4)
	T3(2,2,3) =           - E2(1,2)
     &		+ xx(2)*xx(3)*(E2_2(4,4)-E2(4,4))
     &		+ xx(2)*(E2_2(2,4)-E2(2,4)) - xx(3)*E2(1,4)
	T3(2,2,4) =           - E2(1,3)
     &		+ xx(2)*xx(4)*(E2_2(4,4)-E2(4,4))
     &		+ xx(2)*(E2_2(3,4)-E2(3,4)) - xx(4)*E2(1,4)
	T3(2,3,3) = E2_2(2,2) - E2(2,2)
     &		+ xx(3)**2*(E2_2(4,4)-E2(4,4))
     &		+ 2d0*xx(3)*(E2_2(2,4)-E2(2,4))
	T3(2,3,4) = E2_2(2,3) - E2(2,3)
     &		+ xx(3)*xx(4)*(E2_2(4,4)-E2(4,4))
     &		+ xx(3)*(E2_2(3,4)-E2(3,4)) + xx(4)*(E2_2(2,4)-E2(2,4))
	T3(2,4,4) = E2_2(3,3) - E2(3,3)
     &		+ xx(4)**2*(E2_2(4,4)-E2(4,4))
     &		+ 2d0*xx(4)*(E2_2(3,4)-E2(3,4))

	T3(3,1,1) = E2_3(1,1) - E2sum0
     &		+ xx(1)**2*(E2_3(4,4)-E2(4,4))+ 2d0*xx(1)*E2sum(4)
     &		+ 2d0*xx(1)*E2_3(1,4)
	T3(3,1,2) = E2_3(1,2) + E2sum(1)
     &		+ xx(1)*xx(2)*(E2_3(4,4)-E2(4,4)) + xx(2)*E2sum(4)
     &		+ xx(1)*E2_3(2,4) + xx(2)*E2_3(1,4) - xx(1)*E2(1,4)
	T3(3,1,3) =             E2sum(2)
     &		+ xx(1)*xx(3)*(E2_3(4,4)-E2(4,4)) + xx(3)*E2sum(4)
     &		+ xx(3)*E2_3(1,4) - xx(1)*E2(2,4)
	T3(3,1,4) = E2_3(1,3) + E2sum(3)
     &		+ xx(1)*xx(4)*(E2_3(4,4)-E2(4,4)) + xx(4)*E2sum(4)
     &		+ xx(4)*E2_3(1,4) + xx(1)*(E2_3(3,4)-E2(3,4))
	T3(3,2,2) = E2_3(2,2) - E2(1,1)
     &		+ xx(2)**2*(E2_3(4,4)-E2(4,4))
     &		+2d0*xx(2)*(E2_3(2,4)-E2(1,4))
	T3(3,2,3) =           - E2(1,2)
     &		+ xx(2)*xx(3)*(E2_3(4,4)-E2(4,4))
     &		- xx(2)*E2(2,4) + xx(3)*(E2_3(2,4)-E2(1,4))
	T3(3,2,4) = E2_3(2,3) - E2(1,3)
     &		+ xx(2)*xx(4)*(E2_3(4,4)-E2(4,4))
     &		+ xx(2)*(E2_3(3,4)-E2(3,4)) + xx(4)*(E2_3(2,4)-E2(1,4))
	T3(3,3,3) =           - E2(2,2)
     &		+ xx(3)**2*(E2_3(4,4)-E2(4,4))
     &		- 2d0*xx(3)*E2(2,4)
	T3(3,3,4) =           - E2(2,3)
     &		+ xx(3)*xx(4)*(E2_3(4,4)-E2(4,4))
     &		+ xx(3)*(E2_3(3,4)-E2(3,4)) - xx(4)*E2(2,4)
	T3(3,4,4) = E2_3(3,3) - E2(3,3)
     &		+ xx(4)**2*(E2_3(4,4)-E2(4,4))
     &		+ 2d0*xx(4)*(E2_3(3,4)-E2(3,4))

	T3(4,1,1) = E2_4(1,1) - E2sum0
     &		+ xx(1)**2*(E2_4(4,4)-E2(4,4))+ 2d0*xx(1)*E2sum(4)
     &		+ 2d0*xx(1)*E2_4(1,4)
	T3(4,1,2) = E2_4(1,2) + E2sum(1)
     &		+ xx(1)*xx(2)*(E2_4(4,4)-E2(4,4)) + xx(2)*E2sum(4)
     &		+ xx(1)*E2_4(2,4) + xx(2)*E2_4(1,4) - xx(1)*E2(1,4)
	T3(4,1,3) = E2_4(1,3) + E2sum(2)
     &		+ xx(1)*xx(3)*(E2_4(4,4)-E2(4,4)) + xx(3)*E2sum(4)
     &		+ xx(1)*E2_4(3,4) + xx(3)*E2_4(1,4) - xx(1)*E2(2,4)
	T3(4,1,4) =             E2sum(3)
     &		+ xx(1)*xx(4)*(E2_4(4,4)-E2(4,4)) + xx(4)*E2sum(4)
     &		+ xx(4)*E2_4(1,4) - xx(1)*E2(3,4)
	T3(4,2,2) = E2_4(2,2) - E2(1,1)
     &		+ xx(2)**2*(E2_4(4,4)-E2(4,4))
     &		+ 2d0*xx(2)*(E2_4(2,4)-E2(1,4))
	T3(4,2,3) = E2_4(2,3) - E2(1,2)
     &		+ xx(2)*xx(3)*(E2_4(4,4)-E2(4,4))
     &		+ xx(2)*(E2_4(3,4)-E2(2,4)) + xx(3)*(E2_4(2,4)-E2(1,4))
	T3(4,2,4) =           - E2(1,3)
     &		+ xx(2)*xx(4)*(E2_4(4,4)-E2(4,4))
     &		- xx(2)*E2(3,4) + xx(4)*(E2_4(2,4)-E2(1,4))
	T3(4,3,3) = E2_4(3,3) - E2(2,2)
     &		+ xx(3)**2*(E2_4(4,4)-E2(4,4))
     &		+ 2d0*xx(3)*(E2_4(3,4)-E2(2,4))
	T3(4,3,4) =           - E2(2,3)
     &		+ xx(3)*xx(4)*(E2_4(4,4)-E2(4,4))
     &		- xx(3)*E2(3,4) + xx(4)*(E2_4(3,4)-E2(2,4))
	T3(4,4,4) =           - E2(3,3)
     &		+ xx(4)**2*(E2_4(4,4)-E2(4,4))
     &		- 2d0*xx(4)*E2(3,4)

	do 30 j=1,4
	do 30 l=1,4
	do 30 m=1,l-1
	T3(j,l,m) = T3(j,m,l)
30	continue

	do 3 k=1,4
	do 3 l=1,4
	do 3 m=1,4
	F3(k,l,m) = 0d0
	do 3 j=1,4
	F3(k,l,m) = F3(k,l,m) + ZI(k,j)*( T3(j,l,m) - f(j)*F2(l,m) )
3	continue

	end

***********************************************************************
	SUBROUTINE xF0123_new(p1,p2,p3,p4,p5,p6,s12,s23,s34,s45,s56,s16,
     &	 s123,s234,s345,m02,m12,m22,m32,m42,m52,F0,F1,F2,F3,E0,E1,E2,
     &	 rank,ext)
***********************************************************************
*	6-POINT TENSOR COEFFICIENTS with complex masses
*	direct reduction -> 5-point -> 4-point integrals
*-----------------------------------------------------------------------
*       11.6.04 Stefan Dittmaier
***********************************************************************
        IMPLICIT REAL*8 (A-Z)
        COMPLEX*16 m02,m12,m22,m32,m42,m52,xdet
        COMPLEX*16 D0,D1(3),D2(0:3,0:3),D3(0:3,0:3,3)
        COMPLEX*16 E0,  E1(4),  E2(0:4,0:4),  E3(0:4,0:4,4)
	COMPLEX*16 E4(0:4,0:4,0:4,0:4)
        COMPLEX*16 E0_1,E1_1(4),E2_1(0:4,0:4),E3_1(0:4,0:4,4)
	COMPLEX*16 E4_1(0:4,0:4,0:4,0:4)
        COMPLEX*16 E0_2,E1_2(4),E2_2(0:4,0:4),E3_2(0:4,0:4,4)
	COMPLEX*16 E4_2(0:4,0:4,0:4,0:4)
        COMPLEX*16 E0_3,E1_3(4),E2_3(0:4,0:4),E3_3(0:4,0:4,4)
	COMPLEX*16 E4_3(0:4,0:4,0:4,0:4)
        COMPLEX*16 E0_4,E1_4(4),E2_4(0:4,0:4),E3_4(0:4,0:4,4)
	COMPLEX*16 E4_4(0:4,0:4,0:4,0:4)
        COMPLEX*16 E0_5,E1_5(4),E2_5(0:4,0:4),E3_5(0:4,0:4,4)
	COMPLEX*16 E4_5(0:4,0:4,0:4,0:4)
        COMPLEX*16 Em_sum,Emn_sum,Emn_sum1(4)
	COMPLEX*16 E00m_sum,Emnl_sum,Emnl_sum2(4),Emnl_sum1(4,4)
        COMPLEX*16 F0,F1(5),F2(0:5,0:5),F3(0:5,0:5,5)
	complex*16 y(0:5,0:5),dety(0:5),yy(6,6),detyy
	complex*16 y0(6,6),y1(6,6),y2(6,6),y3(6,6),y4(6,6),y5(6,6)
	INTEGER j,k,l,m,n,rank,ext

	call xE01234_new(p2,p3,p4,p5,s16,s23,s34,s45,s234,s345, m12,m22,
     *	     m32,m42,m52,E0,E1,E2,E3,E4,D0,D1,D2,D3,rank,ext)
 	call xE01234_new(s12,p3,p4,p5,p6,s123,s34,s45,s56,s345, m02,m22,
     *	     m32,m42,m52,E0_1,E1_1,E2_1,E3_1,E4_1,D0,D1,D2,D3,rank,ext)
	call xE01234_new(p1,s23,p4,p5,p6,s123,s234,s45,s56,s16, m02,m12,
     *	     m32,m42,m52,E0_2,E1_2,E2_2,E3_2,E4_2,D0,D1,D2,D3,rank,ext)
	call xE01234_new(p1,p2,s34,p5,p6,s12,s234,s345,s56,s16, m02,m12,
     *	     m22,m42,m52,E0_3,E1_3,E2_3,E3_3,E4_3,D0,D1,D2,D3,rank,ext)
	call xE01234_new(p1,p2,p3,s45,p6,s12,s23,s345,s123,s16, m02,m12,
     *	     m22,m32,m52,E0_4,E1_4,E2_4,E3_4,E4_4,D0,D1,D2,D3,rank,ext)
	call xE01234_new(p1,p2,p3,p4,s56,s12,s23,s34,s123,s234, m02,m12,
     *	     m22,m32,m42,E0_5,E1_5,E2_5,E3_5,E4_5,D0,D1,D2,D3,rank,ext)

	y(0,0) = 2d0*m02
	y(1,1) = 2d0*m12
	y(2,2) = 2d0*m22
	y(3,3) = 2d0*m32
	y(4,4) = 2d0*m42
	y(5,5) = 2d0*m52
	y(0,1) = m02+m12-p1
	y(1,2) = m12+m22-p2
	y(2,3) = m22+m32-p3
	y(3,4) = m32+m42-p4
	y(4,5) = m42+m52-p5
	y(0,5) = m52+m02-p6
	y(0,2) = m02+m22-s12
	y(1,3) = m12+m32-s23
	y(2,4) = m22+m42-s34
	y(3,5) = m32+m52-s45
	y(0,4) = m42+m02-s56
	y(1,5) = m52+m12-s16
	y(0,3) = m02+m32-s123
	y(1,4) = m12+m42-s234
	y(2,5) = m22+m52-s345

	do j=0,4
	  do k=j+1,5
	    y(k,j) = y(j,k)
	  enddo
	enddo

	do j=1,6
	  do k=1,6
	    yy(j,k) = y(j-1,k-1)
	    if (k.ne.1) then
	      y0(j,k) = y(j-1,k-1)
	    else
	      y0(j,k) = 1d0
	    endif
	    if (k.ne.2) then
	      y1(j,k) = y(j-1,k-1)
	    else
	      y1(j,k) = 1d0
	    endif
	    if (k.ne.3) then
	      y2(j,k) = y(j-1,k-1)
	    else
	      y2(j,k) = 1d0
	    endif
	    if (k.ne.4) then
	      y3(j,k) = y(j-1,k-1)
	    else
	      y3(j,k) = 1d0
	    endif
	    if (k.ne.5) then
	      y4(j,k) = y(j-1,k-1)
	    else
	      y4(j,k) = 1d0
	    endif
	    if (k.ne.6) then
	      y5(j,k) = y(j-1,k-1)
	    else
	      y5(j,k) = 1d0
	    endif
	  enddo
	enddo

	detyy   =  xdet(yy,6)
	dety(0) = -xdet(y0,6)/detyy
	dety(1) = -xdet(y1,6)/detyy
	dety(2) = -xdet(y2,6)/detyy
	dety(3) = -xdet(y3,6)/detyy
	dety(4) = -xdet(y4,6)/detyy
	dety(5) = -xdet(y5,6)/detyy

	F0 =  E0*dety(0)  +E0_1*dety(1)+E0_2*dety(2)
     &	     +E0_3*dety(3)+E0_4*dety(4)+E0_5*dety(5)

	if (rank.eq.0) return

	Em_sum = E1(1)+E1(2)+E1(3)+E1(4) + E0

	F1(1)= -Em_sum*dety(0)                 +E1_2(1)*dety(2)
     &	       +E1_3(1)*dety(3)+E1_4(1)*dety(4)+E1_5(1)*dety(5)
	F1(2)=  E1(1)*dety(0)  +E1_1(1)*dety(1)
     &	       +E1_3(2)*dety(3)+E1_4(2)*dety(4)+E1_5(2)*dety(5)
	F1(3)=  E1(2)*dety(0)  +E1_1(2)*dety(1)+E1_2(2)*dety(2)
     &	                       +E1_4(3)*dety(4)+E1_5(3)*dety(5)
	F1(4)=  E1(3)*dety(0)  +E1_1(3)*dety(1)+E1_2(3)*dety(2)
     &	       +E1_3(3)*dety(3)                +E1_5(4)*dety(5)
	F1(5)=  E1(4)*dety(0)  +E1_1(4)*dety(1)+E1_2(4)*dety(2)
     &	       +E1_3(4)*dety(3)+E1_4(4)*dety(4)

	if (rank.eq.1) return

	F2(0,0) = E2(0,0)*dety(0)+E2_1(0,0)*dety(1)+E2_2(0,0)*dety(2)
     &	       +E2_3(0,0)*dety(3)+E2_4(0,0)*dety(4)+E2_5(0,0)*dety(5)

	Emn_sum = Em_sum
	do j=1,4
	  Emn_sum1(j) = E1(j)
	  do k=1,4
	    Emn_sum1(j) = Emn_sum1(j) + E2(j,k)
	  enddo
	  Emn_sum = Emn_sum + Emn_sum1(j)
	enddo

	F2(1,1) = Emn_sum*dety(0)                  +E2_2(1,1)*dety(2)
     &	       +E2_3(1,1)*dety(3)+E2_4(1,1)*dety(4)+E2_5(1,1)*dety(5)
	F2(1,2) = -Emn_sum1(1)*dety(0)
     &	       +E2_3(1,2)*dety(3)+E2_4(1,2)*dety(4)+E2_5(1,2)*dety(5)
	F2(1,3) = -Emn_sum1(2)*dety(0)             +E2_2(1,2)*dety(2)
     &	                         +E2_4(1,3)*dety(4)+E2_5(1,3)*dety(5)
	F2(1,4) = -Emn_sum1(3)*dety(0)             +E2_2(1,3)*dety(2)
     &	       +E2_3(1,3)*dety(3)                  +E2_5(1,4)*dety(5)
	F2(1,5) = -Emn_sum1(4)*dety(0)             +E2_2(1,4)*dety(2)
     &	       +E2_3(1,4)*dety(3)+E2_4(1,4)*dety(4)
	F2(2,2) = E2(1,1)*dety(0)+E2_1(1,1)*dety(1)
     &	       +E2_3(2,2)*dety(3)+E2_4(2,2)*dety(4)+E2_5(2,2)*dety(5)
	F2(2,3) = E2(1,2)*dety(0)+E2_1(1,2)*dety(1)
     &	                         +E2_4(2,3)*dety(4)+E2_5(2,3)*dety(5)
	F2(2,4) = E2(1,3)*dety(0)+E2_1(1,3)*dety(1)
     &	       +E2_3(2,3)*dety(3)                  +E2_5(2,4)*dety(5)
	F2(2,5) = E2(1,4)*dety(0)+E2_1(1,4)*dety(1)
     &	       +E2_3(2,4)*dety(3)+E2_4(2,4)*dety(4)
	F2(3,3) = E2(2,2)*dety(0)+E2_1(2,2)*dety(1)+E2_2(2,2)*dety(2)
     &	                         +E2_4(3,3)*dety(4)+E2_5(3,3)*dety(5)
	F2(3,4) = E2(2,3)*dety(0)+E2_1(2,3)*dety(1)+E2_2(2,3)*dety(2)
     &	                                           +E2_5(3,4)*dety(5)
	F2(3,5) = E2(2,4)*dety(0)+E2_1(2,4)*dety(1)+E2_2(2,4)*dety(2)
     &	                         +E2_4(3,4)*dety(4)
	F2(4,4) = E2(3,3)*dety(0)+E2_1(3,3)*dety(1)+E2_2(3,3)*dety(2)
     &	       +E2_3(3,3)*dety(3)                  +E2_5(4,4)*dety(5)
	F2(4,5) = E2(3,4)*dety(0)+E2_1(3,4)*dety(1)+E2_2(3,4)*dety(2)
     &	       +E2_3(3,4)*dety(3)
	F2(5,5) = E2(4,4)*dety(0)+E2_1(4,4)*dety(1)+E2_2(4,4)*dety(2)
     &	       +E2_3(4,4)*dety(3)+E2_4(4,4)*dety(4)

	do j=1,4
	do k=j+1,5
	  F2(k,j) = F2(j,k)
	enddo
	enddo

	if (rank.eq.2) return

	E00m_sum = E2(0,0)
	Emnl_sum = Emn_sum
	do j=1,4
	  Emnl_sum2(j) = Emn_sum1(j)
	  do k=1,4
            Emnl_sum1(j,k) = E2(j,k)
	    do l=1,4
              Emnl_sum1(j,k) = Emnl_sum1(j,k) + E3(j,k,l)
	    enddo
            Emnl_sum2(j)   = Emnl_sum2(j) + Emnl_sum1(j,k)
	  enddo
	  E00m_sum = E00m_sum + E3(0,0,j)
	  Emnl_sum = Emnl_sum + Emnl_sum2(j)
	enddo

	F3(0,0,1)= -E00m_sum*dety(0)   +E3_2(0,0,1)*dety(2)
     &	    +E3_3(0,0,1)*dety(3)+E3_4(0,0,1)*dety(4)+E3_5(0,0,1)*dety(5)
	F3(0,0,2)=  E3(0,0,1)*dety(0)  +E3_1(0,0,1)*dety(1)
     &	    +E3_3(0,0,2)*dety(3)+E3_4(0,0,2)*dety(4)+E3_5(0,0,2)*dety(5)
	F3(0,0,3)=  E3(0,0,2)*dety(0)  +E3_1(0,0,2)*dety(1)
     &	    +E3_2(0,0,2)*dety(2)+E3_4(0,0,3)*dety(4)+E3_5(0,0,3)*dety(5)
	F3(0,0,4)=  E3(0,0,3)*dety(0)  +E3_1(0,0,3)*dety(1)
     &	    +E3_2(0,0,3)*dety(2)+E3_3(0,0,3)*dety(3)+E3_5(0,0,4)*dety(5)
	F3(0,0,5)=  E3(0,0,4)*dety(0)  +E3_1(0,0,4)*dety(1)
     &	    +E3_2(0,0,4)*dety(2)+E3_3(0,0,4)*dety(3)+E3_4(0,0,4)*dety(4)

	F3(1,1,1) = -Emnl_sum*dety(0)
     &	            +E3_2(1,1,1)*dety(2)+E3_3(1,1,1)*dety(3)
     &	            +E3_4(1,1,1)*dety(4)+E3_5(1,1,1)*dety(5)
	F3(1,1,2) =  Emnl_sum2(1)*dety(0)
     &	                                +E3_3(1,1,2)*dety(3)
     &	            +E3_4(1,1,2)*dety(4)+E3_5(1,1,2)*dety(5)
	F3(1,1,3) =  Emnl_sum2(2)*dety(0)
     &	            +E3_2(1,1,2)*dety(2)
     &	            +E3_4(1,1,3)*dety(4)+E3_5(1,1,3)*dety(5)
	F3(1,1,4) =  Emnl_sum2(3)*dety(0)
     &	            +E3_2(1,1,3)*dety(2)+E3_3(1,1,3)*dety(3)
     &	                                +E3_5(1,1,4)*dety(5)
	F3(1,1,5) =  Emnl_sum2(4)*dety(0)
     &	            +E3_2(1,1,4)*dety(2)+E3_3(1,1,4)*dety(3)
     &	            +E3_4(1,1,4)*dety(4)

	F3(1,2,2) = -Emnl_sum1(1,1)*dety(0)
     &	                                +E3_3(1,2,2)*dety(3)
     &	            +E3_4(1,2,2)*dety(4)+E3_5(1,2,2)*dety(5)
	F3(1,2,3) = -Emnl_sum1(1,2)*dety(0)
     &	            +E3_4(1,2,3)*dety(4)+E3_5(1,2,3)*dety(5)
	F3(1,2,4) = -Emnl_sum1(1,3)*dety(0)
     &	            +E3_3(1,2,3)*dety(3)+E3_5(1,2,4)*dety(5)
	F3(1,2,5) = -Emnl_sum1(1,4)*dety(0)
     &	            +E3_3(1,2,4)*dety(3)+E3_4(1,2,4)*dety(4)
	F3(1,3,3) = -Emnl_sum1(2,2)*dety(0)
     &	                                +E3_2(1,2,2)*dety(2)
     &	            +E3_4(1,3,3)*dety(4)+E3_5(1,3,3)*dety(5)
	F3(1,3,4) = -Emnl_sum1(2,3)*dety(0)
     &	            +E3_2(1,2,3)*dety(2)+E3_5(1,3,4)*dety(5)
	F3(1,3,5) = -Emnl_sum1(2,4)*dety(0)
     &	            +E3_2(1,2,4)*dety(2)+E3_4(1,3,4)*dety(4)
	F3(1,4,4) = -Emnl_sum1(3,3)*dety(0)
     &	                                +E3_2(1,3,3)*dety(2)
     &	            +E3_3(1,3,3)*dety(3)+E3_5(1,4,4)*dety(5)
	F3(1,4,5) = -Emnl_sum1(3,4)*dety(0)
     &	            +E3_2(1,3,4)*dety(2)+E3_3(1,3,4)*dety(3)
	F3(1,5,5) = -Emnl_sum1(4,4)*dety(0)
     &	                                +E3_2(1,4,4)*dety(2)
     &	            +E3_3(1,4,4)*dety(3)+E3_4(1,4,4)*dety(4)

	F3(2,2,2) =  E3(1,1,1)*dety(0)  +E3_1(1,1,1)*dety(1)
     &	                                +E3_3(2,2,2)*dety(3)
     &	            +E3_4(2,2,2)*dety(4)+E3_5(2,2,2)*dety(5)
	F3(2,2,3) =  E3(1,1,2)*dety(0)  +E3_1(1,1,2)*dety(1)
     &	            +E3_4(2,2,3)*dety(4)+E3_5(2,2,3)*dety(5)
	F3(2,2,4) =  E3(1,1,3)*dety(0)  +E3_1(1,1,3)*dety(1)
     &	            +E3_3(2,2,3)*dety(3)+E3_5(2,2,4)*dety(5)
	F3(2,2,5) =  E3(1,1,4)*dety(0)  +E3_1(1,1,4)*dety(1)
     &	            +E3_3(2,2,4)*dety(3)+E3_4(2,2,4)*dety(4)
	F3(2,3,3) =  E3(1,2,2)*dety(0)  +E3_1(1,2,2)*dety(1)
     &	            +E3_4(2,3,3)*dety(4)+E3_5(2,3,3)*dety(5)
	F3(2,3,4) =  E3(1,2,3)*dety(0)  +E3_1(1,2,3)*dety(1)
     &	                                +E3_5(2,3,4)*dety(5)
	F3(2,3,5) =  E3(1,2,4)*dety(0)  +E3_1(1,2,4)*dety(1)
     &	            +E3_4(2,3,4)*dety(4)
	F3(2,4,4) =  E3(1,3,3)*dety(0)  +E3_1(1,3,3)*dety(1)
     &	            +E3_3(2,3,3)*dety(3)+E3_5(2,4,4)*dety(5)
	F3(2,4,5) =  E3(1,3,4)*dety(0)  +E3_1(1,3,4)*dety(1)
     &	            +E3_3(2,3,4)*dety(3)
	F3(2,5,5) =  E3(1,4,4)*dety(0)  +E3_1(1,4,4)*dety(1)
     &	            +E3_3(2,4,4)*dety(3)+E3_4(2,4,4)*dety(4)
	F3(3,3,3) =  E3(2,2,2)*dety(0)  +E3_1(2,2,2)*dety(1)
     &	            +E3_2(2,2,2)*dety(2)
     &	            +E3_4(3,3,3)*dety(4)+E3_5(3,3,3)*dety(5)
	F3(3,3,4) =  E3(2,2,3)*dety(0)  +E3_1(2,2,3)*dety(1)
     &	            +E3_2(2,2,3)*dety(2)+E3_5(3,3,4)*dety(5)
	F3(3,3,5) =  E3(2,2,4)*dety(0)  +E3_1(2,2,4)*dety(1)
     &	            +E3_2(2,2,4)*dety(2)+E3_4(3,3,4)*dety(4)
	F3(3,4,4) =  E3(2,3,3)*dety(0)  +E3_1(2,3,3)*dety(1)
     &	            +E3_2(2,3,3)*dety(2)+E3_5(3,4,4)*dety(5)
	F3(3,4,5) =  E3(2,3,4)*dety(0)  +E3_1(2,3,4)*dety(1)
     &	            +E3_2(2,3,4)*dety(2)
	F3(3,5,5) =  E3(2,4,4)*dety(0)  +E3_1(2,4,4)*dety(1)
     &	            +E3_2(2,4,4)*dety(2)+E3_4(3,4,4)*dety(4)
	F3(4,4,4) =  E3(3,3,3)*dety(0)  +E3_1(3,3,3)*dety(1)
     &	            +E3_2(3,3,3)*dety(2)
     &	            +E3_3(3,3,3)*dety(3)+E3_5(4,4,4)*dety(5)
	F3(4,4,5) =  E3(3,3,4)*dety(0)  +E3_1(3,3,4)*dety(1)
     &	            +E3_2(3,3,4)*dety(2)+E3_3(3,3,4)*dety(3)
	F3(4,5,5) =  E3(3,4,4)*dety(0)  +E3_1(3,4,4)*dety(1)
     &	            +E3_2(3,4,4)*dety(2)+E3_3(3,4,4)*dety(3)
	F3(5,5,5) =  E3(4,4,4)*dety(0)  +E3_1(4,4,4)*dety(1)
     &	            +E3_2(4,4,4)*dety(2)
     &	            +E3_3(4,4,4)*dety(3)+E3_4(4,4,4)*dety(4)

	do j=1,4
	do k=j+1,5
	  F3(k,j,k) = F3(j,k,k)
	  F3(k,k,j) = F3(j,k,k)
          F3(k,j,j) = F3(j,j,k)
          F3(j,k,j) = F3(j,j,k)
	enddo
	enddo

	do j=1,3
	do k=j+1,4
	do l=k+1,5
	  F3(k,l,j) = F3(j,k,l)
	  F3(l,j,k) = F3(j,k,l)
	  F3(l,k,j) = F3(j,k,l)
	  F3(k,j,l) = F3(j,k,l)
	  F3(j,l,k) = F3(j,k,l)
	enddo
	enddo
	enddo

	end

************************************************************************
        FUNCTION xE0_(p1,p2,p3,p4,p5,s12,s23,s34,s45,s15,
     &		     m02,m12,m22,m32,m42,ext)
************************************************************************
*  	scalar 5-pt function with complex masses
*  	pi = p_i^2 = squared external momenta,  s_ij = (p_i+p_j)^2
*-----------------------------------------------------------------------
*	Reduction to five 4-pt functions as described in:
*	  A.Denner, Fortschr. Phys. 41 (1993) 307,
*	  A.Denner, S.Dittmaier and M.Roth, Nucl. Phys. B519 (1998) 39
*-----------------------------------------------------------------------
*  	11.6.04  Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 m02,m12,m22,m32,m42
	complex*16 xe0_,xd0_,d0(0:4),y(0:4,0:4),dety(0:4)
	integer i,ext

	d0(0) = xD0_( p2, p3, p4,s15,s23,s34,m12,m22,m32,m42,ext)
	d0(1) = xD0_(s12, p3, p4, p5,s45,s34,m02,m22,m32,m42,ext)
	d0(2) = xD0_( p1,s23, p4, p5,s45,s15,m02,m12,m32,m42,ext)
	d0(3) = xD0_( p1, p2,s34, p5,s12,s15,m02,m12,m22,m42,ext)
	d0(4) = xD0_( p1, p2, p3,s45,s12,s23,m02,m12,m22,m32,ext)

	y(0,0) = 2d0*m02
	y(1,1) = 2d0*m12
	y(2,2) = 2d0*m22
	y(3,3) = 2d0*m32
	y(4,4) = 2d0*m42
	y(0,1) = m02+m12-p1
	y(1,2) = m12+m22-p2
	y(2,3) = m22+m32-p3
	y(3,4) = m32+m42-p4
	y(0,4) = m42+m02-p5
	y(0,2) = m02+m22-s12
	y(1,3) = m12+m32-s23
	y(2,4) = m22+m42-s34
	y(0,3) = m32+m02-s45
	y(1,4) = m42+m12-s15

        DETYY = Y(0,4)**2*Y(1,3)**2*Y(2,2)-2D0*Y(0,3)*Y(0,4)*Y(1,3)*
     &  Y(1,4)*Y(2,2)+Y(0,3)**2*Y(1,4)**2*Y(2,2)-2D0*Y(0,4)**2*Y(1,
     &  2)*Y(1,3)*Y(2,3)+2D0*Y(0,3)*Y(0,4)*Y(1,2)*Y(1,4)*Y(2,3)+2D0*
     &  Y(0,2)*Y(0,4)*Y(1,3)*Y(1,4)*Y(2,3)-2D0*Y(0,2)*Y(0,3)*Y(1,
     &  4)**2*Y(2,3)+Y(0,4)**2*Y(1,1)*Y(2,3)**2-2D0*Y(0,1)*Y(0,4)*
     &  Y(1,4)*Y(2,3)**2+Y(0,0)*Y(1,4)**2*Y(2,3)**2+2D0*Y(0,3)*Y(0,
     &  4)*Y(1,2)*Y(1,3)*Y(2,4)-2D0*Y(0,2)*Y(0,4)*Y(1,3)**2*Y(2,4)-
     &  2D0*Y(0,3)**2*Y(1,2)*Y(1,4)*Y(2,4)+2D0*Y(0,2)*Y(0,3)*Y(1,3)*
     &  Y(1,4)*Y(2,4)-2D0*Y(0,3)*Y(0,4)*Y(1,1)*Y(2,3)*Y(2,4)+2D0*
     &  Y(0,1)*Y(0,4)*Y(1,3)*Y(2,3)*Y(2,4)+2D0*Y(0,1)*Y(0,3)*Y(1,4)*
     &  Y(2,3)*Y(2,4)-2D0*Y(0,0)*Y(1,3)*Y(1,4)*Y(2,3)*Y(2,4)+Y(0,
     &  3)**2*Y(1,1)*Y(2,4)**2-2D0*Y(0,1)*Y(0,3)*Y(1,3)*Y(2,4)**2+
     &  Y(0,0)*Y(1,3)**2*Y(2,4)**2+Y(0,4)**2*Y(1,2)**2*Y(3,3)-2D0*
     &  Y(0,2)*Y(0,4)*Y(1,2)*Y(1,4)*Y(3,3)+Y(0,2)**2*Y(1,4)**2*Y(3,
     &  3)-Y(0,4)**2*Y(1,1)*Y(2,2)*Y(3,3)+2D0*Y(0,1)*Y(0,4)*Y(1,4)*
     &  Y(2,2)*Y(3,3)-Y(0,0)*Y(1,4)**2*Y(2,2)*Y(3,3)+2D0*Y(0,2)*Y(0,
     &  4)*Y(1,1)*Y(2,4)*Y(3,3)-2D0*Y(0,1)*Y(0,4)*Y(1,2)*Y(2,4)*Y(3,
     &  3)-2D0*Y(0,1)*Y(0,2)*Y(1,4)*Y(2,4)*Y(3,3)+2D0*Y(0,0)*Y(1,2)*
     &  Y(1,4)*Y(2,4)*Y(3,3)+Y(0,1)**2*Y(2,4)**2*Y(3,3)-Y(0,0)*Y(1,
     &  1)*Y(2,4)**2*Y(3,3)-2D0*Y(0,3)*Y(0,4)*Y(1,2)**2*Y(3,4)+2D0*
     &  Y(0,2)*Y(0,4)*Y(1,2)*Y(1,3)*Y(3,4)+2D0*Y(0,2)*Y(0,3)*Y(1,2)*
     &  Y(1,4)*Y(3,4)-2D0*Y(0,2)**2*Y(1,3)*Y(1,4)*Y(3,4)+2D0*Y(0,3)*
     &  Y(0,4)*Y(1,1)*Y(2,2)*Y(3,4)-2D0*Y(0,1)*Y(0,4)*Y(1,3)*Y(2,2)*
     &  Y(3,4)-2D0*Y(0,1)*Y(0,3)*Y(1,4)*Y(2,2)*Y(3,4)+2D0*Y(0,0)*
     &  Y(1,3)*Y(1,4)*Y(2,2)*Y(3,4)-2D0*Y(0,2)*Y(0,4)*Y(1,1)*Y(2,3)*
     &  Y(3,4)+2D0*Y(0,1)*Y(0,4)*Y(1,2)*Y(2,3)*Y(3,4)+2D0*Y(0,1)*
     &  Y(0,2)*Y(1,4)*Y(2,3)*Y(3,4)-2D0*Y(0,0)*Y(1,2)*Y(1,4)*Y(2,3)*
     &  Y(3,4)-2D0*Y(0,2)*Y(0,3)*Y(1,1)*Y(2,4)*Y(3,4)+2D0*Y(0,1)*
     &  Y(0,3)*Y(1,2)*Y(2,4)*Y(3,4)+2D0*Y(0,1)*Y(0,2)*Y(1,3)*Y(2,4)*
     &  Y(3,4)-2D0*Y(0,0)*Y(1,2)*Y(1,3)*Y(2,4)*Y(3,4)-2D0*Y(0,1)**2*
     &  Y(2,3)*Y(2,4)*Y(3,4)+2D0*Y(0,0)*Y(1,1)*Y(2,3)*Y(2,4)*Y(3,4)+
     &  Y(0,2)**2*Y(1,1)*Y(3,4)**2-2D0*Y(0,1)*Y(0,2)*Y(1,2)*Y(3,4)**
     &  2+Y(0,0)*Y(1,2)**2*Y(3,4)**2+Y(0,1)**2*Y(2,2)*Y(3,4)**2-Y(0,
     &  0)*Y(1,1)*Y(2,2)*Y(3,4)**2+Y(0,3)**2*Y(1,2)**2*Y(4,4)-2D0*
     &  Y(0,2)*Y(0,3)*Y(1,2)*Y(1,3)*Y(4,4)+Y(0,2)**2*Y(1,3)**2*Y(4,
     &  4)-Y(0,3)**2*Y(1,1)*Y(2,2)*Y(4,4)+2D0*Y(0,1)*Y(0,3)*Y(1,3)*
     &  Y(2,2)*Y(4,4)-Y(0,0)*Y(1,3)**2*Y(2,2)*Y(4,4)+2D0*Y(0,2)*Y(0,
     &  3)*Y(1,1)*Y(2,3)*Y(4,4)-2D0*Y(0,1)*Y(0,3)*Y(1,2)*Y(2,3)*Y(4,
     &  4)-2D0*Y(0,1)*Y(0,2)*Y(1,3)*Y(2,3)*Y(4,4)+2D0*Y(0,0)*Y(1,2)*
     &  Y(1,3)*Y(2,3)*Y(4,4)+Y(0,1)**2*Y(2,3)**2*Y(4,4)-Y(0,0)*Y(1,
     &  1)*Y(2,3)**2*Y(4,4)-Y(0,2)**2*Y(1,1)*Y(3,3)*Y(4,4)+2D0*Y(0,
     &  1)*Y(0,2)*Y(1,2)*Y(3,3)*Y(4,4)-Y(0,0)*Y(1,2)**2*Y(3,3)*Y(4,
     &  4)-Y(0,1)**2*Y(2,2)*Y(3,3)*Y(4,4)+Y(0,0)*Y(1,1)*Y(2,2)*Y(3,
     &  3)*Y(4,4)
        DETY(0) = -(Y(0,4)*Y(1,3)**2*Y(2,2))+Y(0,3)*Y(1,3)*Y(1,4)*
     &  Y(2,2)+Y(0,4)*Y(1,3)*Y(1,4)*Y(2,2)-Y(0,3)*Y(1,4)**2*Y(2,2)+
     &  2D0*Y(0,4)*Y(1,2)*Y(1,3)*Y(2,3)-Y(0,3)*Y(1,2)*Y(1,4)*Y(2,3)-
     &  Y(0,4)*Y(1,2)*Y(1,4)*Y(2,3)-Y(0,2)*Y(1,3)*Y(1,4)*Y(2,3)-Y(0,
     &  4)*Y(1,3)*Y(1,4)*Y(2,3)+Y(0,2)*Y(1,4)**2*Y(2,3)+Y(0,3)*Y(1,
     &  4)**2*Y(2,3)-Y(0,4)*Y(1,1)*Y(2,3)**2+Y(0,1)*Y(1,4)*Y(2,3)**
     &  2+Y(0,4)*Y(1,4)*Y(2,3)**2-Y(1,4)**2*Y(2,3)**2-Y(0,3)*Y(1,2)*
     &  Y(1,3)*Y(2,4)-Y(0,4)*Y(1,2)*Y(1,3)*Y(2,4)+Y(0,2)*Y(1,3)**2*
     &  Y(2,4)+Y(0,4)*Y(1,3)**2*Y(2,4)+2D0*Y(0,3)*Y(1,2)*Y(1,4)*Y(2,
     &  4)-Y(0,2)*Y(1,3)*Y(1,4)*Y(2,4)-Y(0,3)*Y(1,3)*Y(1,4)*Y(2,4)+
     &  Y(0,3)*Y(1,1)*Y(2,3)*Y(2,4)+Y(0,4)*Y(1,1)*Y(2,3)*Y(2,4)-Y(0,
     &  1)*Y(1,3)*Y(2,3)*Y(2,4)-Y(0,4)*Y(1,3)*Y(2,3)*Y(2,4)-Y(0,1)*
     &  Y(1,4)*Y(2,3)*Y(2,4)-Y(0,3)*Y(1,4)*Y(2,3)*Y(2,4)+2D0*Y(1,3)*
     &  Y(1,4)*Y(2,3)*Y(2,4)-Y(0,3)*Y(1,1)*Y(2,4)**2+Y(0,1)*Y(1,3)*
     &  Y(2,4)**2+Y(0,3)*Y(1,3)*Y(2,4)**2-Y(1,3)**2*Y(2,4)**2-Y(0,
     &  4)*Y(1,2)**2*Y(3,3)+Y(0,2)*Y(1,2)*Y(1,4)*Y(3,3)+Y(0,4)*Y(1,
     &  2)*Y(1,4)*Y(3,3)-Y(0,2)*Y(1,4)**2*Y(3,3)+Y(0,4)*Y(1,1)*Y(2,
     &  2)*Y(3,3)-Y(0,1)*Y(1,4)*Y(2,2)*Y(3,3)-Y(0,4)*Y(1,4)*Y(2,2)*
     &  Y(3,3)+Y(1,4)**2*Y(2,2)*Y(3,3)-Y(0,2)*Y(1,1)*Y(2,4)*Y(3,3)-
     &  Y(0,4)*Y(1,1)*Y(2,4)*Y(3,3)+Y(0,1)*Y(1,2)*Y(2,4)*Y(3,3)+Y(0,
     &  4)*Y(1,2)*Y(2,4)*Y(3,3)+Y(0,1)*Y(1,4)*Y(2,4)*Y(3,3)+Y(0,2)*
     &  Y(1,4)*Y(2,4)*Y(3,3)-2D0*Y(1,2)*Y(1,4)*Y(2,4)*Y(3,3)-Y(0,1)*
     &  Y(2,4)**2*Y(3,3)+Y(1,1)*Y(2,4)**2*Y(3,3)+Y(0,3)*Y(1,2)**2*
     &  Y(3,4)+Y(0,4)*Y(1,2)**2*Y(3,4)-Y(0,2)*Y(1,2)*Y(1,3)*Y(3,4)-
     &  Y(0,4)*Y(1,2)*Y(1,3)*Y(3,4)-Y(0,2)*Y(1,2)*Y(1,4)*Y(3,4)-Y(0,
     &  3)*Y(1,2)*Y(1,4)*Y(3,4)+2D0*Y(0,2)*Y(1,3)*Y(1,4)*Y(3,4)-Y(0,
     &  3)*Y(1,1)*Y(2,2)*Y(3,4)-Y(0,4)*Y(1,1)*Y(2,2)*Y(3,4)+Y(0,1)*
     &  Y(1,3)*Y(2,2)*Y(3,4)+Y(0,4)*Y(1,3)*Y(2,2)*Y(3,4)+Y(0,1)*Y(1,
     &  4)*Y(2,2)*Y(3,4)+Y(0,3)*Y(1,4)*Y(2,2)*Y(3,4)-2D0*Y(1,3)*Y(1,
     &  4)*Y(2,2)*Y(3,4)+Y(0,2)*Y(1,1)*Y(2,3)*Y(3,4)+Y(0,4)*Y(1,1)*
     &  Y(2,3)*Y(3,4)-Y(0,1)*Y(1,2)*Y(2,3)*Y(3,4)-Y(0,4)*Y(1,2)*Y(2,
     &  3)*Y(3,4)-Y(0,1)*Y(1,4)*Y(2,3)*Y(3,4)-Y(0,2)*Y(1,4)*Y(2,3)*
     &  Y(3,4)+2D0*Y(1,2)*Y(1,4)*Y(2,3)*Y(3,4)+Y(0,2)*Y(1,1)*Y(2,4)*
     &  Y(3,4)+Y(0,3)*Y(1,1)*Y(2,4)*Y(3,4)-Y(0,1)*Y(1,2)*Y(2,4)*Y(3,
     &  4)-Y(0,3)*Y(1,2)*Y(2,4)*Y(3,4)-Y(0,1)*Y(1,3)*Y(2,4)*Y(3,4)-
     &  Y(0,2)*Y(1,3)*Y(2,4)*Y(3,4)+2D0*Y(1,2)*Y(1,3)*Y(2,4)*Y(3,4)+
     &  2D0*Y(0,1)*Y(2,3)*Y(2,4)*Y(3,4)-2D0*Y(1,1)*Y(2,3)*Y(2,4)*
     &  Y(3,4)-Y(0,2)*Y(1,1)*Y(3,4)**2+Y(0,1)*Y(1,2)*Y(3,4)**2+Y(0,
     &  2)*Y(1,2)*Y(3,4)**2-Y(1,2)**2*Y(3,4)**2-Y(0,1)*Y(2,2)*Y(3,
     &  4)**2+Y(1,1)*Y(2,2)*Y(3,4)**2-Y(0,3)*Y(1,2)**2*Y(4,4)+Y(0,
     &  2)*Y(1,2)*Y(1,3)*Y(4,4)+Y(0,3)*Y(1,2)*Y(1,3)*Y(4,4)-Y(0,2)*
     &  Y(1,3)**2*Y(4,4)+Y(0,3)*Y(1,1)*Y(2,2)*Y(4,4)-Y(0,1)*Y(1,3)*
     &  Y(2,2)*Y(4,4)-Y(0,3)*Y(1,3)*Y(2,2)*Y(4,4)+Y(1,3)**2*Y(2,2)*
     &  Y(4,4)-Y(0,2)*Y(1,1)*Y(2,3)*Y(4,4)-Y(0,3)*Y(1,1)*Y(2,3)*Y(4,
     &  4)+Y(0,1)*Y(1,2)*Y(2,3)*Y(4,4)+Y(0,3)*Y(1,2)*Y(2,3)*Y(4,4)+
     &  Y(0,1)*Y(1,3)*Y(2,3)*Y(4,4)+Y(0,2)*Y(1,3)*Y(2,3)*Y(4,4)-2D0*
     &  Y(1,2)*Y(1,3)*Y(2,3)*Y(4,4)-Y(0,1)*Y(2,3)**2*Y(4,4)+Y(1,1)*
     &  Y(2,3)**2*Y(4,4)+Y(0,2)*Y(1,1)*Y(3,3)*Y(4,4)-Y(0,1)*Y(1,2)*
     &  Y(3,3)*Y(4,4)-Y(0,2)*Y(1,2)*Y(3,3)*Y(4,4)+Y(1,2)**2*Y(3,3)*
     &  Y(4,4)+Y(0,1)*Y(2,2)*Y(3,3)*Y(4,4)-Y(1,1)*Y(2,2)*Y(3,3)*Y(4,
     &  4)
        DETY(1) = Y(0,3)*Y(0,4)*Y(1,3)*Y(2,2)-Y(0,4)**2*Y(1,3)*Y(2,
     &  2)-Y(0,3)**2*Y(1,4)*Y(2,2)+Y(0,3)*Y(0,4)*Y(1,4)*Y(2,2)-Y(0,
     &  3)*Y(0,4)*Y(1,2)*Y(2,3)+Y(0,4)**2*Y(1,2)*Y(2,3)-Y(0,2)*Y(0,
     &  4)*Y(1,3)*Y(2,3)+Y(0,4)**2*Y(1,3)*Y(2,3)+2D0*Y(0,2)*Y(0,3)*
     &  Y(1,4)*Y(2,3)-Y(0,2)*Y(0,4)*Y(1,4)*Y(2,3)-Y(0,3)*Y(0,4)*Y(1,
     &  4)*Y(2,3)+Y(0,1)*Y(0,4)*Y(2,3)**2-Y(0,4)**2*Y(2,3)**2-Y(0,
     &  0)*Y(1,4)*Y(2,3)**2+Y(0,4)*Y(1,4)*Y(2,3)**2+Y(0,3)**2*Y(1,
     &  2)*Y(2,4)-Y(0,3)*Y(0,4)*Y(1,2)*Y(2,4)-Y(0,2)*Y(0,3)*Y(1,3)*
     &  Y(2,4)+2D0*Y(0,2)*Y(0,4)*Y(1,3)*Y(2,4)-Y(0,3)*Y(0,4)*Y(1,3)*
     &  Y(2,4)-Y(0,2)*Y(0,3)*Y(1,4)*Y(2,4)+Y(0,3)**2*Y(1,4)*Y(2,4)-
     &  Y(0,1)*Y(0,3)*Y(2,3)*Y(2,4)-Y(0,1)*Y(0,4)*Y(2,3)*Y(2,4)+2D0*
     &  Y(0,3)*Y(0,4)*Y(2,3)*Y(2,4)+Y(0,0)*Y(1,3)*Y(2,3)*Y(2,4)-Y(0,
     &  4)*Y(1,3)*Y(2,3)*Y(2,4)+Y(0,0)*Y(1,4)*Y(2,3)*Y(2,4)-Y(0,3)*
     &  Y(1,4)*Y(2,3)*Y(2,4)+Y(0,1)*Y(0,3)*Y(2,4)**2-Y(0,3)**2*Y(2,
     &  4)**2-Y(0,0)*Y(1,3)*Y(2,4)**2+Y(0,3)*Y(1,3)*Y(2,4)**2+Y(0,
     &  2)*Y(0,4)*Y(1,2)*Y(3,3)-Y(0,4)**2*Y(1,2)*Y(3,3)-Y(0,2)**2*
     &  Y(1,4)*Y(3,3)+Y(0,2)*Y(0,4)*Y(1,4)*Y(3,3)-Y(0,1)*Y(0,4)*Y(2,
     &  2)*Y(3,3)+Y(0,4)**2*Y(2,2)*Y(3,3)+Y(0,0)*Y(1,4)*Y(2,2)*Y(3,
     &  3)-Y(0,4)*Y(1,4)*Y(2,2)*Y(3,3)+Y(0,1)*Y(0,2)*Y(2,4)*Y(3,3)+
     &  Y(0,1)*Y(0,4)*Y(2,4)*Y(3,3)-2D0*Y(0,2)*Y(0,4)*Y(2,4)*Y(3,3)-
     &  Y(0,0)*Y(1,2)*Y(2,4)*Y(3,3)+Y(0,4)*Y(1,2)*Y(2,4)*Y(3,3)-Y(0,
     &  0)*Y(1,4)*Y(2,4)*Y(3,3)+Y(0,2)*Y(1,4)*Y(2,4)*Y(3,3)+Y(0,0)*
     &  Y(2,4)**2*Y(3,3)-Y(0,1)*Y(2,4)**2*Y(3,3)-Y(0,2)*Y(0,3)*Y(1,
     &  2)*Y(3,4)-Y(0,2)*Y(0,4)*Y(1,2)*Y(3,4)+2D0*Y(0,3)*Y(0,4)*Y(1,
     &  2)*Y(3,4)+Y(0,2)**2*Y(1,3)*Y(3,4)-Y(0,2)*Y(0,4)*Y(1,3)*Y(3,
     &  4)+Y(0,2)**2*Y(1,4)*Y(3,4)-Y(0,2)*Y(0,3)*Y(1,4)*Y(3,4)+Y(0,
     &  1)*Y(0,3)*Y(2,2)*Y(3,4)+Y(0,1)*Y(0,4)*Y(2,2)*Y(3,4)-2D0*Y(0,
     &  3)*Y(0,4)*Y(2,2)*Y(3,4)-Y(0,0)*Y(1,3)*Y(2,2)*Y(3,4)+Y(0,4)*
     &  Y(1,3)*Y(2,2)*Y(3,4)-Y(0,0)*Y(1,4)*Y(2,2)*Y(3,4)+Y(0,3)*Y(1,
     &  4)*Y(2,2)*Y(3,4)-Y(0,1)*Y(0,2)*Y(2,3)*Y(3,4)-Y(0,1)*Y(0,4)*
     &  Y(2,3)*Y(3,4)+2D0*Y(0,2)*Y(0,4)*Y(2,3)*Y(3,4)+Y(0,0)*Y(1,2)*
     &  Y(2,3)*Y(3,4)-Y(0,4)*Y(1,2)*Y(2,3)*Y(3,4)+Y(0,0)*Y(1,4)*Y(2,
     &  3)*Y(3,4)-Y(0,2)*Y(1,4)*Y(2,3)*Y(3,4)-Y(0,1)*Y(0,2)*Y(2,4)*
     &  Y(3,4)-Y(0,1)*Y(0,3)*Y(2,4)*Y(3,4)+2D0*Y(0,2)*Y(0,3)*Y(2,4)*
     &  Y(3,4)+Y(0,0)*Y(1,2)*Y(2,4)*Y(3,4)-Y(0,3)*Y(1,2)*Y(2,4)*Y(3,
     &  4)+Y(0,0)*Y(1,3)*Y(2,4)*Y(3,4)-Y(0,2)*Y(1,3)*Y(2,4)*Y(3,4)-
     &  2D0*Y(0,0)*Y(2,3)*Y(2,4)*Y(3,4)+2D0*Y(0,1)*Y(2,3)*Y(2,4)*
     &  Y(3,4)+Y(0,1)*Y(0,2)*Y(3,4)**2-Y(0,2)**2*Y(3,4)**2-Y(0,0)*
     &  Y(1,2)*Y(3,4)**2+Y(0,2)*Y(1,2)*Y(3,4)**2+Y(0,0)*Y(2,2)*Y(3,
     &  4)**2-Y(0,1)*Y(2,2)*Y(3,4)**2+Y(0,2)*Y(0,3)*Y(1,2)*Y(4,4)-
     &  Y(0,3)**2*Y(1,2)*Y(4,4)-Y(0,2)**2*Y(1,3)*Y(4,4)+Y(0,2)*Y(0,
     &  3)*Y(1,3)*Y(4,4)-Y(0,1)*Y(0,3)*Y(2,2)*Y(4,4)+Y(0,3)**2*Y(2,
     &  2)*Y(4,4)+Y(0,0)*Y(1,3)*Y(2,2)*Y(4,4)-Y(0,3)*Y(1,3)*Y(2,2)*
     &  Y(4,4)+Y(0,1)*Y(0,2)*Y(2,3)*Y(4,4)+Y(0,1)*Y(0,3)*Y(2,3)*Y(4,
     &  4)-2D0*Y(0,2)*Y(0,3)*Y(2,3)*Y(4,4)-Y(0,0)*Y(1,2)*Y(2,3)*Y(4,
     &  4)+Y(0,3)*Y(1,2)*Y(2,3)*Y(4,4)-Y(0,0)*Y(1,3)*Y(2,3)*Y(4,4)+
     &  Y(0,2)*Y(1,3)*Y(2,3)*Y(4,4)+Y(0,0)*Y(2,3)**2*Y(4,4)-Y(0,1)*
     &  Y(2,3)**2*Y(4,4)-Y(0,1)*Y(0,2)*Y(3,3)*Y(4,4)+Y(0,2)**2*Y(3,
     &  3)*Y(4,4)+Y(0,0)*Y(1,2)*Y(3,3)*Y(4,4)-Y(0,2)*Y(1,2)*Y(3,3)*
     &  Y(4,4)-Y(0,0)*Y(2,2)*Y(3,3)*Y(4,4)+Y(0,1)*Y(2,2)*Y(3,3)*Y(4,
     &  4)
        DETY(2) = -(Y(0,3)*Y(0,4)*Y(1,2)*Y(1,3))+Y(0,4)**2*Y(1,2)*
     &  Y(1,3)+Y(0,2)*Y(0,4)*Y(1,3)**2-Y(0,4)**2*Y(1,3)**2+Y(0,3)**
     &  2*Y(1,2)*Y(1,4)-Y(0,3)*Y(0,4)*Y(1,2)*Y(1,4)-Y(0,2)*Y(0,3)*
     &  Y(1,3)*Y(1,4)-Y(0,2)*Y(0,4)*Y(1,3)*Y(1,4)+2D0*Y(0,3)*Y(0,4)*
     &  Y(1,3)*Y(1,4)+Y(0,2)*Y(0,3)*Y(1,4)**2-Y(0,3)**2*Y(1,4)**2+
     &  Y(0,3)*Y(0,4)*Y(1,1)*Y(2,3)-Y(0,4)**2*Y(1,1)*Y(2,3)-Y(0,1)*
     &  Y(0,4)*Y(1,3)*Y(2,3)+Y(0,4)**2*Y(1,3)*Y(2,3)-Y(0,1)*Y(0,3)*
     &  Y(1,4)*Y(2,3)+2D0*Y(0,1)*Y(0,4)*Y(1,4)*Y(2,3)-Y(0,3)*Y(0,4)*
     &  Y(1,4)*Y(2,3)+Y(0,0)*Y(1,3)*Y(1,4)*Y(2,3)-Y(0,4)*Y(1,3)*Y(1,
     &  4)*Y(2,3)-Y(0,0)*Y(1,4)**2*Y(2,3)+Y(0,3)*Y(1,4)**2*Y(2,3)-
     &  Y(0,3)**2*Y(1,1)*Y(2,4)+Y(0,3)*Y(0,4)*Y(1,1)*Y(2,4)+2D0*Y(0,
     &  1)*Y(0,3)*Y(1,3)*Y(2,4)-Y(0,1)*Y(0,4)*Y(1,3)*Y(2,4)-Y(0,3)*
     &  Y(0,4)*Y(1,3)*Y(2,4)-Y(0,0)*Y(1,3)**2*Y(2,4)+Y(0,4)*Y(1,3)**
     &  2*Y(2,4)-Y(0,1)*Y(0,3)*Y(1,4)*Y(2,4)+Y(0,3)**2*Y(1,4)*Y(2,
     &  4)+Y(0,0)*Y(1,3)*Y(1,4)*Y(2,4)-Y(0,3)*Y(1,3)*Y(1,4)*Y(2,4)-
     &  Y(0,2)*Y(0,4)*Y(1,1)*Y(3,3)+Y(0,4)**2*Y(1,1)*Y(3,3)+Y(0,1)*
     &  Y(0,4)*Y(1,2)*Y(3,3)-Y(0,4)**2*Y(1,2)*Y(3,3)+Y(0,1)*Y(0,2)*
     &  Y(1,4)*Y(3,3)-2D0*Y(0,1)*Y(0,4)*Y(1,4)*Y(3,3)+Y(0,2)*Y(0,4)*
     &  Y(1,4)*Y(3,3)-Y(0,0)*Y(1,2)*Y(1,4)*Y(3,3)+Y(0,4)*Y(1,2)*Y(1,
     &  4)*Y(3,3)+Y(0,0)*Y(1,4)**2*Y(3,3)-Y(0,2)*Y(1,4)**2*Y(3,3)-
     &  Y(0,1)**2*Y(2,4)*Y(3,3)+Y(0,1)*Y(0,4)*Y(2,4)*Y(3,3)+Y(0,0)*
     &  Y(1,1)*Y(2,4)*Y(3,3)-Y(0,4)*Y(1,1)*Y(2,4)*Y(3,3)-Y(0,0)*Y(1,
     &  4)*Y(2,4)*Y(3,3)+Y(0,1)*Y(1,4)*Y(2,4)*Y(3,3)+Y(0,2)*Y(0,3)*
     &  Y(1,1)*Y(3,4)+Y(0,2)*Y(0,4)*Y(1,1)*Y(3,4)-2D0*Y(0,3)*Y(0,4)*
     &  Y(1,1)*Y(3,4)-Y(0,1)*Y(0,3)*Y(1,2)*Y(3,4)-Y(0,1)*Y(0,4)*Y(1,
     &  2)*Y(3,4)+2D0*Y(0,3)*Y(0,4)*Y(1,2)*Y(3,4)-Y(0,1)*Y(0,2)*Y(1,
     &  3)*Y(3,4)+2D0*Y(0,1)*Y(0,4)*Y(1,3)*Y(3,4)-Y(0,2)*Y(0,4)*Y(1,
     &  3)*Y(3,4)+Y(0,0)*Y(1,2)*Y(1,3)*Y(3,4)-Y(0,4)*Y(1,2)*Y(1,3)*
     &  Y(3,4)-Y(0,1)*Y(0,2)*Y(1,4)*Y(3,4)+2D0*Y(0,1)*Y(0,3)*Y(1,4)*
     &  Y(3,4)-Y(0,2)*Y(0,3)*Y(1,4)*Y(3,4)+Y(0,0)*Y(1,2)*Y(1,4)*Y(3,
     &  4)-Y(0,3)*Y(1,2)*Y(1,4)*Y(3,4)-2D0*Y(0,0)*Y(1,3)*Y(1,4)*Y(3,
     &  4)+2D0*Y(0,2)*Y(1,3)*Y(1,4)*Y(3,4)+Y(0,1)**2*Y(2,3)*Y(3,4)-
     &  Y(0,1)*Y(0,4)*Y(2,3)*Y(3,4)-Y(0,0)*Y(1,1)*Y(2,3)*Y(3,4)+Y(0,
     &  4)*Y(1,1)*Y(2,3)*Y(3,4)+Y(0,0)*Y(1,4)*Y(2,3)*Y(3,4)-Y(0,1)*
     &  Y(1,4)*Y(2,3)*Y(3,4)+Y(0,1)**2*Y(2,4)*Y(3,4)-Y(0,1)*Y(0,3)*
     &  Y(2,4)*Y(3,4)-Y(0,0)*Y(1,1)*Y(2,4)*Y(3,4)+Y(0,3)*Y(1,1)*Y(2,
     &  4)*Y(3,4)+Y(0,0)*Y(1,3)*Y(2,4)*Y(3,4)-Y(0,1)*Y(1,3)*Y(2,4)*
     &  Y(3,4)-Y(0,1)**2*Y(3,4)**2+Y(0,1)*Y(0,2)*Y(3,4)**2+Y(0,0)*
     &  Y(1,1)*Y(3,4)**2-Y(0,2)*Y(1,1)*Y(3,4)**2-Y(0,0)*Y(1,2)*Y(3,
     &  4)**2+Y(0,1)*Y(1,2)*Y(3,4)**2-Y(0,2)*Y(0,3)*Y(1,1)*Y(4,4)+
     &  Y(0,3)**2*Y(1,1)*Y(4,4)+Y(0,1)*Y(0,3)*Y(1,2)*Y(4,4)-Y(0,3)**
     &  2*Y(1,2)*Y(4,4)+Y(0,1)*Y(0,2)*Y(1,3)*Y(4,4)-2D0*Y(0,1)*Y(0,
     &  3)*Y(1,3)*Y(4,4)+Y(0,2)*Y(0,3)*Y(1,3)*Y(4,4)-Y(0,0)*Y(1,2)*
     &  Y(1,3)*Y(4,4)+Y(0,3)*Y(1,2)*Y(1,3)*Y(4,4)+Y(0,0)*Y(1,3)**2*
     &  Y(4,4)-Y(0,2)*Y(1,3)**2*Y(4,4)-Y(0,1)**2*Y(2,3)*Y(4,4)+Y(0,
     &  1)*Y(0,3)*Y(2,3)*Y(4,4)+Y(0,0)*Y(1,1)*Y(2,3)*Y(4,4)-Y(0,3)*
     &  Y(1,1)*Y(2,3)*Y(4,4)-Y(0,0)*Y(1,3)*Y(2,3)*Y(4,4)+Y(0,1)*Y(1,
     &  3)*Y(2,3)*Y(4,4)+Y(0,1)**2*Y(3,3)*Y(4,4)-Y(0,1)*Y(0,2)*Y(3,
     &  3)*Y(4,4)-Y(0,0)*Y(1,1)*Y(3,3)*Y(4,4)+Y(0,2)*Y(1,1)*Y(3,3)*
     &  Y(4,4)+Y(0,0)*Y(1,2)*Y(3,3)*Y(4,4)-Y(0,1)*Y(1,2)*Y(3,3)*Y(4,
     &  4)
        DETY(3) = Y(0,3)*Y(0,4)*Y(1,2)**2-Y(0,4)**2*Y(1,2)**2-Y(0,
     &  2)*Y(0,4)*Y(1,2)*Y(1,3)+Y(0,4)**2*Y(1,2)*Y(1,3)-Y(0,2)*Y(0,
     &  3)*Y(1,2)*Y(1,4)+2D0*Y(0,2)*Y(0,4)*Y(1,2)*Y(1,4)-Y(0,3)*Y(0,
     &  4)*Y(1,2)*Y(1,4)+Y(0,2)**2*Y(1,3)*Y(1,4)-Y(0,2)*Y(0,4)*Y(1,
     &  3)*Y(1,4)-Y(0,2)**2*Y(1,4)**2+Y(0,2)*Y(0,3)*Y(1,4)**2-Y(0,
     &  3)*Y(0,4)*Y(1,1)*Y(2,2)+Y(0,4)**2*Y(1,1)*Y(2,2)+Y(0,1)*Y(0,
     &  4)*Y(1,3)*Y(2,2)-Y(0,4)**2*Y(1,3)*Y(2,2)+Y(0,1)*Y(0,3)*Y(1,
     &  4)*Y(2,2)-2D0*Y(0,1)*Y(0,4)*Y(1,4)*Y(2,2)+Y(0,3)*Y(0,4)*Y(1,
     &  4)*Y(2,2)-Y(0,0)*Y(1,3)*Y(1,4)*Y(2,2)+Y(0,4)*Y(1,3)*Y(1,4)*
     &  Y(2,2)+Y(0,0)*Y(1,4)**2*Y(2,2)-Y(0,3)*Y(1,4)**2*Y(2,2)+Y(0,
     &  2)*Y(0,4)*Y(1,1)*Y(2,3)-Y(0,4)**2*Y(1,1)*Y(2,3)-Y(0,1)*Y(0,
     &  4)*Y(1,2)*Y(2,3)+Y(0,4)**2*Y(1,2)*Y(2,3)-Y(0,1)*Y(0,2)*Y(1,
     &  4)*Y(2,3)+2D0*Y(0,1)*Y(0,4)*Y(1,4)*Y(2,3)-Y(0,2)*Y(0,4)*Y(1,
     &  4)*Y(2,3)+Y(0,0)*Y(1,2)*Y(1,4)*Y(2,3)-Y(0,4)*Y(1,2)*Y(1,4)*
     &  Y(2,3)-Y(0,0)*Y(1,4)**2*Y(2,3)+Y(0,2)*Y(1,4)**2*Y(2,3)+Y(0,
     &  2)*Y(0,3)*Y(1,1)*Y(2,4)-2D0*Y(0,2)*Y(0,4)*Y(1,1)*Y(2,4)+Y(0,
     &  3)*Y(0,4)*Y(1,1)*Y(2,4)-Y(0,1)*Y(0,3)*Y(1,2)*Y(2,4)+2D0*Y(0,
     &  1)*Y(0,4)*Y(1,2)*Y(2,4)-Y(0,3)*Y(0,4)*Y(1,2)*Y(2,4)-Y(0,1)*
     &  Y(0,2)*Y(1,3)*Y(2,4)-Y(0,1)*Y(0,4)*Y(1,3)*Y(2,4)+2D0*Y(0,2)*
     &  Y(0,4)*Y(1,3)*Y(2,4)+Y(0,0)*Y(1,2)*Y(1,3)*Y(2,4)-Y(0,4)*Y(1,
     &  2)*Y(1,3)*Y(2,4)+2D0*Y(0,1)*Y(0,2)*Y(1,4)*Y(2,4)-Y(0,1)*Y(0,
     &  3)*Y(1,4)*Y(2,4)-Y(0,2)*Y(0,3)*Y(1,4)*Y(2,4)-2D0*Y(0,0)*Y(1,
     &  2)*Y(1,4)*Y(2,4)+2D0*Y(0,3)*Y(1,2)*Y(1,4)*Y(2,4)+Y(0,0)*Y(1,
     &  3)*Y(1,4)*Y(2,4)-Y(0,2)*Y(1,3)*Y(1,4)*Y(2,4)+Y(0,1)**2*Y(2,
     &  3)*Y(2,4)-Y(0,1)*Y(0,4)*Y(2,3)*Y(2,4)-Y(0,0)*Y(1,1)*Y(2,3)*
     &  Y(2,4)+Y(0,4)*Y(1,1)*Y(2,3)*Y(2,4)+Y(0,0)*Y(1,4)*Y(2,3)*Y(2,
     &  4)-Y(0,1)*Y(1,4)*Y(2,3)*Y(2,4)-Y(0,1)**2*Y(2,4)**2+Y(0,1)*
     &  Y(0,3)*Y(2,4)**2+Y(0,0)*Y(1,1)*Y(2,4)**2-Y(0,3)*Y(1,1)*Y(2,
     &  4)**2-Y(0,0)*Y(1,3)*Y(2,4)**2+Y(0,1)*Y(1,3)*Y(2,4)**2-Y(0,
     &  2)**2*Y(1,1)*Y(3,4)+Y(0,2)*Y(0,4)*Y(1,1)*Y(3,4)+2D0*Y(0,1)*
     &  Y(0,2)*Y(1,2)*Y(3,4)-Y(0,1)*Y(0,4)*Y(1,2)*Y(3,4)-Y(0,2)*Y(0,
     &  4)*Y(1,2)*Y(3,4)-Y(0,0)*Y(1,2)**2*Y(3,4)+Y(0,4)*Y(1,2)**2*
     &  Y(3,4)-Y(0,1)*Y(0,2)*Y(1,4)*Y(3,4)+Y(0,2)**2*Y(1,4)*Y(3,4)+
     &  Y(0,0)*Y(1,2)*Y(1,4)*Y(3,4)-Y(0,2)*Y(1,2)*Y(1,4)*Y(3,4)-Y(0,
     &  1)**2*Y(2,2)*Y(3,4)+Y(0,1)*Y(0,4)*Y(2,2)*Y(3,4)+Y(0,0)*Y(1,
     &  1)*Y(2,2)*Y(3,4)-Y(0,4)*Y(1,1)*Y(2,2)*Y(3,4)-Y(0,0)*Y(1,4)*
     &  Y(2,2)*Y(3,4)+Y(0,1)*Y(1,4)*Y(2,2)*Y(3,4)+Y(0,1)**2*Y(2,4)*
     &  Y(3,4)-Y(0,1)*Y(0,2)*Y(2,4)*Y(3,4)-Y(0,0)*Y(1,1)*Y(2,4)*Y(3,
     &  4)+Y(0,2)*Y(1,1)*Y(2,4)*Y(3,4)+Y(0,0)*Y(1,2)*Y(2,4)*Y(3,4)-
     &  Y(0,1)*Y(1,2)*Y(2,4)*Y(3,4)+Y(0,2)**2*Y(1,1)*Y(4,4)-Y(0,2)*
     &  Y(0,3)*Y(1,1)*Y(4,4)-2D0*Y(0,1)*Y(0,2)*Y(1,2)*Y(4,4)+Y(0,1)*
     &  Y(0,3)*Y(1,2)*Y(4,4)+Y(0,2)*Y(0,3)*Y(1,2)*Y(4,4)+Y(0,0)*Y(1,
     &  2)**2*Y(4,4)-Y(0,3)*Y(1,2)**2*Y(4,4)+Y(0,1)*Y(0,2)*Y(1,3)*
     &  Y(4,4)-Y(0,2)**2*Y(1,3)*Y(4,4)-Y(0,0)*Y(1,2)*Y(1,3)*Y(4,4)+
     &  Y(0,2)*Y(1,2)*Y(1,3)*Y(4,4)+Y(0,1)**2*Y(2,2)*Y(4,4)-Y(0,1)*
     &  Y(0,3)*Y(2,2)*Y(4,4)-Y(0,0)*Y(1,1)*Y(2,2)*Y(4,4)+Y(0,3)*Y(1,
     &  1)*Y(2,2)*Y(4,4)+Y(0,0)*Y(1,3)*Y(2,2)*Y(4,4)-Y(0,1)*Y(1,3)*
     &  Y(2,2)*Y(4,4)-Y(0,1)**2*Y(2,3)*Y(4,4)+Y(0,1)*Y(0,2)*Y(2,3)*
     &  Y(4,4)+Y(0,0)*Y(1,1)*Y(2,3)*Y(4,4)-Y(0,2)*Y(1,1)*Y(2,3)*Y(4,
     &  4)-Y(0,0)*Y(1,2)*Y(2,3)*Y(4,4)+Y(0,1)*Y(1,2)*Y(2,3)*Y(4,4)
        DETY(4) = -(Y(0,3)**2*Y(1,2)**2)+Y(0,3)*Y(0,4)*Y(1,2)**2+
     &  2D0*Y(0,2)*Y(0,3)*Y(1,2)*Y(1,3)-Y(0,2)*Y(0,4)*Y(1,2)*Y(1,3)-
     &  Y(0,3)*Y(0,4)*Y(1,2)*Y(1,3)-Y(0,2)**2*Y(1,3)**2+Y(0,2)*Y(0,
     &  4)*Y(1,3)**2-Y(0,2)*Y(0,3)*Y(1,2)*Y(1,4)+Y(0,3)**2*Y(1,2)*
     &  Y(1,4)+Y(0,2)**2*Y(1,3)*Y(1,4)-Y(0,2)*Y(0,3)*Y(1,3)*Y(1,4)+
     &  Y(0,3)**2*Y(1,1)*Y(2,2)-Y(0,3)*Y(0,4)*Y(1,1)*Y(2,2)-2D0*Y(0,
     &  1)*Y(0,3)*Y(1,3)*Y(2,2)+Y(0,1)*Y(0,4)*Y(1,3)*Y(2,2)+Y(0,3)*
     &  Y(0,4)*Y(1,3)*Y(2,2)+Y(0,0)*Y(1,3)**2*Y(2,2)-Y(0,4)*Y(1,3)**
     &  2*Y(2,2)+Y(0,1)*Y(0,3)*Y(1,4)*Y(2,2)-Y(0,3)**2*Y(1,4)*Y(2,
     &  2)-Y(0,0)*Y(1,3)*Y(1,4)*Y(2,2)+Y(0,3)*Y(1,3)*Y(1,4)*Y(2,2)-
     &  2D0*Y(0,2)*Y(0,3)*Y(1,1)*Y(2,3)+Y(0,2)*Y(0,4)*Y(1,1)*Y(2,3)+
     &  Y(0,3)*Y(0,4)*Y(1,1)*Y(2,3)+2D0*Y(0,1)*Y(0,3)*Y(1,2)*Y(2,3)-
     &  Y(0,1)*Y(0,4)*Y(1,2)*Y(2,3)-Y(0,3)*Y(0,4)*Y(1,2)*Y(2,3)+2D0*
     &  Y(0,1)*Y(0,2)*Y(1,3)*Y(2,3)-Y(0,1)*Y(0,4)*Y(1,3)*Y(2,3)-Y(0,
     &  2)*Y(0,4)*Y(1,3)*Y(2,3)-2D0*Y(0,0)*Y(1,2)*Y(1,3)*Y(2,3)+2D0*
     &  Y(0,4)*Y(1,2)*Y(1,3)*Y(2,3)-Y(0,1)*Y(0,2)*Y(1,4)*Y(2,3)-Y(0,
     &  1)*Y(0,3)*Y(1,4)*Y(2,3)+2D0*Y(0,2)*Y(0,3)*Y(1,4)*Y(2,3)+Y(0,
     &  0)*Y(1,2)*Y(1,4)*Y(2,3)-Y(0,3)*Y(1,2)*Y(1,4)*Y(2,3)+Y(0,0)*
     &  Y(1,3)*Y(1,4)*Y(2,3)-Y(0,2)*Y(1,3)*Y(1,4)*Y(2,3)-Y(0,1)**2*
     &  Y(2,3)**2+Y(0,1)*Y(0,4)*Y(2,3)**2+Y(0,0)*Y(1,1)*Y(2,3)**2-
     &  Y(0,4)*Y(1,1)*Y(2,3)**2-Y(0,0)*Y(1,4)*Y(2,3)**2+Y(0,1)*Y(1,
     &  4)*Y(2,3)**2+Y(0,2)*Y(0,3)*Y(1,1)*Y(2,4)-Y(0,3)**2*Y(1,1)*
     &  Y(2,4)-Y(0,1)*Y(0,3)*Y(1,2)*Y(2,4)+Y(0,3)**2*Y(1,2)*Y(2,4)-
     &  Y(0,1)*Y(0,2)*Y(1,3)*Y(2,4)+2D0*Y(0,1)*Y(0,3)*Y(1,3)*Y(2,4)-
     &  Y(0,2)*Y(0,3)*Y(1,3)*Y(2,4)+Y(0,0)*Y(1,2)*Y(1,3)*Y(2,4)-Y(0,
     &  3)*Y(1,2)*Y(1,3)*Y(2,4)-Y(0,0)*Y(1,3)**2*Y(2,4)+Y(0,2)*Y(1,
     &  3)**2*Y(2,4)+Y(0,1)**2*Y(2,3)*Y(2,4)-Y(0,1)*Y(0,3)*Y(2,3)*
     &  Y(2,4)-Y(0,0)*Y(1,1)*Y(2,3)*Y(2,4)+Y(0,3)*Y(1,1)*Y(2,3)*Y(2,
     &  4)+Y(0,0)*Y(1,3)*Y(2,3)*Y(2,4)-Y(0,1)*Y(1,3)*Y(2,3)*Y(2,4)+
     &  Y(0,2)**2*Y(1,1)*Y(3,3)-Y(0,2)*Y(0,4)*Y(1,1)*Y(3,3)-2D0*Y(0,
     &  1)*Y(0,2)*Y(1,2)*Y(3,3)+Y(0,1)*Y(0,4)*Y(1,2)*Y(3,3)+Y(0,2)*
     &  Y(0,4)*Y(1,2)*Y(3,3)+Y(0,0)*Y(1,2)**2*Y(3,3)-Y(0,4)*Y(1,2)**
     &  2*Y(3,3)+Y(0,1)*Y(0,2)*Y(1,4)*Y(3,3)-Y(0,2)**2*Y(1,4)*Y(3,
     &  3)-Y(0,0)*Y(1,2)*Y(1,4)*Y(3,3)+Y(0,2)*Y(1,2)*Y(1,4)*Y(3,3)+
     &  Y(0,1)**2*Y(2,2)*Y(3,3)-Y(0,1)*Y(0,4)*Y(2,2)*Y(3,3)-Y(0,0)*
     &  Y(1,1)*Y(2,2)*Y(3,3)+Y(0,4)*Y(1,1)*Y(2,2)*Y(3,3)+Y(0,0)*Y(1,
     &  4)*Y(2,2)*Y(3,3)-Y(0,1)*Y(1,4)*Y(2,2)*Y(3,3)-Y(0,1)**2*Y(2,
     &  4)*Y(3,3)+Y(0,1)*Y(0,2)*Y(2,4)*Y(3,3)+Y(0,0)*Y(1,1)*Y(2,4)*
     &  Y(3,3)-Y(0,2)*Y(1,1)*Y(2,4)*Y(3,3)-Y(0,0)*Y(1,2)*Y(2,4)*Y(3,
     &  3)+Y(0,1)*Y(1,2)*Y(2,4)*Y(3,3)-Y(0,2)**2*Y(1,1)*Y(3,4)+Y(0,
     &  2)*Y(0,3)*Y(1,1)*Y(3,4)+2D0*Y(0,1)*Y(0,2)*Y(1,2)*Y(3,4)-Y(0,
     &  1)*Y(0,3)*Y(1,2)*Y(3,4)-Y(0,2)*Y(0,3)*Y(1,2)*Y(3,4)-Y(0,0)*
     &  Y(1,2)**2*Y(3,4)+Y(0,3)*Y(1,2)**2*Y(3,4)-Y(0,1)*Y(0,2)*Y(1,
     &  3)*Y(3,4)+Y(0,2)**2*Y(1,3)*Y(3,4)+Y(0,0)*Y(1,2)*Y(1,3)*Y(3,
     &  4)-Y(0,2)*Y(1,2)*Y(1,3)*Y(3,4)-Y(0,1)**2*Y(2,2)*Y(3,4)+Y(0,
     &  1)*Y(0,3)*Y(2,2)*Y(3,4)+Y(0,0)*Y(1,1)*Y(2,2)*Y(3,4)-Y(0,3)*
     &  Y(1,1)*Y(2,2)*Y(3,4)-Y(0,0)*Y(1,3)*Y(2,2)*Y(3,4)+Y(0,1)*Y(1,
     &  3)*Y(2,2)*Y(3,4)+Y(0,1)**2*Y(2,3)*Y(3,4)-Y(0,1)*Y(0,2)*Y(2,
     &  3)*Y(3,4)-Y(0,0)*Y(1,1)*Y(2,3)*Y(3,4)+Y(0,2)*Y(1,1)*Y(2,3)*
     &  Y(3,4)+Y(0,0)*Y(1,2)*Y(2,3)*Y(3,4)-Y(0,1)*Y(1,2)*Y(2,3)*Y(3,
     &  4)

	xE0_ = 0d0
	do 10 i=0,4
	xE0_ = xE0_ + D0(i)*dety(i)
10	continue
	xE0_ = xE0_/detyy

	end

c=======================================================================
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     scalar 3-point function                                          c
c                                                                      c
c                   O----- q01                                         c
c              m12 /|                                                  c
c                 / |                                                  c
c       q12 -----O  | m02                                              c
c                 \ |                                                  c
c              m22 \|                                                  c
c                   O----- q20                                         c
c                                                                      c
c     m0,m1,m2: squares of masses  q01,q12,q20: squares of momenta     c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function xC0_(q01,q12,q20,m02,m12,m22,ext)
      implicit none
c local variables
      complex*16 xC0_,inta,intb,intir,m02,m12,m22,x1,x2,xli2,y,d1
      complex*16 alpha,i,c0d0,c1d0,pi6,id1,ialpha,iy,ix1,ix2
      complex*16 p01,p12,p20,ip01,ip12,ip20,fct(110),x(16)
      real*8 q01,q12,q20,small,pi,pi2,dev,sgn,acc
      integer ext,i1,i2,i3,k0,k1,k2,j0,j1,j2,n,crit,nzm,nzp,nir
      integer sx1,sx2,sd1,salpha,ny,nx,nrank
      integer coll(0:2),ir(0:2),zm(0:2),zp(0:2)
      logical nocalc
      character*11 name
c fmass
      complex*16 cme2,cmf2(5)
      real*8 me2,rme2,mf2(5),rmf2(5)
c xco
      complex*16 m(0:4,0:2),p(0:4,0:4,0:2),im(0:4,0:2),ip(0:4,0:4,0:2)
      integer sm(0:4),sp(0:4,0:4)
      common/fmass/me2,rme2,cme2,mf2,rmf2,cmf2
      common/xco/m,p,im,ip,sm,sp
      parameter(small=1d-19,acc=1d-10)
      c0d0=dcmplx(0d0)
      x(1)=q01
      x(2)=q12
      x(3)=q20
      x(4)=m02
      x(5)=m12
      x(6)=m22
      x(7)=ext
      nrank=0
      name='xC0_'
      call fastin(fct,x,1,7,5,nrank,name,nocalc)
      if(nocalc)then
        xC0_=fct(1)
        return
      endif
      i=dcmplx(0d0,1d0)
      c1d0=dcmplx(1d0)
      pi=4d0*datan(1d0)
      pi2=pi*pi
      pi6=dcmplx(pi2/6d0)
      m(0,0)=m02
      m(1,0)=m12
      m(2,0)=m22
      p(1,2,0)=dcmplx(q12)
      p(2,0,0)=dcmplx(q20)
      p(0,1,0)=dcmplx(q01)
      do i1=0,2
        coll(i1)=0
        zm(i1)=0
        zp(i1)=0
        ir(i1)=0
      enddo
c recovering small parameters
      do k0=0,2
        sm(k0)=0
        m(k0,2)=0d0
        if(dreal(m(k0,0)).lt.small)then
          sm(k0)=2
          n=nint(dreal(m(k0,0))*1d20)
          dev=dabs(dreal(m(k0,0))*1d20-dble(n))
          if(n.ge.1.and.n.le.5.and.dev.lt.acc)then
            m(k0,2)=mf2(n)
            m(k0,0)=0d0
          elseif(m(k0,2).ne.c0d0)then
            write(*,'(a)')' xC0_: wrong mass parameter'
            stop
          endif
        endif
        k1=mod(k0+1,3)
        sp(k0,k1)=0
        p(k0,k1,2)=0d0
        if(dabs(dreal(p(k0,k1,0))).lt.small)then
          sp(k0,k1)=2
          n=nint(dreal(p(k0,k1,0))*1d20)
          dev=dabs(dreal(p(k0,k1,0))*1d20-dble(n))
          if(n.ge.1.and.n.le.5.and.dev.lt.acc)then
            p(k0,k1,2)=mf2(n)
            p(k0,k1,0)=0d0
          endif
        endif
        sp(k1,k0)=sp(k0,k1)
        p(k1,k0,0)=p(k0,k1,0)
        p(k1,k0,2)=p(k0,k1,2)
      enddo
c determining singularities
      do k0=0,2
        k1=mod(k0+1,3)
        k2=mod(k0+2,3)
        if(sp(k1,k2).gt.0)then
          if(sm(k1).gt.0.and.sm(k2).gt.0)then
c collinear singularities
            coll(k1)=coll(k1)+1
            coll(k2)=coll(k2)+1
          else
c set invariant zero if no singularity present
            p(k1,k2,sp(k1,k2))=0d0
            p(k2,k1,sp(k2,k1))=0d0
          endif
        endif
c vanishing invariants
        if(p(k1,k2,sp(k1,k2)).eq.c0d0)zp(k0)=1
      enddo
      do k0=0,2
        k1=mod(k0+1,3)
        k2=mod(k0+2,3)
        if(p(k2,k0,sp(k2,k0)).eq.m(k2,sm(k2)).and.
     *    p(k0,k1,sp(k0,k1)).eq.m(k1,sm(k1)))then
c infrared singularities
          if(m(k0,sm(k0)).eq.c0d0)ir(k0)=1
        elseif(sm(k0).gt.0.and.coll(k0).eq.0)then
c set mass zero if no singularity present
          m(k0,sm(k0))=0d0
        endif
c vanishing masses
        if(m(k0,sm(k0)).eq.c0d0)zm(k0)=1
      enddo
c number of vanishing masses
      nzm=zm(0)+zm(1)+zm(2)
c number of vanishing invariants
      nzp=zp(0)+zp(1)+zp(2)
c number of infrared singularities
      nir=ir(0)+ir(1)+ir(2)
c defining standard order of arguments
      n=0
      do j0=0,2
      do j1=0,2
      do j2=0,2
      if(j0.ne.j1.and.j0.ne.j2.and.j1.ne.j2)then
        crit=ir(j0)*8+zp(j2)*4+zm(j0)*2+1
        if(crit.gt.n)then
          n=crit
          k0=j0
          k1=j1
          k2=j2
        endif
      endif
      enddo
      enddo
      enddo
c filling common block xco
      do i1=0,2
      do i2=0,2,2
        im(i1,i2)=0d0
        if(m(i1,i2).ne.c0d0)
     *    im(i1,i2)=dcmplx(dreal(m(i1,i2)),-1d0/dble(i1+1))
        do i3=0,2
          ip(i1,i3,i2)=p(i1,i3,i2)
        enddo
      enddo
      enddo
c different cases
      if(zp(k2).eq.1.and.nir.eq.0)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                   O----- 0                                           c
c              m12 /|                                                  c
c                 / |                                                  c
c       q12 -----O  | m02                                              c
c                 \ |                                                  c
c              m22 \|                                                  c
c                   O----- q20                                         c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        alpha=p(k1,k2,0)-p(k2,k0,0)
        sd1=min(sm(k0),sm(k1))
        d1=m(k0,sd1)-m(k1,sd1)
        id1=im(k0,sd1)-im(k1,sd1)
c error messages
        if(alpha.eq.c0d0)then
          write(*,'(a)')' xC0_: alpha=0 or small not implemented'
          print*,'alpha=',alpha,p(k1,k2,0),p(k2,k0,0),k1,k2,k0
          print*,'momenta:',q01,q12,q20,m02,m12,m22
          stop
        endif
        if(d1.eq.c0d0)then
          write(*,'(a)')' xC0_: m02-m12=0 not implemented'
          stop
        endif
c x1,x2
        sx1=-sd1
        if(sx1.eq.0)then
          x1=1d0+alpha/d1
        else
          x1=alpha/d1
        endif
        ix1=dcmplx(dreal(x1),dimag(alpha/id1))
c final calculation
        xC0_=(inta(x1,c1d0,ix1,c1d0,k2,k0,sx1,0)
     *    -inta(x1,c1d0,ix1,c1d0,k2,k1,sx1,0))/alpha
      elseif(zm(k0).eq.1.and.nzp.eq.0)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                   O----- q01                                         c
c              m12 /|                                                  c
c                 / |                                                  c
c       q12 -----O  | 0                                                c
c                 \ |                                                  c
c              m22 \|                                                  c
c                   O----- q20                                         c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c introducing small imaginary part for invariants
        do i1=0,2
        do i2=0,2
        do i3=0,2,2
          if(p(i1,i2,i3).ne.c0d0)
     *      ip(i1,i2,i3)=dcmplx(dreal(p(i1,i2,i3)),1d-1/dble(i1+i2))
        enddo
        enddo
        enddo
        p01=p(k0,k1,sp(k0,k1))
        p12=p(k1,k2,sp(k1,k2))
        p20=p(k2,k0,sp(k2,k0))
        ip01=ip(k0,k1,sp(k0,k1))
        ip12=ip(k1,k2,sp(k1,k2))
        ip20=ip(k2,k0,sp(k2,k0))
c starting the evaluation
        if(sp(k1,k2).eq.sp(k0,k1).and.sp(k1,k2).eq.sp(k2,k0))then
          sgn=dsign(1d0,dreal(p01+p20-p12))
          alpha=sgn*cdsqrt((p12-p01-p20)**2-4d0*p01*p20)
          ialpha=(ip12-ip01-ip20)**2-4d0*ip01*ip20
          ialpha=dcmplx(dreal(alpha),dimag(ialpha/2d0/alpha))
          sx1=0
          x1=(p01+p20-p12+alpha)/2d0/p01
          ix1=(ip01+ip20-ip12+ialpha)/2d0/ip01
          ix1=dcmplx(dreal(x1),dimag(ix1))
          sx2=0
          x2=2d0*p20/(p01+p20-p12+alpha)
          ix2=2d0*ip20/(ip01+ip20-ip12+ialpha)
          ix2=dcmplx(dreal(x2),dimag(ix2))
        elseif(sp(k1,k2).le.sp(k0,k1).and.sp(k1,k2).le.sp(k2,k0))then
          salpha=sp(k1,k2)
          alpha=p(k0,k1,salpha)+p(k2,k0,salpha)-p(k1,k2,salpha)
          ialpha=ip(k0,k1,salpha)+ip(k2,k0,salpha)-ip(k1,k2,salpha)
          sx1=salpha-sp(k0,k1)
          x1=alpha/p01
          ix1=dcmplx(dreal(x1),dimag(ialpha/ip01))
          sx2=sp(k2,k0)-salpha
          x2=p20/alpha
          ix2=dcmplx(dreal(x2),dimag(ip20/ialpha))
        elseif(sp(k1,k2).ge.sp(k0,k1).and.sp(k1,k2).ge.sp(k2,k0))then
          salpha=min(sp(k0,k1),sp(k2,k0))
          alpha=p(k0,k1,salpha)-p(k2,k0,salpha)
          ialpha=ip(k0,k1,salpha)-ip(k2,k0,salpha)
          sx1=0
          x1=1d0
          ix1=1d0
          sx2=sp(k2,k0)-sp(k0,k1)
          x2=p20/p01
          ix2=dcmplx(dreal(x2),dimag(ip20/ip01))
        endif
c error message
        if(cdabs(alpha).lt.acc)then
          write(*,'(a)')' xC0_: alpha=0 or small not implemented'
          print*,'alpha2=',alpha,p(k0,k1,salpha),p(k2,k0,salpha),
     &         k0,k1,salpha
          print*,'momenta:',q01,q12,q20,m02,m12,m22
          stop
        endif
c final calculation
        if(nir.eq.0)then
          xC0_=(inta(x1,x2,ix1,ix2,k2,k1,sx1,sx2)
     *      -intb(x1,x2,ix1,ix2,k2,k0,k1,k0,sx1,sx2)
     *      -xli2(x1,ix1,sx1)+xli2(x2,ix2,sx2))/alpha
        elseif(nir.eq.1.and.ir(k0).eq.1)then
          sgn=dsign(1d0,dreal(x1-x2))
          ix1=dcmplx(dreal(x1),sgn)
          ix2=dcmplx(dreal(x2),-sgn)
          xC0_=(intir(p20,x1,x2,ix1,ix2,sx1,sx2)
     *      -xli2(x1,ix1,sx1)+xli2(x2,ix2,sx2))/alpha
        else
c error message if no case found
          write(*,'(a)')' xC0_: case not implemented'
          write(*,*)q01,q12,q20,m02,m12,m22
          stop
        endif
      elseif(nzm.eq.0.and.nzp.eq.0)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                   O----- q01                                         c
c              m12 /|                                                  c
c                 / |                                                  c
c       q12 -----O  | m02                                              c
c                 \ |                                                  c
c              m22 \|                                                  c
c                   O----- q20                                         c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        alpha=cdsqrt((p(k1,k2,0)-p(k0,k1,0)-p(k2,k0,0))**2
     *    -4d0*p(k0,k1,0)*p(k2,k0,0))
c error messages
        if(cdabs(alpha).eq.c0d0)then
          write(*,'(a)')' xC0_: alpha=small not allowed'
          stop
        endif
        if(dimag(alpha).ne.0d0)then
          write(*,'(a)')' xC0_: complex alpha!'
          stop
        endif
c evaluation
        xC0_=0d0
        do k0=0,2
          k1=mod(k0+1,3)
          k2=mod(k0+2,3)
          if(sp(k1,k2).ne.0)then
            write(*,'(a)')' xC0_: p(k1,k2)=0 or small not implemented'
            stop
          endif
          y=(p(k1,k2,0)*(p(k1,k2,0)-p(k0,k1,0)-p(k2,k0,0)
     *        +2d0*m(k0,0)-m(k1,0)-m(k2,0))
     *      -(p(k0,k1,0)-p(k2,k0,0))*(m(k2,0)-m(k1,0))
     *      +alpha*(p(k1,k2,0)+m(k1,0)-m(k2,0)))
     *      /2d0/alpha/p(k1,k2,0)
          iy=(p(k1,k2,0)*(2d0*im(k0,0)-im(k1,0)-im(k2,0))
     *      -(p(k0,k1,0)-p(k2,k0,0))*(im(k2,0)-im(k1,0))
     *      +alpha*(im(k1,0)-im(k2,0)))/2d0/alpha/p(k1,k2,0)
          iy=dcmplx(dreal(y),dimag(iy))
          if(y.eq.c0d0)then
            write(*,'(a)')' xC0_: y=0 not allowed'
            stop
          endif
          sx1=0
          x1=1d0-1d0/y
          ix1=dcmplx(dreal(x1),dimag(-1d0/iy))
          xC0_=xC0_+(inta(x1,c1d0,ix1,c1d0,k2,k1,sx1,0)
     *      -2d0*xli2(x1,ix1,sx1))/alpha
        enddo
      else
c error message if no case found
        write(*,'(a)')' xC0_: case not implemented'
        write(*,*)q01,q12,q20,m02,m12,m22
        stop
      endif
      fct(1)=xC0_
      call fastout(fct,1,5)
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     scalar 4-point function                                          c
c                                                                      c
c                 m12                                                  c
c       q4 ----O------O---- q1                                         c
c              |      |                                                c
c          m42 |      | m22                                            c
c              |      |                                                c
c       q3 ----O------O---- q2                                         c
c                 m32                                                  c
c                                                                      c
c     m1,m2,m3,m4: squares of masses                                   c
c     q1,q2,q3,q4,q12,q23: squares of momenta                          c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function xD0_(q1,q2,q3,q4,q12,q23,m12,m22,m32,m42,ext)
      implicit none
c local variables
      complex*16 xD0_,m12,m22,m32,m42,x1,x2,xli2
      complex*16 i,c0d0,c1d0,pi6,num,inta,intb,intir,xlog
      complex*16 a(0:6),b(0:6),c(0:6),y(4,4,0:2),sqdety
      complex*16 p1,p2,p3,p4,p12,p23,ip1,ip2,ip3,ip4,ip12,ip23
      complex*16 ix1,ix2,alpha(0:2),alpha2(0:4),fct(110),x(16)
      complex*16 ia(0:6),ib(0:6),ic(0:6),iy(4,4,0:2),inum,al,noal
      real*8 small,acc,pi,pi2,q1,q2,q3,q4,q12,q23,dev,sgn
      integer i1,i2,i3,i4,i5,k1,k2,k3,k4,n,crit,nzm,nzp,nir,ext
      integer coll(4),ir(4),irs(4),zm(4),zp(4,4),j1,j2,j3,j4
      integer sa,sb,sc,sx1,sx2,snum,ny,nx,nrank
      logical la,lb,lc,nocalc
      character*11 name
c xco
      complex*16 m(0:4,0:2),p(0:4,0:4,0:2),im(0:4,0:2),ip(0:4,0:4,0:2)
      integer sm(0:4),sp(0:4,0:4)
c fmass
      complex*16 cme2,cmf2(5)
      real*8 me2,rme2,mf2(5),rmf2(5)
      common/xco/m,p,im,ip,sm,sp
      common/fmass/me2,rme2,cme2,mf2,rmf2,cmf2
      parameter(small=1d-15,acc=1d-10)
      data m/15*0d0/,p/75*0d0/

      c0d0=dcmplx(0d0)
      x(1)=q1
      x(2)=q2
      x(3)=q3
      x(4)=q4
      x(5)=q12
      x(6)=q23
      x(7)=m12
      x(8)=m22
      x(9)=m32
      x(10)=m42
      x(11)=ext
      nrank=0
      name='xD0_'
      call fastin(fct,x,1,11,3,nrank,name,nocalc)
      if(nocalc)then
        xD0_=fct(1)
        return
      endif
      i=dcmplx(0d0,1d0)
      c1d0=dcmplx(1d0)
      pi=4d0*datan(1d0)
      pi2=pi*pi
      pi6=dcmplx(pi2/6d0)
      m(1,0)=m12
      m(2,0)=m22
      m(3,0)=m32
      m(4,0)=m42
      p(1,2,0)=dcmplx(q1)
      p(1,3,0)=dcmplx(q12)
      p(1,4,0)=dcmplx(q4)
      p(2,3,0)=dcmplx(q2)
      p(2,4,0)=dcmplx(q23)
      p(3,4,0)=dcmplx(q3)
      do i1=1,4
        coll(i1)=0
        zm(i1)=0
        ir(i1)=0
        irs(i1)=0
        do i2=1,4
          zp(i1,i2)=0
        enddo
      enddo
c recovering small parameters
      do k1=1,4
        sm(k1)=0
        m(k1,2)=0d0
        if(dreal(m(k1,0)).lt.small)then
          sm(k1)=2
          n=nint(dreal(m(k1,0))*1d20)
          dev=dabs(dreal(m(k1,0))*1d20-dble(n))
          if(n.ge.1.and.n.le.5.and.dev.lt.acc)then
            m(k1,2)=mf2(n)
            m(k1,0)=0d0
          elseif(m(k1,2).ne.c0d0)then
            write(*,'(a)')' xD0_: wrong mass parameter'
            stop
          endif
        endif
        do k2=k1+1,4
          sp(k1,k2)=0
          p(k1,k2,2)=0d0
          if(dabs(dreal(p(k1,k2,0))).lt.small)then
            sp(k1,k2)=2
            n=nint(dreal(p(k1,k2,0))*1d20)
            dev=dabs(dreal(p(k1,k2,0))*1d20-dble(n))
            if(n.ge.1.and.n.le.5.and.dev.lt.acc)then
              p(k1,k2,2)=mf2(n)
              p(k1,k2,0)=0d0
            endif
          endif
          sp(k2,k1)=sp(k1,k2)
          p(k2,k1,0)=p(k1,k2,0)
          p(k2,k1,2)=p(k1,k2,2)
        enddo
      enddo
c determining singularities
      do k1=1,4
      do k2=k1+1,4
        if(sp(k1,k2).gt.0)then
          if(sm(k1).gt.0.and.sm(k2).gt.0)then
c collinear singularities
            coll(k1)=coll(k1)+1
            coll(k2)=coll(k2)+1
          else
c set invariant zero if no singularity present
            p(k1,k2,sp(k1,k2))=0d0
            p(k2,k1,sp(k1,k2))=0d0
          endif
c vanishing invariants
          if(p(k1,k2,sp(k1,k2)).eq.c0d0)then
            zp(k1,k2)=1
            zp(k2,k1)=1
          endif
        endif
      enddo
      enddo
      do k1=1,4
      do k2=1,4
      do k3=k2+1,4
        if(k1.ne.k2.and.k1.ne.k3)then
          if(p(k1,k2,sp(k1,k2)).eq.m(k2,sm(k2)).and.
     *      p(k1,k3,sp(k1,k3)).eq.m(k3,sm(k3)))then
c infrared singularities
            if(m(k1,sm(k1)).eq.c0d0)then
              ir(k1)=1
              irs(k2)=1
              irs(k3)=1
            endif
          elseif(sm(k1).gt.0.and.coll(k1).eq.0)then
c set mass zero if no singularity present
            m(k1,sm(k1))=0d0
          endif
c vanishing masses
          if(m(k1,sm(k1)).eq.c0d0)zm(k1)=1
        endif
      enddo
      enddo
      enddo
c number of vanishing masses
      nzm=zm(1)+zm(2)+zm(3)+zm(4)
c number of vanishing invariants
      nzp=zp(1,2)+zp(1,3)+zp(1,4)+zp(2,3)+zp(2,4)+zp(3,4)
c number of infrared singularities
      nir=ir(1)+ir(2)+ir(3)+ir(4)
c defining standard order of arguments
      n=0
      do j1=1,4
      do j2=1,4
      do j3=1,4
      do j4=1,4
      if(j1.ne.j2.and.j1.ne.j3.and.j1.ne.j4.and.j2.ne.j3.and.
     *  j2.ne.j4.and.j3.ne.j4)then
        crit=(ir(j1)+irs(j2)+irs(j4))*32
     *    +zm(j1)*16+zm(j3)*8+zp(j2,j3)*4+1
        if(crit.gt.n)then
          n=crit
          k1=j1
          k2=j2
          k3=j3
          k4=j4
        endif
      endif
      enddo
      enddo
      enddo
      enddo
c filling common block xco
      do i1=1,4
      do i2=0,2,2
        im(i1,i2)=m(i1,i2)
        if(m(i1,i2).ne.c0d0)
     *    im(i1,i2)=dcmplx(dreal(m(i1,i2)),-1d0/dsqrt(dble(i1+1)))
      enddo
      enddo
      do i1=1,4
      do i2=1,4
      do i3=0,2,2
        ip(i1,i2,i3)=p(i1,i2,i3)
        if(zm(k3).eq.1.or.zp(k2,k3).eq.1)then
          if(p(i1,i2,i3).ne.c0d0)ip(i1,i2,i3)=
     *      dcmplx(dreal(p(i1,i2,i3)),1d-1/dsqrt(dble(i1+i2)))
        endif
        y(i1,i2,i3)=m(i1,i3)+m(i2,i3)-p(i1,i2,i3)
        iy(i1,i2,i3)=im(i1,i3)+im(i2,i3)-ip(i1,i2,i3)
      enddo
      enddo
      enddo
c different cases
      if(nir.ge.1)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                  0                                                   c
c       q4 ----O------O---- q1                                         c
c              |      |                                                c
c          m42 |      | m22                                            c
c              |      |                                                c
c       q3 ----O------O---- q2                                         c
c                 m32                                                  c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c num
        snum=0
        sgn=dsign(1d0,dreal(y(k2,k4,snum)))
        sqdety=cdsqrt(y(k2,k4,snum)**2-4d0*m(k4,snum)*m(k2,snum))
        num=y(k2,k4,snum)+sgn*sqdety
        inum=iy(k2,k4,snum)
     *    +sgn*(iy(k2,k4,snum)**2-4d0*im(k4,snum)*im(k2,snum))
     *      /2d0/sqdety
        inum=dcmplx(dreal(num),dimag(inum))
        sqdety=sqdety*y(k1,k3,0)
c error messages
        if(sqdety.eq.c0d0)then
          write(*,'(a)')' xD0_: dety=small not allowed'
          xD0_=0d0
cacb
c          stop
        endif
        if(num.eq.c0d0)then
          write(*,'(a)')' xD0_: num=small not allowed'
          stop
        endif
c x1,x2
        sx1=-sm(k4)
        x1=num/2d0/m(k4,sm(k4))
        ix1=dcmplx(dreal(x1),dimag(inum/2d0/im(k4,sm(k4))))
        sx2=sm(k2)
        x2=2d0*m(k2,sm(k2))/num
        ix2=dcmplx(dreal(x2),dimag(2d0*im(k2,sm(k2))/inum))
        p12=p(k1,k3,sp(k1,k3))
        p1=p(k1,k2,sp(k1,k2))
        if(nir.eq.1.and.zm(k3).eq.1)then
          xD0_=-sgn*(xlog(-p12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      +intir(p1,x1,x2,ix1,ix2,sx1,sx2)
     *      -intb(x1,x2,ix1,ix2,k2,k3,k4,k3,sx1,sx2))/sqdety
        elseif(zp(k2,k3).eq.1)then
          xD0_=-sgn*(intir(p1,x1,x2,ix1,ix2,sx1,sx2)
     *      -inta(x1,x2,ix1,ix2,k3,k4,sx1,sx2)
     *      +intb(x1,x2,ix1,ix2,k3,k1,k4,k1,sx1,sx2))/sqdety
        else
c error message if case not found
          write(*,'(a)')' xD0_: case not implemented'
          write(*,*)q1,q2,q3,q4,q12,q23,m12,m22,m32,m42
          stop
        endif
      elseif(nzm.ge.2)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                  0                                                   c
c       q4 ----O------O---- q1                                         c
c              |      |                                                c
c          m42 |      | m22                                            c
c              |      |                                                c
c       q3 ----O------O---- q2                                         c
c                  0                                                   c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        la=.true.
        lb=.true.
        lc=.true.
        do i1=0,4,2
        if(la.or.lb.or.lc)then
          a(i1)=0d0
          ia(i1)=0d0
          b(i1)=0d0
          ib(i1)=0d0
          c(i1)=0d0
          ic(i1)=0d0
          do i2=0,2,2
          do i3=0,2,2
          if(i1.eq.i2+i3)then
            if(la)then
              a(i1)=a(i1)+y(k1,k4,i2)*y(k3,k4,i3)
     *          +p(k1,k3,i2)*m(k4,i3)
              ia(i1)=ia(i1)+iy(k1,k4,i2)*iy(k3,k4,i3)
     *          +ip(k1,k3,i2)*im(k4,i3)
            endif
            if(lb)then
              b(i1)=b(i1)+y(k1,k2,i2)*y(k3,k4,i3)
     *          +y(k1,k4,i2)*y(k2,k3,i3)+p(k1,k3,i2)*y(k2,k4,i3)
              ib(i1)=ib(i1)+iy(k1,k2,i2)*iy(k3,k4,i3)
     *          +iy(k1,k4,i2)*iy(k2,k3,i3)+ip(k1,k3,i2)*iy(k2,k4,i3)
            endif
            if(lc)then
              c(i1)=c(i1)+y(k1,k2,i2)*y(k2,k3,i3)
     *          +p(k1,k3,i2)*m(k2,i3)
              ic(i1)=ic(i1)+iy(k1,k2,i2)*iy(k2,k3,i3)
     *          +ip(k1,k3,i2)*im(k2,i3)
            endif
          endif
          enddo
          enddo
          ia(i1)=dcmplx(dreal(a(i1)),dimag(ia(i1)))
          ib(i1)=dcmplx(dreal(b(i1)),dimag(ib(i1)))
          ic(i1)=dcmplx(dreal(c(i1)),dimag(ic(i1)))
          if(la.and.a(i1).ne.0d0)sa=i1
          if(lb.and.b(i1).ne.0d0)sb=i1
          if(lc.and.c(i1).ne.0d0)sc=i1
          if(la)la=a(i1).eq.c0d0
          if(lb)lb=b(i1).eq.c0d0
          if(lc)lc=c(i1).eq.c0d0
        endif
        enddo
c num
        snum=0
        sgn=dsign(1d0,dreal(b(snum)))
        sqdety=cdsqrt(b(snum)**2-4d0*a(snum)*c(snum))
        num=b(snum)+sgn*sqdety
        inum=ib(snum)
     *    +sgn*(ib(snum)**2-4d0*ia(snum)*ic(snum))/2d0/sqdety
        inum=dcmplx(dreal(num),dimag(inum))
c error messages
        if(sqdety.eq.c0d0)then
          write(*,'(a)')' xD0_: dety=small not allowed'
          xD0_=0d0
cacb
c          stop
        endif
        if(num.eq.c0d0)then
          write(*,'(a)')' xD0_: num=small not allowed'
          stop
        endif
        if(sa.gt.4.or.sb.gt.4.or.sc.gt.4)then
          write(*,'(a)')' xD0_: sa, sb, or sc > 4'
          stop
        endif
        if(cdabs(a(sa)).lt.acc)write(*,'(a)')' xD0_: |a| < acc'
        if(cdabs(b(sb)).lt.acc)write(*,'(a)')' xD0_: |b| < acc'
        if(cdabs(c(sc)).lt.acc)write(*,'(a)')' xD0_: |c| < acc'
c x1,x2
        sx1=-sa
        x1=num/2d0/a(sa)
        ix1=dcmplx(dreal(x1),dimag(inum/2d0/ia(sa)))
        sx2=sc
        x2=2d0*c(sc)/num
        ix2=dcmplx(dreal(x2),dimag(2d0*ic(sc)/inum))
        p12=p(k1,k3,sp(k1,k3))
        xD0_=sgn*(xlog(-p12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *    +inta(x1,x2,ix1,ix2,k2,k4,sx1,sx2)
c     *    +xli2(x1,ix1,sx1)-xli2(x2,ix2,sx2)
     *    -intb(x1,x2,ix1,ix2,k2,k1,k4,k1,sx1,sx2)
     *    -intb(x1,x2,ix1,ix2,k2,k3,k4,k3,sx1,sx2))/sqdety
      elseif(nzm.eq.1.and.zp(k2,k3).eq.1)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                 0                                                    c
c       q4 ----O------O---- q1                                         c
c              |      |                                                c
c          m42 |      | m22                                            c
c              |      |                                                c
c       q3 ----O------O---- 0                                          c
c                 m32                                                  c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        la=.true.
        lb=.true.
        lc=.true.
        do i1=0,4,2
        if(la.or.lb.or.lc)then
          a(i1)=0d0
          ia(i1)=0d0
          b(i1)=0d0
          ib(i1)=0d0
          c(i1)=0d0
          ic(i1)=0d0
          do i2=0,2,2
          do i3=0,2,2
          if(i1.eq.i2+i3)then
            p1=p(k1,k2,i3)
            p2=p(k2,k3,i3)
            p3=p(k3,k4,i3)
            p4=p(k1,k4,i3)
            p12=p(k1,k3,i3)
            p23=p(k2,k4,i3)
            ip1=ip(k1,k2,i3)
            ip2=ip(k2,k3,i3)
            ip3=ip(k3,k4,i3)
            ip4=ip(k1,k4,i3)
            ip12=ip(k1,k3,i3)
            ip23=ip(k2,k4,i3)
            if(la)then
              a(i1)=a(i1)+y(k1,k2,i2)*(-p4)+y(k1,k3,i2)*p4
     *          +y(k1,k4,i2)*(p1+p3-p12-p23)
              ia(i1)=ia(i1)+iy(k1,k2,i2)*(-ip4)+iy(k1,k3,i2)*ip4
     *          +iy(k1,k4,i2)*(ip1+ip3-ip12-ip23)
            endif
            if(lb)then
              b(i1)=b(i1)+y(k1,k2,i2)*(p3-p4-p12)
     *          +y(k1,k3,i2)*(p1-p23+p4)+y(k1,k4,i2)*(p1-p12)
              ib(i1)=ib(i1)+iy(k1,k2,i2)*(ip3-ip4-ip12)
     *          +iy(k1,k3,i2)*(ip1-ip23+ip4)+iy(k1,k4,i2)*(ip1-ip12)
            endif
            if(lc)then
              c(i1)=c(i1)+m(k2,i2)*(-p12)+m(k3,i2)*p1
              ic(i1)=ic(i1)+im(k2,i2)*(-ip12)+im(k3,i2)*ip1
            endif
          endif
          enddo
          enddo
          ia(i1)=dcmplx(dreal(a(i1)),dimag(ia(i1)))
          ib(i1)=dcmplx(dreal(b(i1)),dimag(ib(i1)))
          ic(i1)=dcmplx(dreal(c(i1)),dimag(ic(i1)))
          if(la.and.a(i1).ne.0d0)sa=i1
          if(lb.and.b(i1).ne.0d0)sb=i1
          if(lc.and.c(i1).ne.0d0)sc=i1
          if(la)la=a(i1).eq.c0d0
          if(lb)lb=b(i1).eq.c0d0
          if(lc)lc=c(i1).eq.c0d0
        endif
        enddo
c num
        snum=0
        sgn=dsign(1d0,dreal(b(snum)))
        sqdety=cdsqrt(b(snum)**2-4d0*a(snum)*c(snum))
        num=b(snum)+sgn*sqdety
        inum=ib(snum)
     *    +sgn*(ib(snum)**2-4d0*ia(snum)*ic(snum))/2d0/sqdety
        inum=dcmplx(dreal(num),dimag(inum))
c error messages
        if(sqdety.eq.c0d0)then
          write(*,'(a)')' xD0_: dety=small not allowed'
          xD0_=0d0
cacb
c          stop
        endif
        if(num.eq.c0d0)then
          write(*,'(a)')' xD0_: num=small not allowed'
          stop
        endif
        if(sa.gt.4.or.sb.gt.4.or.sc.gt.4)then
          write(*,'(a)')' xD0_: sa,sb, or sc > 4'
          stop
        endif
        if(cdabs(a(sa)).lt.acc)write(*,'(a)')' xD0_: |a| < acc'
        if(cdabs(b(sb)).lt.acc)write(*,'(a)')' xD0_: |b| < acc'
        if(cdabs(c(sc)).lt.acc)write(*,'(a)')' xD0_: |c| < acc'
c x1,x2
        sx1=-sa
        x1=num/2d0/a(sa)
        ix1=dcmplx(dreal(x1),dimag(inum/2d0/ia(sa)))
        sx2=sc
        x2=2d0*c(sc)/num
        ix2=dcmplx(dreal(x2),dimag(2d0*ic(sc)/inum))
        xD0_=-sgn*(inta(x1,x2,ix1,ix2,k2,k4,sx1,sx2)
     *    -intb(x1,x2,ix1,ix2,k2,k1,k4,k1,sx1,sx2)
     *    -inta(x1,x2,ix1,ix2,k3,k4,sx1,sx2)
     *    +intb(x1,x2,ix1,ix2,k3,k1,k4,k1,sx1,sx2))/sqdety
      elseif(nzm.eq.1.and.zp(k2,k3).eq.0)then
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c                 0                                                    c
c       q4 ----O------O---- q1                                         c
c              |      |                                                c
c          m42 |      | m22                                            c
c              |      |                                                c
c       q3 ----O------O---- q2                                         c
c                 m32                                                  c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c determination of alpha
        do i1=0,4,2
          alpha2(i1)=0d0
          do i2=0,2,2
          do i3=0,2,2
          if(i1.eq.i2+i3)then
            alpha2(i1)=alpha2(i1)
     *        +(p(k2,k4,i2)-p(k2,k3,i2)-p(k3,k4,i2))
     *          *(p(k2,k4,i3)-p(k2,k3,i3)-p(k3,k4,i3))
     *        -4d0*p(k2,k3,i2)*p(k3,k4,i3)
          endif
          enddo
          enddo
        enddo
        if(alpha2(0).ne.c0d0)then
          alpha(0)=cdsqrt(alpha2(0))
          alpha(1)=alpha2(2)/2d0/alpha(0)
          alpha(2)=-(alpha2(4)/alpha(0)+alpha(1))/4d0
        elseif(alpha2(2).ne.c0d0)then
          alpha(0)=0d0
          alpha(1)=cdsqrt(alpha2(2))
          alpha(2)=alpha2(4)/2d0/alpha(1)
        else
          alpha(0)=0d0
          alpha(1)=0d0
          alpha(2)=cdsqrt(alpha2(4))
        endif
        if(dimag(alpha(0)).ne.0d0.or.dimag(alpha(1)).ne.0d0.or.
     *    dimag(alpha(2)).ne.0d0)then
          write(*,'(a)')' xD0_: complex alpha!'
          stop
        endif
c starting the calculation
        xD0_=0d0
        do i1=1,3
c rotating indices
          i2=k2
          k2=k3
          k3=k4
          k4=i2
c starting then evaluation
          la=.true.
          lb=.true.
          lc=.true.
          do i2=0,6
          if(la.or.lb.or.lc)then
            a(i2)=0d0
            ia(i2)=0d0
            b(i2)=0d0
            ib(i2)=0d0
            c(i2)=0d0
            ic(i2)=0d0
            do i3=0,2,2
            do i4=0,2
            do i5=0,2,2
            if(i2.eq.i3+i4+i5)then
              noal=p(k3,k4,i4)/p(k3,k4,sp(k3,k4))
              al=(p(k2,k4,i4)-p(k2,k3,i4)-p(k3,k4,i4)+alpha(i4))
     *          /2d0/p(k3,k4,sp(k3,k4))
              p1=p(k1,k2,i5)
              p2=p(k2,k3,i5)
              p3=p(k3,k4,i5)
              p4=p(k1,k4,i5)
              p12=p(k1,k3,i5)
              p23=p(k2,k4,i5)
              if(la)then
                a(i2)=a(i2)+y(k1,k2,i3)*noal*p12
     *            +y(k1,k3,i3)*(noal*(p2-p1)+al*(p3-p4))
     *            +y(k1,k4,i3)*al*p12
                ia(i2)=ia(i2)+iy(k1,k2,i3)*noal*p12
     *            +iy(k1,k3,i3)*(noal*(p2-p1)+al*(p3-p4))
     *            +iy(k1,k4,i3)*al*p12
              endif
              if(lb)then
                b(i2)=b(i2)+y(k1,k2,i3)*noal*(-p3+p4+p12)
     *            +y(k1,k3,i3)*(noal*(-p1-p4+p23)-2d0*al*p4)
     *            +y(k1,k4,i3)*(noal*(-p1+p2+p12)+2d0*al*p12)
                ib(i2)=ib(i2)+iy(k1,k2,i3)*noal*(-p3+p4+p12)
     *            +iy(k1,k3,i3)*(noal*(-p1-p4+p23)-2d0*al*p4)
     *            +iy(k1,k4,i3)*(noal*(-p1+p2+p12)+2d0*al*p12)
              endif
              if(lc)then
                c(i2)=c(i2)+y(k1,k2,i3)*noal*p4
     *            +y(k1,k3,i3)*(-noal-al)*p4
     *            +y(k1,k4,i3)*(noal*(p23-p1)+(-noal-al)*(p3-p12))
                ic(i2)=ic(i2)+iy(k1,k2,i3)*noal*p4
     *            +iy(k1,k3,i3)*(-noal-al)*p4
     *            +iy(k1,k4,i3)*(noal*(p23-p1)+(-noal-al)*(p3-p12))
              endif
            endif
            enddo
            enddo
            enddo
            ia(i2)=dcmplx(dreal(a(i2)),dimag(ia(i2)))
            ib(i2)=dcmplx(dreal(b(i2)),dimag(ib(i2)))
            ic(i2)=dcmplx(dreal(c(i2)),dimag(ic(i2)))
            if(la.and.a(i2).ne.0d0)sa=i2
            if(lb.and.b(i2).ne.0d0)sb=i2
            if(lc.and.c(i2).ne.0d0)sc=i2
            if(la)la=a(i2).eq.c0d0
            if(lb)lb=b(i2).eq.c0d0
            if(lc)lc=c(i2).eq.c0d0
          endif
          enddo
c num
          snum=0
          sgn=dsign(1d0,dreal(b(snum)))
          sqdety=cdsqrt(b(snum)**2-4d0*a(snum)*c(snum))
          num=b(snum)+sgn*sqdety
          inum=ib(snum)
     *      +sgn*(ib(snum)**2-4d0*ia(snum)*ic(snum))/2d0/sqdety
          inum=dcmplx(dreal(num),dimag(inum))
c error messages
          if(sqdety.eq.c0d0)then
            write(*,'(a)')' xD0_: dety=small not allowed'
            xD0_=0d0
cacb
c            stop
          endif
          if(num.eq.c0d0)then
            write(*,'(a)')' xD0_: num=small not allowed'
            stop
          endif
          if(sa.gt.6.or.sb.gt.6.or.sc.gt.6)then
            write(*,'(a)')' xD0_: sa,sb, or sc > 6'
            stop
          endif
          if(cdabs(a(sa)).lt.acc)write(*,'(a)')' xD0_: |a| < acc'
          if(cdabs(b(sb)).lt.acc)write(*,'(a)')' xD0_: |b| < acc'
          if(cdabs(c(sc)).lt.acc)write(*,'(a)')' xD0_: |c| < acc'
c x1,x2
          sx1=-sa
          x1=num/2d0/a(sa)
          ix1=dcmplx(dreal(x1),dimag(inum/2d0/ia(sa)))
          sx2=sc
          x2=2d0*c(sc)/num
          ix2=dcmplx(dreal(x2),dimag(2d0*ic(sc)/inum))
          xD0_=xD0_-sgn*(inta(x1,x2,ix1,ix2,k4,k3,sx1,sx2)
     *      -intb(x1,x2,ix1,ix2,k4,k1,k3,k1,sx1,sx2)
     *      -xli2(x1,ix1,sx1)+xli2(x2,ix2,sx2))/sqdety
        enddo
      else
c error message if case not found
        write(*,'(a)')' xD0_: case not implemented'
        write(*,*)q1,q2,q3,q4,q12,q23,m12,m22,m32,m42
        stop
      endif
      fct(1)=xD0_
      call fastout(fct,1,3)
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     calculation of the basic integral                                c
c                                                                      c
c     inta=int_0^infinity dx [1/(x+x1)-1/(x+x2)]                       c
c                            *log[m12+(m12+m22-s)*x+m22*x^2]           c
c                                                                      c
c     Im(m12),Im(m22)<0,Im(s)>0                                        c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function inta(x1,x2,ix1,ix2,k1,k2,sx1,sx2)
      implicit none
c local variables
      complex*16 inta,x1,x2,alpha,r1,r2,c0d0,pi6,id1,id2,i,m12,m22,s
      complex*16 xeta,xlog,xli2t,xli2,d1,d2,eta,ialpha,ix1,ix2,ir1,ir2
      complex*16 im12,im22,is
      real*8 pi,pi2,xmin,sgn
      integer k1,k2,sx1,sx2,sd1,sd2,sr1,sr2,salpha
c xco
      complex*16 m(0:4,0:2),p(0:4,0:4,0:2),im(0:4,0:2),ip(0:4,0:4,0:2)
      integer sm(0:4),sp(0:4,0:4)
      common/xco/m,p,im,ip,sm,sp
      if(x1.eq.x2)then
        write(*,'(a)')' inta: x1=x2 not allowed'
        inta=0d0
        stop
      endif
      i=dcmplx(0d0,1d0)
      c0d0=dcmplx(0d0)
      pi=4d0*datan(1d0)
      pi2=pi*pi
      pi6=dcmplx(pi2/6d0)
      m12=m(k1,sm(k1))
      m22=m(k2,sm(k2))
      s=p(k1,k2,sp(k1,k2))
      im12=im(k1,sm(k1))
      im22=im(k2,sm(k2))
      is=ip(k1,k2,sp(k1,k2))
      if(m12.eq.c0d0)then
        if(m22.eq.c0d0)then
          inta=xlog(-s,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      -0.5d0*xlog(x1,ix1)**2+0.5d0*xlog(x2,ix2)**2
        elseif(s.eq.c0d0)then
          inta=xlog(m22,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      -0.5d0*xlog(x1,ix1)**2+0.5d0*xlog(x2,ix2)**2
     *      +xli2(x1,ix1,sx1)-xli2(x2,ix2,sx2)
        elseif(m22.eq.s)then
          inta=xlog(m22,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      -xlog(x1,ix1)**2+xlog(x2,ix2)**2
        else
          sd2=min(sm(k2),sp(k1,k2))
          d2=m(k2,sd2)-p(k1,k2,sd2)
          id2=im(k2,sd2)-ip(k1,k2,sd2)
          sr1=sm(k2)-sd2
          r1=m22/d2
          ir1=dcmplx(dreal(r1),dimag(im22/id2))
          inta=xlog(d2,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      -0.5d0*xlog(x1,ix1)**2+0.5d0*xlog(x2,ix2)**2
     *      +xli2t(x1,r1,ix1,ir1,sx1,sr1)
     *      -xli2t(x2,r1,ix2,ir1,sx2,sr1)
        endif
      elseif(m22.eq.c0d0)then
        if(m12.eq.s)then
          inta=xlog(m12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
        elseif(s.eq.c0d0)then
          inta=xlog(m12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      +xli2(x1,ix1,sx1)-xli2(x2,ix2,sx2)
        else
          sd1=min(sm(k1),sp(k1,k2))
          d1=m(k1,sd1)-p(k1,k2,sd1)
          id1=im(k1,sd1)-ip(k1,k2,sd1)
          sr1=sd1-sm(k1)
          r1=d1/m12
          ir1=dcmplx(dreal(r1),dimag(id1/im12))
          inta=xlog(m12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *      +xli2t(x1,r1,ix1,ir1,sx1,sr1)
     *      -xli2t(x2,r1,ix2,ir1,sx2,sr1)
        endif
      elseif(s.eq.c0d0)then
        sr1=sm(k2)-sm(k1)
        r1=m22/m12
        ir1=dcmplx(dreal(r1),dimag(im22/im12))
        inta=xlog(m12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *    +xli2t(x1,r1,ix1,ir1,sx1,sr1)-xli2t(x2,r1,ix2,ir1,sx2,sr1)
     *    +xli2(x1,ix1,sx1)-xli2(x2,ix2,sx2)
      else
        if(sp(k1,k2).eq.sm(k1).and.sp(k1,k2).eq.sm(k2))then
          sgn=dsign(1d0,dreal(m12+m22-s))
          alpha=sgn*cdsqrt((s-m12-m22)**2-4d0*m12*m22)
          ialpha=(is-im12-im22)**2-4d0*im12*im22
          ialpha=dcmplx(dreal(alpha),dimag(ialpha/2d0/alpha))
          sr1=0
          r1=(m12+m22-s+alpha)/2d0/m12
          ir1=(im12+im22-is+ialpha)/2d0/im12
          ir1=dcmplx(dreal(r1),dimag(ir1))
          sr2=0
          r2=2d0*m22/(m12+m22-s+alpha)
          ir2=2d0*im22/(im12+im22-is+ialpha)
          ir2=dcmplx(dreal(r2),dimag(ir2))
        elseif(sp(k1,k2).le.sm(k1).and.sp(k1,k2).le.sm(k2))then
          salpha=sp(k1,k2)
          alpha=m(k1,salpha)+m(k2,salpha)-p(k1,k2,salpha)
          if(alpha.ne.c0d0)then
            ialpha=im(k1,salpha)+im(k2,salpha)-ip(k1,k2,salpha)
            sr1=salpha-sm(k1)
            r1=alpha/m12
            ir1=dcmplx(dreal(r1),dimag(ialpha/im12))
            sr2=sm(k2)-salpha
            r2=m22/alpha
            ir2=dcmplx(dreal(r2),dimag(im22/ialpha))
          else
            alpha=2d0*i*cdsqrt(m12*m22)
            sr1=(sm(k2)-sm(k1))/2
            r1=alpha/2d0/m12**2
            ir1=dcmplx(dreal(r1),dimag(ialpha/2d0/im12**2))
            sr2=(sm(k2)-sm(k1))/2
            r2=-alpha/2d0/m12**2
            ir2=dcmplx(dreal(r2),dimag(-ialpha/2d0/im12**2))
          endif
        elseif(sp(k1,k2).ge.sm(k1).and.sp(k1,k2).ge.sm(k2))then
          sr1=0
          r1=1d0
          ir1=1d0
          sr2=sm(k2)-sm(k1)
          r2=m22/m12
          ir2=dcmplx(dreal(r2),dimag(im22/im12))
        endif
        inta=xlog(m12,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *    +xli2t(x1,r1,ix1,ir1,sx1,sr1)-xli2t(x2,r1,ix2,ir1,sx2,sr1)
     *    +xli2t(x1,r2,ix1,ir2,sx1,sr2)-xli2t(x2,r2,ix2,ir2,sx2,sr2)
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     calculation of the basic integral                                c
c                                                                      c
c     intb=int_0^infinity dx [1/(x+x1)-1/(x+x2)]                       c
c                            *log[(m12-s1)+(m22-s2)*x]                 c
c                                                                      c
c     Im(m12-s1),Im(m22-s2)<0                                          c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function intb(x1,x2,ix1,ix2,k1,j1,k2,j2,sx1,sx2)
      implicit none
c local variables
      complex*16 intb,x1,x2,d1,d2,r1,ix1,ix2,ir1,xli2t,xlog
      complex*16 c0d0,id1,id2,i
      integer k1,k2,j1,j2,sd1,sd2,sx1,sx2,sr1
c xco
      complex*16 m(0:4,0:2),p(0:4,0:4,0:2),im(0:4,0:2),ip(0:4,0:4,0:2)
      integer sm(0:4),sp(0:4,0:4)
      common/xco/m,p,im,ip,sm,sp
      if(x1.eq.x2)then
        write(*,'(a)')' intb: x1=x2 not allowed'
        intb=0d0
        stop
      endif
      i=dcmplx(0d0,1d0)
      c0d0=dcmplx(0d0)
      sd1=min(sm(k1),sp(k1,j1))
      d1=m(k1,sd1)-p(k1,j1,sd1)
      sd2=min(sm(k2),sp(k2,j2))
      d2=m(k2,sd2)-p(k2,j2,sd2)
      if(d1.eq.c0d0)then
        intb=xlog(d2,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *    -0.5d0*xlog(x1,ix1)**2+0.5d0*xlog(x2,ix2)**2
      elseif(d2.eq.c0d0)then
        intb=xlog(d1,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
      else
        id1=im(k1,sd1)-ip(k1,j1,sd1)
        id2=im(k2,sd2)-ip(k2,j2,sd2)
        sr1=sd2-sd1
        r1=d2/d1
        ir1=dcmplx(dreal(r1),dimag(id2/id1))
        intb=xlog(d1,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *    +xli2t(x1,r1,ix1,ir1,sx1,sr1)-xli2t(x2,r1,ix2,ir1,sx2,sr1)
        if(dimag(r1).eq.0d0.and.dimag(ir1).eq.0d0)then
          write(*,'(a)')' intb: imaginary part of r1 undefined'
          stop
        endif
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     calculation of the basic infrared-divergent integral             c
c     (for definition see infrared-divergent 3-point function)         c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function intir(s,x1,x2,ix1,ix2,sx1,sx2)
      implicit none
c local variables
      complex*16 intir,s,x1,x2,ix1,ix2,r1,ir1,r2,ir2,y,iy,z,iz,i
      complex*16 xli2,xlog,xeta,c0d0,eta
      real*8 xmin
      integer sx1,sx2,sr1,sr2,sz
c ir
      real*8 lambda,lambda2
      common/ir/lambda,lambda2
      i=dcmplx(0d0,1d0)
      c0d0=dcmplx(0d0)
      sr1=-sx1
      r1=1d0/x1
      ir1=dcmplx(dreal(r1),dimag(1d0/ix1))
      sr2=-sx2
      r2=1d0/x2
      ir2=dcmplx(dreal(r2),dimag(1d0/ix2))
      sz=sx1-sx2
      z=x1/x2
      iz=dcmplx(dreal(z),dimag(ix1/ix2))
      if(sz.eq.0)then
        y=2d0-z-1d0/z
        iy=2d0-iz-1d0/iz
      elseif(sz.lt.0)then
        y=-z
        iy=-iz
      else
        y=-1d0/z
        iy=-1d0/iz
      endif
      iy=dcmplx(dreal(y),dimag(iy))
      intir=0.5d0*xlog(s/lambda2,-i)*(xlog(x2,ix2)-xlog(x1,ix1))
     *  +xli2(z,iz,sz)+0.25d0*xlog(z,iz)**2
     *  +0.5d0*xeta(x1,r2,ix1,ir2)*xlog(y,iy)
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     logarithm                                                        c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function xlog(z,iz)
      implicit none
c local variables
      complex*16 z,iz,xlog
      real*8 x,pi
      pi=4d0*datan(1d0)
      if(dimag(z).ne.0d0.or.dreal(z).gt.0d0)then
        xlog=cdlog(z)
      else
        x=-dreal(z)
        if(dimag(iz).gt.0d0)then
          xlog=dcmplx(dlog(x),pi)
        elseif(dimag(iz).lt.0d0)then
          xlog=dcmplx(dlog(x),-pi)
        else
          write(*,'(a)')' xlog: argument on cut'
          stop
        endif
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     generalized dilogarithm                                          c
c                                                                      c
c     xli2t(x,r)=li2(1-x*r)+eta(x,r)*log(1-x*r)                        c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function xli2t(x,r,ix,ir,sx,sr)
      implicit none
c local variables
      complex*16 xli2t,xli2,eta,xeta,xlog,x,r,z,ix,ir,iz,pi6,c0d0
      integer sx,sr,sz
      c0d0=dcmplx(0d0)
      pi6=dcmplx(datan(1d0)**2*8d0/3d0)
      sz=sx+sr
      iz=ix*ir
      z=x*r
      if(sz.gt.0)then
        xli2t=pi6
      else
        xli2t=xli2(z,iz,sz)
        eta=xeta(x,r,ix,ir)
        if(eta.ne.c0d0.and.sz.eq.0)then
          xli2t=xli2t+eta*xlog(1d0-z,-iz)
        elseif(eta.ne.c0d0.and.sz.lt.0)then
          xli2t=xli2t+eta*xlog(-z,-iz)
        endif
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     dilogarithm                                                      c
c                                                                      c
c     warning: different input as usual                                c
c              xli2(z)=Spen(1-z)                                       c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function xli2(z,iz,sz)
      implicit none
c local variables
      complex*16 xli2,cspen,xlog,pi6,z,iz
      integer sz
      pi6=dcmplx(datan(1d0)**2*8d0/3d0)
      if(sz.gt.0)then
        xli2=pi6
      elseif(sz.lt.0)then
        xli2=-0.5d0*xlog(z,iz)**2-pi6
      elseif(dreal(z).ge.0d0)then
        xli2=cspen(1d0-z)
      else
        xli2=-cspen(z)-xlog(1d0-z,-iz)*xlog(z,iz)+pi6
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     eta function                                                     c
c                                                                      c
c     eta(z1,z2)=log(z1*z2)-log(z1)-log(z2)                            c
c                                                                      c
c     written by Markus Roth, 20.10.2004                               c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      function xeta(z1,z2,iz1,iz2)
      implicit none
c local variables
      complex*16 xeta,z1,z2,iz1,iz2
      real*8 im1,im2,im12,pi
      pi=4d0*datan(1d0)
      xeta=0d0
      im12=dimag(z1*z2)
      im1=dimag(z1)
      im2=dimag(z2)
      if(im1.eq.0d0.and.im2.eq.0d0)im12=dimag(iz1*iz2)
      if(im1.eq.0d0)im1=dimag(iz1)
      if(im2.eq.0d0)im2=dimag(iz2)
      if(im1.lt.0d0.and.im2.lt.0d0.and.im12.ge.0d0)then
        xeta=dcmplx(0d0,2d0*pi)
      elseif(im1.ge.0d0.and.im2.ge.0d0.and.im12.lt.0d0)then
        xeta=-dcmplx(0d0,2d0*pi)
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     no fast evaluation of tensor integrals                           c
c                                                                      c
c     written by Markus Roth, 3.1.2005                                 c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine fastoff
      implicit none
c local variables
      integer maxt,maxf,maxn,maxfct,maxx
      parameter(maxt=10,maxf=500,maxn=2000,maxfct=110,maxx=16)
c fast
      complex*16 f(maxt,maxf,maxfct),arg(maxt,maxf,maxx)
      integer rank(maxt,maxf),nf(maxt),calc(maxt,maxn)
      integer pointer(maxt,maxn),nt(maxt),n
      character*11 name(maxt)
      logical nofast
      common/fast/f,arg,rank,nf,calc,pointer,nt,name         
      common/fast2/n,nofast
      nofast=.true.
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     initialization of fast evaluation of tensor integrals            c
c                                                                      c
c     written by Markus Roth, 3.1.2005                                 c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine fastinit(nout)
      implicit none
c local variables
      integer maxt,maxf,maxn,maxfct,maxx
      parameter(maxt=10,maxf=500,maxn=2000,maxfct=110,maxx=16)
      integer i1,i2,nout
c fast
      complex*16 f(maxt,maxf,maxfct),arg(maxt,maxf,maxx)
      integer rank(maxt,maxf),nf(maxt),calc(maxt,maxn)
      integer pointer(maxt,maxn),nt(maxt),n
      character*11 name(maxt)
      logical nofast
      common/fast/f,arg,rank,nf,calc,pointer,nt,name         
      common/fast2/n,nofast
      data nofast/.true./,n/0/
c initialization for the first call
      if(n.eq.0)then
        do i1=1,maxt
          nf(i1)=0
          do i2=1,maxn
            calc(i1,i2)=0
          enddo
        enddo
      endif
c output of initialization
c      if(nout.ne.0.and.n.le.maxt.and..not.nofast)then
c        if(n.eq.0)then
c          write(nout,'(a)')' '
c          write(nout,'(a)')
c     *      ' Tensor integral      Type        Calls     Calculations'
c        elseif(nf(n).ne.0)then
c          write(nout,'(1x,a11,3(4x,i9))')name(n),n,nt(n),nf(n)
c        endif
c      endif
c initialization for each event
      if(.not.nofast)n=n+1
      nofast=.false.
      do i1=1,maxt
        nt(i1)=0
      enddo
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     fast evaluation of tensor integrals                              c
c                                                                      c
c     written by Markus Roth, 3.1.2005                                 c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine fastin(fct,x,nfct,nx,type,nrank,namefct,nocalc)
      implicit none
c local variables
      integer maxt,maxf,maxn,maxfct,maxx
      parameter(maxt=10,maxf=500,maxn=2000,maxfct=110,maxx=16)
      complex*16 fct(maxfct),x(maxx)
      integer i1,i2,nfct,nx,type,nrank
      logical nocalc,same
      character*11 namefct
c fast
      complex*16 f(maxt,maxf,maxfct),arg(maxt,maxf,maxx)
      integer rank(maxt,maxf),nf(maxt),calc(maxt,maxn)
      integer pointer(maxt,maxn),nt(maxt),n
      character*11 name(maxt)
      logical nofast
      common/fast/f,arg,rank,nf,calc,pointer,nt,name         
      common/fast2/n,nofast
      nocalc=.false.
      if(nofast)return
      if(n.gt.type)then
c no optimization
        nt(type)=nt(type)+1
        if(calc(type,nt(type)).eq.1)then
c calculation of function
          nrank=rank(type,pointer(type,nt(type)))
        else
c using already calculated result to fasten computation
          nocalc=.true.
          do i1=1,nfct
            fct(i1)=f(type,pointer(type,nt(type)),i1)
          enddo
        endif
      elseif(n.eq.type)then
c error messages
        if(type.gt.maxt)then
          write(*,'(a)')' fastin: maxt too small'
          stop
        endif
        if(nt(type)+1.gt.maxn)then
          write(*,'(a)')' fastin: maxn too small'
          stop
        endif
        if(nf(type)+1.gt.maxf)then
          write(*,'(a)')' fastin: maxf too small'
          stop
        endif
        if(nfct.gt.maxfct)then
          write(*,'(a)')' fastin: maxfct too small'
          stop
        endif
        if(nx.gt.maxx)then
          write(*,'(a)')' fastin: maxx too small'
          stop
        endif
c initialization for function type
        nt(type)=nt(type)+1
        do i1=1,nf(type)
          same=.true.
          do i2=1,nx
            if(arg(type,i1,i2).ne.x(i2))same=.false.
          enddo
          if(same)then
            pointer(type,nt(type))=i1
            calc(type,nt(type))=0
            rank(type,i1)=max(rank(type,i1),nrank)
            return
          endif
        enddo
        nf(type)=nf(type)+1
        pointer(type,nt(type))=nf(type)
        calc(type,nt(type))=1
        rank(type,nf(type))=nrank
        name(type)=namefct
        do i1=1,nx
          arg(type,nf(type),i1)=x(i1)
        enddo
      endif
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                      c
c     fast evaluation of tensor integrals                              c
c                                                                      c
c     written by Markus Roth, 3.1.2005                                 c
c                                                                      c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine fastout(fct,nfct,type)
      implicit none
c local variables
      integer maxt,maxf,maxn,maxfct,maxx
      parameter(maxt=10,maxf=500,maxn=2000,maxfct=110,maxx=16)
      complex*16 fct(maxfct)
      integer i1,nfct,type
c fast
      complex*16 f(maxt,maxf,maxfct),arg(maxt,maxf,maxx)
      integer rank(maxt,maxf),nf(maxt),calc(maxt,maxn)
      integer pointer(maxt,maxn),nt(maxt),n
      character*11 name(maxt)
      logical nofast
      common/fast/f,arg,rank,nf,calc,pointer,nt,name         
      common/fast2/n,nofast
      if(nofast.or.n.le.type)return
      if(calc(type,nt(type)).eq.1)then
c storing result of function type
        do i1=1,nfct
          f(type,pointer(type,nt(type)),i1)=fct(i1)
        enddo
      endif
      end

